
# https://cmake.org/pipermail/cmake/2007-March/013081.html
set(CMAKE_CONFIGURATION_TYPES Debug Release CACHE TYPE INTERNAL FORCE)

OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." OFF)
IF(BUILD_SHARED_LIBS)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_WINDLL")
	SET(NLIB_DLL 1)  # for compatibility, to be removed
ENDIF()

# VS_USER_PROPS requires cmake 3.8
cmake_minimum_required(VERSION 3.8 FATAL_ERROR)
IF (CMAKE_VERSION VERSION_LESS 3.10.0)
	MESSAGE(WARNING "You should use the latest version of cmake")
ENDIF()

# VS2013 or above
IF(MSVC_VERSION LESS 1800)
	MESSAGE(FATAL_ERROR "!!!!! Please use Visual Studio 2013 or later. !!!!!")
ELSEIF(MSVC_VERSION EQUAL 1800)
	SET(CAT_DIR "vc120")
ELSEIF(MSVC_VERSION EQUAL 1900)
	SET(CAT_DIR "vc140")
ELSEIF((MSVC_VERSION GREATER 1909) AND (MSVC_VERSION LESS 1920))
	SET(CAT_DIR "vc150") # to replace it with vc141
ELSE()
	MESSAGE(WARNING "Sorry, Unknown Visual studio")
	SET(CAT_DIR "vcXXX")
ENDIF()
IF(BUILD_SHARED_LIBS)
	STRING(CONCAT CAT_DIR ${CAT_DIR} "dll")
ENDIF()
IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
	STRING(CONCAT CAT_DIR ${CAT_DIR} "amd64")
ENDIF()

# Only 64bit Windows
SET(ProgramW6432 "$ENV{ProgramW6432}")
IF(NOT ProgramW6432)
	MESSAGE(FATAL_ERROR "!!!!! Please use 64bit Windows !!!!!") 
ENDIF()

# no nmake support
IF(${CMAKE_GENERATOR} MATCHES "NMake Makefiles.*")
	MESSAGE(FATAL_ERROR "!!!!! Don't use nmake !!!!!") 
ENDIF()

# Only 64bit cmake
IF($ENV{PROCESSOR_ARCHITECTURE} MATCHES "x86")
	MESSAGE(WARNING "Please use 64bit cmake")
ENDIF()

# From cmake 3.8, you can install the binary version of boost into C:\boost
# https://sourceforge.net/projects/boost/files/boost-binaries/
IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
	# PROGRAMFILES has to be chagned if cmake is invoked from 32bit shell.
	SET(ENV{PROGRAMFILES} "$ENV{ProgramW6432}")
	IF(CMAKE_VERSION VERSION_LESS "3.8.0")
		IF(NOT DEFINED BOOSTROOT)
			SET(BOOSTROOT "D:/Program Files/boost") # workaround for D-drive users (use junction)
		ENDIF()
	ENDIF()
ELSE()
	# PROGRAMFILES has to be chagned if cmake is invoked from 64bit shell.
	SET(ENV{PROGRAMFILES} "$ENV{ProgramFiles\(x86\)}")
	IF(CMAKE_VERSION VERSION_LESS "3.8.0")
		IF(NOT DEFINED BOOSTROOT)
			SET(BOOSTROOT "D:/Program Files (x86)/boost") # workaround for D-drive users (use junction)
		ENDIF()
	ENDIF()
ENDIF()



# setup/Update-MsvcLibraries.ps1 installs zlib and curl using vcpkg by Microsoft
# does not use vcpkg toolchain file because it misdetects boost at C:\boost
IF (EXISTS "$ENV{VCPKG_ROOT}")
	IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
		SET(_root "$ENV{VCPKG_ROOT}/installed/x64-windows")
	ELSE()
		SET(_root "$ENV{VCPKG_ROOT}/installed/x86-windows")
	ENDIF()
	# from vcpkg.cmake
	IF(CMAKE_BUILD_TYPE MATCHES "^Debug$" OR NOT DEFINED CMAKE_BUILD_TYPE)
		LIST(APPEND CMAKE_PREFIX_PATH ${_root}/debug)
		LIST(APPEND CMAKE_LIBRARY_PATH ${_root}/debug/lib/manual-link)
		LIST(APPEND CMAKE_FIND_ROOT_PATH ${_root}/debug)
	endif()
	LIST(APPEND CMAKE_PREFIX_PATH ${_root})
	LIST(APPEND CMAKE_LIBRARY_PATH ${_root}/lib/manual-link)
	LIST(APPEND CMAKE_FIND_ROOT_PATH ${_root})
ELSEIF(EXISTS "${CMAKE_CURRENT_LIST_DIR}/../setup/.cache/vcpkg")
	IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
		SET(_root "${CMAKE_CURRENT_LIST_DIR}/../setup/.cache/vcpkg/installed/x64-windows")
	ELSE()
		SET(_root "${CMAKE_CURRENT_LIST_DIR}/../setup/.cache/vcpkg/installed/x86-windows")
	ENDIF()
	# from vcpkg.cmake
	IF(CMAKE_BUILD_TYPE MATCHES "^Debug$" OR NOT DEFINED CMAKE_BUILD_TYPE)
		LIST(APPEND CMAKE_PREFIX_PATH ${_root}/debug)
		LIST(APPEND CMAKE_LIBRARY_PATH ${_root}/debug/lib/manual-link)
		LIST(APPEND CMAKE_FIND_ROOT_PATH ${_root}/debug)
	endif()
	LIST(APPEND CMAKE_PREFIX_PATH ${_root})
	LIST(APPEND CMAKE_LIBRARY_PATH ${_root}/lib/manual-link)
	LIST(APPEND CMAKE_FIND_ROOT_PATH ${_root})
ELSE()
	# use system wide zlib and curl installed at C:\Program Files\ or C:\Program Files (x86)\
	IF(NOT DEFINED ZLIB_ROOT)
		SET(ZLIB_ROOT "$ENV{PROGRAMFILES}/zlib")
	ENDIF()
	SET(CMAKE_INCLUDE_PATH "$ENV{PROGRAMFILES}/CURL/include;${CMAKE_INCLUDE_PATH}")
	SET(CMAKE_LIBRARY_PATH "$ENV{PROGRAMFILES}/CURL/lib;${CMAKE_LIBRARY_PATH}")
ENDIF()

SET(CMAKE_DEBUG_POSTFIX "-${CAT_DIR}_d")
SET(CMAKE_RELEASE_POSTFIX "-${CAT_DIR}")

#
# Compiler options
#
# http://www.cmake.org/Wiki/CMake_FAQ#Dynamic_Replace
FOREACH(flag_var
         CMAKE_CXX_FLAGS CMAKE_CXX_FLAGS_DEBUG CMAKE_CXX_FLAGS_RELEASE
         CMAKE_CXX_FLAGS_MINSIZEREL CMAKE_CXX_FLAGS_RELWITHDEBINFO
         CMAKE_C_FLAGS CMAKE_C_FLAGS_DEBUG CMAKE_C_FLAGS_RELEASE
         CMAKE_C_FLAGS_MINSIZEREL CMAKE_C_FLAGS_RELWITHDEBINFO)
	STRING(REPLACE "/Zi" "/Z7" ${flag_var} "${${flag_var}}")
	STRING(REPLACE "/W3" "/W4" ${flag_var} "${${flag_var}}")
ENDFOREACH()
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NOMINMAX;WIN32_LEAN_AND_MEAN;_SCL_SECURE_NO_WARNINGS;_USE_MATH_DEFINES")

IF(NOT CMAKE_SIZEOF_VOID_P EQUAL 8)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/arch:SSE2")
ENDIF()
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/volatile:iso")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/bigobj;/Oi")
# replace /sdl with them (remove /we4995;/we4996)
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/we4156;/we4308;/we4532;/we4533;/we4700;/we4703;/we4789")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<CONFIG:Debug>:/Ob1>")
IF((MSVC_VERSION EQUAL 1900) OR (MSVC_VERSION GREATER 1900))
	# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/std:c++latest")
	# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_NOBOOST")
	IF(MSVC_VERSION EQUAL 1900)
		# disable C4592 temporarily for VS2015 Update1
		# see https://stackoverflow.com/questions/34013930/error-c4592-symbol-will-be-dynamically-initialized-vs2015-1-static-const-std
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/wd4592")
	ENDIF()
	IF((MSVC_VERSION EQUAL 1910) OR (MSVC_VERSION GREATER 1910))
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/permissive-")
		IF(MSVC_VERSION GREATER 1910)
			# Workaround for VS2017 15.3+
			SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/Zc:twoPhase-")
		ENDIF()
		IF((MSVC_VERSION EQUAL 1914) OR (MSVC_VERSION GREATER 1914))
			# VS 2017 15.7+
			# https://blogs.msdn.microsoft.com/vcblog/2018/04/09/msvc-now-correctly-reports-__cplusplus/
			# not enabled for now
			# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/Zc:__cplusplus")
		ENDIF()
	ENDIF()
ELSE()
	# VS2013
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/wd4351")
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/Zc:rvalueCast")
ENDIF()

# To remove _MBCS
ADD_DEFINITIONS(-D_SBCS)

GET_FILENAME_COMPONENT(NLIB_DIR "${NLIB_RELEASE_ROOT}/lib/cmake/msvc" ABSOLUTE)
IF(BUILD_SHARED_LIBS)
	LIST(APPEND CMAKE_MSVCIDE_RUN_PATH "${NLIB_DIR}")
ENDIF()

FUNCTION(library_default_build_rule target)
	IF(BUILD_SHARED_LIBS)
		IF(EXISTS ${NLIB_PKGROOT_DIR}/../_common/cmake/version.rc.in)
			configure_file(
				${NLIB_PKGROOT_DIR}/../_common/cmake/version.rc.in
				${PROJECT_BINARY_DIR}/version.rc
				@ONLY)
			ADD_LIBRARY(${target} SHARED ${LIB_SRC} ${LIB_HEADER} ${PROJECT_BINARY_DIR}/version.rc)
		ELSE()
			ADD_LIBRARY(${target} SHARED ${LIB_SRC} ${LIB_HEADER})
		ENDIF()
	ELSE()
		ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
	ENDIF()
	TARGET_INCLUDE_DIRECTORIES(${target} BEFORE PRIVATE "${NLIB_PKGROOT_DIR}/include" "${NLIB_PKGROOT_DIR}/sources")
	TARGET_INCLUDE_DIRECTORIES(${target} BEFORE INTERFACE "${NLIB_PKGROOT_DIR}/include")
	SET(win_no_def NOGDICAPMASKS NOVIRTUALKEYCODES NOWINMESSAGES NOWINSTYLES
		 NOSYSMETRICS NOMENUS NOICONS NOKEYSTATES NOSYSCOMMANDS NORASTEROPS NOSHOWWINDOW
		 OEMRESOURCE NOATOM NOCLIPBOARD NOCTRLMGR NODRAWTEXT NOGDI NOMB NOMETAFILE NOSCROLL
		 NOSERVICE NOSOUND NOTEXTMETRIC NOWH NOWINOFFSETS NOCOMM NOKANJI NOHELP
		 NODEFFERWINDOWPOS NOMCX)
	SET_PROPERTY(TARGET ${target} APPEND PROPERTY COMPILE_DEFINITIONS "${win_no_def}")
	INSTALL(TARGETS ${target} DESTINATION ${NLIB_DIR})
	INSTALL(DIRECTORY ${NLIB_PKGROOT_DIR}/include DESTINATION . USE_SOURCE_PERMISSIONS)
ENDFUNCTION()

FUNCTION(target_link_libraries_default target)
	IF(WIN32 AND (MSVC_VERSION EQUAL 1900))
		# https://connect.microsoft.com/VisualStudio/feedback/details/2486566/std-atomic-udt-gives-errors-about-alignment-when-aligned-correctly
		SET_PROPERTY(TARGET ${target} APPEND PROPERTY COMPILE_DEFINITIONS "_ENABLE_ATOMIC_ALIGNMENT_FIX")
	ENDIF()
ENDFUNCTION()
