﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_BASE64_H_
#define INCLUDE_NN_NLIB_BASE64_H_

#include <utility>
#include "nn/nlib/UniquePtr.h"
#ifdef __cpp_rvalue_references
#include <tuple>
#endif

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC Base64Encoder NLIB_FINAL {
 public:
    enum CharOption {
        kBase64PlusSlash = 0,
        kBase64PlusMinus,
        kBase64MinusUnderscore,
        kBase64DotMinus,
        kBase64UnderscoreColon,
        kBase64UnderscoreMinus,
        kBase64DotUnderscore,
        kBase64ExclamationMinus,
        kBase64Max,
        kBase64Default = kBase64PlusSlash,
        kBase64UrlSafe = kBase64MinusUnderscore
    };

 public:
    static size_t GetRequiredSize(size_t srcsize) NLIB_NOEXCEPT {
        return ((srcsize + 2) / 3) * 4 + 1;
    }
    static errno_t
    Encode(char* dst, size_t dstsize, const void* src, size_t srcsize,
           CharOption char_option = kBase64Default, bool padding = false) NLIB_NOEXCEPT;
    template<size_t N>
    static errno_t
    Encode(char (&dst)[N], const void* src, size_t srcsize, CharOption char_option = kBase64Default,
           bool padding = false) NLIB_NOEXCEPT {
        return Encode(dst, N, src, srcsize, char_option, padding);
    }

#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    static errno_t
    Encode(UniquePtr<char[]>& dst, const void* src, size_t srcsize,
           CharOption char_option = kBase64Default, bool padding = false) NLIB_NOEXCEPT {
        // to avoid crossing the DLL boundary
        size_t dstsize = GetRequiredSize(srcsize);
        // do not use UniquePtr to avoid C4714
        char* p = new (std::nothrow) char[dstsize];
        if (!p) return ENOMEM;
        errno_t e = Encode(p, dstsize, src, srcsize, char_option, padding);
        if (e == 0)
            dst.reset(p);
        else
            delete[] p;
        return e;
    }
#ifdef __cpp_rvalue_references
    static std::pair<errno_t, std::unique_ptr<char[]> >
    Encode(const void* src, size_t srcsize, CharOption char_option = kBase64Default,
           bool padding = false) NLIB_NOEXCEPT {
        size_t dstsize = GetRequiredSize(srcsize);
        char* p        = new (std::nothrow) char[dstsize];
        if (!p) return ::std::make_pair(ENOMEM, nullptr);
        errno_t e = Encode(p, dstsize, src, srcsize, char_option, padding);
        if (e == 0) {
            return ::std::make_pair(0, UniquePtr<char[]>(p));
        } else {
            return ::std::make_pair(e, nullptr);
        }
    }
#endif
    NLIB_CEXPR Base64Encoder() NLIB_NOEXCEPT : rem_(-1), buf_(), table_() {}
    errno_t Init(CharOption char_option = kBase64Default) NLIB_NOEXCEPT;

    // note that dst is NOT null terminated
    errno_t StepEncode(size_t* written, char* dst, size_t dstsize, const void* src,
                       size_t srcsize) NLIB_NOEXCEPT NLIB_NONNULL;
    // note that dst is NOT null terminated
    template<size_t N>
    errno_t
    StepEncode(size_t* written, char (&dst)[N], const void* src, size_t srcsize) NLIB_NOEXCEPT {
        return StepEncode(written, dst, N, src, srcsize);
    }
    std::pair<errno_t, size_t>
    StepEncode(char* dst, size_t dstsize, const void* src, size_t srcsize) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = StepEncode(&written, dst, dstsize, src, srcsize);
        return std::make_pair(e, written);
    }
    template<size_t N>
    std::pair<errno_t, size_t>
    StepEncode(char (&dst)[N], const void* src, size_t srcsize) NLIB_NOEXCEPT {
        return StepEncode(&dst[0], N, src, srcsize);
    }

    // note that dst is NOT null terminated
    errno_t
    Close(size_t* written, char* dst, size_t dstsize, bool padding) NLIB_NOEXCEPT NLIB_NONNULL;
    // note that dst is NOT null terminated
    template<size_t N>
    errno_t Close(size_t* written, char (&dst)[N], bool padding) NLIB_NOEXCEPT {
        return Close(written, dst, N, padding);
    }
    std::pair<errno_t, size_t> Close(char* dst, size_t dstsize, bool padding) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = Close(&written, dst, dstsize, padding);
        return std::make_pair(e, written);
    }
    template<size_t N>
    std::pair<errno_t, size_t> Close(char (&dst)[N], bool padding) NLIB_NOEXCEPT {
        return Close(dst, N, padding);
    }
    NLIB_SAFE_BOOL(Base64Encoder, (rem_ >= 0))

 private:
    int rem_;
    uint8_t buf_[2];
    char table_[64];
    NLIB_DISALLOW_COPY_AND_ASSIGN(Base64Encoder);
};

class NLIB_VIS_PUBLIC Base64Decoder NLIB_FINAL {
 public:
    enum CharOption {
        kBase64PlusSlash = 0,
        kBase64PlusMinus,
        kBase64MinusUnderscore,
        kBase64DotMinus,
        kBase64UnderscoreColon,
        kBase64UnderscoreMinus,
        kBase64DotUnderscore,
        kBase64ExclamationMinus,
        kBase64Max,
        kBase64Default = kBase64PlusSlash,
        kBase64UrlSafe = kBase64MinusUnderscore
    };

 public:
    static size_t GetRequiredSize(size_t srcsize) NLIB_NOEXCEPT { return ((srcsize + 3) / 4) * 3; }
    static errno_t Decode(size_t* written, void* dst, size_t dstsize, const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        return Decode_(written, static_cast<nlib_byte_t*>(dst), dstsize, src, nlib_strlen(src),
                       char_option);
    }
    template<size_t N>
    static errno_t Decode(size_t* written, nlib_byte_t (&dst)[N], const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        return Decode_(written, dst, N, src, nlib_strlen(src), char_option);
    }
    static std::pair<errno_t, size_t>
    Decode(void* dst, size_t dstsize, const char* src,
           CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = Decode(&written, dst, dstsize, src, char_option);
        return std::make_pair(e, written);
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    Decode(nlib_byte_t (&dst)[N], const char* src,
           CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        return Decode(dst, N, src, char_option);
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    static errno_t Decode(UniquePtr<uint8_t[]>& dst, size_t* dstsize, const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        // to avoid crossing the DLL boundary
        size_t srcsize  = nlib_strlen(src);
        size_t dstsize_ = GetRequiredSize(srcsize);
        // do not use UniquePtr to avoid C4714
        uint8_t* p = new (std::nothrow) uint8_t[dstsize_];
        if (!p) return ENOMEM;
        size_t written;
        errno_t e = Decode_(&written, reinterpret_cast<nlib_byte_t*>(p), dstsize_, src, srcsize,
                            char_option);
        if (e == 0) {
            dst.reset(p);
            *dstsize = written;
        } else {
            delete[] p;
        }
        return e;
    }
#ifdef __cpp_rvalue_references
    static std::tuple<errno_t, std::unique_ptr<nlib_byte_t[]>, size_t>
    Decode(const char* src, CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        size_t srcsize = nlib_strlen(src);
        size_t dstsize = GetRequiredSize(srcsize);
        nlib_byte_t* p = new (std::nothrow) nlib_byte_t[dstsize];
        if (!p) return ::std::make_tuple(ENOMEM, nullptr, 0);
        size_t written;
        errno_t e = Decode_(&written, p, dstsize, src, srcsize, char_option);
        if (e == 0) {
            return ::std::make_tuple(0, UniquePtr<nlib_byte_t[]>(p), written);
        } else {
            delete[] p;
            return ::std::make_tuple(e, nullptr, 0);
        }
    }
#endif
    static errno_t
    DecodeInplace(size_t* written, char* src,
                  CharOption char_option = kBase64Default) NLIB_NOEXCEPT NLIB_NONNULL;

#if defined(NLIB_SIMD) && defined(NLIB_LITTLE_ENDIAN)
    NLIB_CEXPR Base64Decoder() NLIB_NOEXCEPT : rem_(-1), buf_(), c62_(), c63_(), table_() {}
#else
    NLIB_CEXPR Base64Decoder() NLIB_NOEXCEPT : rem_(-1), buf_(), table_() {}
#endif
    errno_t Init(CharOption char_option = kBase64Default) NLIB_NOEXCEPT;

    errno_t StepDecode(size_t* written, void* dst, size_t dstsize, const char* src,
                       size_t srcsize) NLIB_NOEXCEPT NLIB_NONNULL;
    template<size_t N>
    errno_t StepDecode(size_t* written, nlib_byte_t (&dst)[N], const char* src,
                       size_t srcsize) NLIB_NOEXCEPT {
        return StepDecode(written, dst, N, src, srcsize);
    }
    std::pair<errno_t, size_t>
    StepDecode(void* dst, size_t dstsize, const char* src, size_t srcsize) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = StepDecode(&written, dst, dstsize, src, srcsize);
        return std::make_pair(e, written);
    }
    template<size_t N>
    std::pair<errno_t, size_t>
    StepDecode(nlib_byte_t (&dst)[N], const char* src, size_t srcsize) NLIB_NOEXCEPT {
        return StepDecode(dst, N, src, srcsize);
    }
    errno_t Close(size_t* written, void* dst, size_t dstsize) NLIB_NOEXCEPT NLIB_NONNULL;
    template<size_t N>
    errno_t Close(size_t* written, nlib_byte_t (&dst)[N]) NLIB_NOEXCEPT {
        return Close(written, dst, N);
    }
    std::pair<errno_t, size_t> Close(void* dst, size_t dstsize) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = Close(&written, dst, dstsize);
        return std::make_pair(e, written);
    }
    template<size_t N>
    std::pair<errno_t, size_t> Close(nlib_byte_t (&dst)[N]) NLIB_NOEXCEPT {
        return Close(&dst[0], N);
    }
    NLIB_SAFE_BOOL(Base64Decoder, (rem_ >= 0))

 private:
    static errno_t Decode_(size_t* written, nlib_byte_t* dst, size_t dstsize, const char* src,
                           size_t srcsize, CharOption char_option) NLIB_NOEXCEPT NLIB_NONNULL;

    int rem_;
    int8_t buf_[4];
#if defined(NLIB_SIMD) && defined(NLIB_LITTLE_ENDIAN)
    char c62_;
    char c63_;
#endif
    int8_t table_[256];
    NLIB_DISALLOW_COPY_AND_ASSIGN(Base64Decoder);
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_BASE64_H_
