﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_BINARYWRITER_H_
#define INCLUDE_NN_NLIB_BINARYWRITER_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/OutputStream.h"

NLIB_NAMESPACE_BEGIN
NLIB_EXPIMP_TEMPLATE(detail::MiniBufOut<512>);

class NLIB_VIS_PUBLIC BinaryWriter NLIB_FINAL {
 public:
    enum EndianSetting { kEndianDefault = 0, kEndianLittle, kEndianBig };

 public:
    BinaryWriter() NLIB_NOEXCEPT : stream_(nullptr), errno_(-1), swap_endian_(false) {}
    ~BinaryWriter() NLIB_NOEXCEPT { this->Close(); }
    errno_t Init(EndianSetting endian) NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT { return Init(kEndianDefault); }
    errno_t Open(OutputStream* stream) NLIB_NOEXCEPT;

#define NLIB_BINW(tp) return this->Write_(static_cast<tp>(x))
    bool Write(char x) NLIB_NOEXCEPT { NLIB_BINW(uint8_t); }
    bool Write(signed char x) NLIB_NOEXCEPT { NLIB_BINW(uint8_t); }
    bool Write(unsigned char x) NLIB_NOEXCEPT { NLIB_BINW(uint8_t); }
    bool Write(short x) NLIB_NOEXCEPT { NLIB_BINW(uint16_t); }
    bool Write(unsigned short x) NLIB_NOEXCEPT { NLIB_BINW(uint16_t); }
    bool Write(int x) NLIB_NOEXCEPT { NLIB_BINW(uint32_t); }
    bool Write(unsigned int x) NLIB_NOEXCEPT { NLIB_BINW(uint32_t); }
    bool Write(long x) NLIB_NOEXCEPT { NLIB_BINW(nlib_ulong_compatible_t); }
    bool Write(unsigned long x) NLIB_NOEXCEPT { NLIB_BINW(nlib_ulong_compatible_t); }
    bool Write(long long x) NLIB_NOEXCEPT { NLIB_BINW(uint64_t); }
    bool Write(unsigned long long x) NLIB_NOEXCEPT { NLIB_BINW(uint64_t); }
    bool Write(float x) NLIB_NOEXCEPT { NLIB_BINW(float); }
    bool Write(double x) NLIB_NOEXCEPT { NLIB_BINW(double); }
#undef NLIB_BINW

    template<class T>
    bool Write(T x) NLIB_NOEXCEPT;

    bool WriteArray(const unsigned char* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const unsigned short* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const unsigned int* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const unsigned long long* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const unsigned long* x, size_t n) NLIB_NOEXCEPT {
#ifdef NLIB_LP64
        return this->WriteArray(reinterpret_cast<const unsigned long long*>(x), n);
#else
        return this->WriteArray(reinterpret_cast<const unsigned int*>(x), n);
#endif
    }

    bool WriteArray(const float* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const double* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArray(const signed char* x, size_t n) NLIB_NOEXCEPT {
        return this->WriteArray(reinterpret_cast<const unsigned char*>(x), n);
    }
    bool WriteArray(const char* x, size_t n) NLIB_NOEXCEPT {
        return this->WriteArray(reinterpret_cast<const unsigned char*>(x), n);
    }
    bool WriteArray(const short* x, size_t n) NLIB_NOEXCEPT {
        return this->WriteArray(reinterpret_cast<const unsigned short*>(x), n);
    }
    bool WriteArray(const int* x, size_t n) NLIB_NOEXCEPT {
        return this->WriteArray(reinterpret_cast<const unsigned int*>(x), n);
    }
    bool WriteArray(const long* x, size_t n) NLIB_NOEXCEPT {
#ifdef NLIB_LP64
        return this->WriteArray(reinterpret_cast<const unsigned long long*>(x), n);
#else
        return this->WriteArray(reinterpret_cast<const unsigned int*>(x), n);
#endif
    }
    bool WriteArray(const long long* x, size_t n) NLIB_NOEXCEPT {
        return this->WriteArray(reinterpret_cast<const unsigned long long*>(x), n);
    }
    template<class T>
    bool WriteArray(const T* x, size_t n) NLIB_NOEXCEPT;
    template<class T, size_t N>
    bool WriteArray(const T (&a)[N]) NLIB_NOEXCEPT;
    bool Flush() NLIB_NOEXCEPT;
    bool Close() NLIB_NOEXCEPT;
    void SetError(errno_t e) NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }
    OutputStream* GetStream() NLIB_NOEXCEPT { return stream_; }

    NLIB_SAFE_BOOL(BinaryWriter, GetErrorValue() == 0)

 private:
    bool Write_(uint8_t x) NLIB_NOEXCEPT;
    bool Write_(uint16_t x) NLIB_NOEXCEPT;
    bool Write_(uint32_t x) NLIB_NOEXCEPT;
    bool Write_(uint64_t x) NLIB_NOEXCEPT;
    bool Write_(float x) NLIB_NOEXCEPT;
    bool Write_(double x) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool WriteStream() NLIB_NOEXCEPT;

 private:
    OutputStream* stream_;
    ErrnoT errno_;
    bool swap_endian_;
    detail::MiniBufOut<512> mini_buf_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(BinaryWriter);
};

namespace binary_writer {

template<class T>
NLIB_ALWAYS_INLINE bool Write(BinaryWriter* w, T x) {
    return w->Write(x);
}

template<class T>
NLIB_ALWAYS_INLINE bool WriteArray(BinaryWriter* w, const T* x, size_t n) {
    return w->WriteArray(x, n);
}
}  // namespace binary_writer

template<class T>
NLIB_ALWAYS_INLINE bool BinaryWriter::Write(T x) NLIB_NOEXCEPT {
    return binary_writer::Write(this, x);
}

template<class T>
NLIB_ALWAYS_INLINE bool BinaryWriter::WriteArray(const T* x, size_t n) NLIB_NOEXCEPT {
    return binary_writer::WriteArray(this, x, n);
}

template<class T, size_t N>
NLIB_ALWAYS_INLINE bool BinaryWriter::WriteArray(const T (&a)[N]) NLIB_NOEXCEPT {
    return this->WriteArray(&a[0], N);
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_BINARYWRITER_H_
