﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_INPUTCONVERTERSTREAM_H_
#define INCLUDE_NN_NLIB_INPUTCONVERTERSTREAM_H_

#include "nn/nlib/InputStream.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC InputTransform {
    // define a derived class and use it from InputConverterStream
 public:
    virtual ~InputTransform() {}

    // is, p, n are not nullptr, and size > 0 when called.
    // p can be the buffer memory the user specified.
    // The data size to write can be less than 'size'.
    // *n becomes the size of written data.
    //
    // return 0 if it is all ok. otherwise, an error has occured.
    // *n == 0 if and only if EOF or error.
    virtual errno_t Transform(InputStream* is, void* p, size_t size, size_t* n) = 0;

    // called only once at first, and pass the workbuffer which OutputConverterStream uses.
    // The derived class has to manage the memory.
    virtual void* GetWorkBuffer(size_t* n) = 0;

    // called when InputConverterStream::SetStream() is called.
    virtual errno_t OnSetStream(InputStream* is) = 0;

    // called when the stream is closing.
    virtual errno_t OnClose() = 0;
};

class NLIB_VIS_PUBLIC InputConverterStream : public InputStream {
 public:
    NLIB_CEXPR InputConverterStream() NLIB_NOEXCEPT : InputStream(true, false),
                                                      stream_(nullptr),
                                                      cvtr_(nullptr),
                                                      ics_bufsize_(0),
                                                      ics_buf_(nullptr) {}
    errno_t SetStream(InputStream* istr) NLIB_NOEXCEPT;
    InputStream* GetStream() const NLIB_NOEXCEPT { return stream_; }

    // NOTE:
    // The destructor does not close its stream_, nor call InputTransform::OnClose().
    // it is because OutputConverterStream does not own stream_ nor cvtr_.
    virtual ~InputConverterStream() NLIB_NOEXCEPT NLIB_OVERRIDE {}

 protected:
    // Note that SetTransform returns ENOMEM when cvtr->GetWorkBuffer() fails
    errno_t SetTransform(InputTransform* cvtr) NLIB_NOEXCEPT;

 private:
    virtual size_t FillBuffer_(void* p, size_t nbytes) NLIB_NOEXCEPT NLIB_OVERRIDE NLIB_FINAL;
    virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE NLIB_FINAL;

 private:
    InputStream* stream_;
    InputTransform* cvtr_;
    size_t ics_bufsize_;
    unsigned char* ics_buf_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(InputConverterStream);
};

template<class TRANSFORM>
class InputConverterStreamTempl : public InputConverterStream {
 public:
    typedef TRANSFORM TransformType;

    InputConverterStreamTempl() NLIB_NOEXCEPT { this->SetTransform(&m_Transform); }
    virtual ~InputConverterStreamTempl() NLIB_NOEXCEPT NLIB_OVERRIDE { m_Transform.OnClose(); }

 private:
    void SetTransform(InputTransform* cvtr) NLIB_NOEXCEPT {
        // to hide SetTransform()
        errno_t e = InputConverterStream::SetTransform(cvtr);
        if (e != 0) this->SetError(e);
    }

 protected:
    TRANSFORM m_Transform;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(InputConverterStreamTempl);
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_INPUTCONVERTERSTREAM_H_
