﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_INPUTSTREAM_H_
#define INCLUDE_NN_NLIB_INPUTSTREAM_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

// code snippets:
//   InputStream& stream = 'stream that inherits InputStream';
//   if (!stream.Read(dataptr, nbyte)) { error = stream.GetErrorValue(); ..... }
//   while ((c = stream.Read()) >= 0) { c is [0x00 - 0xFF] }
//   if (!stream) { error = stream.GetErrorValue(); ..... }
class NLIB_VIS_PUBLIC InputStream {
 public:
    NLIB_CEXPR InputStream() NLIB_NOEXCEPT : pos_(0),
                                             is_buf_curidx_(0),
                                             is_buf_endidx_(0),
                                             is_buf_markidx_(static_cast<size_t>(-1)),
                                             is_buf_(NULL),
                                             is_bufsize_(0),
                                             is_buf_readonly_(false),
                                             is_mark_supported_(false),
                                             errno_(0) {}
    NLIB_CEXPR InputStream(bool is_mark_supported, bool is_buf_readonly) :
        pos_(0),
        is_buf_curidx_(0),
        is_buf_endidx_(0),
        is_buf_markidx_(static_cast<size_t>(-1)),
        is_buf_(NULL),
        is_bufsize_(0),
        is_buf_readonly_(is_buf_readonly),
        is_mark_supported_(is_mark_supported),
        errno_(0) {}
    virtual ~InputStream() NLIB_NOEXCEPT {
        // NOTE:
        // cannot call Close_() from this destructor.
        // the destructor of the derived class has to call Close_() if needed.
    }
    // returns true if there has been no errors
    // you can also write the codes such that 'if (!!stream) { ... }'
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }
    // use Pos64() if you need a 64bit value on 32bit arch
    size_t Pos() const NLIB_NOEXCEPT { return static_cast<size_t>(Pos64()); }
    uint64_t Pos64() const NLIB_NOEXCEPT { return pos_ + is_buf_curidx_; }
    bool IsEos() NLIB_NOEXCEPT {
        if (*this && is_buf_curidx_ == is_buf_endidx_) {
            CheckBuffer_();
            return *this && is_buf_curidx_ == is_buf_endidx_;
        }
        return false;
    }
    // < 0 if EOS or error, otherwise returns a byte and ++pos
    int Read() NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(is_buf_curidx_ == is_buf_endidx_)) {
            CheckBuffer_();
            if (NLIB_UNLIKELY(is_buf_curidx_ == is_buf_endidx_)) return -1;
        }
        return is_buf_[is_buf_curidx_++];
    }
    // < 0 if EOS or error, otherwise returns a byte and pos is unchanged
    NLIB_CHECK_RESULT int Peek() NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(is_buf_curidx_ == is_buf_endidx_)) {
            CheckBuffer_();
            if (NLIB_UNLIKELY(is_buf_curidx_ == is_buf_endidx_)) return -1;
        }
        return is_buf_[is_buf_curidx_];
    }
    // returns skipped number of bytes
    size_t Skip(size_t nbytes) NLIB_NOEXCEPT;
    size_t Read(void* ptr, size_t nbytes) NLIB_NOEXCEPT {
#ifndef NLIB_NONNULL_ENABLED
        if (!ptr) {
            this->SetError(EINVAL);
            return 0;
        }
#endif
        if (nbytes > RSIZE_MAX) {  // INT01-C
            this->SetError(EINVAL);
            return 0;
        }
        if (is_buf_curidx_ + nbytes <= is_buf_endidx_) {
            nlib_memcpy(ptr, nbytes, &is_buf_[is_buf_curidx_], nbytes);
            is_buf_curidx_ += nbytes;
            return nbytes;
        }
        return this->Read_(ptr, nbytes);
    }
    template<size_t N>
    size_t Read(nlib_byte_t (&buf)[N]) NLIB_NOEXCEPT {
        return Read(&buf[0], N);
    }
    bool Close() NLIB_NOEXCEPT;
    bool Mark(size_t readlimit) NLIB_NOEXCEPT;
    bool GoBackToMark() NLIB_NOEXCEPT;
    bool IsMarkSupported() const NLIB_NOEXCEPT { return is_mark_supported_; }
    NLIB_SAFE_BOOL(InputStream, errno_ == 0)

 protected:
    void
    SetBuffer(void* p, size_t nbytes, bool is_mark_supported, bool is_buf_readonly) NLIB_NOEXCEPT {
        NLIB_ASSERT(is_buf_readonly || !is_mark_supported);
        NLIB_ASSERT(p);
        NLIB_ASSERT(!is_mark_supported_ || is_buf_readonly_ || nbytes >= 1024);
        is_buf_            = static_cast<unsigned char*>(p);
        is_bufsize_        = static_cast<int>(nbytes);
        is_mark_supported_ = is_mark_supported;
        is_buf_readonly_   = is_buf_readonly;
    }
    void SetBuffer(void* p, size_t nbytes) NLIB_NOEXCEPT {
        NLIB_ASSERT(p);
        NLIB_ASSERT(!is_mark_supported_ || is_buf_readonly_ || nbytes >= 1024);
        is_buf_     = static_cast<unsigned char*>(p);
        is_bufsize_ = static_cast<int>(nbytes);
    }
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }

 private:
    virtual size_t FillBuffer_(void* p, size_t nbytes) NLIB_NOEXCEPT = 0;
    virtual bool Close_() NLIB_NOEXCEPT                              = 0;
    // By default, Skip_() reads data on is_buf_.
    // you can override it if you can implement more efficient code.
    virtual size_t Skip_(size_t nbytes) NLIB_NOEXCEPT;

 private:
    size_t Read_(void* ptr, size_t nbytes) NLIB_NOEXCEPT;
    void CheckBuffer_() NLIB_NOEXCEPT;

 private:
    uint64_t pos_;
    size_t is_buf_curidx_;
    size_t is_buf_endidx_;
    size_t is_buf_markidx_;
    unsigned char* is_buf_;
    size_t is_bufsize_;
    bool is_buf_readonly_;
    bool is_mark_supported_;
    mutable ErrnoT errno_;
    // errno_ is changed only from derived classes.
    // FillBuffer_() or Close_() changes errno_ in other words.

    NLIB_DISALLOW_COPY_AND_ASSIGN(InputStream);
};

class NLIB_VIS_PUBLIC NullInputStream NLIB_FINAL : public InputStream {
 public:
    NullInputStream() NLIB_NOEXCEPT;
    virtual ~NullInputStream() NLIB_NOEXCEPT {}

 private:
    char dummy_buf_[256];
    virtual size_t FillBuffer_(void* p, size_t nbytes) NLIB_NOEXCEPT NLIB_OVERRIDE;
    virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE { return true; }
    virtual size_t Skip_(size_t nbytes) NLIB_NOEXCEPT NLIB_OVERRIDE;
};

namespace detail {

template<size_t N>
class MiniBufIn {
 public:
    NLIB_STATIC_ASSERT(N >= 16);
    static NLIB_CEXPR const size_t buf_size = N;
    NLIB_ALWAYS_INLINE MiniBufIn() NLIB_NOEXCEPT { cur_ = end_ = &buf_[0]; }
    template<class IS>
    NLIB_ALWAYS_INLINE bool Prefetch(IS* stream, size_t n) NLIB_NOEXCEPT {
        if (cur_ + n > end_) {
            NLIB_ASSERT(n <= N);
            size_t rem = end_ - cur_;
            nlib_memmove(static_cast<void*>(&buf_[0]), N, cur_, rem);
            size_t nread = stream->Read(&buf_[rem], N - rem);
            cur_         = &buf_[0];
            end_         = &buf_[0] + rem + nread;
            return (nread + rem >= n);
        }
        return true;
    }
    template<class IS>
    NLIB_ALWAYS_INLINE int Peek(IS* stream) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(cur_ == end_)) {
            size_t nread = stream->Read(&buf_[0], N);
            if (NLIB_UNLIKELY(nread == 0)) return -1;
            cur_ = &buf_[0];
            end_ = &buf_[0] + nread;
        }
        return *cur_;
    }
    template<class IS>
    NLIB_ALWAYS_INLINE int Read(IS* stream) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(cur_ == end_)) {
            size_t nread = stream->Read(&buf_[0], N);
            if (NLIB_UNLIKELY(nread == 0)) return -1;
            cur_ = &buf_[0] + 1;
            end_ = &buf_[0] + nread;
            return buf_[0];
        } else {
            int c = *cur_;
            ++cur_;
            return c;
        }
    }

    NLIB_ALWAYS_INLINE void Advance(size_t n) NLIB_NOEXCEPT {
        // Prefetch(stream, n) must be successful beforehand
        cur_ += n;
    }
    NLIB_ALWAYS_INLINE const uint8_t& operator[](size_t n) NLIB_NOEXCEPT {
        // Prefetch(stream, n + 1) must be successful beforehand
        return cur_[n];
    }
    template<class IS>
    NLIB_ALWAYS_INLINE size_t ReadBytes(IS* stream, void* data, size_t nbytes) NLIB_NOEXCEPT {
        if (cur_ + nbytes <= end_) {
            nlib_memcpy(data, nbytes, cur_, nbytes);
            cur_ += nbytes;
            return nbytes;
        }
        uint8_t* p = static_cast<uint8_t*>(data);
        size_t rem = end_ - cur_;
        nlib_memcpy(static_cast<void*>(p), nbytes, cur_, rem);
        cur_ = end_  = &buf_[0];
        size_t nread = stream->Read(p + rem, nbytes - rem);
        return nread + rem;
    }
    template<class IS, class T>
    size_t Read(IS* stream, T* data, size_t count) NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(*data) <= N);
        NLIB_STATIC_ASSERT(IsPod<T>::value);
        size_t nbytes = sizeof(*data) * count;
        size_t nread  = ReadBytes(stream, data, nbytes);
        if (NLIB_LIKELY(nread == nbytes)) return count;
        size_t mod = nread % sizeof(*data);
        uint8_t* p = static_cast<uint8_t*>(static_cast<void*>(data));
        nlib_memcpy(static_cast<void*>(&buf_[0]), N, p + nread - mod, mod);
        end_ = &buf_[mod];
        return nread / sizeof(*data);
    }
    NLIB_ALWAYS_INLINE size_t size() const NLIB_NOEXCEPT { return end_ - cur_; }

 private:
    uint8_t* cur_;
    uint8_t* end_;
    uint8_t buf_[N];
};

}  // namespace detail
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_INPUTSTREAM_H_
