﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_NLIST_H_
#define INCLUDE_NN_NLIB_NLIST_H_

#include <string.h>
#include <utility>
#include <algorithm>
#include <iterator>
#include <memory>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

template<class T, class AL>
class Nlist;

namespace nlist {

class NLIB_VIS_PUBLIC NlistBase {
 protected:
    struct Blk {
        Blk* next;
        void* item;
    };

 protected:
    NLIB_CEXPR NlistBase() NLIB_NOEXCEPT : curblk_(nullptr),
                                           curlevel_(0),
                                           curidx_(0),
                                           firstblk_(nullptr) {}
#ifdef __cpp_rvalue_references
    NlistBase(NlistBase&& rhs) NLIB_NOEXCEPT : curblk_(rhs.curblk_),
                                               curlevel_(rhs.curlevel_),
                                               curidx_(rhs.curidx_),
                                               firstblk_(rhs.firstblk_) {
        rhs.curblk_   = nullptr;
        rhs.curlevel_ = 0U;
        rhs.curidx_   = 0U;
        rhs.firstblk_ = nullptr;
    }
    NlistBase& operator=(NlistBase&& rhs) NLIB_NOEXCEPT {
        curblk_       = rhs.curblk_;
        curlevel_     = rhs.curlevel_;
        curidx_       = rhs.curidx_;
        firstblk_     = rhs.firstblk_;
        rhs.curblk_   = nullptr;
        rhs.curlevel_ = 0U;
        rhs.curidx_   = 0U;
        rhs.firstblk_ = nullptr;
        return *this;
    }
#endif
    NlistBase(NlistBase& rhs, move_tag) NLIB_NOEXCEPT : curblk_(rhs.curblk_),
                                                        curlevel_(rhs.curlevel_),
                                                        curidx_(rhs.curidx_),
                                                        firstblk_(rhs.firstblk_) {
        rhs.curblk_   = nullptr;
        rhs.curlevel_ = 0U;
        rhs.curidx_   = 0U;
        rhs.firstblk_ = nullptr;
    }
    NlistBase& assign(NlistBase& rhs, move_tag) NLIB_NOEXCEPT {
        curblk_       = rhs.curblk_;
        curlevel_     = rhs.curlevel_;
        curidx_       = rhs.curidx_;
        firstblk_     = rhs.firstblk_;
        rhs.curblk_   = nullptr;
        rhs.curlevel_ = 0U;
        rhs.curidx_   = 0U;
        rhs.firstblk_ = nullptr;
        return *this;
    }
    static size_t BlkBaseIdx(size_t level) NLIB_NOEXCEPT { return 8 * ((1 << level) - 1); }
    static size_t BlkSize(size_t level) NLIB_NOEXCEPT { return 8 * (1 << level); }
    Blk* LastBlock(size_t* level) const NLIB_NOEXCEPT;
    bool ConfirmBack() const NLIB_NOEXCEPT;
    void ClearSimple() NLIB_NOEXCEPT {
        curlevel_ = 0;
        curidx_   = 0;
        curblk_   = firstblk_;
    }
    const Blk* AtSub(size_t n, size_t* base) const NLIB_NOEXCEPT;

 protected:
    mutable Blk* curblk_;
    mutable size_t curlevel_;
    mutable size_t curidx_;
    mutable Blk* firstblk_;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(NlistBase);
};

template<class T>
class NlistConstIterator;
template<class T>
class NlistIterator;

template<class T>
class NlistBaseT : public NlistBase {
 public:
    typedef T& reference;
    typedef const T& const_reference;
    typedef NlistIterator<T> iterator;
    typedef NlistConstIterator<T> const_iterator;
    typedef size_t size_type;
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef ptrdiff_t difference_type;

 protected:
    NLIB_CEXPR NlistBaseT() {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    NlistBaseT(NlistBaseT&& rhs) = default;
    NlistBaseT& operator=(NlistBaseT&& rhs) = default;
#else
    NlistBaseT(NlistBaseT&& rhs) NLIB_NOEXCEPT : NlistBase(std::move(rhs)) {}
    NlistBaseT& operator=(NlistBaseT&& rhs) NLIB_NOEXCEPT {
        NlistBase::operator=(std::move(rhs));
        return *this;
    }
#endif
#endif
    NlistBaseT(NlistBaseT& rhs, move_tag) NLIB_NOEXCEPT : NlistBase(rhs, move_tag()) {}
    NlistBaseT& assign(NlistBaseT& rhs, move_tag) NLIB_NOEXCEPT {
        NlistBase::assign(rhs, move_tag());
        return *this;
    }

    static T* GetItem(Blk* blk) NLIB_NOEXCEPT { return static_cast<T*>(blk->item); }
    static const T* GetItem(const Blk* blk) NLIB_NOEXCEPT {
        return static_cast<const T*>(blk->item);
    }
    T* At(size_type n) const NLIB_NOEXCEPT {
        size_type base;
        const Blk* p = this->AtSub(n, &base);
        return p ? const_cast<T*>(&(GetItem(p)[n - base])) : nullptr;
    }
    iterator Begin() NLIB_NOEXCEPT { return iterator(this, TrueType()); }
    iterator End() NLIB_NOEXCEPT { return iterator(this, FalseType()); }
    const_iterator Begin() const NLIB_NOEXCEPT { return const_iterator(this, TrueType()); }
    const_iterator End() const NLIB_NOEXCEPT { return const_iterator(this, FalseType()); }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(NlistBaseT);
    friend class NlistConstIterator<T>;
    friend class NlistIterator<T>;
};

template<class T>
class NlistConstIterator : public std::iterator<std::forward_iterator_tag, T> {
 private:
    typedef NlistBaseT<T> ContainerType;
    typedef typename ContainerType::Blk Blk;

 public:
    typedef typename std::iterator<std::forward_iterator_tag, T> BaseType;
    typedef typename BaseType::iterator_category iterator_category;
    typedef typename BaseType::value_type value_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::reference reference;

 public:
    NLIB_CEXPR NlistConstIterator() NLIB_NOEXCEPT : idx_(0),
                                                    elemptr_(nullptr),
                                                    blkend_(nullptr),
                                                    blk_(nullptr) {}

 protected:
    NlistConstIterator(const ContainerType* p, TrueType) NLIB_NOEXCEPT {
        ContainerType* rhs = const_cast<ContainerType*>(p);
        idx_               = 0;
        blk_               = rhs->firstblk_;
        elemptr_           = GetItem(blk_);
        blkend_            = elemptr_ + 8;
    }
    NlistConstIterator(const ContainerType* p, FalseType) NLIB_NOEXCEPT {
        ContainerType* rhs = const_cast<ContainerType*>(p);
        idx_               = rhs->BlkBaseIdx(rhs->curlevel_) + rhs->curidx_;
        blk_               = rhs->curblk_;
        elemptr_           = &(GetItem(blk_)[rhs->curidx_]);
        blkend_            = GetItem(blk_) + rhs->BlkSize(rhs->curlevel_);
        if (elemptr_ == blkend_) this->Normalize();
    }

 public:
    const T& operator*() const NLIB_NOEXCEPT { return *elemptr_; }
    const T* operator->() const NLIB_NOEXCEPT { return elemptr_; }
    NlistConstIterator& operator++() NLIB_NOEXCEPT {
        NLIB_ASSERT(elemptr_ != blkend_);
        ++idx_;
        if (++elemptr_ == blkend_) this->Normalize();
        return *this;
    }
    NlistConstIterator operator++(int) NLIB_NOEXCEPT {
        NlistConstIterator rval(*this);
        this->operator++();
        return rval;
    }
    bool operator==(const NlistConstIterator& rhs) const NLIB_NOEXCEPT {
        return idx_ == rhs.idx_ && elemptr_ == rhs.elemptr_;
    }
    bool operator!=(const NlistConstIterator& rhs) const NLIB_NOEXCEPT { return !operator==(rhs); }
    void Advance(difference_type n) NLIB_NOEXCEPT;
    difference_type Distance(NlistConstIterator to) const NLIB_NOEXCEPT { return to.idx_ - idx_; }

 private:
    static T* GetItem(Blk* blk) NLIB_NOEXCEPT { return static_cast<T*>(blk->item); }
    static const T* GetItem(const Blk* blk) NLIB_NOEXCEPT {
        return reinterpret_cast<const T*>(blk->item);
    }
    bool Normalize() NLIB_NOEXCEPT;

 private:
    size_t idx_;
    T* elemptr_;
    T* blkend_;
    Blk* blk_;
    friend class NlistBaseT<T>;
};

template<class T>
bool NlistConstIterator<T>::Normalize() NLIB_NOEXCEPT {
    if (blk_ && blk_->next) {
        Blk* next        = blk_->next;
        size_t next_size = 2 * static_cast<size_t>(blkend_ - GetItem(blk_));
        T* next_elemptr  = GetItem(next);
        T* next_end      = next_elemptr + next_size;
        blk_             = next;
        elemptr_         = next_elemptr;
        blkend_          = next_end;
        return true;
    }
    return false;
}

template<class T>
void NlistConstIterator<T>::Advance(difference_type n) NLIB_NOEXCEPT {
    while (elemptr_ + n >= blkend_) {
        n -= blkend_ - elemptr_;
        idx_ += blkend_ - elemptr_;
        elemptr_ = blkend_;
        if (!this->Normalize()) return;
    }
    elemptr_ += n;
    idx_ += n;
}

template<class T>
class NlistIterator : public NlistConstIterator<T> {
 private:
    typedef NlistBaseT<T> ContainerType;
    typedef typename ContainerType::Blk Blk;
    typedef NlistConstIterator<T> MyBase;

 public:
    typedef typename NlistConstIterator<T>::BaseType BaseType;
    typedef typename BaseType::iterator_category iterator_category;
    typedef typename BaseType::value_type value_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::reference reference;

 public:
    NLIB_CEXPR NlistIterator() NLIB_NOEXCEPT : MyBase() {}

 private:
    NlistIterator(const ContainerType* p, TrueType) NLIB_NOEXCEPT : MyBase(p, TrueType()) {}
    NlistIterator(const ContainerType* p, FalseType) NLIB_NOEXCEPT : MyBase(p, FalseType()) {}

 public:
    T& operator*() const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        return const_cast<T&>(**tmp);
    }
    T* operator->() const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        return const_cast<T*>(tmp->operator->());
    }
    NlistIterator& operator++() NLIB_NOEXCEPT {
        MyBase* tmp = static_cast<MyBase*>(this);
        ++(*tmp);
        return *this;
    }
    NlistIterator operator++(int) NLIB_NOEXCEPT {
        NlistIterator rval(*this);
        ++(*this);
        return rval;
    }
    bool operator==(const NlistIterator& rhs) const NLIB_NOEXCEPT {
        const MyBase* tmp  = static_cast<const MyBase*>(this);
        const MyBase* tmp2 = static_cast<const MyBase*>(&rhs);
        return *tmp == *tmp2;
    }
    bool operator!=(const NlistIterator& rhs) const NLIB_NOEXCEPT { return !operator==(rhs); }
    difference_type Distance(NlistIterator to) const NLIB_NOEXCEPT {
        const MyBase* tmp  = static_cast<const MyBase*>(this);
        const MyBase* tmp2 = static_cast<const MyBase*>(&to);
        return tmp->Distance(*tmp2);
    }

 private:
    friend class NlistBaseT<T>;
};

template<bool is_empty, class AL>
class NlistAlloc {
 public:
    NlistAlloc() NLIB_NOEXCEPT : alloc_(AL()) {}
    explicit NlistAlloc(const AL& al) NLIB_NOEXCEPT : alloc_(al) {}

 protected:
    AL& _GetNlistAlloc() const NLIB_NOEXCEPT { return const_cast<AL&>(alloc_); }
    void _SwapNlistAlloc(NlistAlloc& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(alloc_, rhs.alloc_);
    }
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    NlistAlloc(NlistAlloc&& rhs) = default;
    NlistAlloc& operator=(NlistAlloc&& rhs) = default;
#else
    NlistAlloc(NlistAlloc&& rhs) NLIB_NOEXCEPT : alloc_(std::move(rhs.alloc_)) {}
    NlistAlloc& operator=(NlistAlloc&& rhs) NLIB_NOEXCEPT {
        alloc_ = std::move(rhs.alloc_);
        return *this;
    }
#endif
#endif
    NlistAlloc(NlistAlloc& rhs, move_tag) NLIB_NOEXCEPT : alloc_() {
        using std::swap;
        swap(alloc_, rhs.alloc_);
    }
    NlistAlloc& assign(NlistAlloc& rhs, move_tag) NLIB_NOEXCEPT {
        using std::swap;
        AL x(*this, move_tag());
        swap(alloc_, rhs.alloc_);
        return *this;
    }

 private:
    AL alloc_;
};

template<class AL>
class NlistAlloc<true, AL> {
 public:
    NlistAlloc() NLIB_NOEXCEPT {}
    explicit NlistAlloc(const AL& al) NLIB_NOEXCEPT { NLIB_UNUSED(al); }

 protected:
    AL _GetNlistAlloc() const NLIB_NOEXCEPT { return AL(); }
    void _SwapNlistAlloc(NlistAlloc& rhs) NLIB_NOEXCEPT { NLIB_UNUSED(rhs); }
#ifdef __cpp_rvalue_references
    NlistAlloc(NlistAlloc&&) NLIB_NOEXCEPT {}
    NlistAlloc& operator=(NlistAlloc&&) NLIB_NOEXCEPT { return *this; }
#endif
    NlistAlloc(NlistAlloc&, move_tag) NLIB_NOEXCEPT {}
    NlistAlloc& assign(NlistAlloc&, move_tag) NLIB_NOEXCEPT { return *this; }
};

}  // namespace nlist

template<class T, class AL = std::allocator<char> >
class Nlist NLIB_FINAL : public nlist::NlistBaseT<T>,
                         public nlist::NlistAlloc<IsEmpty<AL>::value, AL> {
    // NOTE:
    // To develop custom allocators:
    //
    // Nlist uses only the subset of std::allocator<T>.
    //  1) rebind() is not used. the custom allocator does not have to be a class template.
    //  2) construct(), destroy() are not used.
    //
    // You have to implement:
    //  a) void* allocate(size_t nbytes);
    //  b) void deallocate(void* ptr, size_t nbytes);
    // allocate has to return 8 bytes aligned memory.
    // And the custom allocator cannot have non-static data members.
    /*
    class MyAllocator
    {
    public:
        MyAllocator();
        MyAllocator(const MyAllocator& rhs);
        void* allocate(size_t nbytes, void*);
        void deallocate(void* p, size_t nbytes);
    };
    */

 private:
    typedef typename nlist::NlistBaseT<T> BaseType;
    typedef typename nlist::NlistAlloc<IsEmpty<AL>::value, AL> AllocType;
    using BaseType::curblk_;
    using BaseType::curidx_;
    using BaseType::curlevel_;
    using BaseType::firstblk_;
    using BaseType::ConfirmBack;
    using BaseType::ClearSimple;
    using BaseType::BlkBaseIdx;
    using BaseType::BlkSize;
    using BaseType::GetItem;
    using BaseType::Begin;
    using BaseType::End;
    typedef typename BaseType::Blk Blk;

 public:
    typedef typename BaseType::value_type value_type;
    typedef AL allocator_type;
    typedef typename BaseType::size_type size_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::reference reference;
    typedef typename BaseType::reference const_reference;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::const_pointer const_pointer;
    typedef typename BaseType::iterator iterator;
    typedef typename BaseType::const_iterator const_iterator;

 public:
    Nlist() NLIB_NOEXCEPT : AllocType(AL()) {}
    explicit Nlist(const AL& al) NLIB_NOEXCEPT : AllocType(al) {}
    ~Nlist() NLIB_NOEXCEPT;
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Nlist(Nlist&& rhs) = default;
#else
    Nlist(Nlist&& rhs) NLIB_NOEXCEPT : BaseType(std::move(rhs)), AllocType(std::move(rhs)) {}
#endif
    Nlist& operator=(Nlist&& rhs) NLIB_NOEXCEPT {
        this->clear();
        this->DeallocBlkList(firstblk_, 0);
        BaseType::operator =(std::move(rhs));
        AllocType::operator=(std::move(rhs));
        return *this;
    }
#endif
    Nlist(Nlist& rhs, move_tag) NLIB_NOEXCEPT : BaseType(rhs, move_tag()),
                                                AllocType(rhs, move_tag()) {}
    Nlist& assign(Nlist& rhs, move_tag) NLIB_NOEXCEPT {
        this->clear();
        this->DeallocBlkList(firstblk_, 0);
        BaseType::assign(rhs, move_tag());
        AllocType::assign(rhs, move_tag());
        return *this;
    }
    size_type size() const NLIB_NOEXCEPT { return BlkBaseIdx(curlevel_) + curidx_; }
    size_type capacity() const NLIB_NOEXCEPT {
        size_type lv;
        return (nullptr != this->LastBlock(&lv)) ? BlkBaseIdx(lv + 1) : 0;
    }
    bool empty() const NLIB_NOEXCEPT { return curlevel_ == 0 && curidx_ == 0; }
    bool resize(size_type n) NLIB_NOEXCEPT {
        return Resize(n, typename IsTriviallyDefaultConstructible<T>::type());
    }
    bool reserve(size_type n) NLIB_NOEXCEPT;
    pointer push_back() {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = MyCtor(p, typename IsTriviallyDefaultConstructible<T>::type());  // may throw
        ++curidx_;
        return ptr;
    }
#ifdef __cpp_rvalue_references
    pointer push_back(T&& rhs) NLIB_NOEXCEPT {
#if defined(NLIB_HAS_NATIVE_TYPETRAITS) && !defined(NLIB_HAS_TR1_TYPETRAITS) && \
    (!defined(_MSC_VER) || _MSC_VER > 1800)
#ifdef __cpp_exceptions
        // with -fno-exceptions, MpObjectKv cannot compile
        NLIB_STATIC_ASSERT(std::is_nothrow_move_constructible<T>::value);
#endif
#endif
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(std::forward<T>(rhs));
        ++curidx_;
        return ptr;
    }
#endif
    pointer push_back(const T& rhs) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(rhs);  // may throw
        ++curidx_;
        return ptr;
    }
    bool pop_back() NLIB_NOEXCEPT {
        return this->pop_back_(typename IsTriviallyDestructible<T>::type());
    }
    void clear() NLIB_NOEXCEPT { Clear(typename IsTriviallyDestructible<T>::type()); }
    allocator_type get_allocator() const NLIB_NOEXCEPT { return AllocType::_GetNlistAlloc(); }

    reference back() {
        NLIB_ASSERT(!this->empty());
        if (curidx_ == 0) ConfirmBack();
        return GetItem(curblk_)[curidx_ - 1];
    }
    const_reference back() const {
        NLIB_ASSERT(!this->empty());
        if (curidx_ == 0) ConfirmBack();
        return GetItem(curblk_)[curidx_ - 1];
    }
    reference front() {
        NLIB_ASSERT(!this->empty());
        return GetItem(firstblk_)[0];
    }
    const_reference front() const {
        NLIB_ASSERT(!this->empty());
        return GetItem(firstblk_)[0];
    }
    reference operator[](size_type n) {
        NLIB_ASSERT(n < this->size());
        return *this->At(n);
    }
    const_reference operator[](size_type n) const {
        NLIB_ASSERT(n < this->size());
        return *this->At(n);
    }
    iterator begin() NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return iterator();
        return Begin();
    }
    const_iterator begin() const NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return const_iterator();
        return Begin();
    }
    const_iterator cbegin() const NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return const_iterator();
        return Begin();
    }

    iterator end() NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return iterator();
        return End();
    }
    const_iterator end() const NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return const_iterator();
        return End();
    }
    const_iterator cend() const NLIB_NOEXCEPT {
        if (!firstblk_ && !ConfirmFirstBlk()) return const_iterator();
        return End();
    }

    void shrink_to_fit() NLIB_NOEXCEPT {
        if (!curblk_) return;
        Blk* p        = curblk_->next;
        curblk_->next = nullptr;
        DeallocBlkList(p, static_cast<int>(curlevel_ + 1));
    }
#if defined(__cpp_rvalue_references) && defined(__cpp_variadic_templates)
    template<class... Args>
    pointer EmplaceBack(Args&&... args) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(std::forward<Args>(args)...);  // may throw
        ++curidx_;
        return ptr;
    }
#else
    template<class A1>
    pointer EmplaceBack(const A1& a1) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1));
        ++curidx_;
        return ptr;
    }
    template<class A1, class A2>
    pointer EmplaceBack(const A1& a1, const A2& a2) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2));
        ++curidx_;
        return ptr;
    }
    template<class A1, class A2, class A3>
    pointer EmplaceBack(const A1& a1, const A2& a2, const A3& a3) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3));
        ++curidx_;
        return ptr;
    }
    template<class A1, class A2, class A3, class A4>
    pointer EmplaceBack(const A1& a1, const A2& a2, const A3& a3, const A4& a4) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3),
                                const_cast<typename RemoveConst<A4>::type&>(a4));
        ++curidx_;
        return ptr;
    }
    template<class A1, class A2, class A3, class A4, class A5>
    pointer EmplaceBack(const A1& a1, const A2& a2, const A3& a3, const A4& a4, const A5& a5) {
        void* p = this->PushBack();
        if (!p) return nullptr;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3),
                                const_cast<typename RemoveConst<A4>::type&>(a4),
                                const_cast<typename RemoveConst<A5>::type&>(a5));
        ++curidx_;
        return ptr;
    }
#endif
    void Clobber() NLIB_NOEXCEPT {
        curblk_   = nullptr;
        curlevel_ = 0;
        curidx_   = 0;
        firstblk_ = nullptr;
    }

 private:
    // The # of elements in the block increases as 8, 16, 32, 64, ....
    Blk* AllocBlk(size_t n) const NLIB_NOEXCEPT;
    void DeallocBlkList(Blk* p, int lv) const NLIB_NOEXCEPT;
    void DestroyBlk(Blk*, size_t, TrueType) const NLIB_NOEXCEPT {}
    void DestroyBlk(Blk* p, size_t n, FalseType) const NLIB_NOEXCEPT;
    void DestroyBlk(Blk* p, size_t n) const NLIB_NOEXCEPT {
        DestroyBlk(p, n, typename IsTriviallyDestructible<T>::type());
    }
    bool ConfirmFirstBlk() const NLIB_NOEXCEPT;
    void* PushBack() NLIB_NOEXCEPT;
    void Clear(FalseType) NLIB_NOEXCEPT;
    void Clear(TrueType) NLIB_NOEXCEPT { ClearSimple(); }
    bool Resize(size_t n, FalseType) NLIB_NOEXCEPT;
    bool Resize(size_t n, TrueType) NLIB_NOEXCEPT;
    T* MyCtor(void* p, TrueType) NLIB_NOEXCEPT {
        nlib_memset(p, 0, sizeof(T));
        return static_cast<T*>(p);
    }
    T* MyCtor(void* p, FalseType) {
        T* ptr = new (p) T;  // may throw
        return ptr;
    }
    bool pop_back_(FalseType) NLIB_NOEXCEPT {
        if (curidx_ == 0) {
            if (!ConfirmBack()) return false;
        }
        reference item = GetItem(curblk_)[--curidx_];
        NLIB_UNUSED(item);
        item.~T();
        return true;
    }
    bool pop_back_(TrueType) NLIB_NOEXCEPT {
        if (curidx_ == 0) {
            if (!ConfirmBack()) return false;
        }
        --curidx_;
        return true;
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Nlist);
};

template<class T, class AL>
inline Nlist<T, AL>::~Nlist() NLIB_NOEXCEPT {
    this->clear();
    this->DeallocBlkList(firstblk_, 0);
}

namespace nlist {

template<class T, class AL>
class ResizeRewinder {
 public:
    ResizeRewinder(Nlist<T, AL>& l) NLIB_NOEXCEPT : count_(0), list_(l) {}
    void PushBack(size_t n) {
        for (; count_ < n; ++count_) {
            list_.push_back();  // may throw in the default constructor of T
        }
        count_ = 0;
    }
    ~ResizeRewinder() NLIB_NOEXCEPT {
        for (; count_ > 0; --count_) {
            list_.pop_back();
        }
    }

 private:
    size_t count_;
    Nlist<T, AL>& list_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ResizeRewinder);
};

}  // namespace nlist

template<class T, class AL>
bool Nlist<T, AL>::Resize(size_t n, FalseType) NLIB_NOEXCEPT {
    // NOTE: This can be optimized.
    size_t sz = this->size();
    if (n == sz) return true;
    if (n < sz) {
        for (size_t i = sz; i > n; --i) {
            this->pop_back();
        }
    } else {
        if (!this->reserve(n)) return false;
        nlist::ResizeRewinder<T, AL> rewinder(*this);
        rewinder.PushBack(n);
    }
    return true;
}

template<class T, class AL>
bool Nlist<T, AL>::Resize(size_t n, TrueType) NLIB_NOEXCEPT {
    size_type sz = this->size();
    if (n == sz) return true;
    if (n < sz) {
        for (size_type i = sz; i > n; --i) {
            this->pop_back();
        }
    } else {
        if (!this->reserve(n)) return false;
        // NOTE:
        // Just set curlevel_, curblk_, curidx_
        // because we don't have to invoke the constuctor of T.
        size_t lower = this->BlkBaseIdx(curlevel_);
        size_t upper = this->BlkBaseIdx(curlevel_ + 1);
        while (upper <= n) {
            size_t curmax_idx = BlkSize(curlevel_);
            void* buf         = reinterpret_cast<nlib_byte_t*>(curblk_->item) + curidx_ * sizeof(T);
            size_t fill_size  = (curmax_idx - curidx_) * sizeof(T);
            nlib_memset(buf, 0, fill_size);
            ++curlevel_;
            curidx_ = 0;
            curblk_ = curblk_->next;
            lower   = upper;
            upper   = this->BlkBaseIdx(curlevel_ + 1);
        }
        size_t old_curidx = curidx_;
        curidx_           = static_cast<size_t>(n - lower);
        void* buf         = reinterpret_cast<nlib_byte_t*>(curblk_->item) + old_curidx * sizeof(T);
        size_t fill_size  = (curidx_ - old_curidx) * sizeof(T);
        nlib_memset(buf, 0, fill_size);
    }
    return true;
}

template<class T, class AL>
bool Nlist<T, AL>::reserve(size_type n) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(n == 0)) return true;
    if (!firstblk_ && !this->ConfirmFirstBlk()) return false;
    if (n < BlkBaseIdx(curlevel_ + 1)) return true;
    size_t lv;
    Blk* blk = this->LastBlock(&lv);
    for (;;) {
        if (n < BlkBaseIdx(lv + 1)) return true;
        ++lv;
        Blk* p = this->AllocBlk(BlkSize(lv));
        if (!p) return false;
        blk->next = p;
        blk       = p;
    }
}

template<class T, class AL>
typename Nlist<T, AL>::Blk* Nlist<T, AL>::AllocBlk(size_t n) const NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(NLIB_ALIGNOF(T) == 1 || NLIB_ALIGNOF(T) == 2 || NLIB_ALIGNOF(T) == 4 ||
                       NLIB_ALIGNOF(T) == 8);
    NLIB_STATIC_ASSERT(sizeof(Blk) % 8 == 0);
    size_t sz = sizeof(Blk) + n * sizeof(T);
    void* pp;
    NLIB_TRY {
        pp = AllocType::_GetNlistAlloc().allocate(sz);  // must be 8 bytes aligned
    }
#ifdef __cpp_exceptions
    NLIB_CATCH(const std::bad_alloc&) { return nullptr; }
#endif
    if (!pp) return nullptr;
    Blk* p  = static_cast<Blk*>(pp);
    p->next = nullptr;
    p->item = reinterpret_cast<T*>(p + 1);
    return p;
}

template<class T, class AL>
void Nlist<T, AL>::DeallocBlkList(Blk* p, int lv) const NLIB_NOEXCEPT {
    while (p) {
        size_t n = BlkSize(lv);
        Blk* pp  = p;
        p        = pp->next;

        // DeallocBlock
        AllocType::_GetNlistAlloc().deallocate(reinterpret_cast<char*>(pp),
                                               sizeof(Blk) + n * sizeof(T));
        ++lv;
    }
}

template<class T, class AL>
void Nlist<T, AL>::DestroyBlk(Blk* p, size_t n, FalseType) const NLIB_NOEXCEPT {
    // NOTE:
    // Do not use std::allocator::destroy() because destroy() is not required.
    // invoke the destructor directly.
    T* base = static_cast<T*>(p->item);
    NLIB_UNUSED(base);
    for (size_t i = 0; i < n; ++i) {
        (base + i)->~T();
    }
}

template<class T, class AL>
bool Nlist<T, AL>::ConfirmFirstBlk() const NLIB_NOEXCEPT {
    Blk* p = AllocBlk(BlkSize(0));
    if (!p) return false;
    firstblk_ = p;
    curblk_   = p;
    curidx_   = 0;
    curlevel_ = 0;
    return true;
}

template<class T, class AL>
void* Nlist<T, AL>::PushBack() NLIB_NOEXCEPT {
    if (!firstblk_ && !ConfirmFirstBlk()) return nullptr;
    if (curidx_ == BlkSize(curlevel_)) {
        if (!curblk_->next) {
            Blk* p = AllocBlk(BlkSize(curlevel_ + 1));
            if (!p) return nullptr;
            curblk_->next = p;
        }
        ++curlevel_;
        curblk_ = curblk_->next;
        curidx_ = 0;
    }

    // NOTE:
    // If curidx_ is incremented here,
    // The state will be inconsisitent when T's constructor throws an exception.
    T& item = GetItem(curblk_)[curidx_];
    return &item;
}

template<class T, class AL>
void Nlist<T, AL>::Clear(FalseType) NLIB_NOEXCEPT {
    int lv = 0;
    Blk* p = firstblk_;
    if (!p) return;
    while (p != curblk_) {
        size_t cnt = BlkSize(lv);
        DestroyBlk(p, cnt);
        ++lv;
        p = p->next;
    }
    if (p == curblk_) {
        DestroyBlk(p, curidx_);
    }
    ClearSimple();
}

template<class T, class AL1, class AL2>
inline bool operator==(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return lhs.size() == rhs.size() && std::equal(lhs.begin(), lhs.end(), rhs.begin());
}

template<class T, class AL1, class AL2>
inline bool operator!=(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}

template<class T, class AL1, class AL2>
inline bool operator<(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return std::lexicographical_compare(lhs.begin(), lhs.end(), rhs.begin(), rhs.end());
}

template<class T, class AL1, class AL2>
inline bool operator>(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (rhs < lhs);
}

template<class T, class AL1, class AL2>
inline bool operator<=(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (!(rhs < lhs));
}

template<class T, class AL1, class AL2>
inline bool operator>=(const Nlist<T, AL1>& lhs, const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (!(lhs < rhs));
}

template<class AL = std::allocator<char> >
class StringList {
    typedef Nlist<char*, AL> ContainerType;

 public:
    typedef AL allocator_type;
    typedef typename ContainerType::iterator iterator;
    typedef typename ContainerType::const_iterator const_iterator;

 public:
    ~StringList() NLIB_NOEXCEPT;
    size_t size() const NLIB_NOEXCEPT { return list_.size(); }
    size_t capacity() const NLIB_NOEXCEPT { return list_.capacity(); }
    bool empty() const NLIB_NOEXCEPT { return list_.empty(); }
    bool reserve(size_t n) { return list_.reserve(n); }
    bool resize(size_t n) { return list_.resize(n); }

    char** push_back(const char* str);
    template<class Iterator>
    char** push_back(Iterator first, Iterator last);

    bool pop_back() NLIB_NOEXCEPT;
    void clear() NLIB_NOEXCEPT { list_.clear(); }
    allocator_type get_allocator() const { return list_.get_allocator(); }

    const char* back() const { return list_.back(); }
    char* back() { return list_.back(); }
    const char* front() const { return list_.front(); }
    char* front() { return list_.front(); }
    const char* operator[](size_t idx) const { return list_[idx]; }
    char* operator[](size_t idx) { return list_[idx]; }
    iterator begin() NLIB_NOEXCEPT { return list_.begin(); }
    const_iterator begin() const NLIB_NOEXCEPT { return list_.begin(); }
    const_iterator cbegin() NLIB_NOEXCEPT { return list_.cbegin(); }

    iterator end() NLIB_NOEXCEPT { return list_.end(); }
    const_iterator end() const NLIB_NOEXCEPT { return list_.end(); }
    const_iterator cend() NLIB_NOEXCEPT { return list_.cend(); }

    void shrink_to_fit() { list_.shrink_to_fit(); }
    void Clobber() NLIB_NOEXCEPT { list_.clobber(); }

    char** Replace(iterator it, const char* str);
    char** Replace(size_t idx, const char* str) {
        if (idx >= list_.size()) return nullptr;
        iterator it = list_.begin();
        it.Advance(idx);
        return this->Replace(it, str);
    }

 private:
    void Dealloc(const char* ptr) const {
        if (ptr) {
            // NOTE: note that size info is not supplied.
            this->get_allocator().deallocate(const_cast<char*>(ptr), 0);
        }
    }
    char* StrDup(const char* ptr) const;
    char* Alloc(size_t nbytes) const { return this->get_allocator().allocate(nbytes); }

 private:
    ContainerType list_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(StringList);
};

template<class AL>
char* StringList<AL>::StrDup(const char* ptr) const {
    size_t nbytes = nlib_strlen(ptr) + 1;
    void* p       = this->Alloc(nbytes);
    if (!p) return nullptr;
    nlib_memcpy(p, nbytes, ptr, nbytes);
    return reinterpret_cast<char*>(p);
}

template<class AL>
StringList<AL>::~StringList() NLIB_NOEXCEPT {
    const_iterator it     = this->cbegin();
    const_iterator it_end = this->cend();
    while (it != it_end) {
        Dealloc(*it);
        ++it;
    }
}

template<class AL>
char** StringList<AL>::push_back(const char* str) {
    char* dup;
    if (str) {
        dup = this->StrDup(str);
        if (!dup) return nullptr;
    } else {
        dup = nullptr;
    }
    char** ptr_str = list_.push_back(dup);
    if (!ptr_str) this->Dealloc(dup);
    return ptr_str;
}

template<class AL>
template<class Iterator>
char** StringList<AL>::push_back(Iterator first, Iterator last) {
    typename std::iterator_traits<Iterator>::difference_type n = std::distance(first, last);
    char* str                                                  = this->Alloc(n + 1);
    if (!str) return nullptr;
    std::copy(first, last, str);
    str[n]      = '\0';
    char** rval = list_.push_back(str);
    if (!rval) this->Dealloc(str);
    return rval;
}

template<class AL>
bool StringList<AL>::pop_back() NLIB_NOEXCEPT {
    if (list_.empty()) return false;
    Dealloc(list_.back());
    return list_.pop_back();
}

template<class AL>
char** StringList<AL>::Replace(iterator it, const char* str) {
    char* dup;
    if (str) {
        dup = this->StrDup(str);
        if (!dup) return nullptr;
    } else {
        dup = nullptr;
    }
    Dealloc(*it);
    *it = dup;
    return &*it;
}

template<class AL>
bool AppendString(Nlist<char, AL>* obj, const char* str) {
    size_t n = nlib_strlen(str);
    if (!obj->reserve(obj->size() + n)) return false;
    while (*str) {
        obj->push_back(*str);
        ++str;
    }
    return true;
}

NLIB_NAMESPACE_END

namespace std {

template<class T, class Distance>
inline void advance(::nlib_ns::nlist::NlistIterator<T>& pos, Distance n) {
    pos.Advance(static_cast<size_t>(n));
}

template<class T>
inline std::ptrdiff_t distance(const ::nlib_ns::nlist::NlistIterator<T>& first,
                               const ::nlib_ns::nlist::NlistIterator<T>& last) {
    return first.Distance(last);
}

}  // namespace std

NLIB_DEFINE_STD_SWAP_T_BEGIN2(nn, nlib)
NLIB_DEFINE_STD_SWAP_T2(T, AL, NLIB_NS::Nlist)
NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::StringList)
NLIB_DEFINE_STD_SWAP_T_END2(nn, nlib)

#endif  // INCLUDE_NN_NLIB_NLIST_H_
