
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_STRINGHOLDER_H_
#define INCLUDE_NN_NLIB_STRINGHOLDER_H_

#include <iterator>
#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN

template<size_t N, class Ch = char>
class StringHolder NLIB_FINAL {
 public:
    NLIB_CEXPR14 StringHolder() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(Ch) == 1);
        NLIB_STATIC_ASSERT(sizeof(Ch*) <= 8);
        NLIB_STATIC_ASSERT(N >= 24);
        NLIB_STATIC_ASSERT(N % 8 == 0);
        str_.ptr        = nullptr;
        str_.sso[N - 1] = static_cast<Ch>(-1);
    }
    ~StringHolder() NLIB_NOEXCEPT {
        if (str_.sso[N - 1] && str_.ptr) nlib_free(str_.ptr);
    }
#ifdef __cpp_rvalue_references
    StringHolder(StringHolder&& rhs) NLIB_NOEXCEPT : str_(rhs.str_) {
        rhs.str_.ptr        = nullptr;
        rhs.str_.sso[N - 1] = static_cast<Ch>(-1);
    }
    StringHolder& operator=(StringHolder&& rhs) NLIB_NOEXCEPT {
        nlib_memcpy(&str_.sso[0], N, &rhs.str_.sso[0], N);
        rhs.str_.ptr        = nullptr;
        rhs.str_.sso[N - 1] = static_cast<Ch>(-1);
        return *this;
    }
#endif
    StringHolder(StringHolder& rhs, move_tag) NLIB_NOEXCEPT : str_(rhs.str_) {
        rhs.str_.ptr        = nullptr;
        rhs.str_.sso[N - 1] = static_cast<Ch>(-1);
    }
    StringHolder& assign(StringHolder& rhs, move_tag) NLIB_NOEXCEPT {
        nlib_memcpy(static_cast<void*>(&str_.sso[0]), N, static_cast<const void*>(&rhs.str_.sso[0]),
                    N);
        rhs.str_.ptr        = nullptr;
        rhs.str_.sso[N - 1] = static_cast<Ch>(-1);
        return *this;
    }
    // NOTE: add move ctor/assign
    errno_t Init(const Ch* first, const Ch* last) NLIB_NOEXCEPT;
    errno_t Init(const Ch* str) NLIB_NOEXCEPT {
        NLIB_ASSERT(str);
        return Init(str, str + nlib_strlen(str));
    }
    size_t length() const NLIB_NOEXCEPT {
        if (str_.sso[N - 1]) {
            return str_.ptr ? static_cast<size_t>(str_.len.val) : 0U;
        } else {
            return nlib_strlen(&str_.sso[0]);
        }
    }
    const Ch* c_str() const NLIB_NOEXCEPT { return str_.sso[N - 1] ? str_.ptr : &str_.sso[0]; }
    const Ch* get() const NLIB_NOEXCEPT {
        // for unique_ptr<char[]> compatible
        return c_str();
    }
    operator const Ch*() const NLIB_NOEXCEPT { return c_str(); }
    void reset() NLIB_NOEXCEPT {
        // for unique_ptr<char[]> compatible
        if (str_.sso[N - 1]) {
            if (str_.ptr) nlib_free(str_.ptr);
        } else {
            str_.sso[N - 1] = static_cast<Ch>(-1);
        }
        str_.ptr = nullptr;
    }

 private:
    union {
        Ch sso[N];
        Ch* ptr;
        struct lentype {
            uint64_t _;
            uint64_t val;
        } len;
    } str_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(StringHolder);
};

template<size_t N, class Ch>
inline errno_t StringHolder<N, Ch>::Init(const Ch* first, const Ch* last) NLIB_NOEXCEPT {
    NLIB_ASSERT(first);
    size_t n = std::distance(first, last);
    Ch* p;
    if (n + 1 > N) {
        p = static_cast<Ch*>(nlib_malloc(n + 1));
        if (!p) return ENOMEM;
        if (str_.sso[N - 1]) {
            if (str_.ptr) nlib_free(str_.ptr);
        } else {
            str_.sso[N - 1] = static_cast<Ch>(-1);
        }
        str_.ptr     = p;
        str_.len.val = n;
        // cast for CTR
        nlib_memcpy(static_cast<void*>(p), n + 1, static_cast<const void*>(first), n);
        p[n] = static_cast<Ch>('\0');
    } else {
        p = &str_.sso[0];
        if (str_.sso[N - 1]) {
            if (str_.ptr) nlib_free(str_.ptr);
            str_.sso[N - 1] = static_cast<Ch>('\0');
        }
        // cast for CTR
        nlib_memcpy(static_cast<void*>(p), N - 1, static_cast<const void*>(first), n);
        if (n != N - 1) {
            p[n] = static_cast<Ch>('\0');
        }
    }
    return 0;
}

template<class Ch>
class StringHolder<16, Ch> NLIB_FINAL {
 public:
    NLIB_CEXPR14 StringHolder() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(Ch) == 1);
        NLIB_STATIC_ASSERT(sizeof(Ch*) <= 8);
        str_.ptr     = nullptr;
        str_.sso[15] = static_cast<Ch>(-1);
    }
    ~StringHolder() NLIB_NOEXCEPT {
        if (str_.sso[15] && str_.ptr) nlib_free(str_.ptr);
    }
#ifdef __cpp_rvalue_references
    StringHolder(StringHolder&& rhs) NLIB_NOEXCEPT : str_(rhs.str_) {
        rhs.str_.ptr     = nullptr;
        rhs.str_.sso[15] = static_cast<Ch>(-1);
    }
    StringHolder& operator=(StringHolder&& rhs) NLIB_NOEXCEPT {
        nlib_memcpy(&str_.sso[0], 16, &rhs.str_.sso[0], 16);
        rhs.str_.ptr     = nullptr;
        rhs.str_.sso[15] = static_cast<Ch>(-1);
        return *this;
    }
#endif
    StringHolder(StringHolder& rhs, move_tag) NLIB_NOEXCEPT : str_(rhs.str_) {
        rhs.str_.ptr     = nullptr;
        rhs.str_.sso[15] = static_cast<Ch>(-1);
    }
    StringHolder& assign(StringHolder& rhs, move_tag) NLIB_NOEXCEPT {
        nlib_memcpy(static_cast<void*>(&str_.sso[0]), 16,
                    static_cast<const void*>(&rhs.str_.sso[0]), 16);
        rhs.str_.ptr     = nullptr;
        rhs.str_.sso[15] = static_cast<Ch>(-1);
        return *this;
    }
    // NOTE: add move ctor/assign
    errno_t Init(const Ch* first, const Ch* last) NLIB_NOEXCEPT;
    errno_t Init(const Ch* str) NLIB_NOEXCEPT {
        NLIB_ASSERT(str);
        return Init(str, str + nlib_strlen(str));
    }
    size_t length() const NLIB_NOEXCEPT {
        if (str_.sso[15]) {
            if (!str_.ptr) return 0U;
            int64_t len = str_.len.minus;
#ifndef NLIB_LITTLE_ENDIAN
            nlib_swapendian_64(reinterpret_cast<uint64_t*>(&len), 1);
#endif
            return static_cast<size_t>(-len);
        } else {
            return nlib_strlen(&str_.sso[0]);
        }
    }
    const Ch* c_str() const NLIB_NOEXCEPT { return str_.sso[15] ? str_.ptr : &str_.sso[0]; }
    const Ch* get() const NLIB_NOEXCEPT {
        // for unique_ptr<char[]> compatible
        return c_str();
    }
    operator const Ch*() const NLIB_NOEXCEPT { return c_str(); }
    void reset() NLIB_NOEXCEPT {
        // for unique_ptr<char[]> compatible
        if (str_.sso[15]) {
            if (str_.ptr) nlib_free(str_.ptr);
        } else {
            str_.sso[15] = static_cast<Ch>(-1);
        }
        str_.ptr = nullptr;
    }

 private:
    union {
        Ch sso[16];
        Ch* ptr;
        struct lentype {
            uint64_t _;
            int64_t minus;
        } len;
    } str_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(StringHolder);
};

template<class Ch>
inline errno_t StringHolder<16, Ch>::Init(const Ch* first, const Ch* last) NLIB_NOEXCEPT {
    NLIB_ASSERT(first);
    ptrdiff_t n = std::distance(first, last);
    Ch* p;
    if (n > 15) {
        p = static_cast<Ch*>(nlib_malloc(n + 1));
        if (!p) return ENOMEM;
        if (str_.sso[15]) {
            if (str_.ptr) nlib_free(str_.ptr);
        }
        str_.ptr       = p;
        str_.len.minus = -n;
#ifndef NLIB_LITTLE_ENDIAN
        nlib_swapendian_64(reinterpret_cast<uint64_t*>(&str_.len.minus), 1);
#endif
        // cast for CTR
        nlib_memcpy(static_cast<void*>(p), n + 1, static_cast<const void*>(first), n);
        p[n] = static_cast<Ch>('\0');
    } else {
        p = &str_.sso[0];
        if (str_.sso[15]) {
            if (str_.ptr) nlib_free(str_.ptr);
            str_.sso[15] = static_cast<Ch>('\0');
        }
        // cast for CTR
        nlib_memcpy(static_cast<void*>(p), 15, static_cast<const void*>(first), n);
        if (n != 15) {
            p[n] = static_cast<Ch>('\0');
        }
    }
    return 0;
}

template<class BYTE = nlib_byte_t, size_t N = 232>
class BytePool {
 public:
    BytePool() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(BYTE) == sizeof(nlib_byte_t));
        cur_         = &init_buf_[0];
        end_         = &init_buf_[0] + N;
        chunk_stack_ = nullptr;
    }
    ~BytePool() NLIB_NOEXCEPT { Reset(); }
    void Reset() NLIB_NOEXCEPT;
    BYTE* Alloc(size_t n) NLIB_NOEXCEPT;

 private:
    struct Chunk {
        Chunk* next;
        BYTE buf[1];  // hacked
    };
    BYTE* cur_;
    BYTE* end_;
    Chunk* chunk_stack_;
    BYTE init_buf_[N];
    NLIB_DISALLOW_COPY_AND_ASSIGN(BytePool);
};

template<class BYTE, size_t N>
inline void BytePool<BYTE, N>::Reset() NLIB_NOEXCEPT {
    Chunk* p = chunk_stack_;
    while (p) {
        Chunk* tmp = p->next;
        nlib_free(p);
        p = tmp;
    }
    cur_         = &init_buf_[0];
    end_         = &init_buf_[0] + N;
    chunk_stack_ = nullptr;
}

template<class BYTE, size_t N>
inline BYTE* BytePool<BYTE, N>::Alloc(size_t n) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(n > static_cast<size_t>(end_ - cur_))) {
        size_t real_n = (n + sizeof(Chunk*) + 4095) & ~4095;
        Chunk* chk    = static_cast<Chunk*>(nlib_malloc(real_n));
        if (NLIB_UNLIKELY(!chk)) return nullptr;
        chk->next    = chunk_stack_;
        chunk_stack_ = chk;
        cur_         = &chk->buf[0];
        end_         = reinterpret_cast<BYTE*>(chk) + real_n;
    }
    BYTE* p;
    p = cur_;
    cur_ += n;
    return p;
}

NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP_T_BEGIN2(nn, nlib)
#ifndef __cpp_rvalue_references
template<size_t N, class Ch>
NLIB_ALWAYS_INLINE void swap(nlib_ns::StringHolder<N, Ch>& lhs, nlib_ns::StringHolder<N, Ch>& rhs) {
    nlib_ns::StringHolder<N, Ch> x(lhs, nlib_ns::move_tag());
    lhs.assign(rhs, nlib_ns::move_tag());
    rhs.assign(x, nlib_ns::move_tag());
}
#endif
NLIB_DEFINE_STD_SWAP_T_END2(nn, nlib)

#endif  // INCLUDE_NN_NLIB_STRINGHOLDER_H_
