﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_
#define INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

class TaggedTextParser NLIB_FINAL {
 public:
    enum Event {
        kNone         = -1,
        kStartElement = 1,
        kEndElement,
        kCharacters,
        kEndDocument,
        kSyntaxError
    };
    explicit TaggedTextParser(char* tagged_text) {
        // change tagged_text in parsing
        text_             = tagged_text;
        cur_              = reinterpret_cast<unsigned char*>(tagged_text);
        chr_              = *tagged_text;
        *cur_++           = '\0';
        is_elemtag_empty_ = false;
    }
    const char* GetText() const {
        // tag name if kStartElement, kEndElement
        // trimmed string if kCharacters
        return text_;
    }
    Event Next() {
    NLIB_TAGGEDTEXTPARSER_NEXT:
        if (is_elemtag_empty_) {
            is_elemtag_empty_ = false;
            return kEndElement;
        }
        // SkipSpace
        if (IsSpc(chr_)) {
            for (;;) {
                if (!IsSpc(*cur_)) {
                    chr_ = *cur_++;
                    break;
                }
                ++cur_;
            }
        }
        if (chr_ == '\0') return kEndDocument;
        if (chr_ == '<') {
            unsigned char* p = cur_;
            int c            = *p;
            if (c == '?' || c == '!') {
                // skip until '>' if '<!' or '<?'.
                ++p;
                unsigned char* p2 = GotoGT(p);
                if (*p2 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                ++p2;  // skip '>'
                chr_ = *p2;
                cur_ = *p2 != '\0' ? p2 + 1 : p2;
                goto NLIB_TAGGEDTEXTPARSER_NEXT;
            } else if (c == '/') {
                // ETag
                ++p;  // move after '/'
                // skip until space, '/' or '>'.
                unsigned char* p2 = SpotName(p);
                // skip until '>', even if there are attributes.
                unsigned char* p3 = GotoGT(p2);
                if (*p3 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;

                *p2   = '\0';
                text_ = reinterpret_cast<const char*>(p);
                ++p3;  // skip after '>'
                chr_ = *p3;
                cur_ = *p3 != '\0' ? p3 + 1 : p3;
                return kEndElement;
            } else {
                // skip until space, '/' or '>'.
                unsigned char* p2 = SpotName(p);
                // skip until '/' or '>'.
                unsigned char* p3 = GotoSlashGT(p2);
                if (*p3 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                if (*p3 == '/') {  // determine '>' or '/>'
                    // EmptyElemTag
                    ++p3;  // move after '/'
                    if (*p3 != '>') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                    is_elemtag_empty_ = true;
                }
                // STag or EmptyElemTag
                *p2   = '\0';
                text_ = reinterpret_cast<const char*>(p);
                ++p3;  // move after '>'
                chr_ = *p3;
                cur_ = *p3 != '\0' ? p3 + 1 : p3;
                return kStartElement;
            }
        } else {
            // it must be kCharacters
            unsigned char* p = cur_ - 1;
            for (;;) {
                int c = *p;
                if (c == '<' || c == '\0') {
                    text_             = reinterpret_cast<const char*>(cur_ - 1);
                    chr_              = *p;
                    cur_              = (chr_ != '\0') ? p + 1 : p;
                    unsigned char* pp = p - 1;
                    while (IsSpc(*pp)) --pp;
                    *(pp + 1) = '\0';  // trim the last space chars
                    return kCharacters;
                }
                ++p;
            }
        }
    NLIB_TAGGEDTEXTPARSER_ERROR:
        chr_  = '\0';
        cur_  = nullptr;
        text_ = "";
        return kSyntaxError;
    }

 private:
    static bool IsSpc(int c) {
        // S = ( #x20 | #x9 | #xD | #xA )
        static const char table[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0,  // 0x00 - 0x0F
                                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  // 0x10 - 0x1F
                                     1};
        return (c <= 0x20 && table[c] != 0);
    }
    static unsigned char* SpotName(unsigned char* p) {
        static const char table[] = {
            1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0,  // 0x00 - 0x0F
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  // 0x10 - 0x1F
            1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,  // 0x20 - 0x2F
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0   // 0x30 - 0x3F
        };
        for (;;) {
            int c = *p;
            // (c == '/' || c == '>' || c == '\0' || IsSpc(c))
            if (c < 0x40 && table[c] != 0) return p;
            ++p;
        }
    }
    static unsigned char* GotoGT(unsigned char* p) {
        for (;;) {
            int c = *p;
            if (c == '>' || c == '\0') return p;
            ++p;
        }
    }
    static unsigned char* GotoSlashGT(unsigned char* p) {
        for (;;) {
            int c = *p;
            if (c == '/' || c == '>' || c == '\0') return p;
            ++p;
        }
    }

 private:
    const char* text_;
    unsigned char* cur_;
    int chr_;
    bool is_elemtag_empty_;
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_
