﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_URI_H_
#define INCLUDE_NN_NLIB_URI_H_

#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/Nlist.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN

class InputStream;
class OutputStream;
class MemoryOutputStream;

template<size_t N = 1024>
class UriQueryEncoder NLIB_FINAL {
 public:
    UriQueryEncoder() NLIB_NOEXCEPT { cur_ = &buf_[0]; }
    errno_t Append(const char* key, const char* value) NLIB_NOEXCEPT;
    const char* c_str() const NLIB_NOEXCEPT {
        *cur_ = '\0';
        return &buf_[0];
    }
    operator const char*() const NLIB_NOEXCEPT { return c_str(); }
    std::pair<const char*, const char*> GetData() const NLIB_NOEXCEPT {
        return std::make_pair(&buf_[0], cur_);
    }

 private:
    char* cur_;
    char buf_[N];
};

struct UriPrivate;
class NLIB_VIS_PUBLIC Uri NLIB_FINAL {
 public:
    static bool
    IsHostName(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsHostName(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsHostName(str, str + nlib_strlen(str));
    }
    static bool
    IsIpv4(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsIpv4(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsIpv4(str, str + nlib_strlen(str));
    }
    static bool
    IsIpv6(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsIpv6(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsIpv6(str, str + nlib_strlen(str));
    }
    static bool
    IsEmailAddress(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsEmailAddress(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsEmailAddress(str, str + nlib_strlen(str));
    }
    static bool IsUri(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsUri(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsUri(str, str + nlib_strlen(str));
    }
    static bool
    IsUriReference(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool IsUriReference(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsUriReference(str, str + nlib_strlen(str));
    }

    // reserved    = gen-delims / sub-delims
    // gen - delims = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    // sub - delims = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
    static bool IsReserved(int c) NLIB_NOEXCEPT;
    // unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
    static bool IsUnreserved(int c) NLIB_NOEXCEPT;
    // 0-9 / A-F / a-f
    static bool IsHexDigit(int c) NLIB_NOEXCEPT;

    static errno_t DecodeUriComponent(size_t* written, char* buf, size_t n, const char* first,
                                      const char* last) NLIB_NOEXCEPT;
    template<size_t N>
    static errno_t
    DecodeUriComponent(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, N, s);
    }
    static errno_t
    DecodeUriComponent(size_t* written, char* buf, size_t n, const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, n, s, s + nlib_strlen(s));
    }
    template<size_t N>
    static errno_t DecodeUriComponent(size_t* written, char (&buf)[N], const char* first,
                                      const char* last) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, N, first, last);
    }
    static std::pair<errno_t, size_t>
    DecodeUriComponent(char* buf, size_t n, const char* first, const char* last) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = DecodeUriComponent(&written, buf, n, first, last);
        return std::make_pair(e, written);
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    DecodeUriComponent(char (&buf)[N], const char* first, const char* last) NLIB_NOEXCEPT {
        return DecodeUriComponent(buf, N, first, last);
    }
    static std::pair<errno_t, size_t>
    DecodeUriComponent(char* buf, size_t n, const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(buf, n, s, s + nlib_strlen(s));
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    DecodeUriComponent(char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(buf, N, s, s + nlib_strlen(s));
    }
    static errno_t DecodePath(size_t* written, char* buf, size_t n, const char* s) NLIB_NOEXCEPT;
    template<size_t N>
    static errno_t DecodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodePath(written, buf, N, s);
    }
    static std::pair<errno_t, size_t> DecodePath(char* buf, size_t n, const char* s) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = DecodePath(&written, buf, n, s);
        return std::make_pair(e, written);
    }
    template<size_t N>
    static std::pair<errno_t, size_t> DecodePath(char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodePath(buf, N, s);
    }

    static errno_t EncodeUriComponent(size_t* written, char* buf, size_t n, const char* first,
                                      const char* last, bool fragment_mode) NLIB_NOEXCEPT;
    template<size_t N>
    static errno_t EncodeUriComponent(size_t* written, char (&buf)[N], const char* first,
                                      const char* last, bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, N, first, last, fragment_mode);
    }
    static errno_t EncodeUriComponent(size_t* written, char* buf, size_t n, const char* s,
                                      bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, n, s, s + nlib_strlen(s), fragment_mode);
    }
    template<size_t N>
    static errno_t EncodeUriComponent(size_t* written, char (&buf)[N], const char* s,
                                      bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, N, s, fragment_mode);
    }
    static std::pair<errno_t, size_t>
    EncodeUriComponent(char* buf, size_t n, const char* first, const char* last,
                       bool fragment_mode) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = EncodeUriComponent(&written, buf, n, first, last, fragment_mode);
        return std::make_pair(e, written);
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    EncodeUriComponent(char (&buf)[N], const char* first, const char* last,
                       bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(buf, N, first, last, fragment_mode);
    }
    static std::pair<errno_t, size_t>
    EncodeUriComponent(char* buf, size_t n, const char* s, bool fragment_mode) {
        return EncodeUriComponent(buf, n, s, s + nlib_strlen(s), fragment_mode);
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    EncodeUriComponent(char (&buf)[N], const char* s, bool fragment_mode) {
        return EncodeUriComponent(buf, N, s, s + nlib_strlen(s), fragment_mode);
    }

    static errno_t EncodePath(size_t* written, char* buf, size_t n, const char* s) NLIB_NOEXCEPT;
    template<size_t N>
    static errno_t EncodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return EncodePath(written, buf, N, s);
    }
    static std::pair<errno_t, size_t> EncodePath(char* buf, size_t n, const char* s) NLIB_NOEXCEPT {
        size_t written;
        errno_t e = EncodePath(&written, buf, n, s);
        return std::make_pair(e, written);
    }
    template<size_t N>
    static std::pair<errno_t, size_t> EncodePath(char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return EncodePath(buf, N, s);
    }

 public:
    NLIB_CEXPR Uri() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~Uri() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(Uri);
    bool Parse(const char* first, const char* last) NLIB_NOEXCEPT;
    bool Parse(const char* str) NLIB_NOEXCEPT { return Parse(str, str + nlib_strlen(str)); }
    // for example:
    //   http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag
    // scheme = "http"
    // userinfo = "foobar:pass"
    // host = "www.example.com"
    // port = "80"
    // path = "/my/dir/index.html"
    // query = "var=value"
    // fragment = "frag"
    bool SetUri(const char* scheme, const char* userinfo, const char* host, const char* port,
                const char* path, const char* query, const char* fragment) NLIB_NOEXCEPT;
    bool IsAbsolute() const NLIB_NOEXCEPT { return GetScheme() != nullptr; }
    bool ComposeString(char* buf, size_t size) const NLIB_NOEXCEPT;
    template<size_t N>
    bool ComposeString(char (&buf)[N]) const NLIB_NOEXCEPT {
        return this->ComposeString(buf, N);
    }

    // AddBaseUri("g/", "http://a/b/c/d;p?q") -> http://a/b/c/g/
    // AddBaseUri("/g", "http://a/b/c/d;p?q") -> http://a/g
    // AddBaseUri("//g", "http://a/b/c/d;p?q") -> http://g
    bool AddBaseUri(const Uri& relative, const Uri& base) NLIB_NOEXCEPT;
    // "http" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetScheme() const NLIB_NOEXCEPT;
    // "foobar:pass" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetUserInfo() const NLIB_NOEXCEPT;
    // "www.example.com" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetHost() const NLIB_NOEXCEPT;
    // "80" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPort() const NLIB_NOEXCEPT;
    int GetPortNumber() const NLIB_NOEXCEPT;
    // "/my/dir/index.html" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPath() const NLIB_NOEXCEPT;
    // "var=value" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetQuery() const NLIB_NOEXCEPT;
    // "frag" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetFragment() const NLIB_NOEXCEPT;

    bool SetScheme(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetUserInfo(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetHost(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetPort(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetPath(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetQuery(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetFragment(const char* first, const char* last) NLIB_NOEXCEPT;
    bool SetScheme(const char* scheme) NLIB_NOEXCEPT {
        return scheme ? SetScheme(scheme, scheme + nlib_strlen(scheme))
                      : SetScheme(nullptr, nullptr);
    }
    bool SetUserInfo(const char* userinfo) NLIB_NOEXCEPT {
        return userinfo ? SetUserInfo(userinfo, userinfo + nlib_strlen(userinfo))
                        : SetUserInfo(nullptr, nullptr);
    }
    bool SetHost(const char* host) NLIB_NOEXCEPT {
        return host ? SetHost(host, host + nlib_strlen(host)) : SetHost(nullptr, nullptr);
    }
    bool SetPort(const char* port) NLIB_NOEXCEPT {
        return port ? SetPort(port, port + nlib_strlen(port)) : SetPort(nullptr, nullptr);
    }
    bool SetPortNumber(int port) NLIB_NOEXCEPT {
        size_t cnt;
        char buf[16];
        nlib_snprintf(&cnt, buf, "%d", port);
        return SetPort(&buf[0], &buf[0] + cnt);
    }
    bool SetPath(const char* path) NLIB_NOEXCEPT {
        return path ? SetPath(path, path + nlib_strlen(path)) : SetPath(nullptr, nullptr);
    }
    bool SetQuery(const char* query) NLIB_NOEXCEPT {
        return query ? SetQuery(query, query + nlib_strlen(query)) : SetQuery(nullptr, nullptr);
    }
    bool SetFragment(const char* fragment) NLIB_NOEXCEPT {
        return fragment ? SetFragment(fragment, fragment + nlib_strlen(fragment))
                        : SetFragment(nullptr, nullptr);
    }
    template<size_t N>
    bool SetQuery(const UriQueryEncoder<N>& encoder) NLIB_NOEXCEPT {
        std::pair<const char*, const char*> x = encoder.GetData();
        return SetQueryByQueryEncoder_(x.first, x.second);
    }

    void Reset() NLIB_NOEXCEPT;
    enum Error {
        kOk = 0,
        kInvalidParam,
        kOutOfMemory,
        kNotInitialized,
        kInvalidUri,
        kAbsoluteUriRequired,
        kPathNormalizationError,
        kUriTooLong,
        kBufferSizeNotEnough
    };
    Error GetError() const NLIB_NOEXCEPT;
    NLIB_SAFE_BOOL(Uri, GetError() == kOk);

 private:
    bool SetQueryByQueryEncoder_(const char* first, const char* last) NLIB_NOEXCEPT;
    UriPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Uri);
    friend NLIB_VIS_PUBLIC bool operator==(const Uri& lhs, const Uri& rhs) NLIB_NOEXCEPT;
    friend NLIB_VIS_PUBLIC bool operator<(const Uri& lhs, const Uri& rhs) NLIB_NOEXCEPT;
};

NLIB_VIS_PUBLIC bool operator==(const Uri& lhs, const Uri& rhs) NLIB_NOEXCEPT;
NLIB_EQUAL_OPERATOR(Uri)
NLIB_VIS_PUBLIC bool operator<(const Uri& lhs, const Uri& rhs) NLIB_NOEXCEPT;
NLIB_COMPARE_OPERATOR(Uri)

template<size_t N>
errno_t UriQueryEncoder<N>::Append(const char* key, const char* value) NLIB_NOEXCEPT {
    // cur_ unchanged if not successful
    char* p = cur_;
    if (p != &buf_[0]) {
        if (p == &buf_[N - 1]) return ERANGE;
        *p++ = '&';
    }
    errno_t e;
    size_t written;
    if (key) {
        e = Uri::EncodeUriComponent(&written, p, &buf_[N] - p, key, false);
        if (e != 0) return e;
        p += written;
    }
    if (p == &buf_[N - 1]) return ERANGE;
    *p++ = '=';
    if (value) {
        e = Uri::EncodeUriComponent(&written, p, &buf_[N] - p, value, false);
        if (e != 0) return e;
        p += written;
    }
    cur_ = p;
    return 0;
}

template<size_t K = 64, size_t V = 192>
class UriQueryDecoder NLIB_FINAL {
 public:
    NLIB_CEXPR14 UriQueryDecoder() NLIB_NOEXCEPT : ptr_beg_(nullptr),
                                                   ptr_end_(nullptr),
                                                   ptr_key_(nullptr),
                                                   ptr_eq_(nullptr),
                                                   ptr_and_(nullptr),
                                                   buf_key_valid_(false),
                                                   buf_value_valid_(false) {
        buf_key_[0]   = '\0';
        buf_value_[0] = '\0';
    }
    errno_t Init(const Uri& uri) NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool HasNext() NLIB_NOEXCEPT;
    errno_t MoveNext() NLIB_NOEXCEPT;
    std::pair<errno_t, const char*> GetKey() NLIB_NOEXCEPT;
    std::pair<errno_t, const char*> GetValue() NLIB_NOEXCEPT;
    std::pair<errno_t, const char*> GetFirstValueByName(const char* key) NLIB_NOEXCEPT;
    void Rewind() NLIB_NOEXCEPT;

 private:
    static std::pair<errno_t, const char*> Error(errno_t e) NLIB_NOEXCEPT {
        return std::make_pair(e, static_cast<const char*>(nullptr));
    }
    errno_t GetKeyWithCompare_(const char* key) NLIB_NOEXCEPT {
        std::pair<errno_t, const char*> x = GetKey();
        if (x.first != 0) return x.first;
        if (strcmp(key, x.second) == 0) return 0;
        return ESRCH;
    }
    void SpotKeyValue() NLIB_NOEXCEPT;

    const char* ptr_beg_;
    const char* ptr_end_;
    const char* ptr_key_;  // key or p_end_ or nullptr
    const char* ptr_eq_;   // '=' or ptr_and_
    const char* ptr_and_;  // '&' or p_end_ or nullptr
    bool buf_key_valid_;
    bool buf_value_valid_;
    char buf_key_[K];
    char buf_value_[V];
};

template<size_t K, size_t V>
errno_t UriQueryDecoder<K, V>::Init(const Uri& uri) NLIB_NOEXCEPT {
    if (!uri) return EINVAL;
    const char* p = uri.GetQuery();
    if (p) {
        ptr_beg_ = ptr_key_ = p;
        ptr_end_            = p + nlib_strlen(p);
    } else {
        ptr_beg_ = ptr_end_ = &buf_key_[0];
    }
    ptr_eq_ = ptr_and_ = nullptr;
    buf_key_valid_ = buf_value_valid_ = false;
    return 0;
}

template<size_t K, size_t V>
void UriQueryDecoder<K, V>::Reset() NLIB_NOEXCEPT {
    ptr_beg_ = ptr_end_ = ptr_key_ = ptr_eq_ = ptr_and_ = nullptr;
    buf_key_valid_ = buf_value_valid_ = false;
}

template<size_t K, size_t V>
bool UriQueryDecoder<K, V>::HasNext() NLIB_NOEXCEPT {
    if (ptr_key_ == ptr_end_) return false;
    if (NLIB_LIKELY(*ptr_key_ != '&')) return true;
    ptr_eq_ = ptr_and_ = nullptr;
    buf_key_valid_ = buf_value_valid_ = false;
    do {
        ++ptr_key_;
    } while (ptr_key_ != ptr_end_ && *ptr_key_ == '&');
    return ptr_key_ != ptr_end_;
}

template<size_t K, size_t V>
errno_t UriQueryDecoder<K, V>::MoveNext() NLIB_NOEXCEPT {
    if (!HasNext()) return ENOENT;
    if (ptr_and_) {
        ptr_key_ = (ptr_and_ != ptr_end_) ? ptr_and_ + 1 : ptr_end_;
    } else {
        const void* p = nlib_memchr(ptr_key_, '&', std::distance(ptr_key_, ptr_end_));
        ptr_key_      = (p != nullptr) ? static_cast<const char*>(p) + 1 : ptr_end_;
    }
    ptr_eq_ = ptr_and_ = nullptr;
    buf_key_valid_ = buf_value_valid_ = false;
    return 0;
}

template<size_t K, size_t V>
std::pair<errno_t, const char*> UriQueryDecoder<K, V>::GetKey() NLIB_NOEXCEPT {
    if (buf_key_valid_) return std::make_pair(0, &buf_key_[0]);
    if (!ptr_key_) return Error(EBADF);
    if (!HasNext()) return Error(ENOENT);
    if (!ptr_eq_) SpotKeyValue();
    size_t w;
    errno_t e = Uri::DecodeUriComponent(&w, buf_key_, K, ptr_key_, ptr_eq_);
    if (e != 0) return Error(e);
    buf_key_valid_ = true;
    return std::make_pair(0, &buf_key_[0]);
}

template<size_t K, size_t V>
std::pair<errno_t, const char*> UriQueryDecoder<K, V>::GetValue() NLIB_NOEXCEPT {
    if (buf_value_valid_) return std::make_pair(0, &buf_value_[0]);
    if (!ptr_key_) return Error(EBADF);
    if (!HasNext()) return Error(ENOENT);
    if (!ptr_eq_) SpotKeyValue();
    if (ptr_eq_ != ptr_and_) {
        size_t w;
        errno_t e = Uri::DecodeUriComponent(&w, buf_value_, V, ptr_eq_ + 1, ptr_and_);
        if (e != 0) return Error(e);
    } else {
        buf_value_[0] = '\0';
    }
    buf_value_valid_ = true;
    return std::make_pair(0, &buf_value_[0]);
}

template<size_t K, size_t V>
std::pair<errno_t, const char*>
UriQueryDecoder<K, V>::GetFirstValueByName(const char* key) NLIB_NOEXCEPT {
    errno_t e;
    do {
        e = GetKeyWithCompare_(key);
        if (e == 0) return std::make_pair(0, &buf_key_[0]);
        if (e != ESRCH) return Error(e);
    } while (MoveNext() == 0);
    return Error(ESRCH);
}

template<size_t K, size_t V>
void UriQueryDecoder<K, V>::Rewind() NLIB_NOEXCEPT {
    ptr_key_ = ptr_beg_;
    ptr_eq_ = ptr_and_ = nullptr;
    buf_key_valid_ = buf_value_valid_ = false;
}

template<size_t K, size_t V>
void UriQueryDecoder<K, V>::SpotKeyValue() NLIB_NOEXCEPT {
    const void* p_and = nlib_memchr(ptr_key_, '&', std::distance(ptr_key_, ptr_end_));
    ptr_and_          = p_and ? static_cast<const char*>(p_and) : ptr_end_;
    const void* p_eq  = nlib_memchr(ptr_key_, '=', std::distance(ptr_key_, ptr_and_));
    ptr_eq_           = p_eq ? static_cast<const char*>(p_eq) : ptr_and_;
}

NLIB_NAMESPACE_END
#ifndef __cpp_rvalue_references
NLIB_DEFINE_STD_SWAP(::nlib_ns::Uri)
#endif
#endif  // INCLUDE_NN_NLIB_URI_H_
