﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_EXI_XMLSTREAMWRITER_H_
#define INCLUDE_NN_NLIB_EXI_XMLSTREAMWRITER_H_

#include "nn/nlib/exi/ExiErrorStatus.h"
#include "nn/nlib/exi/Types.h"

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
class OutputStream;

namespace exi {

// See:
// http://java.sun.com/javase/ja/6/docs/ja/api/javax/xml/stream/XMLStreamWriter.html

struct XmlStreamWriterSettings {
    bool encodeCookie;
    bool encodeOptions;
    Alignment alignment;
    bool fragment;
    Preserve preserve;
    int version;
    XmlProcessor processor;

 public:
    NLIB_CEXPR XmlStreamWriterSettings() NLIB_NOEXCEPT : encodeCookie(false),
                                                         encodeOptions(false),
                                                         alignment(ALIGNMENT_BIT_PACKED),
                                                         fragment(false),
                                                         preserve(),
                                                         version(0),
                                                         processor(XML_PROCESSOR_EXI) {}
};

class NLIB_VIS_PUBLIC XmlStreamWriter {
 public:
    static XmlStreamWriter* Create(OutputStream* stream, const XmlStreamWriterSettings& settings,
                                   const ExiAllocatorEx& al) NLIB_NOEXCEPT;
    static XmlStreamWriter*
    Create(OutputStream* stream, const XmlStreamWriterSettings& settings) NLIB_NOEXCEPT {
        ExiAllocatorEx al;
        return Create(stream, settings, al);
    }
    static XmlStreamWriter* Create(OutputStream* stream) NLIB_NOEXCEPT {
        NLIB_CEXPR XmlStreamWriterSettings settings;
        return Create(stream, settings);
    }
    static XmlStreamWriter* Create(OutputStream* stream, const ExiAllocatorEx& al) NLIB_NOEXCEPT {
        NLIB_CEXPR XmlStreamWriterSettings settings;
        return Create(stream, settings, al);
    }
    virtual ~XmlStreamWriter() NLIB_NOEXCEPT {
        if (is_local_al_) ExiAllocator::Swap(al_);
    }
    void Close() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        this->Close_();
    }
    void Flush() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        this->Flush_();
    }
    const ExiChar* GetPrefix(const ExiChar* uri) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetPrefix_(uri ? uri : NLIB_EXI_ESTR);
    }
    void WriteAttribute(const ExiChar* local_name, const ExiChar* value) NLIB_NOEXCEPT {
        this->WriteAttribute(NLIB_EXI_ESTR, NLIB_EXI_ESTR, local_name, value);
    }
    void WriteAttribute(const ExiChar* prefix, const ExiChar* namespace_uri,
                        const ExiChar* local_name, const ExiChar* value) NLIB_NOEXCEPT;
    void WriteCData(const ExiChar* data) NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteCData_(data ? data : NLIB_EXI_ESTR);
    }
    void WriteCharacters(const ExiChar* text) NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteCharacters_(text ? text : NLIB_EXI_ESTR);
    }
    void WriteComment(const ExiChar* data) NLIB_NOEXCEPT;
    void WriteProcessingInstruction(const ExiChar* target, const ExiChar* data) NLIB_NOEXCEPT;
    void WriteNamespace(const ExiChar* prefix, const ExiChar* namespace_uri) NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteNamespace_(prefix ? prefix : NLIB_EXI_ESTR,
                              namespace_uri ? namespace_uri : NLIB_EXI_ESTR);
    }
    void WriteDefaultNamespace(const ExiChar* namespace_uri) NLIB_NOEXCEPT {
        this->WriteNamespace(NLIB_EXI_ESTR, namespace_uri);
    }
    // virtual void WriteDtd(const ExiChar* dtd)=0;
    void WriteEmptyElement(const ExiChar* local_name) NLIB_NOEXCEPT {
        this->WriteEmptyElement(NLIB_EXI_ESTR, NLIB_EXI_ESTR, local_name);
    }
    void WriteEmptyElement(const ExiChar* prefix, const ExiChar* namespace_uri,
                           const ExiChar* local_name) NLIB_NOEXCEPT;
    void WriteStartDocument() NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteStartDocument_();
    }
    // virtual void WriteStartDocument(const ExiChar* version)=0;
    // virtual void WriteStartDocument(const ExiChar* encoding, const ExiChar* version)=0;
    void WriteEndDocument() NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteEndDocument_();
    }
    void WriteEndElement() NLIB_NOEXCEPT {
        if (NEXI_IS_ERROR) return;
        ExiAllocatorSwapper swapper(al_);
        this->WriteEndElement_();
    }
    // virtual void WriteEntityRef(const ExiChar* name)=0;
    void WriteStartElement(const ExiChar* local_name) NLIB_NOEXCEPT {
        this->WriteStartElement(NLIB_EXI_ESTR, NLIB_EXI_ESTR, local_name);
    }
    void WriteStartElement(const ExiChar* prefix, const ExiChar* namespace_uri,
                           const ExiChar* local_name) NLIB_NOEXCEPT;
    ExiErrorStatus::ErrorValue GetError() const NLIB_NOEXCEPT { return error_status_->GetError(); }
    ExiErrorStatus* GetErrorStatus() const NLIB_NOEXCEPT { return error_status_; }
    NLIB_SAFE_BOOL(XmlStreamWriter, !!*error_status_)

 private:
    virtual void Close_() NLIB_NOEXCEPT                                                         = 0;
    virtual void Flush_() NLIB_NOEXCEPT                                                         = 0;
    virtual const ExiChar* GetPrefix_(const ExiChar* uri) NLIB_NOEXCEPT                         = 0;
    virtual void WriteAttribute_(const ExiChar* prefix, const ExiChar* namespace_uri,
                                 const ExiChar* local_name, const ExiChar* value) NLIB_NOEXCEPT = 0;
    virtual void WriteCData_(const ExiChar* data) NLIB_NOEXCEPT                                 = 0;
    virtual void WriteCharacters_(const ExiChar* text) NLIB_NOEXCEPT                            = 0;
    virtual void WriteComment_(const ExiChar* data) NLIB_NOEXCEPT                               = 0;
    virtual void
    WriteProcessingInstruction_(const ExiChar* target, const ExiChar* data) NLIB_NOEXCEPT = 0;
    virtual void WriteStartElement_(const ExiChar* prefix, const ExiChar* namespace_uri,
                                    const ExiChar* local_name) NLIB_NOEXCEPT              = 0;
    virtual void WriteEmptyElement_(const ExiChar* prefix, const ExiChar* namespace_uri,
                                    const ExiChar* local_name) NLIB_NOEXCEPT {
        this->WriteStartElement_(prefix, namespace_uri, local_name);
        this->WriteEndElement();
    }
    virtual void
    WriteNamespace_(const ExiChar* prefix, const ExiChar* namespace_uri) NLIB_NOEXCEPT = 0;
    virtual void WriteStartDocument_() NLIB_NOEXCEPT                                   = 0;
    virtual void WriteEndDocument_() NLIB_NOEXCEPT                                     = 0;
    virtual void WriteEndElement_() NLIB_NOEXCEPT                                      = 0;

 protected:
    static bool IsNullOrEmptyString(const ExiChar* p) NLIB_NOEXCEPT { return (!p) || (*p == '\0'); }
    void SetExiErrorStatus(ExiErrorStatus* p) NLIB_NOEXCEPT { error_status_ = p; }
    ExiErrorStatus* GetExiErrorStatus() const NLIB_NOEXCEPT { return error_status_; }
    explicit XmlStreamWriter(const ExiAllocatorEx& al) NLIB_NOEXCEPT : error_status_(nullptr),
                                                                       al_(al),
                                                                       is_local_al_(al) {}

 private:
    ExiErrorStatus* error_status_;

 protected:
    ExiAllocatorEx al_;
    bool is_local_al_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(XmlStreamWriter);
};

}  // namespace exi
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_EXI_XMLSTREAMWRITER_H_
