﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_
#define INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_

#include "nn/nlib/Config.h"

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace heap {

class NLIB_VIS_PUBLIC FrameHeap NLIB_FINAL {
 public:
    typedef void (*DestructorHandler)(void* start, void* end) NLIB_NOEXCEPT_FUNCPTR;
    enum AllocOption { kAllocOptionAllocHead = 0x00000000, kAllocOptionAllocTail = 0x00000001 };

    struct State {
        size_t _[2];
        bool operator==(const State& rhs) NLIB_NOEXCEPT {
            return _[0] == rhs._[0] && _[1] == rhs._[1];
        }
        bool operator!=(const State& rhs) NLIB_NOEXCEPT {
            return _[0] != rhs._[0] || _[1] != rhs._[1];
        }
    };
    FrameHeap() NLIB_NOEXCEPT : heap_option_(0), start_(nullptr), end_(nullptr), handler_(nullptr) {
        cur_.interlocked._[0] = cur_.interlocked._[1] = 0;
    }
    ~FrameHeap() NLIB_NOEXCEPT {
        if (handler_) {
            (*handler_)(start_, end_);
        }
    }
    errno_t
    Init(void* start, size_t size, DestructorHandler handler, uint32_t heap_option) NLIB_NOEXCEPT;
    void FreeAll() NLIB_NOEXCEPT {
        HeapStatus to;
        to.tb.topidx = 0;
        to.tb.botidx = end_ - start_;
        cur_         = to;
    }
    void FreeTop() NLIB_NOEXCEPT { cur_.tb.topidx = 0; }
    void FreeBot() NLIB_NOEXCEPT {
        size_t bot     = end_ - start_;
        cur_.tb.botidx = bot;
    }
    errno_t ResizeTop(void* p, size_t from_size, size_t to_size) NLIB_NOEXCEPT;
    errno_t Adjust(void** start, size_t* size) NLIB_NOEXCEPT;

    State SaveState() NLIB_NOEXCEPT { return cur_.interlocked; }
    void RestoreState(State state) NLIB_NOEXCEPT { cur_.interlocked = state; }
    void* Alloc(size_t n) NLIB_NOEXCEPT { return this->Alloc(n, 8, 0); }
    void* Alloc(size_t n, size_t algn) NLIB_NOEXCEPT { return this->Alloc(n, algn, 0); }
    void* Alloc(size_t n, size_t algn, uint32_t alloc_option) NLIB_NOEXCEPT;
    void Dump() NLIB_NOEXCEPT;

 private:
    union HeapStatus {
        State interlocked;
        struct topbot_ {
            size_t topidx;
            size_t botidx;
        } tb;
    };
    static uint8_t* RoundUp(uint8_t* ptr, size_t algn) NLIB_NOEXCEPT {
        // algn is power of 2.
        intptr_t tmp = reinterpret_cast<intptr_t>(ptr);
        return reinterpret_cast<uint8_t*>(tmp + (-tmp & (algn - 1)));
    }
    static uint8_t* RoundDown(uint8_t* ptr, size_t algn) NLIB_NOEXCEPT {
        // algn is power of 2.
        intptr_t tmp = reinterpret_cast<intptr_t>(ptr);
        return reinterpret_cast<uint8_t*>(tmp & (~algn + 1));
    }
    void ChangeHeapRange(void* start, void* end) NLIB_NOEXCEPT {
        start_ = static_cast<uint8_t*>(start);
        end_   = static_cast<uint8_t*>(end);
    }

 private:
    HeapStatus cur_;
    uint32_t heap_option_;
    uint8_t* start_;
    uint8_t* end_;
    DestructorHandler handler_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FrameHeap);
};

}  // namespace heap
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_
