﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HEAP_NMALLOC_H_
#define INCLUDE_NN_NLIB_HEAP_NMALLOC_H_

#include <errno.h>
#include <stddef.h>
#include <stdlib.h>  // for malloc/free

#ifdef __cplusplus
#include <new>
#include "nn/nlib/Config.h"
#else
#include "nn/nlib/Platform.h"
#endif

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

#ifndef INCLUDE_NN_NLIB_PLATFORM_H_
// some compilers complain about double declaration
typedef int errno_t;
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct NLIB_VIS_PUBLIC {
    size_t alloc_count;
    size_t alloc_size;
    size_t hash;
} HeapHash;

typedef struct NLIB_VIS_PUBLIC {
    void* addr;
    size_t size;
    unsigned int heap_option;
} NMallocSettings;

#define NMALLOC_HEAPOPTION_ENABLE_ENV (0x00000001)
#define NMALLOC_HEAPOPTION_CACHE_DISABLE (0x00000004)

#ifdef _MSC_VER
NLIB_VIS_PUBLIC void nmalloc_get_settings_default(NMallocSettings* settings);
#ifndef NLIB_WINDLL
void nmalloc_get_settings(NMallocSettings* settings);  // weak function
#ifndef _WIN64
#pragma comment(linker, "/alternatename:_nmalloc_get_settings=_nmalloc_get_settings_default")
#else
#pragma comment(linker, "/alternatename:nmalloc_get_settings=nmalloc_get_settings_default")
#endif
#else
NLIB_VIS_PUBLIC extern void (*g_nmalloc_get_settings_func)(NMallocSettings*);
#endif
#else
NLIB_VIS_PUBLIC void nmalloc_get_settings(NMallocSettings* settings);  // weak function
#endif

typedef enum {
    kNmallocDumpBasic       = 0,
    kNmallocDumpSpans       = 1,
    kNmallocDumpPointers    = 2,
    kNmallocDumpPageSummary = 4,
    kNmallocDumpAll         = kNmallocDumpSpans | kNmallocDumpPointers | kNmallocDumpPageSummary
} NMallocDumpMode;

typedef enum {
    kNmallocQueryDump = 0,
    kNmallocQueryPageSize,
    kNmallocQueryAllocatedSize,
    kNmallocQueryFreeSize,
    kNmallocQuerySystemSize,
    kNmallocQueryMaxAllocatableSize,
    kNmallocQueryIsClean,
    kNmallocQueryHeapHash,
    kNmallocQueryUnifyFreelist,
    kNmallocQuerySetColor,
    kNmallocQueryGetColor,
    kNmallocQuerySetName,
    kNmallocQueryGetName,
    kNmallocQueryCacheMin_,
    kNmallocQueryCheckCache,
    kNmallocQueryClearCache,
    kNmallocQueryFinalizeCache,
    kNmallocQueryFreeSizeNx,
    kNmallocQueryMaxAllocatableSizeNx,
    kNmallocQueryDumpJson
} NMallocQuery;

typedef int (*nmalloc_heapwalk_callback)(void* allocated_ptr, size_t size, void* user_ptr);

NLIB_VIS_PUBLIC errno_t nmalloc_query(int query, ...);
NLIB_CHECK_RESULT NLIB_VIS_PUBLIC void* nrealloc(void* ptr, size_t size) NLIB_ATTRIBUTE_MALLOC
    NLIB_ATTRIBUTE_ALLOC_SIZE1(2) NLIB_ATTRIBUTE_ASSUME_ALIGNED(8);
NLIB_CHECK_RESULT NLIB_VIS_PUBLIC void* nmalloc(size_t size) NLIB_ATTRIBUTE_MALLOC
    NLIB_ATTRIBUTE_ALLOC_SIZE1(1) NLIB_ATTRIBUTE_ASSUME_ALIGNED(8);
NLIB_CHECK_RESULT NLIB_VIS_PUBLIC void* ncalloc(size_t nmemb, size_t size) NLIB_ATTRIBUTE_MALLOC
    NLIB_ATTRIBUTE_ALLOC_SIZE2(1, 2) NLIB_ATTRIBUTE_ASSUME_ALIGNED(8);
NLIB_CHECK_RESULT NLIB_VIS_PUBLIC void*
nmalloc_aligned(size_t size, size_t algn) NLIB_ATTRIBUTE_MALLOC NLIB_ATTRIBUTE_ALLOC_SIZE1(1)
    NLIB_ATTRIBUTE_ALLOC_ALIGN(2) NLIB_ATTRIBUTE_ASSUME_ALIGNED(8);
NLIB_VIS_PUBLIC size_t nmalloc_size(const void* ptr) NLIB_NONNULL;
NLIB_VIS_PUBLIC void nfree(void* p);
NLIB_VIS_PUBLIC void nfree_size(void* p, size_t size);

NLIB_VIS_PUBLIC errno_t nmalloc_walk_allocated_ptrs(nmalloc_heapwalk_callback func, void* user_ptr);

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus

inline bool operator==(const HeapHash& rhs, const HeapHash& lhs) {
    return rhs.alloc_count == lhs.alloc_count && rhs.alloc_size == lhs.alloc_size &&
           rhs.hash == lhs.hash;
}

inline bool operator!=(const HeapHash& rhs, const HeapHash& lhs) {
    return rhs.alloc_count != lhs.alloc_count || rhs.alloc_size != lhs.alloc_size ||
           rhs.hash != lhs.hash;
}

NLIB_NAMESPACE_BEGIN
namespace heap {

template<typename T>
struct NMallocStlAllocator {
    typedef size_t size_type;
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T& reference;
    typedef const T& const_reference;
    typedef ptrdiff_t difference_type;
    template<typename U>
    struct rebind {
        typedef NMallocStlAllocator<U> other;
    };
    NMallocStlAllocator() {}
    explicit NMallocStlAllocator(const NMallocStlAllocator<T>&) {}
    template<class _Other>
    NMallocStlAllocator(const NMallocStlAllocator<_Other>&) {}
    pointer allocate(size_type n, const void* p = 0);
    void deallocate(pointer p, size_type n) {
#if 1
        nfree_size(p, n * sizeof(T));
#else
        (void)n;
        nfree(p);
#endif
    }
    void construct(pointer p, const T& t) { new ((void*)p) T(t); }
    void construct(pointer p) { new ((void*)p) T(); }
#if defined(__cpp_rvalue_references) && defined(__cpp_variadic_templates)
    template<class X, class... ARGS>
    void construct(X* p, ARGS&&... args) {
        new ((void*)p) X(std::forward<ARGS>(args)...);
    }
#endif
    void destroy(pointer ptr) {
        (void)ptr;
        (*ptr).~T();
    }
    template<class X>
    void destroy(X* p) {
        NLIB_UNUSED(p);  // workaround for VS2012
        p->~X();
    }
    size_t max_size() const {
        size_t cnt = (size_t)-1 / sizeof(T);
        return 0 < cnt ? cnt : 1;
    }
    pointer address(reference value) const { return &value; }
    const_pointer address(const_reference value) const { return &value; }
};

template<typename T>
typename NMallocStlAllocator<T>::pointer
NMallocStlAllocator<T>::allocate(size_type n, const void* p) {
    (void)p;
    void* pp = nmalloc(n * sizeof(T));
    return reinterpret_cast<T*>(pp);
}

template<class T1, class T2>
inline bool operator==(const NMallocStlAllocator<T1>&, const NMallocStlAllocator<T2>&) {
    return true;
}

template<class T1, class T2>
inline bool operator!=(const NMallocStlAllocator<T1>&, const NMallocStlAllocator<T2>&) {
    return false;
}

}  // namespace heap
NLIB_NAMESPACE_END

#endif

#ifdef __cplusplus
#define NLIB_REPLACE_MALLOC                                                       \
    extern "C" {                                                                  \
    void* nlib_malloc(size_t n) { return nmalloc(n); }                            \
    void nlib_free(void* p) { nfree(p); }                                         \
    void* nlib_calloc(size_t nmemb, size_t size) { return ncalloc(nmemb, size); } \
    void* nlib_realloc(void* ptr, size_t size) { return nrealloc(ptr, size); }    \
    void* nlib_memalign(size_t alignment, size_t size) {                          \
        return nmalloc_aligned(size, alignment);                                  \
    }                                                                             \
    void nlib_free_size(void* ptr, size_t size) { nfree_size(ptr, size); }        \
    }

#ifndef NN_PLATFORM_CTR
#define NLIB_REPLACE_MALLOC_NEW                                                                 \
    NLIB_REPLACE_MALLOC                                                                         \
    void* operator new(size_t size) { return nlib_malloc(size); }                               \
    void operator delete(void* ptr)NLIB_NOEXCEPT { nlib_free(ptr); }                            \
    void* operator new(size_t size, const std::nothrow_t&) NLIB_NOEXCEPT {                      \
        return nlib_malloc(size);                                                               \
    }                                                                                           \
    void operator delete(void* ptr, const std::nothrow_t&)NLIB_NOEXCEPT { nlib_free(ptr); }     \
    void* operator new[](size_t size) { return nlib_malloc(size); }                             \
    void operator delete[](void* ptr) NLIB_NOEXCEPT { nlib_free(ptr); }                         \
    void* operator new[](size_t size, const std::nothrow_t&) NLIB_NOEXCEPT {                    \
        return nlib_malloc(size);                                                               \
    }                                                                                           \
    void operator delete[](void* ptr, const std::nothrow_t&) NLIB_NOEXCEPT { nlib_free(ptr); }  \
    void operator delete(void* ptr, size_t size)NLIB_NOEXCEPT { nlib_free_size(ptr, size); }    \
    void operator delete(void* ptr, size_t size, const std::nothrow_t&)NLIB_NOEXCEPT {          \
        nlib_free_size(ptr, size);                                                              \
    }                                                                                           \
    void operator delete[](void* ptr, size_t size) NLIB_NOEXCEPT { nlib_free_size(ptr, size); } \
    void operator delete[](void* ptr, size_t size, const std::nothrow_t&) NLIB_NOEXCEPT {       \
        nlib_free_size(ptr, size);                                                              \
    }
#else
#define NLIB_REPLACE_MALLOC_NEW                                                                  \
    NLIB_REPLACE_MALLOC                                                                          \
    void* operator new(size_t size) throw(std::bad_alloc) { return nlib_malloc(size); }          \
    void operator delete(void* ptr) throw() { nlib_free(ptr); }                                  \
    void* operator new(size_t size, const std::nothrow_t&) throw() { return nlib_malloc(size); } \
    void* operator new[](size_t size) throw(std::bad_alloc) { return nlib_malloc(size); }        \
    void operator delete[](void* ptr) throw() { nlib_free(ptr); }                                \
    void* operator new[](size_t size, const std::nothrow_t&) throw() { return nlib_malloc(size); }
#endif
#else
#define NLIB_REPLACE_MALLOC                                                       \
    void* nlib_malloc(size_t n) { return nmalloc(n); }                            \
    void nlib_free(void* p) { nfree(p); }                                         \
    void* nlib_calloc(size_t nmemb, size_t size) { return ncalloc(nmemb, size); } \
    void* nlib_realloc(void* ptr, size_t size) { return nrealloc(ptr, size); }    \
    void* nlib_memalign(size_t alignment, size_t size) {                          \
        return nmalloc_aligned(size, alignment);                                  \
    }                                                                             \
    void nlib_free_size(void* ptr, size_t size) { nfree_size(ptr, size); }
#define NLIB_REPLACE_MALLOC_NEW NLIB_REPLACE_MALLOC
#endif

#ifdef NLIB_WINDLL
#define NMALLOC_GET_SETTINGS                                                   \
    void nmalloc_get_settings(NMallocSettings*);                               \
    void __stdcall _nlib_tls_callback_heap(void* h, DWORD dwReason, void* r) { \
        (void)h;                                                               \
        (void)r;                                                               \
        static __declspec(thread) int nlib_dummy_nmalloc;                      \
        if (dwReason == DLL_PROCESS_ATTACH) {                                  \
            extern void (*g_nmalloc_get_settings_func)(NMallocSettings*);      \
            nlib_dummy_nmalloc          = 1;                                   \
            g_nmalloc_get_settings_func = nmalloc_get_settings;                \
        }                                                                      \
    }                                                                          \
    __pragma(section(".CRT$XLB", long, read)) __declspec(allocate(".CRT$XLB")) \
        PIMAGE_TLS_CALLBACK _xl_b = _nlib_tls_callback_heap;                   \
    void nmalloc_get_settings(NMallocSettings* settings)
#else
#ifdef __cplusplus
#define NMALLOC_GET_SETTINGS extern "C" void nmalloc_get_settings(NMallocSettings* settings)
#else
#define NMALLOC_GET_SETTINGS void nmalloc_get_settings(NMallocSettings* settings)
#endif
#endif

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_HEAP_NMALLOC_H_
