﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMGENERATOR_H_
#define INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMGENERATOR_H_

#include <utility>
#include "nn/nlib/Swap.h"
#include "nn/nlib/msgpack/MpObject.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif
NLIB_NAMESPACE_BEGIN
class OutputStream;
namespace msgpack {

struct JsonStreamGeneratorSettings NLIB_FINAL {
    bool msgpack;
    bool pretty_print;
    bool omit_utf8_checking;
    bool strict;
    NLIB_CEXPR JsonStreamGeneratorSettings() NLIB_NOEXCEPT : msgpack(false),
                                                             pretty_print(false),
                                                             omit_utf8_checking(false),
                                                             strict(false) {}
};

class NLIB_VIS_PUBLIC JsonStreamGenerator NLIB_FINAL {
 public:
    enum Error {
        kOk = 0,
        kErrorKeyRequired,
        kErrorKeyForbidden,
        kErrorEndArrayRequired,
        kErrorEndArrayForbidden,
        kErrorEndMapRequired,
        kErrorEndMapForbidden,
        kErrorStringUtf8,
        kErrorStream,
        kErrorOutOfMemory,
        kErrorArrayOrMapRequired,
        kErrorDocumentEnd,
        kErrorBinaryNotAvailable,
        kErrorExtNotAvailable,
        kErrorInvalidParam
    };

 public:
    static errno_t
    Generate(size_t* written, nlib_utf8_t* str, size_t n, const MpObject& obj,
             const JsonStreamGeneratorSettings& settings) NLIB_NOEXCEPT NLIB_NONNULL_2;
    static errno_t
    Generate(size_t* written, nlib_utf8_t* str, size_t n, const MpObject& obj) NLIB_NOEXCEPT {
        JsonStreamGeneratorSettings settings;
        return Generate(written, str, n, obj, settings);
    }
    template<size_t N>
    static errno_t Generate(size_t* written, nlib_utf8_t (&str)[N], const MpObject& obj,
                            const JsonStreamGeneratorSettings& settings) NLIB_NOEXCEPT {
        return Generate(written, str, N, obj, settings);
    }
    template<size_t N>
    static errno_t
    Generate(size_t* written, nlib_utf8_t (&str)[N], const MpObject& obj) NLIB_NOEXCEPT {
        JsonStreamGeneratorSettings settings;
        return Generate(written, str, N, obj, settings);
    }

    static std::pair<errno_t, size_t>
    Generate(nlib_utf8_t* str, size_t n, const MpObject& obj,
             const JsonStreamGeneratorSettings& settings) NLIB_NOEXCEPT {
        std::pair<errno_t, size_t> rval;
        rval.first = Generate(&rval.second, str, n, obj, settings);
        return rval;
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    Generate(nlib_utf8_t (&str)[N], const MpObject& obj,
             const JsonStreamGeneratorSettings& settings) NLIB_NOEXCEPT {
        return Generate(&str[0], N, obj, settings);
    }
    static std::pair<errno_t, size_t>
    Generate(nlib_utf8_t* str, size_t n, const MpObject& obj) NLIB_NOEXCEPT {
        JsonStreamGeneratorSettings settings;
        return Generate(str, n, obj, settings);
    }
    template<size_t N>
    static std::pair<errno_t, size_t>
    Generate(nlib_utf8_t (&str)[N], const MpObject& obj) NLIB_NOEXCEPT {
        JsonStreamGeneratorSettings settings;
        return Generate(&str[0], N, obj, settings);
    }

    NLIB_CEXPR JsonStreamGenerator() NLIB_NOEXCEPT : prv_(nullptr), error_(kOk) {}
    ~JsonStreamGenerator() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(JsonStreamGenerator);
    void Reset() NLIB_NOEXCEPT;
    errno_t Init(const JsonStreamGeneratorSettings& settings) NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT {
        JsonStreamGeneratorSettings settings;
        return Init(settings);
    }
    errno_t Open(OutputStream* stream) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Flush() NLIB_NOEXCEPT;
    errno_t Close() NLIB_NOEXCEPT;

    JsonStreamGenerator& StartArray(size_t count) NLIB_NOEXCEPT;
    JsonStreamGenerator& StartMap(size_t count) NLIB_NOEXCEPT;
    JsonStreamGenerator& Key(const nlib_utf8_t* key) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& EndArray() NLIB_NOEXCEPT;
    JsonStreamGenerator& EndMap() NLIB_NOEXCEPT;
    JsonStreamGenerator& Null() NLIB_NOEXCEPT;
    JsonStreamGenerator& Boolean(bool value) NLIB_NOEXCEPT;
    JsonStreamGenerator& Int8(int8_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Uint8(uint8_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Int16(int16_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Uint16(uint16_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Int32(int32_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Uint32(uint32_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Int64(int64_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Uint64(uint64_t num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Float(float num) NLIB_NOEXCEPT;
    JsonStreamGenerator& Double(double num) NLIB_NOEXCEPT;
    JsonStreamGenerator& String(const nlib_utf8_t* str) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Binary(const void* bin, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Ext(int8_t tp, const void* bin, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Timestamp(nlib_time t) NLIB_NOEXCEPT;
    JsonStreamGenerator& Object(const MpObject& obj) NLIB_NOEXCEPT;

    // NOTE: nlib_utf8_t str[32];  ...  String(str) calls this...
    // template<class STDSTRING>
    // JsonStreamGenerator& String(const STDSTRING& str) NLIB_NOEXCEPT {
    //     return String(str.c_str());
    // }
    JsonStreamGenerator& EmptyArray() NLIB_NOEXCEPT { return StartArray(0).EndArray(); }
    JsonStreamGenerator& EmptyMap() NLIB_NOEXCEPT { return StartMap(0).EndMap(); }
    JsonStreamGenerator& Int8Array(const int8_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Uint8Array(const uint8_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Int16Array(const int16_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Uint16Array(const uint16_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Int32Array(const int32_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Uint32Array(const uint32_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Int64Array(const int64_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& Uint64Array(const uint64_t* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& FloatArray(const float* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    JsonStreamGenerator& DoubleArray(const double* p, size_t count) NLIB_NOEXCEPT NLIB_NONNULL;
    Error GetError() const NLIB_NOEXCEPT { return error_; }
    NLIB_SAFE_BOOL(JsonStreamGenerator, error_ == kOk);

 private:
    struct State;
    NLIB_VIS_HIDDEN State* WriteStartSeq_(size_t count) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool JsonIndent_(size_t level) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool CheckAndStateChange_() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool
    WriteJsonString_(const nlib_utf8_t* str, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_VIS_HIDDEN bool
    WriteMsgpackString_(const nlib_utf8_t* str, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_VIS_HIDDEN void String_(const nlib_utf8_t* str, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_VIS_HIDDEN void Key_(const nlib_utf8_t* key, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;

    struct JsonStreamGeneratorPrivate;
    JsonStreamGeneratorPrivate* prv_;
    Error error_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonStreamGenerator);
};

}  // namespace msgpack
NLIB_NAMESPACE_END
NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::JsonStreamGenerator)
#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMGENERATOR_H_
