﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
#define INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_

#include <map>
#include <utility>
#include <vector>
#include <string>
#include <iterator>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/Cstring.h"
#include "nn/nlib/Nlist.h"

#ifdef __cpp_rvalue_references
#include <array>
#include <unordered_map>
#include <tuple>
#endif

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {

class MpObject;

template<class T>
errno_t Box(MpObject* obj, const T& v);

template<class T>
errno_t Unbox(const MpObject* obj, T* v);

struct nil {};

template<class T>
inline bool operator==(const nil& lhs, const T& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
template<class T>
inline bool operator==(const T& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
inline bool operator==(const nil& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template<class T>
inline bool operator!=(const nil& rhs, const T& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template<class T>
inline bool operator!=(const T& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
inline bool operator!=(const nil& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}

struct MpObjectKv;
class MpObjectAsArray;
class MpObjectAsMap;
class MpObjectAsString;
class MpObjectAsBinary;
class MpObjectAsExt;

class NLIB_VIS_PUBLIC MpObject NLIB_FINAL {
 private:
    union UnionType {
        uint64_t u64;
        int64_t i64;
        float f32;
        double f64;
        Nlist<MpObject>* ary;
        Nlist<MpObjectKv>* mp;
        nlib_utf8_t* str;
        void* binary;
    };

 public:
    static bool IsJsonPointer(const nlib_utf8_t* first, const nlib_utf8_t* last) NLIB_NOEXCEPT;
    static bool IsJsonPointer(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return IsJsonPointer(str, str + nlib_strlen(str));
    }
    static errno_t ResolveJsonPointer(MpObject** result, MpObject* root,
                                      const nlib_utf8_t* json_pointer) NLIB_NOEXCEPT;
    static errno_t DigByJsonPointer(MpObject** result, MpObject* root,
                                    const nlib_utf8_t* json_pointer) NLIB_NOEXCEPT;
    static errno_t RemoveByJsonPointer(MpObjectKv* removed, MpObject* root,
                                       const nlib_utf8_t* json_pointer) NLIB_NOEXCEPT;

 public:
    MpObjectAsArray AsArray() NLIB_NOEXCEPT;
    const MpObjectAsArray AsArray() const NLIB_NOEXCEPT;
    MpObjectAsMap AsMap() NLIB_NOEXCEPT;
    const MpObjectAsMap AsMap() const NLIB_NOEXCEPT;
    MpObjectAsString AsString() NLIB_NOEXCEPT;
    const MpObjectAsString AsString() const NLIB_NOEXCEPT;
    MpObjectAsBinary AsBinary() NLIB_NOEXCEPT;
    const MpObjectAsBinary AsBinary() const NLIB_NOEXCEPT;
    MpObjectAsExt AsExt() NLIB_NOEXCEPT;
    const MpObjectAsExt AsExt() const NLIB_NOEXCEPT;

    MpObject* GetArrayItem(size_t n) NLIB_NOEXCEPT;
    const MpObject* GetArrayItem(size_t n) const NLIB_NOEXCEPT;
    MpObject& operator[](size_t n) {
        NLIB_ASSERT(this->IsArray());
        NLIB_ASSERT(n < size_);
        return (*via_.ary)[n];
    }
    const MpObject& operator[](size_t n) const {
        NLIB_ASSERT(this->IsArray());
        NLIB_ASSERT(n < size_);
        return (*via_.ary)[n];
    }
    const MpObject* GetMapItem(const nlib_utf8_t* str) const NLIB_NOEXCEPT;
    MpObject* GetMapItem(const nlib_utf8_t* str) NLIB_NOEXCEPT;
    template<size_t N>
    MpObject* GetMapItem(const nlib_utf8_t (&str)[N]) NLIB_NOEXCEPT {
        return GetMapItem(&str[0]);
    }
    template<size_t N>
    const MpObject* GetMapItem(const nlib_utf8_t (&str)[N]) const NLIB_NOEXCEPT {
        return GetMapItem(&str[0]);
    }
    template<class STDSTRING>
    const MpObject* GetMapItem(const STDSTRING& str) const NLIB_NOEXCEPT {
        return GetMapItem(&str[0], str.size());
    }
    template<class STDSTRING>
    MpObject* GetMapItem(const STDSTRING& str) NLIB_NOEXCEPT {
        return GetMapItem(&str[0], str.size());
    }
    errno_t Resize(uint32_t n);
    MpObject* AppendArrayItem() NLIB_NOEXCEPT;
    MpObject* InsertArrayItem(size_t n) NLIB_NOEXCEPT;
    MpObjectKv* AppendMapItem() NLIB_NOEXCEPT;
    NLIB_OVERRIDE_NEW;

 public:
    MpObject() NLIB_NOEXCEPT : type_(kNil) {}  // only do minimum memory access
    ~MpObject() NLIB_NOEXCEPT { ReleaseIfNeeded(); }
#ifdef __cpp_rvalue_references
    MpObject(MpObject&& rhs) NLIB_NOEXCEPT {
        nlib_memcpy(this, sizeof(*this), &rhs, sizeof(rhs));
        rhs.type_ = kNil;
    }
    MpObject& operator=(MpObject&& rhs) NLIB_NOEXCEPT {
        ReleaseIfNeeded();
        nlib_memcpy(this, sizeof(*this), &rhs, sizeof(rhs));
        rhs.type_ = kNil;
        return *this;
    }
#endif
    MpObject(MpObject& rhs, move_tag) NLIB_NOEXCEPT {
        nlib_memcpy(this, sizeof(*this), &rhs, sizeof(rhs));
        rhs.type_ = kNil;
    }
    MpObject& assign(MpObject& rhs, move_tag) NLIB_NOEXCEPT {
        ReleaseIfNeeded();
        nlib_memcpy(this, sizeof(*this), &rhs, sizeof(rhs));
        rhs.type_ = kNil;
        return *this;
    }

    // NOTE: implicit call was error in gcc or clang....
    template<class T>
    MpObject(const T& x) {  // no explicit
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Construct(this, x);
    }
    errno_t GetArrayItem(size_t n, MpObject** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetArrayItem(size_t n, const MpObject** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n, MpObjectKv** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n, const MpObjectKv** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t
    GetMapItem(const nlib_utf8_t* str, size_t n, MpObjectKv** obj) NLIB_NOEXCEPT NLIB_NONNULL {
        const MpObjectKv* kv;
        errno_t e = this->GetMapItem(str, n, &kv);
        if (e != 0) return e;
        *obj = const_cast<MpObjectKv*>(kv);
        return 0;
    }
    errno_t GetMapItem(const nlib_utf8_t* str, size_t n,
                       const MpObjectKv** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    nlib_utf8_t* GetString() NLIB_NOEXCEPT {
        if (type_ == kString) return via_.str;
        if (NLIB_LIKELY(type_ == kStringInline)) return reinterpret_cast<nlib_utf8_t*>(&raw14data_);
        return nullptr;
    }
    errno_t RemoveArrayItem(size_t n, MpObject* obj) NLIB_NOEXCEPT;
    errno_t RemoveMapItem(size_t n, MpObjectKv* kv) NLIB_NOEXCEPT;
    errno_t RemoveMapItem(const nlib_utf8_t* key, MpObjectKv* kv) NLIB_NOEXCEPT;
    template<class STDSTRING>
    errno_t RemoveMapItem(const STDSTRING& str, MpObjectKv* kv) NLIB_NOEXCEPT {
        return RemoveMapItem(str.c_str(), kv);
    }

    const nlib_utf8_t* GetString() const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetString();
    }
    void* GetBinary(uint32_t* n) NLIB_NOEXCEPT;
    const void* GetBinary(uint32_t* n) const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetBinary(n);
    }
    void* GetExt(int8_t* tp, uint32_t* n) NLIB_NOEXCEPT;
    const void* GetExt(int8_t* tp, uint32_t* n) const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetExt(tp, n);
    }

    // I/F which returns std::pair, std::tuple
#ifdef __cpp_rvalue_references
    std::tuple<errno_t, const nlib_byte_t*, uint32_t> GetBinary() const NLIB_NOEXCEPT;
    std::tuple<errno_t, int8_t, const nlib_byte_t*, uint32_t> GetExt() const NLIB_NOEXCEPT;
    std::pair<errno_t, MpObject> RemoveArrayItem(size_t n) NLIB_NOEXCEPT {
        std::pair<errno_t, MpObject> rval;
        rval.first = RemoveArrayItem(n, &rval.second);
        return rval;
    }
    std::pair<errno_t, MpObjectKv> RemoveMapItem(size_t n) NLIB_NOEXCEPT;
#endif
    std::pair<errno_t, MpObjectKv*> GetMapItem(const nlib_utf8_t* str, size_t n) NLIB_NOEXCEPT {
        MpObjectKv* kv;
        return std::make_pair(GetMapItem(str, n, &kv), kv);
    }
    std::pair<errno_t, const MpObjectKv*>
    GetMapItem(const nlib_utf8_t* str, size_t n) const NLIB_NOEXCEPT {
        const MpObjectKv* kv;
        return std::make_pair(GetMapItem(str, n, &kv), kv);
    }
    // std::pair<errno_t, int32_t> GetInt32, GetInt64, GetUint32, GetUint64,
    // GetBoolean(), GetFloat(), GetDouble()
    std::pair<errno_t, nlib_time> GetTimestamp() const NLIB_NOEXCEPT;

    errno_t InitArray(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitMap(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(const nlib_utf8_t* str, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        uint32_t n = static_cast<uint32_t>(nlib_strlen(str));
        return InitString(str, n);
    }
    errno_t InitBinary(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitBinary(const void* p, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitExt(int8_t tp, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitExt(int8_t tp, const void* p, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitTimestamp(nlib_time t) NLIB_NOEXCEPT;

 public:  // Box
    errno_t Box(const nlib_utf8_t* str) NLIB_NOEXCEPT NLIB_NONNULL;
    template<uint32_t n>
    errno_t Box(const nlib_utf8_t (&str)[n]) NLIB_NOEXCEPT;
    template<class T, uint32_t n>
    errno_t Box(const T (&vec)[n]);
#ifdef __cpp_rvalue_references
    template<class T, size_t n>
    errno_t Box(const std::array<T, n>& vec);
#endif
    template<class T>
    errno_t Box(const T& v) {
        return MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Box(this, v);
    }
    // Use BoxBinary(p, n) or Box(reinterpret_cast<const nlib_utf8_t*>(p)) instead
    errno_t BoxBinary(const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t BoxExt(int8_t tp, const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;

 public:  // Unbox
    template<class T, size_t n>
    errno_t Unbox(T (&a)[n]) const {
        return this->Unbox(&a[0], n);
    }
#ifdef __cpp_rvalue_references
    template<class T, size_t n>
    errno_t Unbox(std::array<T, n>& a) {
        return this->Unbox(reinterpret_cast<T*>(&*a.begin()), n);
    }
#endif
    template<size_t n>
    errno_t Unbox(nlib_utf8_t (&str)[n]) const NLIB_NOEXCEPT {
        return this->Unbox(&str[0], n);
    }
    template<class T>
    errno_t Unbox(T* a, size_t n) const NLIB_NONNULL;
    errno_t Unbox(nlib_utf8_t* str, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t UnboxBinary(void* p, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t UnboxExt(int8_t* tp, void* p, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    template<class T>
    errno_t Unbox(T v) const {
        // T cannot be T* though v is only pointer type
        // because of conflict against array types.
        return MyBox<typename RemovePtr<T>::type,
                     typename IsIntegral<typename RemovePtr<T>::type>::type,
                     typename IsSigned<typename RemovePtr<T>::type>::type>::Unbox(this, v);
    }

 public:
    // You can use operator=() only for basic types(never fails).
    // Use Box() otherwise.
    template<class T>
    MpObject& operator=(const T& x) {
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Assign(this, x);
        return *this;
    }

    MpObject* Clone() const NLIB_NOEXCEPT;

 public:
    enum ObjectType {
        kNil = 0,
        kBoolean,
        kUint64,
        kInt64,
        kFloat,
        kDouble,
        kString,
        kArray,
        kMap,
        kBinary,
        kExt,
        kStringInline = 0x80 | kString,
        kBinaryInline = 0x80 | kBinary,
        kBooleanFalse = 0x00 | kBoolean,
        kBooleanTrue  = 0x80 | kBoolean
    };
    ObjectType GetType() const NLIB_NOEXCEPT { return static_cast<ObjectType>(type_ & 0x7F); }
    uint32_t GetSize() const NLIB_NOEXCEPT {
        if (type_ == kStringInline || type_ == kBinaryInline)
            return raw14size_;
        else
            return size_;
    }
    bool IsNil() const NLIB_NOEXCEPT { return GetType() == kNil; }
    bool IsBoolean() const NLIB_NOEXCEPT { return GetType() == kBoolean; }
    bool IsInteger() const NLIB_NOEXCEPT {
        ObjectType tp = GetType();
        return tp == kUint64 || tp == kInt64;
    }
    bool IsFloat() const NLIB_NOEXCEPT { return GetType() == kFloat; }
    bool IsDouble() const NLIB_NOEXCEPT { return GetType() == kDouble; }
    bool IsString() const NLIB_NOEXCEPT { return GetType() == kString; }
    bool IsBinary() const NLIB_NOEXCEPT { return GetType() == kBinary; }
    bool IsExt() const NLIB_NOEXCEPT { return GetType() == kExt; }
    bool IsArray() const NLIB_NOEXCEPT { return GetType() == kArray; }
    bool IsMap() const NLIB_NOEXCEPT { return GetType() == kMap; }

 private:
    errno_t Clone(MpObject* p) const NLIB_NOEXCEPT;
    template<class T, class IS_INTEGRAL, class IS_SIGNED>
    struct MyBox {
        static errno_t Box(MpObject* This, const T& v) { return ::nlib_ns::msgpack::Box(This, v); }
        static errno_t Unbox(const MpObject* This, T* v) {
            if (!v) return EINVAL;
            return ::nlib_ns::msgpack::Unbox(This, v);
        }
        static void Assign(MpObject* This, const T& x);
        static void Construct(MpObject* This, const T& x);
    };
    void ReleaseIfNeeded() NLIB_NOEXCEPT {
        if (type_ >= kString && type_ <= kExt) ReleaseIfNeeded_();
    }
    void ReleaseIfNeeded_() NLIB_NOEXCEPT;

 private:
    uint8_t type_;
    uint8_t raw14size_;
    uint8_t raw14data_;
    // uint8_t _;
    uint32_t size_;
    UnionType via_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObject);
    friend NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
    friend class MpObjectAsArray;
    friend class MpObjectAsMap;
    friend class MpObjectAsString;
    friend class MpObjectAsBinary;
    friend class MpObjectAsExt;
};

struct MpObjectKv NLIB_FINAL {
    MpObject first;
    MpObject second;

 public:
    MpObjectKv() NLIB_NOEXCEPT {}
    ~MpObjectKv() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    MpObjectKv(MpObjectKv&& rhs) = default;
    MpObjectKv& operator=(MpObjectKv&& rhs) = default;
#else
    MpObjectKv(MpObjectKv&& rhs) NLIB_NOEXCEPT : first(std::move(rhs.first)),
                                                 second(std::move(rhs.second)) {}
    MpObjectKv& operator=(MpObjectKv&& rhs) NLIB_NOEXCEPT {
        first  = std::move(rhs.first);
        second = std::move(rhs.second);
        return *this;
    }
#endif
#endif
    MpObjectKv(MpObjectKv& rhs, move_tag) NLIB_NOEXCEPT : first(rhs.first, move_tag()),
                                                          second(rhs.second, move_tag()) {}
    MpObjectKv& assign(MpObjectKv& rhs, move_tag) NLIB_NOEXCEPT {
        first.assign(rhs.first, move_tag());
        second.assign(rhs.second, move_tag());
        return *this;
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObjectKv);
};

class MpObjectAsArray NLIB_FINAL {
 public:
    typedef Nlist<MpObject>::iterator iterator;
    typedef Nlist<MpObject>::const_iterator const_iterator;
    typedef MpObject& reference;
    typedef const MpObject& const_reference;
    typedef MpObject* pointer;
    typedef const MpObject* const_pointer;
#if defined(_MSC_VER) && _MSC_VER == 1800
    MpObjectAsArray(const MpObjectAsArray& rhs) NLIB_NOEXCEPT : obj_(rhs.obj_) {}
    MpObjectAsArray& operator=(const MpObjectAsArray& rhs) NLIB_NOEXCEPT { obj_ = rhs.obj_; }
#endif
    explicit MpObjectAsArray(MpObject& obj) NLIB_NOEXCEPT : obj_(obj) {
        NLIB_ASSERT(obj.IsArray());
    }
    iterator begin() NLIB_NOEXCEPT { return obj_.via_.ary->begin(); }
    iterator end() NLIB_NOEXCEPT { return obj_.via_.ary->end(); }
    const_iterator begin() const NLIB_NOEXCEPT { return obj_.via_.ary->begin(); }
    const_iterator end() const NLIB_NOEXCEPT { return obj_.via_.ary->end(); }
    reference front() NLIB_NOEXCEPT { return obj_.via_.ary->front(); }
    const_reference front() const NLIB_NOEXCEPT { return obj_.via_.ary->front(); }
    reference back() NLIB_NOEXCEPT { return obj_.via_.ary->back(); }
    const_reference back() const NLIB_NOEXCEPT { return obj_.via_.ary->back(); }
    reference operator[](size_t n) NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return (*obj_.via_.ary)[n];
    }
    const_reference operator[](size_t n) const NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return (*obj_.via_.ary)[n];
    }
    size_t GetSize() const NLIB_NOEXCEPT { return obj_.size_; }
    errno_t Resize(size_t n) NLIB_NOEXCEPT {
#ifdef NLIB_64BIT
        NLIB_ASSERT(n < 0x100000000ULL);
#endif
        if (!(*obj_.via_.ary).resize(n)) return ENOMEM;
        obj_.size_ = static_cast<uint32_t>(n);
        return 0;
    }
#ifdef __cpp_rvalue_references
    errno_t PushBack(MpObject&& rhs) NLIB_NOEXCEPT {
        MpObject* p = obj_.via_.ary->push_back(std::forward<MpObject>(rhs));
        if (NLIB_UNLIKELY(!p)) return ENOMEM;
        ++obj_.size_;
        NLIB_ASSERT(obj_.size_ == obj_.via_.ary->size());
        return 0;
    }
    errno_t Insert(iterator it, MpObject&& rhs) NLIB_NOEXCEPT;
    errno_t Insert(size_t n, MpObject&& rhs) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(n > GetSize())) return ERANGE;
        iterator it = begin();
        it.Advance(n);
        return Insert(it, std::move(rhs));
    }
    errno_t Erase(iterator it) NLIB_NOEXCEPT;
    errno_t Erase(size_t n) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(n >= GetSize())) return ERANGE;
        iterator it = begin();
        it.Advance(n);
        return Erase(it);
    }
#endif
    errno_t PopBack() NLIB_NOEXCEPT {
        if (!obj_.via_.ary->pop_back()) {
            return ENOENT;
        }
        --obj_.size_;
        NLIB_ASSERT(obj_.size_ == obj_.via_.ary->size());
        return 0;
    }

 private:
    MpObject& obj_;
};
inline MpObjectAsArray MpObject::AsArray() NLIB_NOEXCEPT {
    return MpObjectAsArray(*this);
}
inline const MpObjectAsArray MpObject::AsArray() const NLIB_NOEXCEPT {
    return MpObjectAsArray(const_cast<MpObject&>(*this));
}
inline MpObject* MpObject::GetArrayItem(size_t n) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsArray())) return NULL;
    MpObjectAsArray ar = AsArray();
    if (NLIB_UNLIKELY(n >= ar.GetSize())) return nullptr;
    return &ar[n];
}
inline const MpObject* MpObject::GetArrayItem(size_t n) const NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsArray())) return NULL;
    const MpObjectAsArray ar = AsArray();
    if (NLIB_UNLIKELY(n >= ar.GetSize())) return nullptr;
    return &ar[n];
}

class MpObjectAsMap NLIB_FINAL {
 public:
    typedef Nlist<MpObjectKv>::iterator iterator;
    typedef Nlist<MpObjectKv>::const_iterator const_iterator;
    typedef MpObjectKv& reference;
    typedef const MpObjectKv& const_reference;
    typedef MpObjectKv* pointer;
    typedef const MpObjectKv* const_pointer;
#if defined(_MSC_VER) && _MSC_VER == 1800
    MpObjectAsMap(const MpObjectAsMap& rhs) NLIB_NOEXCEPT : obj_(rhs.obj_) {}
    MpObjectAsMap& operator=(const MpObjectAsMap& rhs) NLIB_NOEXCEPT { obj_ = rhs.obj_; }
#endif
    explicit MpObjectAsMap(MpObject& obj) NLIB_NOEXCEPT : obj_(obj) { NLIB_ASSERT(obj.IsMap()); }
    iterator begin() NLIB_NOEXCEPT { return obj_.via_.mp->begin(); }
    iterator end() NLIB_NOEXCEPT { return obj_.via_.mp->end(); }
    const_iterator begin() const NLIB_NOEXCEPT { return obj_.via_.mp->begin(); }
    const_iterator end() const NLIB_NOEXCEPT { return obj_.via_.mp->end(); }
    reference operator[](size_t n) NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return (*obj_.via_.mp)[n];
    }
    const_reference operator[](size_t n) const NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return (*obj_.via_.mp)[n];
    }
    iterator GetByKey(const MpObject& key) NLIB_NOEXCEPT;
    const_iterator GetByKey(const MpObject& key) const NLIB_NOEXCEPT;
    template<class StringIterator>
    iterator GetByKey(StringIterator first, StringIterator last) NLIB_NOEXCEPT;
    template<class StringIterator>
    const_iterator GetByKey(StringIterator first, StringIterator last) const NLIB_NOEXCEPT;
    iterator GetByKey(const nlib_utf8_t* key) NLIB_NOEXCEPT;
    const_iterator GetByKey(const nlib_utf8_t* key) const NLIB_NOEXCEPT;

    uint32_t GetSize() const NLIB_NOEXCEPT { return obj_.size_; }
    errno_t Resize(size_t n) NLIB_NOEXCEPT {
#ifdef NLIB_64BIT
        NLIB_ASSERT(n < 0x100000000ULL);
#endif
        if (!(*obj_.via_.mp).resize(n)) return ENOMEM;
        obj_.size_ = static_cast<uint32_t>(n);
        return 0;
    }
#ifdef __cpp_rvalue_references
    errno_t PushBack(MpObjectKv&& rhs) NLIB_NOEXCEPT {
        MpObjectKv* p = obj_.via_.mp->push_back(std::forward<MpObjectKv>(rhs));
        if (NLIB_UNLIKELY(!p)) return ENOMEM;
        ++obj_.size_;
        return 0;
    }
    errno_t PushBack(const nlib_utf8_t* key, MpObject&& value) NLIB_NOEXCEPT {
        MpObjectKv kv;
        errno_t e = kv.first.InitString(key);
        if (NLIB_UNLIKELY(e != 0)) return e;
        kv.second = std::move(value);
        return PushBack(std::move(kv));
    }
#endif
    errno_t PopBack() NLIB_NOEXCEPT {
        if (!obj_.via_.mp->pop_back()) {
            return ENOENT;
        }
        --obj_.size_;
        return 0;
    }
#ifdef __cpp_rvalue_references
    errno_t Erase(iterator it) NLIB_NOEXCEPT;
    errno_t Erase(size_t n) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(n >= GetSize())) return ERANGE;
        iterator it = begin();
        it.Advance(n);
        return Erase(it);
    }
    errno_t Erase(const nlib_utf8_t* key) NLIB_NOEXCEPT;
#endif

 private:
    MpObject& obj_;
};
inline MpObjectAsMap MpObject::AsMap() NLIB_NOEXCEPT {
    return MpObjectAsMap(*this);
}
inline const MpObjectAsMap MpObject::AsMap() const NLIB_NOEXCEPT {
    return MpObjectAsMap(const_cast<MpObject&>(*this));
}

class MpObjectAsString NLIB_FINAL {
 public:
    typedef nlib_utf8_t* iterator;
    typedef const nlib_utf8_t* const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    typedef nlib_utf8_t& reference;
    typedef const nlib_utf8_t& const_reference;
#if defined(_MSC_VER) && _MSC_VER == 1800
    MpObjectAsString(const MpObjectAsString& rhs) NLIB_NOEXCEPT : obj_(rhs.obj_) {}
    MpObjectAsString& operator=(const MpObjectAsString& rhs) NLIB_NOEXCEPT { obj_ = rhs.obj_; }
#endif
    explicit MpObjectAsString(MpObject& obj) NLIB_NOEXCEPT : obj_(obj) {
        NLIB_ASSERT(obj.IsString());
    }
    iterator begin() NLIB_NOEXCEPT { return ptr(); }
    iterator end() NLIB_NOEXCEPT { return ptr() + GetSize(); }
    const_iterator begin() const NLIB_NOEXCEPT { return ptr(); }
    const_iterator end() const NLIB_NOEXCEPT { return ptr() + GetSize(); }
    reverse_iterator rbegin() NLIB_NOEXCEPT { return std::reverse_iterator<iterator>(end()); }
    reverse_iterator rend() NLIB_NOEXCEPT { return std::reverse_iterator<iterator>(begin()); }
    const_reverse_iterator rbegin() const NLIB_NOEXCEPT {
        return std::reverse_iterator<const_iterator>(end());
    }
    const_reverse_iterator rend() const NLIB_NOEXCEPT {
        return std::reverse_iterator<const_iterator>(begin());
    }

    reference operator[](size_t n) NLIB_NOEXCEPT { return *(begin() + n); }
    const_reference operator[](size_t n) const NLIB_NOEXCEPT { return *(begin() + n); }
    reference at(size_t n) NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return *(begin() + n);
    }
    const_reference at(size_t n) const NLIB_NOEXCEPT {
        NLIB_ASSERT(n < GetSize());
        return *(begin() + n);
    }

    uint32_t GetSize() const NLIB_NOEXCEPT {
        return (obj_.type_ == MpObject::kStringInline) ? obj_.raw14size_ : obj_.size_;
    }
    size_t size() const NLIB_NOEXCEPT { return GetSize(); }
    size_t length() const NLIB_NOEXCEPT { return GetSize(); }
    size_t max_size() const NLIB_NOEXCEPT { return GetSize(); }
    bool empty() const NLIB_NOEXCEPT { return GetSize() == 0; }

    const nlib_utf8_t* c_str() const NLIB_NOEXCEPT { return ptr(); }
    const nlib_utf8_t* data() const NLIB_NOEXCEPT { return ptr(); }

 private:
    const nlib_utf8_t* ptr() const NLIB_NOEXCEPT {
        if (obj_.type_ == MpObject::kString) return obj_.via_.str;
        NLIB_ASSERT(obj_.type_ == MpObject::kStringInline);
        return reinterpret_cast<const nlib_utf8_t*>(&obj_.raw14data_);
    }
    nlib_utf8_t* ptr() NLIB_NOEXCEPT {
        if (obj_.type_ == MpObject::kString) return obj_.via_.str;
        NLIB_ASSERT(obj_.type_ == MpObject::kStringInline);
        return reinterpret_cast<nlib_utf8_t*>(&obj_.raw14data_);
    }

 private:
    MpObject& obj_;
};
inline MpObjectAsString MpObject::AsString() NLIB_NOEXCEPT {
    return MpObjectAsString(*this);
}
inline const MpObjectAsString MpObject::AsString() const NLIB_NOEXCEPT {
    return MpObjectAsString(const_cast<MpObject&>(*this));
}

class MpObjectAsBinary NLIB_FINAL {
 public:
    typedef nlib_byte_t* iterator;
    typedef const nlib_byte_t* const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    typedef nlib_byte_t& reference;
    typedef const nlib_byte_t& const_reference;

#if defined(_MSC_VER) && _MSC_VER == 1800
    MpObjectAsBinary(const MpObjectAsBinary& rhs) NLIB_NOEXCEPT : obj_(rhs.obj_) {}
    MpObjectAsBinary& operator=(const MpObjectAsBinary& rhs) NLIB_NOEXCEPT { obj_ = rhs.obj_; }
#endif
    explicit MpObjectAsBinary(MpObject& obj) NLIB_NOEXCEPT : obj_(obj) {
        NLIB_ASSERT(obj.IsBinary());
    }
    iterator begin() NLIB_NOEXCEPT { return ptr(); }
    iterator end() NLIB_NOEXCEPT { return ptr() + GetSize(); }
    const_iterator begin() const NLIB_NOEXCEPT { return ptr(); }
    const_iterator end() const NLIB_NOEXCEPT { return ptr() + GetSize(); }
    reverse_iterator rbegin() NLIB_NOEXCEPT { return std::reverse_iterator<iterator>(end()); }
    reverse_iterator rend() NLIB_NOEXCEPT { return std::reverse_iterator<iterator>(begin()); }
    const_reverse_iterator rbegin() const NLIB_NOEXCEPT {
        return std::reverse_iterator<const_iterator>(end());
    }
    const_reverse_iterator rend() const NLIB_NOEXCEPT {
        return std::reverse_iterator<const_iterator>(begin());
    }
    reference operator[](size_t n) NLIB_NOEXCEPT { return *(begin() + n); }
    const_reference operator[](size_t n) const NLIB_NOEXCEPT { return *(begin() + n); }
    uint32_t GetSize() const NLIB_NOEXCEPT {
        return (obj_.type_ == MpObject::kBinaryInline) ? obj_.raw14size_ : obj_.size_;
    }

 private:
    const nlib_byte_t* ptr() const NLIB_NOEXCEPT {
        if (obj_.type_ == MpObject::kBinary)
            return static_cast<const nlib_byte_t*>(obj_.via_.binary);
        NLIB_ASSERT(obj_.type_ == MpObject::kBinaryInline);
        return reinterpret_cast<const nlib_byte_t*>(&obj_.raw14data_);
    }
    nlib_byte_t* ptr() NLIB_NOEXCEPT {
        if (obj_.type_ == MpObject::kBinary) return static_cast<nlib_byte_t*>(obj_.via_.binary);
        NLIB_ASSERT(obj_.type_ == MpObject::kBinaryInline);
        return reinterpret_cast<nlib_byte_t*>(&obj_.raw14data_);
    }

 private:
    MpObject& obj_;
};
inline MpObjectAsBinary MpObject::AsBinary() NLIB_NOEXCEPT {
    return MpObjectAsBinary(*this);
}
inline const MpObjectAsBinary MpObject::AsBinary() const NLIB_NOEXCEPT {
    return MpObjectAsBinary(const_cast<MpObject&>(*this));
}
inline void* MpObject::GetBinary(uint32_t* n) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsBinary())) return nullptr;
    MpObjectAsBinary bin(*this);
    *n = bin.GetSize();
    return &*bin.begin();
}
#ifdef __cpp_rvalue_references
inline std::tuple<errno_t, const nlib_byte_t*, uint32_t> MpObject::GetBinary() const NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsBinary())) return std::make_tuple(EACCES, nullptr, 0);
    MpObjectAsBinary bin = AsBinary();
    return std::make_tuple(0, &*bin.begin(), bin.GetSize());
}
#endif

class MpObjectAsExt NLIB_FINAL {
 public:
#if defined(_MSC_VER) && _MSC_VER == 1800
    MpObjectAsExt(const MpObjectAsExt& rhs) NLIB_NOEXCEPT : obj_(rhs.obj_) {}
    MpObjectAsExt& operator=(const MpObjectAsExt& rhs) NLIB_NOEXCEPT { obj_ = rhs.obj_; }
#endif
    explicit MpObjectAsExt(MpObject& obj) NLIB_NOEXCEPT : obj_(obj) { NLIB_ASSERT(obj.IsExt()); }
    void* GetExtData() NLIB_NOEXCEPT { return obj_.via_.binary; }
    const void* GetExtData() const NLIB_NOEXCEPT { return obj_.via_.binary; }
    int8_t GetExtType() const NLIB_NOEXCEPT { return static_cast<int8_t>(obj_.raw14data_); }
    uint32_t GetSize() const NLIB_NOEXCEPT { return obj_.size_; }

 private:
    MpObject& obj_;
};
inline MpObjectAsExt MpObject::AsExt() NLIB_NOEXCEPT {
    return MpObjectAsExt(*this);
}
inline const MpObjectAsExt MpObject::AsExt() const NLIB_NOEXCEPT {
    return MpObjectAsExt(const_cast<MpObject&>(*this));
}
inline void* MpObject::GetExt(int8_t* tp, uint32_t* n) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsExt())) return nullptr;
    MpObjectAsExt ext(*this);
    *n  = ext.GetSize();
    *tp = ext.GetExtType();
    return ext.GetExtData();
}
#ifdef __cpp_rvalue_references
inline std::tuple<errno_t, int8_t, const nlib_byte_t*, uint32_t>
MpObject::GetExt() const NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(!IsExt())) return std::make_tuple(EACCES, static_cast<int8_t>(0), nullptr, 0);
    MpObjectAsExt ext = AsExt();
    return std::make_tuple(0, ext.GetExtType(), static_cast<const nlib_byte_t*>(ext.GetExtData()),
                           ext.GetSize());
}
#endif

template<class StringIterator>
MpObjectAsMap::iterator
MpObjectAsMap::GetByKey(StringIterator first, StringIterator last) NLIB_NOEXCEPT {
    MpObjectAsMap::iterator itend = end();
    MpObjectAsMap::iterator it    = begin();
    for (; it != itend; ++it) {
        if (!it->first.IsString()) continue;
        uint32_t len = it->first.GetSize();
        if (len != last - first) continue;
        const nlib_utf8_t* k = it->first.GetString();
        while (first != last) {
            if (*k++ != *first++) break;
        }
        if (first == last) break;
    }
    return it;
}
template<class StringIterator>
MpObjectAsMap::const_iterator
MpObjectAsMap::GetByKey(StringIterator first, StringIterator last) const NLIB_NOEXCEPT {
    MpObjectAsMap::const_iterator itend = end();
    MpObjectAsMap::const_iterator it    = begin();
    for (; it != itend; ++it) {
        if (!it->first.IsString()) continue;
        uint32_t len = it->first.GetSize();
        if (len != last - first) continue;
        const nlib_utf8_t* k = it->first.GetString();
        while (first != last) {
            if (*k++ != *first++) break;
        }
        if (first == last) break;
    }
    return it;
}

#ifdef __cpp_rvalue_references
inline std::pair<errno_t, MpObjectKv> MpObject::RemoveMapItem(size_t n) NLIB_NOEXCEPT {
    std::pair<errno_t, MpObjectKv> rval;
    rval.first = RemoveMapItem(n, &rval.second);
    return rval;
}
#endif

template<>
struct MpObject::MyBox<nil, FalseType, FalseType> {
    static errno_t Box(MpObject* This, nil v) NLIB_NOEXCEPT {
        NLIB_UNUSED(v);
        This->ReleaseIfNeeded();
        This->type_ = kNil;
        return 0;
    }
    static void Assign(MpObject* This, nil x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, nil) NLIB_NOEXCEPT { This->type_ = kNil; }
};

template<>
struct MpObject::MyBox<bool, TrueType, FalseType> {
    static errno_t Box(MpObject* This, bool v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = v ? static_cast<uint8_t>(kBooleanTrue) : static_cast<uint8_t>(kBooleanFalse);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, bool* v) NLIB_NOEXCEPT {
        if (This->type_ == kBooleanFalse) {
            *v = false;
            return 0;
        } else if (NLIB_LIKELY(This->type_ == kBooleanTrue)) {
            *v = true;
            return 0;
        }
        *v = false;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, bool x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, bool x) NLIB_NOEXCEPT {
        This->type_ = x ? static_cast<uint8_t>(kBooleanTrue) : static_cast<uint8_t>(kBooleanFalse);
    }
};

template<class T>
struct MpObject::MyBox<T, TrueType, TrueType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_    = kInt64;
        This->via_.i64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->type_ == kInt64) {
            *v = static_cast<T>(This->via_.i64);
            return 0;
        } else if (NLIB_LIKELY(This->type_ == kUint64)) {
            *v = static_cast<T>(This->via_.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->type_    = kInt64;
        This->via_.i64 = x;
    }
};

template<class T>
struct MpObject::MyBox<T, TrueType, FalseType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_    = kUint64;
        This->via_.u64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->type_ == kInt64) {
            *v = static_cast<T>(This->via_.i64);
            return 0;
        } else if (NLIB_LIKELY(This->type_ == kUint64)) {
            *v = static_cast<T>(This->via_.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->type_    = kUint64;
        This->via_.u64 = x;
    }
};

template<class TR1_WORKAROUND>
struct MpObject::MyBox<float, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, float v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_    = kFloat;
        This->via_.f32 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, float* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->type_) {
        case kFloat:
            *v = This->via_.f32;
            break;
        case kDouble:
            *v = static_cast<float>(This->via_.f64);
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, float x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, float x) NLIB_NOEXCEPT {
        This->type_    = kFloat;
        This->via_.f32 = x;
    }
};

template<class TR1_WORKAROUND>
struct MpObject::MyBox<double, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, double v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_    = kDouble;
        This->via_.f64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, double* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->type_) {
        case kFloat:
            *v = This->via_.f32;
            break;
        case kDouble:
            *v = This->via_.f64;
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, double x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, double x) NLIB_NOEXCEPT {
        This->type_    = kDouble;
        This->via_.f64 = x;
    }
};

template<>
struct MpObject::MyBox<std::string, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::string& str) NLIB_NOEXCEPT {
        uint32_t n = static_cast<uint32_t>(str.size());
        ErrnoT e   = This->InitString(n);
        if (NLIB_UNLIKELY(e != 0)) return e;
        nlib_memcpy(static_cast<void*>(This->GetString()), n, static_cast<const void*>(&str[0]), n);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::string* str) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!str)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kString) && NLIB_UNLIKELY(This->type_ != kStringInline))
            return EACCES;
        const nlib_utf8_t* chr = This->GetString();
        NLIB_TRY { (*str).assign(chr, This->GetSize()); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        return 0;
    }
};

template<class T, class ALLOC>
struct MpObject::MyBox<std::vector<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::vector<T, ALLOC>& vec) {
        uint32_t n = static_cast<uint32_t>(vec.size());
        MpObject tmp;
        ErrnoT e = tmp.InitArray(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        for (uint32_t i = 0; i < n; ++i) {
            e = (*tmp.via_.ary)[i].Box(vec[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::vector<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kArray)) return EACCES;
        NLIB_TRY { vec->resize(This->size_); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        for (uint32_t i = 0; i < This->size_; ++i) {
            ErrnoT e = (*This->via_.ary)[i].Unbox(&(*vec)[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template<class T, class ALLOC>
struct MpObject::MyBox<Nlist<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const Nlist<T, ALLOC>& vec) {
        size_t n = vec.size();
        MpObject tmp;
        ErrnoT e = tmp.InitArray(static_cast<uint32_t>(n));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typename Nlist<T, ALLOC>::const_iterator it = vec.begin();
        for (size_t i = 0; i < n; ++it, ++i) {
            e = (*tmp.via_.ary)[i].Box(*it);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, Nlist<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kArray)) return EACCES;
        vec->clear();
        if (NLIB_UNLIKELY(!vec->reserve(This->size_))) return ENOMEM;
        for (uint32_t i = 0; i < This->size_; ++i) {
            T* item  = vec->push_back();
            ErrnoT e = (*This->via_.ary)[i].Unbox(item);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template<class T1, class T2>
struct MpObject::MyBox<std::pair<T1, T2>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::pair<T1, T2>& p) {
        MpObject tmp;
        ErrnoT e = tmp.InitArray(2);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.via_.ary)[0].Box(p.first);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.via_.ary)[1].Box(p.second);
        if (NLIB_UNLIKELY(e != 0)) return e;
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::pair<T1, T2>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kArray)) return EACCES;
        if (NLIB_UNLIKELY(This->size_ != 2)) return EACCES;
        ErrnoT e;
        e = (*This->via_.ary)[0].Unbox(&p->first);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        e = (*This->via_.ary)[1].Unbox(&p->second);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        return 0;
    }
};

#if defined(__cpp_rvalue_references) && defined(__cpp_variadic_templates)
template<class... Args>
struct MpObject::MyBox<std::tuple<Args...>, FalseType, FalseType> {
 private:
    template<size_t N, class HEAD>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head) {
        return (*tmp->via_.ary)[N].Box(head);
    }
    template<size_t N, class HEAD, class... TAIL>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head, TAIL&... tail) {
        ErrnoT e = (*tmp->via_.ary)[N].Box(head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return BoxHelper<N + 1, TAIL...>(tmp, tail...);
    }
    template<size_t N, class HEAD>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head) {
        return (*This->via_.ary)[N].Unbox(&head);
    }
    template<size_t N, class HEAD, class... TAIL>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head, TAIL&... tail) {
        ErrnoT e = (*This->via_.ary)[N].Unbox(&head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return UnboxHelper<N + 1, TAIL...>(This, tail...);
    }
    template<int...>
    struct seq {};
    template<int N, int... S>
    struct gens : gens<N - 1, N - 1, S...> {};
    template<int... S>
    struct gens<0, S...> {
        typedef seq<S...> type;
    };
    template<int... S>
    static errno_t Box(MpObject* tmp, seq<S...>, const std::tuple<Args...>& p) {
        return BoxHelper<0>(tmp, std::get<S>(p)...);
    }
    template<int... S>
    static errno_t Unbox(const MpObject* This, seq<S...>, std::tuple<Args...>& p) {
        return UnboxHelper<0>(This, std::get<S>(p)...);
    }

 public:
    static errno_t Box(MpObject* This, const std::tuple<Args...>& p) {
        size_t count = std::tuple_size<std::tuple<Args...> >::value;
        MpObject tmp;
        ErrnoT e = tmp.InitArray(static_cast<uint32_t>(count));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        e = Box(&tmp, MySeq(), p);
        if (NLIB_UNLIKELY(e != 0)) return e;
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::tuple<Args...>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kArray)) return EACCES;
        if (NLIB_UNLIKELY(This->size_ != std::tuple_size<std::tuple<Args...> >::value))
            return EACCES;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        return Unbox(This, MySeq(), *p);
    }
};
#endif

template<class K, class V, class Pr, class Alloc>
struct MpObject::MyBox<std::map<K, V, Pr, Alloc>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::map<K, V, Pr, Alloc>& m) {
        typedef typename std::map<K, V, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        ErrnoT e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.via_.mp)[i];
            e              = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::map<K, V, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kMap)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->size_; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->via_.mp)[i];
            ErrnoT e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};

#ifdef __cpp_rvalue_references
template<class K, class V, class Hash, class Pr, class Alloc>
struct MpObject::MyBox<std::unordered_map<K, V, Hash, Pr, Alloc>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::unordered_map<K, V, Hash, Pr, Alloc>& m) {
        typedef typename std::unordered_map<K, V, Hash, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        ErrnoT e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.via_.mp)[i];
            e              = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        This->assign(tmp, move_tag());
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::unordered_map<K, V, Hash, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != kMap)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->size_; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->via_.mp)[i];
            ErrnoT e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};
#endif

NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
inline bool operator==(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return lhs.first == rhs.first && lhs.second == rhs.second;
}
inline bool operator!=(const MpObject& lhs, const MpObject& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}
inline bool operator!=(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}

template<uint32_t n>
errno_t MpObject::Box(const nlib_utf8_t (&str)[n]) NLIB_NOEXCEPT {
    uint32_t nn = static_cast<uint32_t>(nlib_strlen(str));
    ErrnoT e    = this->InitString(nn);
    if (NLIB_UNLIKELY(e != 0)) return e;
    nlib_memcpy(this->GetString(), nn, static_cast<const void*>(&str[0]), nn);
    return 0;
}

template<class T, uint32_t n>
errno_t MpObject::Box(const T (&vec)[n]) {
    MpObject tmp;
    ErrnoT e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.via_.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->assign(tmp, move_tag());
    return 0;
}

#ifdef __cpp_rvalue_references
template<class T, size_t n>
errno_t MpObject::Box(const std::array<T, n>& vec) {
    MpObject tmp;
    ErrnoT e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.via_.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->assign(tmp, move_tag());
    return 0;
}
#endif

template<class T>
errno_t MpObject::Unbox(T* a, size_t n) const {
    NLIB_EINVAL_IFNULL(a);
    if (NLIB_UNLIKELY(n == 0)) return EINVAL;
    if (NLIB_UNLIKELY(type_ != kArray)) return EACCES;
    if (NLIB_UNLIKELY(n != size_)) return ERANGE;
    for (uint32_t i = 0; i < size_; ++i) {
        ErrnoT e = (*via_.ary)[i].Unbox(&a[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    return 0;
}

}  // namespace msgpack
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObject)
NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObjectKv)

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
