﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_BP_H_
#define INCLUDE_NN_NLIB_SUCCINCT_BP_H_

#include "nn/nlib/succinct/Sbv.h"
#include "nn/nlib/ReallocVec.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class NLIB_VIS_PUBLIC Bp NLIB_FINAL {
 public:
    typedef uint32_t NodeId;
    typedef uint32_t Pos;
    NLIB_CEXPR Bp() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~Bp() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(Bp);
    template<class TREEITER>
    bool Init(TREEITER& root, uint32_t num_nodes) NLIB_NOEXCEPT;
    bool InitDirect(const uint32_t* bv, uint32_t bv_size) NLIB_NOEXCEPT;
    int ToNodeId(Pos pos) const NLIB_NOEXCEPT;
    int ToPos(NodeId nodeid) const NLIB_NOEXCEPT;
    int FindOpen(Pos pos) const NLIB_NOEXCEPT;
    int FindClose(Pos pos) const NLIB_NOEXCEPT;
    int Enclose(Pos pos) const NLIB_NOEXCEPT;
    int FirstChild(Pos pos) const NLIB_NOEXCEPT;
    int LastChild(Pos pos) const NLIB_NOEXCEPT;
    int NextSibling(Pos pos) const NLIB_NOEXCEPT;
    int PrevSibling(Pos pos) const NLIB_NOEXCEPT;
    int Parent(Pos pos) const NLIB_NOEXCEPT { return this->Enclose(pos); }

    std::pair<bool, NodeId> GetNodeIdByPos(Pos pos) const NLIB_NOEXCEPT {
        int r = ToNodeId(pos);
        return std::make_pair((r >= 0), static_cast<NodeId>(r));
    }
    std::pair<bool, Pos> GetPosByNodeId(NodeId nodeid) const NLIB_NOEXCEPT {
        int r = ToPos(nodeid);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }

    std::pair<bool, Pos> GetPosOfFirstChild(Pos pos) const NLIB_NOEXCEPT {
        int r = FirstChild(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfLastChild(Pos pos) const NLIB_NOEXCEPT {
        int r = LastChild(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfNextSibling(Pos pos) const NLIB_NOEXCEPT {
        int r = NextSibling(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfPrevSibling(Pos pos) const NLIB_NOEXCEPT {
        int r = PrevSibling(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfParent(Pos pos) const NLIB_NOEXCEPT {
        int r = Parent(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }

    std::pair<bool, NodeId> GetNodeIdOfFirstChild(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = FirstChild(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfLastChild(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = LastChild(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfNextSibling(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = NextSibling(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfPrevSibling(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = PrevSibling(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfParent(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, uint32_t> pos = GetPosByNodeId(nodeid);
        int r                         = Parent(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }

    size_t MemSize() const NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    Set* Get1stLevelSet() NLIB_NOEXCEPT;
    bool Build(const uint32_t* bv, uint32_t bvsize) NLIB_NOEXCEPT;

 private:
    struct BpPrivate;
    BpPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Bp);
};

template<class TREEITER>
bool Bp::Init(TREEITER& root, uint32_t num_nodes) NLIB_NOEXCEPT {
    TREEITER& iter = root;
    // TREEITER has following methods:
    //   const void* NodePtr();        // ptr to node
    //   const void* ParentNodePtr();  // ptr to parent node
    //   bool MoveNext();  // move to the next node by depth first search
    if (num_nodes >= 0x80000000) return false;
    if (prv_) return false;
    Set* bv = Get1stLevelSet();  // bv = &bp_.b_
    if (!bv) {
        this->Reset();
        return false;
    }
    bv->Init(num_nodes * 2);

    uint32_t idx = 0;
    ReallocVec<const void*> stack;
    do {
        const void* parent = iter.ParentNodePtr();
        const void* node   = iter.NodePtr();
        if (stack.empty() || stack.back() == parent) {
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) {
                this->Reset();
                return false;
            }
        } else {
            while (stack.back() != parent) {
                stack.pop_back();
                ++idx;
                if (stack.empty()) {
                    // Has multiple root nodes
                    this->Reset();
                    return false;
                }
            }
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) {
                this->Reset();
                return false;
            }
        }
    } while (iter.MoveNext());

    if (!this->Build(bv->GetBitVector(), bv->GetBitVectorSize())) {
        this->Reset();
        return false;
    }
    return true;
}

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Bp)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_BP_H_
