﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_
#define INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_

#include <utility>
#include "nn/nlib/Swap.h"
#include "nn/nlib/succinct/Sbv.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class NLIB_VIS_PUBLIC Louds NLIB_FINAL {
 public:
    typedef uint32_t NodeId;
    typedef uint32_t Pos;
    NLIB_CEXPR Louds() NLIB_NOEXCEPT {}
    ~Louds() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
    Louds(Louds&& rhs) NLIB_NOEXCEPT : sbv_(std::move(rhs.sbv_)) {}
    Louds& operator=(Louds&& rhs) NLIB_NOEXCEPT {
        sbv_ = std::move(rhs.sbv_);
        return *this;
    }
#endif
    Louds(Louds& rhs, move_tag) NLIB_NOEXCEPT : sbv_(rhs.sbv_, move_tag()) {}
    Louds& assign(Louds& rhs, move_tag) NLIB_NOEXCEPT {
        sbv_.assign(rhs.sbv_, move_tag());
        return *this;
    }
    template<class TREEITER>
    bool Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT;

    bool InitDirect(const uint32_t* bv, uint32_t bv_size) NLIB_NOEXCEPT;
    int ToNodeId(Pos pos) const NLIB_NOEXCEPT;
    int ToPos(NodeId nodeid) const NLIB_NOEXCEPT;
    std::pair<bool, NodeId> GetNodeIdByPos(Pos pos) const NLIB_NOEXCEPT {
        int r = ToNodeId(pos);
        return std::make_pair((r >= 0), static_cast<NodeId>(r));
    }
    std::pair<bool, Pos> GetPosByNodeId(NodeId nodeid) const NLIB_NOEXCEPT {
        int r = ToPos(nodeid);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }

    int FirstChild(Pos pos) const NLIB_NOEXCEPT;
    int NextSibling(Pos pos) const NLIB_NOEXCEPT;
    int PrevSibling(Pos pos) const NLIB_NOEXCEPT;
    int Parent(Pos pos) const NLIB_NOEXCEPT;

    std::pair<bool, Pos> GetPosOfFirstChild(Pos pos) const NLIB_NOEXCEPT {
        int r = FirstChild(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfNextSibling(Pos pos) const NLIB_NOEXCEPT {
        int r = NextSibling(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfPrevSibling(Pos pos) const NLIB_NOEXCEPT {
        int r = PrevSibling(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }
    std::pair<bool, Pos> GetPosOfParent(Pos pos) const NLIB_NOEXCEPT {
        int r = Parent(pos);
        return std::make_pair((r >= 0), static_cast<Pos>(r));
    }

    std::pair<bool, NodeId> GetNodeIdOfFirstChild(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = FirstChild(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfNextSibling(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = NextSibling(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfPrevSibling(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, NodeId> pos = GetPosByNodeId(nodeid);
        int r                       = PrevSibling(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }
    std::pair<bool, NodeId> GetNodeIdOfParent(NodeId nodeid) const NLIB_NOEXCEPT {
        std::pair<bool, uint32_t> pos = GetPosByNodeId(nodeid);
        int r                         = Parent(pos.second);
        return GetNodeIdByPos(static_cast<Pos>(r));
    }

    bool IsLeaf(Pos pos) const NLIB_NOEXCEPT { return this->FirstChild(pos) < 0; }
    size_t MemSize() const NLIB_NOEXCEPT { return sbv_.MemSize(); }
    void Reset() NLIB_NOEXCEPT { sbv_.Reset(); }
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT { return sbv_.Export(w); }
    bool Import(BinaryReader* r) NLIB_NOEXCEPT { return sbv_.Import(r); }

 private:
    Sbv sbv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Louds);
};

template<class TREEITER>
bool Louds::Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT {
    // TREEITER has following methods:
    //   int GetNumChildren();
    //   bool BfsMoveNext();  // move to the next node by breadth first search
    sbv_.Init(2 * num_nodes + 1);

    // diffrent from the Louds paper: reversed bits
    sbv_.TurnOn(1);
    uint32_t idx = 2;  // consider super-root 01
    do {
        idx += iter.GetNumChildren();
        sbv_.TurnOn(idx++);
    } while (iter.BfsMoveNext());

    NLIB_ASSERT(idx == 2 * num_nodes + 1);
    if (!sbv_.Build()) {
        sbv_.Reset();
        return false;
    }
    return true;
}

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Louds)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_
