﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_TRIE_H_
#define INCLUDE_NN_NLIB_SUCCINCT_TRIE_H_

#include <string.h>
#include <utility>

#include "nn/nlib/succinct/Louds.h"
#include "nn/nlib/succinct/Sbv.h"
#include "nn/nlib/ReallocVec.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/Nlist.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class NLIB_VIS_PUBLIC Trie NLIB_FINAL {
 public:
    NLIB_CEXPR Trie() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~Trie() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(Trie);
    bool Init() NLIB_NOEXCEPT;
    typedef bool (*MatchCallback)(const char* first, const char* last, uint32_t nodeid,
                                  void* user_obj);
    void Match(const char* cstr, MatchCallback callback, void* user_obj) NLIB_NOEXCEPT {
        this->Match(cstr, nlib_strlen(cstr), callback, user_obj);
    }
    void Match(const char* cstr, MatchCallback callback) NLIB_NOEXCEPT {
        this->Match(cstr, nlib_strlen(cstr), callback, nullptr);
    }
    void Match(const void* data, size_t n, MatchCallback callback, void* user_obj) NLIB_NOEXCEPT;
    void Match(const void* data, size_t n, MatchCallback callback) NLIB_NOEXCEPT {
        Match(data, n, callback, nullptr);
    }

    void
    MatchBackward(const void* data, size_t n, MatchCallback callback, void* user_obj) NLIB_NOEXCEPT;
    void MatchBackward(const void* data, size_t n, MatchCallback callback) NLIB_NOEXCEPT {
        MatchBackward(data, n, callback, nullptr);
    }
    errno_t
    GetCommonPrefixWords(const void* prefix, size_t n, ReallocCstringVec* vec) NLIB_NOEXCEPT;
#ifdef __cpp_rvalue_references
    std::pair<errno_t, Nlist<uint32_t> >
    GetCommonPrefixWords(const void* prefix, size_t n) NLIB_NOEXCEPT;
#endif

    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;
    size_t MemSize() const NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN int Match_(const nlib_byte_t* data, const unsigned char* p, uint32_t pos,
                               MatchCallback callback, void* user_obj, bool isfwd) NLIB_NOEXCEPT;

 private:
    struct TriePrivate;
    TriePrivate* prv_;
    friend class TrieBuilder;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Trie);
};

class NLIB_VIS_PUBLIC TrieBuilder NLIB_FINAL {
 public:
    NLIB_CEXPR TrieBuilder() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~TrieBuilder() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(TrieBuilder);
    void Reset() NLIB_NOEXCEPT;
    bool Init() NLIB_NOEXCEPT;
    Trie* Build(ReallocVec<uint32_t>* keyids = nullptr) NLIB_NOEXCEPT;
#ifdef __cpp_rvalue_references
    std::unique_ptr<Trie> Build2() NLIB_NOEXCEPT { return decltype(Build2())(Build(nullptr)); }
#endif
    bool AddWord(const char* str) NLIB_NOEXCEPT;
    bool AddPattern(const void* p, size_t n) NLIB_NOEXCEPT;
    bool AddWords(const char* str, size_t len) NLIB_NOEXCEPT;
    bool AddWords(const char* str) NLIB_NOEXCEPT { return AddWords(str, strlen(str)); }

 private:
    struct TrieBuilderPrivate;
    TrieBuilderPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(TrieBuilder);
};

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Trie)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::TrieBuilder)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_TRIE_H_
