﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_PRINTER_H_
#define INCLUDE_NN_NLIB_TESTING_PRINTER_H_

#include <string>
#include <utility>
#include <vector>

#include "nn/nlib/Config.h"

#if defined(_MSC_VER) && defined(nx_testing_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace testing {
namespace detail_func {

class NLIB_VIS_PUBLIC DummyStream NLIB_FINAL {
 public:
    DummyStream() : buf_(nullptr), cur_(0U), cap_(0U) {}
    ~DummyStream();
    const char* c_str() const { return buf_ ? buf_ : ""; }
#ifdef __cpp_rvalue_references
    DummyStream(DummyStream&& rhs) NLIB_NOEXCEPT : buf_(rhs.buf_), cur_(rhs.cur_), cap_(rhs.cap_) {
        rhs.buf_ = nullptr;
        rhs.cur_ = rhs.cap_ = 0U;
    }
    DummyStream& operator=(DummyStream&& rhs) NLIB_NOEXCEPT;
#endif
    DummyStream(DummyStream& rhs, move_tag) NLIB_NOEXCEPT : buf_(rhs.buf_),
                                                            cur_(rhs.cur_),
                                                            cap_(rhs.cap_) {
        rhs.buf_ = nullptr;
        rhs.cur_ = rhs.cap_ = 0U;
    }
    DummyStream& assign(DummyStream& rhs, move_tag) NLIB_NOEXCEPT;

 public:
    DummyStream& append(const char* str);

    DummyStream& operator<<(char c);
    DummyStream& operator<<(signed char c) {
        *this << static_cast<int>(c);
        return *this;
    }
    DummyStream& operator<<(unsigned char c) {
        *this << static_cast<int>(c);
        return *this;
    }
    DummyStream& operator<<(wchar_t c);
    DummyStream& operator<<(int value);
    DummyStream& operator<<(unsigned int value);
    DummyStream& operator<<(long value);
    DummyStream& operator<<(unsigned long value);
    DummyStream& operator<<(long long value);
    DummyStream& operator<<(unsigned long long value);
    DummyStream& operator<<(float value);
    DummyStream& operator<<(double value);
    DummyStream& operator<<(bool value);
    DummyStream& operator<<(const char* str);
    DummyStream& operator<<(const wchar_t* str);
    DummyStream& operator<<(const nlib_utf16_t* str);
    DummyStream& operator<<(const nlib_utf32_t* str);
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    DummyStream& operator<<(const std::string& str) {
        *this << str.c_str();
        return *this;
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    DummyStream& operator<<(const std::wstring& str) {
        *this << str.c_str();
        return *this;
    }
    template<class T>
    DummyStream& operator<<(const T& rhs) {
        PrintTo(rhs, this);
        return *this;
    }

 private:
    char* buf_;
    uint32_t cur_;
    uint32_t cap_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(DummyStream);
};

template<class T>
void PrintToArray(const T* a, size_t n, DummyStream* ostr) {
    if (n == 0) {
        *ostr << "{}";
    } else {
        *ostr << "{" << a[0];
        size_t nn = (n > 8) ? 8 : n;
        for (size_t i = 1; i < nn; ++i) {
            *ostr << ", " << a[i];
        }
        if (nn == n) {
            *ostr << "}";
        } else {
            *ostr << ", ...}";
        }
    }
}

template<class T, size_t N>
inline void PrintTo(const T (&a)[N], DummyStream* ostr) {
    PrintToArray(&a[0], N, ostr);
}

template<size_t N>
inline void PrintTo(const char (&str)[N], DummyStream* ostr) {
    *ostr << str;
}

template<size_t N>
inline void PrintTo(const wchar_t (&str)[N], DummyStream* ostr) {
    *ostr << str;
}

template<class T, class Alloc>
inline void PrintTo(const std::vector<T, Alloc>& value, DummyStream* ostr) {
    PrintToArray(&value[0], value.size(), ostr);
}

template<class T1, class T2>
void PrintTo(const std::pair<T1, T2>& value, DummyStream* ostr) {
    PrintTo("(", ostr);
    PrintTo(value.first, ostr);
    PrintTo(", ", ostr);
    PrintTo(value.second, ostr);
    PrintTo(")", ostr);
}

template<class T>
inline void PrintTo(const T& value, DummyStream* ostr) {
    NLIB_UNUSED(value);
    *ostr << "?????";
}

template<class T, class DUMMY>
inline void PrintTo(const T& value, DUMMY* ostr) {
    // NOTE:
    // DUMMY must be a template, not DummyStream.
    // It is to use user-defined PrintTo(UserType, STREAM* ostr) prior to this function.
    //
    // Though you can write as follows using gtest:
    //   void PrintTo(const Bar& bar, std::ostream* os) { ... }
    // using nlib::testing, we have to write like this:
    //   template<class OSTREAM> void PrintTo(const Bar& bar, OSTREAM* os) { ... }
    *ostr << value;
}

}  // namespace detail_func

template<class T>
inline std::string PrintToString(const T& value) {
    // NOTE:
    // we have to make nlib::testing work if we cannot use std::ostream.
    ::nlib_ns::testing::detail_func::DummyStream str;

    // NOTE:
    // ADL is performed, if PrintTo for type T is defined.
    // otherwise, ::nlib_ns::testing::detail_func::PrintTo is used.
    using ::nlib_ns::testing::detail_func::PrintTo;
    PrintTo(value, &str);
    return std::string(str.c_str());
}

}  // namespace testing
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_testing_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_TESTING_PRINTER_H_
