﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_CONDVAR_H_
#define INCLUDE_NN_NLIB_THREADING_CONDVAR_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/DateTime.h"
#include "nn/nlib/threading/CriticalSection.h"

#include <new>

NLIB_NAMESPACE_BEGIN
namespace threading {

#ifndef NLIB_CONDVAR_USE_FALLBACK
// code snippets:
//  Initialization:
//    CondVar cond;
//    SimpleCriticalSection m;
//    bool flag = false;
//  Wating thread:
//    m.lock();
//    while (!flag)
//      if (cond.Wait(m) != 0) { error; .... }
//      # note that cond.Wait() may return without signal notified
//    flag = false;
//    m.unlock();
//  Notifying thread:
//    m.lock();
//    flag = true;
//    cond.NotifyAll();
//    m.unlock();
class CondVar NLIB_FINAL {
 public:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    constexpr CondVar() NLIB_NOEXCEPT = default;
#else
    CondVar() NLIB_NOEXCEPT {
        errno_t e = nlib_cond_init(&cond_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~CondVar() NLIB_NOEXCEPT {
        errno_t e = nlib_cond_destroy(&cond_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    void Notify() NLIB_NOEXCEPT {
        errno_t e;
        e = nlib_cond_signal(&cond_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void NotifyAll() NLIB_NOEXCEPT {
        errno_t e;
        e = nlib_cond_broadcast(&cond_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }

    template<class lock_type>
    errno_t Wait(lock_type& lock) NLIB_NOEXCEPT NLIB_REQUIRES(lock) {
        errno_t e;
        e = nlib_cond_wait(&cond_, detail::GetRawMutex(lock));
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
        return 0;
    }
    template<class lock_type>
    NLIB_CHECK_RESULT errno_t WaitFor(lock_type& lock,
                                      const nlib_ns::TimeSpan& timeout) NLIB_NOEXCEPT
        NLIB_REQUIRES(lock) {
        return nlib_cond_wait_for(&cond_, detail::GetRawMutex(lock), timeout.ToTimeValue().tick);
    }
    template<class lock_type>
    NLIB_CHECK_RESULT errno_t WaitUntil(lock_type& lock,
                                        const nlib_ns::DateTime& datetime) NLIB_NOEXCEPT
        NLIB_REQUIRES(lock) {
        return nlib_cond_wait_until(&cond_, detail::GetRawMutex(lock), datetime.ToTimeValue().tick);
    }

 private:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    nlib_cond cond_ = NLIB_COND_INITIALIZER;
#else
    nlib_cond cond_;
#endif
    NLIB_DISALLOW_COPY_AND_ASSIGN(CondVar);
};

#ifndef NLIB_DOXYGEN
template<>
NLIB_CHECK_RESULT inline errno_t
CondVar::WaitFor(SimpleCriticalSection& lock, const nlib_ns::TimeSpan& timeout) NLIB_NOEXCEPT
    NLIB_REQUIRES(lock) {
    return nlib_cond_wait_for(&cond_, detail::GetRawMutex(lock), timeout.ToTimeValue().tick);
}

template<>
NLIB_CHECK_RESULT inline errno_t
CondVar::WaitUntil(SimpleCriticalSection& lock, const nlib_ns::DateTime& datetime) NLIB_NOEXCEPT
    NLIB_REQUIRES(lock) {
    return nlib_cond_wait_until(&cond_, detail::GetRawMutex(lock), datetime.ToTimeValue().tick);
}
#endif

#endif

}  // namespace threading
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_THREADING_CONDVAR_H_
