﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_FUTURE_H_
#define INCLUDE_NN_NLIB_THREADING_FUTURE_H_

#include <utility>

#include "nn/nlib/Swap.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/DateTime.h"
#include "nn/nlib/TypeTraits.h"
#include "nn/nlib/threading/CriticalSection.h"
#include "nn/nlib/threading/LimitedSharedPtr.h"
#include "nn/nlib/threading/Thread.h"

NLIB_NAMESPACE_BEGIN
class TimeSpan;
class DateTime;
namespace threading {

template<class FutureResult>
inline errno_t GetFutureError(FutureResult& result) NLIB_NOEXCEPT {
    NLIB_UNUSED(result);
    return 0;
}

template<class R>
class Future;
template<class R>
class Promise;

namespace detail {

class NLIB_VIS_PUBLIC FutureContinuationBase {
 public:
    FutureContinuationBase() NLIB_NOEXCEPT {}
    virtual ~FutureContinuationBase() NLIB_NOEXCEPT {}
    virtual void DoContinuation() {}

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(FutureContinuationBase);
};

template<class FUTURE, class R, typename CONT>
class FutureContinuation;

template<class Derived>
class FutureBase {
 public:
    FutureBase() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    FutureBase(FutureBase&& rhs) = default;
    FutureBase& operator=(FutureBase&& rhs) = default;
#endif
#endif
    FutureBase(FutureBase&, move_tag) NLIB_NOEXCEPT {}
    FutureBase& assign(FutureBase&, move_tag) NLIB_NOEXCEPT { return *this; }
    errno_t Wait() NLIB_NOEXCEPT;
    NLIB_CHECK_RESULT errno_t WaitFor(const TimeSpan& span) NLIB_NOEXCEPT;
    NLIB_CHECK_RESULT errno_t WaitUntil(const DateTime& datetime) NLIB_NOEXCEPT;
    bool IsValid() const NLIB_NOEXCEPT {
        const Derived& ref = static_cast<const Derived&>(*this);
        return !!ref.value_;  // for explicit operator bool
    }
    bool IsReady() const NLIB_NOEXCEPT;
    errno_t MakeSharedFrom(const Derived& f) NLIB_NOEXCEPT {
        Derived& ref = static_cast<Derived&>(*this);
        return ref.value_.MakeSharedFrom(f.value_);
    }
    template<class R>
    errno_t Then(Future<R>* next, R (*cont)(Derived&)) NLIB_NOEXCEPT;

 private:
    bool IsReady_() const NLIB_NOEXCEPT {
        return static_cast<const Derived&>(*this).value_->common.IsReady();
    }
};

template<class Derived>
bool FutureBase<Derived>::IsReady() const NLIB_NOEXCEPT {
    const Derived& ref = static_cast<const Derived&>(*this);
    if (ref.value_) {
        ref.value_->common.Lock();
        bool rval = IsReady_();
        ref.value_->common.Unlock();
        return rval;
    } else {
        return false;
    }
}

template<class Derived>
errno_t FutureBase<Derived>::Wait() NLIB_NOEXCEPT {
    Derived& ref = static_cast<Derived&>(*this);
    NLIB_ASSERT(ref.value_);
    if (!ref.value_) return EINVAL;
    ref.value_->common.Lock();
    while (!IsReady_()) {
        errno_t e = ref.value_->common.Wait();
        if (e != 0) {
            ref.value_->common.Unlock();
            NLIB_ASSERT_NOERR(e);
            return e;
        }
    }
    ref.value_->common.Unlock();
    return 0;
}

template<class Derived>
errno_t FutureBase<Derived>::WaitFor(const TimeSpan& span) NLIB_NOEXCEPT {
    Derived& ref = static_cast<Derived&>(*this);
    NLIB_ASSERT(ref.value_);
    if (!ref.value_) return EINVAL;
    DateTime abstime;
    errno_t e = DateTime::GetNow(&abstime);
    abstime += span;
    if (e != 0) return e;
    ref.value_->common.Lock();
    while (!IsReady_()) {
        // NOTE:
        // Because of spurious wakeup, WaitUntil is better than WaitFor.
        // Also note that CAFE's nlib_cond_wait_for
        // always makes spurious wakeup and returns 0 if duration >= 0.
        // CAFE's nlib_cond_wait_until can return ETIMEDOUT on the other hand.
        e = nlib_cond_wait_until(&ref.value_->common.cond, &ref.value_->common.lock,
                                 abstime.ToTimeValue().tick);
        if (e != 0) {
            ref.value_->common.Unlock();
            return e;
        }
    }
    ref.value_->common.Unlock();
    return 0;
}

template<class Derived>
errno_t FutureBase<Derived>::WaitUntil(const DateTime& datetime) NLIB_NOEXCEPT {
    Derived& ref = static_cast<Derived&>(*this);
    NLIB_ASSERT(ref.value_);
    if (!ref.value_) return EINVAL;
    ref.value_->common.Lock();
    while (!IsReady_()) {
        errno_t e = nlib_cond_wait_until(&ref.value_->common.cond, &ref.value_->common.lock,
                                         datetime.ToTimeValue().tick);
        if (e != 0) {
            ref.value_->common.Unlock();
            return e;
        }
    }
    ref.value_->common.Unlock();
    return 0;
}

}  // namespace detail

namespace detail {
template<class Derived>
class PromiseBase {
 public:
    PromiseBase() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PromiseBase(PromiseBase&& /* rhs */) = default;
    PromiseBase& operator=(PromiseBase&&) = default;
#endif
#endif
    errno_t Init() NLIB_NOEXCEPT;

    bool IsValid() const NLIB_NOEXCEPT {
        const Derived& ref = static_cast<const Derived&>(*this);
        return !!ref.value_;
    }
    errno_t SetError(errno_t e) NLIB_NOEXCEPT;
    template<class F>
    errno_t GetFuture(F* p) NLIB_NOEXCEPT;

 protected:
    struct NLIB_CAPABILITY("mutex") CommonStatus {
        int32_t status;
        detail::FutureContinuationBase* cont;
        errno_t err;
        nlib_mutex lock;
        nlib_cond cond;
        CommonStatus() NLIB_NOEXCEPT : status(0), cont(NULL), err(0) {
            errno_t e;
            NLIB_UNUSED(e);
            e = nlib_mutex_init(&lock);
            NLIB_ASSERT_NOERR(e);
            e = nlib_cond_init(&cond);
            NLIB_ASSERT_NOERR(e);
        }
        ~CommonStatus() NLIB_NOEXCEPT {
            if (cont) delete cont;
            errno_t e;
            NLIB_UNUSED(e);
            e = nlib_mutex_destroy(&lock);
            NLIB_ASSERT_NOERR(e);
            e = nlib_cond_destroy(&cond);
            NLIB_ASSERT_NOERR(e);
        }
        void Lock() NLIB_NOEXCEPT NLIB_ACQUIRE() NLIB_NO_THREAD_SAFETY_ANALYSIS {
            errno_t e;
            NLIB_UNUSED(e);
            e = nlib_mutex_lock(&lock);
            NLIB_ASSERT_NOERR(e);
        }
        void Unlock() NLIB_NOEXCEPT NLIB_RELEASE() NLIB_NO_THREAD_SAFETY_ANALYSIS {
            errno_t e;
            NLIB_UNUSED(e);
            e = nlib_mutex_unlock(&lock);
            NLIB_ASSERT_NOERR(e);
        }
        errno_t Wait() NLIB_NOEXCEPT NLIB_REQUIRES(*this) NLIB_NO_THREAD_SAFETY_ANALYSIS {
            return nlib_cond_wait(&cond, &lock);
        }
        bool IsReady() NLIB_NOEXCEPT { return status == 20120915L; }
        bool SetReady() NLIB_NOEXCEPT {
            if (status != 20120915L) {
                status = 20120915L;
                return true;
            } else {
                return false;
            }
        }
        void NotifyAndContinuation() NLIB_NOEXCEPT {
            // have to notify all the threads waiting
            errno_t e = nlib_cond_broadcast(&cond);
            NLIB_ASSERT_NOERR(e);
            NLIB_UNUSED(e);
            if (cont) cont->DoContinuation();
        }
    };
    void NotifyAndContinuation() NLIB_NOEXCEPT {
        static_cast<Derived&>(*this).value_->common.NotifyAndContinuation();
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(PromiseBase);
};

template<class Derived>
errno_t PromiseBase<Derived>::Init() NLIB_NOEXCEPT {
    // NOTE:
    // This is not thread safe
    // Note that Promise<R> itself is not thread safe.
    Derived& ref = static_cast<Derived&>(*this);
    if (ref.value_) return EALREADY;
    ref.value_.reset(new (std::nothrow) typename Derived::Status());
    if (!ref.value_) return ENOMEM;
    return 0;
}

template<class Derived>
errno_t PromiseBase<Derived>::SetError(errno_t e) NLIB_NOEXCEPT {
    if (e == 0) return 0;
    Derived& ref = static_cast<Derived&>(*this);
    if (!ref.value_) {
        errno_t ee = ref.Init();
        if (ee != 0) return ee;
    }
    {
        ref.value_->common.Lock();
        if (ref.value_->common.SetReady()) {
            ref.value_->common.err = e;
            ref.value_->common.Unlock();
        } else {
            ref.value_->common.Unlock();
            return EALREADY;
        }
    }
    this->NotifyAndContinuation();
    return 0;
}

template<class Derived>
template<class F>
errno_t PromiseBase<Derived>::GetFuture(F* p) NLIB_NOEXCEPT {
    if (!p) return EINVAL;
    Derived& ref = static_cast<Derived&>(*this);
    if (!ref.value_) {
        errno_t e = this->Init();
        if (e != 0) return e;
    }
    errno_t e = p->value_.MakeSharedFrom(ref.value_);
    if (e != 0) return e;
    return 0;
}

}  // namespace detail

template<class R>
class Promise : public detail::PromiseBase<Promise<R> > {
    typedef detail::PromiseBase<Promise<R> > BaseType;

 public:
    typedef Future<R> FutureType;
    Promise() NLIB_NOEXCEPT {}
    ~Promise() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Promise(Promise&& rhs) = default;
    Promise& operator=(Promise&& rhs) = default;
#else
    Promise(Promise&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Promise& operator=(Promise&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Promise(Promise& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Promise& assign(Promise& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t SetValue(const R& value);

    // for doxygen :(
    errno_t Init() NLIB_NOEXCEPT { return BaseType::Init(); }
    errno_t SetError(errno_t e) NLIB_NOEXCEPT { return BaseType::SetError(e); }
    errno_t GetFuture(Future<R>* p) NLIB_NOEXCEPT { return BaseType::GetFuture(p); }

 private:
    struct Status {
        typename BaseType::CommonStatus common;
        R data;
    };
    LimitedSharedPtr<Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Promise);
    template<class RR>
    friend class Future;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class R>
errno_t Promise<R>::SetValue(const R& value) {
    if (!value_) {
        errno_t e = this->Init();
        if (e != 0) return e;
    }
    {
        ScopedLock<nlib_mutex> lock(value_->common.lock);
        if (value_->common.SetReady()) {
            NLIB_TRY {
                value_->data = value;  // may throw
            }
#ifdef __cpp_exceptions
            NLIB_RETHROW_UNWIND {
                this->SetError(ECANCELED);
                throw;
            }
            NLIB_CATCH(...) { return this->SetError(EINVAL); }
#endif
        } else {
            return EALREADY;
        }
    }
    this->NotifyAndContinuation();
    return 0;
}

template<class T, class DEL>
class Promise<UniquePtr<T, DEL> > : public detail::PromiseBase<Promise<UniquePtr<T, DEL> > > {
    typedef detail::PromiseBase<Promise<UniquePtr<T, DEL> > > BaseType;

 public:
    typedef Future<UniquePtr<T, DEL> > FutureType;
    Promise() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Promise(Promise&& rhs) = default;
    Promise& operator=(Promise&& rhs) = default;
#else
    Promise(Promise&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Promise& operator=(Promise&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Promise(Promise& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Promise& assign(Promise& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t SetValue(T* value) NLIB_NOEXCEPT;

 private:
    struct Status {
        typename BaseType::CommonStatus common;
        UniquePtr<T, DEL> data;
    };
    LimitedSharedPtr<Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Promise);
    template<class RR>
    friend class Future;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class T, class DEL>
inline errno_t Promise<UniquePtr<T, DEL> >::SetValue(T* value) NLIB_NOEXCEPT {
    if (!value_) {
        errno_t e = this->Init();
        if (e != 0) return e;
    }
    {
        value_->common.Lock();
        if (value_->common.SetReady()) {
            value_->data.reset(value);
            value_->common.Unlock();
        } else {
            value_->common.Unlock();
            return EALREADY;
        }
    }
    this->NotifyAndContinuation();
    return 0;
}

template<class T, class DEL>
class Promise<UniquePtr<T[], DEL> > : public detail::PromiseBase<Promise<UniquePtr<T[], DEL> > > {
    typedef detail::PromiseBase<Promise<UniquePtr<T[], DEL> > > BaseType;

 public:
    typedef Future<UniquePtr<T[], DEL> > FutureType;
    Promise() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Promise(Promise&& rhs) = default;
    Promise& operator=(Promise&& rhs) = default;
#else
    Promise(Promise&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Promise& operator=(Promise&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Promise(Promise& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Promise& assign(Promise& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t SetValue(T* value) NLIB_NOEXCEPT;

 private:
    struct Status {
        typename BaseType::CommonStatus common;
        UniquePtr<T[], DEL> data;
    };
    LimitedSharedPtr<Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Promise);
    template<class RR>
    friend class Future;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class T, class DEL>
inline errno_t Promise<UniquePtr<T[], DEL> >::SetValue(T* value) NLIB_NOEXCEPT {
    if (!value_) {
        errno_t e = this->Init();
        if (e != 0) return e;
    }
    {
        value_->common.Lock();
        if (value_->common.SetReady()) {
            value_->data.reset(value);
            value_->common.Unlock();
        } else {
            value_->common.Unlock();
            return EALREADY;
        }
    }
    this->NotifyAndContinuation();
    return 0;
}

template<>
class Promise<void> : public detail::PromiseBase<Promise<void> > {
    typedef detail::PromiseBase<Promise<void> > BaseType;

 public:
    typedef Future<void> FutureType;
    Promise() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Promise(Promise&& rhs) = default;
    Promise& operator=(Promise&& rhs) = default;
#else
    Promise(Promise&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Promise& operator=(Promise&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Promise(Promise& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Promise& assign(Promise& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    NLIB_VIS_PUBLIC errno_t SetValue() NLIB_NOEXCEPT;

 private:
    struct Status {
        BaseType::CommonStatus common;
    };
    LimitedSharedPtr<Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Promise);
    template<class RR>
    friend class Future;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class R>
class Promise<Future<R> > : public detail::PromiseBase<Promise<Future<R> > > {
    typedef detail::PromiseBase<Promise<Future<R> > > BaseType;

 public:
    typedef Future<Future<R> > FutureType;
    Promise() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Promise(Promise&& rhs) = default;
    Promise& operator=(Promise&& rhs) = default;
#else
    Promise(Promise&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Promise& operator=(Promise&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Promise(Promise& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Promise& assign(Promise& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t SetValue(const Future<R>& value) NLIB_NOEXCEPT;

 private:
    struct Status {
        typename BaseType::CommonStatus common;
        Future<R> data;
    };
    LimitedSharedPtr<Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Promise);
    template<class RR>
    friend class Future;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class R>
inline errno_t Promise<Future<R> >::SetValue(const Future<R>& value) NLIB_NOEXCEPT {
    if (!value_) {
        errno_t e = this->Init();
        if (e != 0) return e;
    }
    {
        value_->common.Lock();
        if (value_->common.SetReady()) {
            errno_t e = value_->data.MakeSharedFrom(value);
            value_->common.Unlock();
            if (e != 0) {
                return e;
            }
        } else {
            value_->common.Unlock();
            return EALREADY;
        }
    }
    this->NotifyAndContinuation();
    return 0;
}

template<class R>
class Future : public detail::FutureBase<Future<R> > {
    typedef detail::FutureBase<Future<R> > BaseType;

 public:
    Future() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Future(Future&& rhs) = default;
    Future& operator=(Future&& rhs) = default;
#else
    Future(Future&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Future& operator=(Future&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Future(Future& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Future& assign(Future& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    R Get() {
        R tmp     = R();  // may throw
        errno_t e = this->Get(&tmp);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
        return tmp;
    }
    errno_t Get(R* ptr) {
        if (!ptr) return EINVAL;
        BaseType::Wait();
        errno_t e = value_->common.err;
        if (e == 0) {
            NLIB_TRY {
                *ptr = value_->data;  // may throw
            }
#ifdef __cpp_exceptions
            NLIB_RETHROW_UNWIND { throw; }
            NLIB_CATCH(...) { return ENOMEM; }
#endif
        }
        return e;
    }
    // for doxygen :(
    bool IsValid() const NLIB_NOEXCEPT { return BaseType::IsValid(); }
    bool IsReady() const NLIB_NOEXCEPT { return BaseType::IsReady(); }
    errno_t Wait() NLIB_NOEXCEPT { return BaseType::Wait(); }
    NLIB_CHECK_RESULT errno_t WaitFor(const TimeSpan& span) NLIB_NOEXCEPT {
        return BaseType::WaitFor(span);
    }
    NLIB_CHECK_RESULT errno_t WaitUntil(const DateTime& datetime) NLIB_NOEXCEPT {
        return BaseType::WaitUntil(datetime);
    }
    template<class RNEXT>
    errno_t Then(Future<RNEXT>* next, RNEXT (*cont)(Future<R>&)) NLIB_NOEXCEPT {
        return BaseType::Then(next, cont);
    }
    errno_t MakeSharedFrom(const Future& f) NLIB_NOEXCEPT { return BaseType::MakeSharedFrom(f); }

 private:
    LimitedSharedPtr<typename Promise<R>::Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Future);
    template<class RR>
    friend class Promise;
    template<class RR>
    friend class detail::FutureBase;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class T, class DEL>
class Future<UniquePtr<T, DEL> > : public detail::FutureBase<Future<UniquePtr<T, DEL> > > {
    typedef detail::FutureBase<Future<UniquePtr<T, DEL> > > BaseType;

 public:
    Future() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Future(Future&& rhs) = default;
    Future& operator=(Future&& rhs) = default;
#else
    Future(Future&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Future& operator=(Future&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Future(Future& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Future& assign(Future& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    T* Get() NLIB_NOEXCEPT {
        BaseType::Wait();
        return value_->data.get();
    }
    errno_t Get(T** ptr) NLIB_NOEXCEPT {
        if (!ptr) return EINVAL;
        BaseType::Wait();
        errno_t e = value_->common.err;
        if (e == 0) *ptr = value_->data.get();
        return e;
    }
    T* Release() NLIB_NOEXCEPT {
        BaseType::Wait();
        return value_->data.release();
    }

 private:
    LimitedSharedPtr<typename Promise<UniquePtr<T, DEL> >::Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Future);
    template<class RR>
    friend class Promise;
    template<class RR>
    friend class detail::FutureBase;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class T, class DEL>
class Future<UniquePtr<T[], DEL> > : public detail::FutureBase<Future<UniquePtr<T[], DEL> > > {
    typedef detail::FutureBase<Future<UniquePtr<T[], DEL> > > BaseType;

 public:
    Future() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Future(Future&& rhs) = default;
    Future& operator=(Future&& rhs) = default;
#else
    Future(Future&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Future& operator=(Future&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Future(Future& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Future& assign(Future& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    T* Get() NLIB_NOEXCEPT {
        BaseType::Wait();
        return value_->data.get();
    }
    errno_t Get(T** ptr) NLIB_NOEXCEPT {
        if (!ptr) return EINVAL;
        BaseType::Wait();
        errno_t e = value_->common.err;
        if (e == 0) *ptr = value_->data.get();
        return e;
    }
    T* Release() NLIB_NOEXCEPT {
        BaseType::Wait();
        return value_->common.data.release();
    }

 private:
    LimitedSharedPtr<typename Promise<UniquePtr<T[], DEL> >::Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Future);
    template<class RR>
    friend class Promise;
    template<class RR>
    friend class detail::FutureBase;
    template<class RR>
    friend class detail::PromiseBase;
};

template<>
class Future<void> : public detail::FutureBase<Future<void> > {
    typedef detail::FutureBase<Future<void> > BaseType;

 public:
    Future() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Future(Future&& rhs) = default;
    Future& operator=(Future&& rhs) = default;
#else
    Future(Future&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Future& operator=(Future&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Future(Future& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Future& assign(Future& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t Get() NLIB_NOEXCEPT {
        BaseType::Wait();
        return value_->common.err;
    }

 private:
    LimitedSharedPtr<Promise<void>::Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Future);
    template<class RR>
    friend class Promise;
    template<class RR>
    friend class detail::FutureBase;
    template<class RR>
    friend class detail::PromiseBase;
};

template<class R>
class Future<Future<R> > : public detail::FutureBase<Future<Future<R> > > {
    typedef detail::FutureBase<Future<Future<R> > > BaseType;

 public:
    Future() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    Future(Future&& rhs) = default;
    Future& operator=(Future&& rhs) = default;
#else
    Future(Future&& rhs) NLIB_NOEXCEPT : value_(std::move(rhs.value_)) {}
    Future& operator=(Future&& rhs) NLIB_NOEXCEPT {
        value_ = std::move(rhs.value_);
        return *this;
    }
#endif
#endif
    Future(Future& rhs, move_tag) NLIB_NOEXCEPT : value_(rhs.value_, move_tag()) {}
    Future& assign(Future& rhs, move_tag) NLIB_NOEXCEPT {
        value_.assign(rhs.value_, move_tag());
        return *this;
    }
    errno_t Get(Future<R>* inner) NLIB_NOEXCEPT {
        if (!inner) return EINVAL;
        BaseType::Wait();
        errno_t e = value_->common.err;
        if (e != 0) return e;
        return inner->MakeSharedFrom(value_->data);
    }

 private:
    LimitedSharedPtr<typename Promise<Future<R> >::Status> value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Future);
    template<class RR>
    friend class Promise;
    template<class RR>
    friend class detail::FutureBase;
    template<class RR>
    friend class detail::PromiseBase;
};

namespace detail {

template<class R>
class FuncHolderBase0 {
 public:
    virtual ~FuncHolderBase0() {}
    virtual R operator()() = 0;
};

template<class R, class T1>
class FuncHolderBase1 {
 public:
    virtual ~FuncHolderBase1() {}
    virtual R operator()(T1 NLIB_RREF arg1) = 0;
};

template<class R, class T1, class T2>
class FuncHolderBase2 {
 public:
    virtual ~FuncHolderBase2() {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2) = 0;
};

template<class R, class T1, class T2, class T3>
class FuncHolderBase3 {
 public:
    virtual ~FuncHolderBase3() {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3) = 0;
};

template<class R, class T1, class T2, class T3, class T4>
class FuncHolderBase4 {
 public:
    virtual ~FuncHolderBase4() {}
    virtual R
    operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4) = 0;
};

template<class R, class T1, class T2, class T3, class T4, class T5>
class FuncHolderBase5 {
 public:
    virtual ~FuncHolderBase5() {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4,
                         T5 NLIB_RREF arg5) = 0;
};

template<class FUNC, bool USESWAP>
class FuncWrap_ {
 public:
    FuncWrap_(FUNC& func) NLIB_NOEXCEPT : func_() {
        using std::swap;
        swap(func, func_);
    }
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap_);
};

template<class FUNC>
class FuncWrap_<FUNC, false> {
 public:
    FuncWrap_(FUNC& func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap_);
};

template<class FUNC>
class FuncWrap : public FuncWrap_<FUNC, IsSwappable<FUNC>::value> {
    typedef FuncWrap_<FUNC, IsSwappable<FUNC>::value> BaseType;

 public:
    FuncWrap(FUNC& func) : BaseType(func) {}
};

template<class R>
class FuncWrap<R()> {
 public:
    typedef R (*FUNC)();
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class R, class T1>
class FuncWrap<R(T1)> {
 public:
    typedef R (*FUNC)(T1);
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class R, class T1, class T2>
class FuncWrap<R(T1, T2)> {
 public:
    typedef R (*FUNC)(T1, T2);
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class R, class T1, class T2, class T3>
class FuncWrap<R(T1, T2, T3)> {
 public:
    typedef R (*FUNC)(T1, T2, T3);
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class R, class T1, class T2, class T3, class T4>
class FuncWrap<R(T1, T2, T3, T4)> {
 public:
    typedef R (*FUNC)(T1, T2, T3, T4);
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class R, class T1, class T2, class T3, class T4, class T5>
class FuncWrap<R(T1, T2, T3, T4, T5)> {
 public:
    typedef R (*FUNC)(T1, T2, T3, T4, T5);
    explicit FuncWrap(FUNC func) : func_(func) {}
    FUNC& get() NLIB_NOEXCEPT { return func_; }

 protected:
    FUNC func_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FuncWrap);
};

template<class FUNC, class R>
class FuncHolder0 : public FuncHolderBase0<R> {
 public:
    FuncHolder0(FUNC& f) : func_(f) {}
    virtual ~FuncHolder0() NLIB_OVERRIDE {}
    virtual R operator()() NLIB_OVERRIDE { return func_.get()(); }

 private:
    FuncWrap<FUNC> func_;
};

template<class FUNC, class R, class T1>
class FuncHolder1 : public FuncHolderBase1<R, T1> {
 public:
    FuncHolder1(FUNC& f) : func_(f) {}
    virtual ~FuncHolder1() NLIB_OVERRIDE {}
    virtual R operator()(T1 NLIB_RREF arg1) NLIB_OVERRIDE {
        return func_.get()(NLIB_FWD(T1, arg1));
    }

 private:
    FuncWrap<FUNC> func_;
};

template<class FUNC, class R, class T1, class T2>
class FuncHolder2 : public FuncHolderBase2<R, T1, T2> {
 public:
    FuncHolder2(FUNC& f) : func_(f) {}
    virtual ~FuncHolder2() NLIB_OVERRIDE {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2) NLIB_OVERRIDE {
        return func_.get()(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2));
    }

 private:
    FuncWrap<FUNC> func_;
};

template<class FUNC, class R, class T1, class T2, class T3>
class FuncHolder3 : public FuncHolderBase3<R, T1, T2, T3> {
 public:
    FuncHolder3(FUNC& f) : func_(f) {}
    virtual ~FuncHolder3() NLIB_OVERRIDE {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3) NLIB_OVERRIDE {
        return func_.get()(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3));
    }

 private:
    FuncWrap<FUNC> func_;
};

template<class FUNC, class R, class T1, class T2, class T3, class T4>
class FuncHolder4 : public FuncHolderBase4<R, T1, T2, T3, T4> {
 public:
    FuncHolder4(FUNC& f) : func_(f) {}
    virtual ~FuncHolder4() NLIB_OVERRIDE {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3,
                         T4 NLIB_RREF arg4) NLIB_OVERRIDE {
        return func_.get()(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3),
                           NLIB_FWD(T4, arg4));
    }

 private:
    FuncWrap<FUNC> func_;
};

template<class FUNC, class R, class T1, class T2, class T3, class T4, class T5>
class FuncHolder5 : public FuncHolderBase5<R, T1, T2, T3, T4, T5> {
 public:
    FuncHolder5(FUNC& f) : func_(f) {}
    virtual ~FuncHolder5() NLIB_OVERRIDE {}
    virtual R operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4,
                         T5 NLIB_RREF arg5) NLIB_OVERRIDE {
        return func_.get()(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3),
                           NLIB_FWD(T4, arg4), NLIB_FWD(T5, arg5));
    }

 private:
    FuncWrap<FUNC> func_;
};

}  // namespace detail

template<class T>
class PackagedTask {  // defined for doygen ....
 public:
    typedef PackagedTask<T> ThisType;
    PackagedTask() NLIB_NOEXCEPT;
    template<class FUNC>
    errno_t Init(FUNC& func);
    template<class R>
    errno_t GetFuture(Future<R>* p) NLIB_NOEXCEPT;
};

template<class R>
class PackagedTask<R()> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType()> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder0<FUNC, R>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()() { Exec<R>(); }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type Exec() {
        NLIB_TRY {
            R rval    = (*func_)();
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type Exec(){NLIB_TRY{(*func_)();
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}  // namespace threading

UniquePtr<detail::FuncHolderBase0<R> > func_;
Promise<R> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

template<class R, class T1>
class PackagedTask<R(T1)> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType(T1)> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder1<FUNC, R, T1>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()(T1 NLIB_RREF arg1) { Exec<R>(NLIB_FWD(T1, arg1)); }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type Exec(T1 NLIB_RREF arg1) {
        NLIB_TRY {
            R rval    = (*func_)(NLIB_FWD(T1, arg1));
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type
    Exec(T1 NLIB_RREF arg1){NLIB_TRY{(*func_)(NLIB_FWD(T1, arg1));
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}

UniquePtr<detail::FuncHolderBase1<R, T1> > func_;
Promise<R> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

template<class R, class T1, class T2>
class PackagedTask<R(T1, T2)> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType(T1, T2)> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder2<FUNC, R, T1, T2>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2) {
        Exec<R>(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2));
    }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2) {
        NLIB_TRY {
            R rval    = (*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2));
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type
    Exec(T1 NLIB_RREF arg1,
         T2 NLIB_RREF arg2){NLIB_TRY{(*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2));
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}

UniquePtr<detail::FuncHolderBase2<R, T1, T2> > func_;
Promise<ReturnType> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

template<class R, class T1, class T2, class T3>
class PackagedTask<R(T1, T2, T3)> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType(T1, T2, T3)> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder3<FUNC, R, T1, T2, T3>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3) {
        Exec<R>(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3));
    }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3) {
        NLIB_TRY {
            R rval    = (*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3));
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3){
        NLIB_TRY{(*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3));
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}

UniquePtr<detail::FuncHolderBase3<R, T1, T2, T3> > func_;
Promise<ReturnType> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

template<class R, class T1, class T2, class T3, class T4>
class PackagedTask<R(T1, T2, T3, T4)> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType(T1, T2, T3, T4)> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder4<FUNC, R, T1, T2, T3, T4>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4) {
        Exec<R>(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3), NLIB_FWD(T4, arg4));
    }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4) {
        NLIB_TRY {
            R rval    = (*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3),
                              NLIB_FWD(T4, arg4));
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4){NLIB_TRY{
        (*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3), NLIB_FWD(T4, arg4));
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}

UniquePtr<detail::FuncHolderBase4<R, T1, T2, T3, T4> > func_;
Promise<ReturnType> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

template<class R, class T1, class T2, class T3, class T4, class T5>
class PackagedTask<R(T1, T2, T3, T4, T5)> {
 public:
    typedef R ReturnType;
    typedef PackagedTask<ReturnType(T1, T2, T3, T4, T5)> ThisType;
    PackagedTask() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    PackagedTask(PackagedTask&& rhs) = default;
    PackagedTask& operator=(PackagedTask&& rhs) = default;
#else
    PackagedTask(PackagedTask&& rhs) NLIB_NOEXCEPT : func_(std::move(rhs.func_)),
                                                     promise_(std::move(rhs.promise_)) {}
    PackagedTask& operator=(PackagedTask&& rhs) NLIB_NOEXCEPT {
        func_    = std::move(rhs.func_);
        promise_ = std::move(rhs.promise_);
        return *this;
    }
#endif
#endif
    PackagedTask(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT : func_(rhs.func_.release()),
                                                              promise_(rhs.promise_, move_tag()) {}
    PackagedTask& assign(PackagedTask& rhs, move_tag) NLIB_NOEXCEPT {
        func_.reset(rhs.func_.release());
        promise_.assign(rhs.promise_, move_tag());
        return *this;
    }
    template<class FUNC>
    errno_t Init(FUNC& func) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        func_.reset(new (std::nothrow) detail::FuncHolder5<FUNC, R, T1, T2, T3, T4, T5>(func));
        return !func_ ? ENOMEM : 0;
    }
    errno_t GetFuture(Future<ReturnType>* p) NLIB_NOEXCEPT { return promise_.GetFuture(p); }
    void operator()(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4,
                    T5 NLIB_RREF arg5) {
        Exec<R>(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3), NLIB_FWD(T4, arg4),
                NLIB_FWD(T5, arg5));
    }

 private:
    template<class DUMMY>
    typename EnableIf<!IsSame<DUMMY, void>::value, void>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4,
         T5 NLIB_RREF arg5) {
        NLIB_TRY {
            R rval    = (*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3),
                              NLIB_FWD(T4, arg4), NLIB_FWD(T5, arg5));
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                promise_.SetValue(rval);
            } else {
                promise_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            promise_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { promise_.SetError(EINVAL); }
#endif
    }
    template<class DUMMY>
    typename EnableIf<IsSame<DUMMY, void>::value, DUMMY>::type
    Exec(T1 NLIB_RREF arg1, T2 NLIB_RREF arg2, T3 NLIB_RREF arg3, T4 NLIB_RREF arg4,
         T5 NLIB_RREF arg5){
        NLIB_TRY{(*func_)(NLIB_FWD(T1, arg1), NLIB_FWD(T2, arg2), NLIB_FWD(T3, arg3),
                          NLIB_FWD(T4, arg4), NLIB_FWD(T5, arg5));
    promise_.SetValue();
}
#ifdef __cpp_exceptions
NLIB_RETHROW_UNWIND {
    promise_.SetError(ECANCELED);
    throw;
}
NLIB_CATCH(...) {
    promise_.SetError(EINVAL);
}
#endif
}

UniquePtr<detail::FuncHolderBase5<R, T1, T2, T3, T4, T5> > func_;
Promise<ReturnType> promise_;
NLIB_DISALLOW_COPY_AND_ASSIGN(PackagedTask);
}
;

namespace detail {

template<class T, bool B>
struct XHolder {
    explicit XHolder(const T& value_) : value() {
        using std::swap;
        swap(const_cast<T&>(value_), value);
    }
    T value;
};

template<class T>
struct XHolder<T, false> {
    explicit XHolder(const T& value_) : value(value_) {}
    T value;
};

template<class R>
struct XHolder<R(), false> {
    XHolder(R (*value_)()) : value(value_) {}
    R (*value)();
};

template<class R, class T1>
struct XHolder<R(T1), false> {
    XHolder(R (*value_)(T1)) : value(value_) {}
    R (*value)(T1);
};

template<class R, class T1, class T2>
struct XHolder<R(T1, T2), false> {
    XHolder(R (*value_)(T1, T2)) : value(value_) {}
    R (*value)(T1, T2);
};

template<class R, class T1, class T2, class T3>
struct XHolder<R(T1, T2, T3), false> {
    XHolder(R (*value_)(T1, T2, T3)) : value(value_) {}
    R (*value)(T1, T2, T3);
};

template<class R, class T1, class T2, class T3, class T4>
struct XHolder<R(T1, T2, T3, T4), false> {
    XHolder(R (*value_)(T1, T2, T3, T4)) : value(value_) {}
    R (*value)(T1, T2, T3, T4);
};

template<class R, class T1, class T2, class T3, class T4, class T5>
struct XHolder<R(T1, T2, T3, T4, T5), false> {
    XHolder(R (*value_)(T1, T2, T3, T4, T5)) : value(value_) {}
    R (*value)(T1, T2, T3, T4, T5);
};

#define NLIB_ASYNC_HOLDER(T, B) \
    XHolder<T, B::value && IsSwappable<T>::value && !IsArithmetic<T>::value>
#ifdef __cpp_rvalue_references
#define NLIB_ASYNC_HOLDER_F(FUNC, B) \
    XHolder<FUNC, B::value && IsSwappable<FUNC>::value && !IsPointer<FUNC>::value>
#else
#define NLIB_ASYNC_HOLDER_F(FUNC, B) \
    XHolder<FUNC, B::value && IsSwappable<FUNC>::value && !IsPointer<FUNC>::value>
#endif

template<class FUNC, class R, class BF>
errno_t Async_(Future<R>* future, const FUNC& func, const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R()> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    e = th.Start(settings, task, move_tag());
    return e;
}

template<class FUNC, class R, class BF, class T1>
errno_t
Async_(Future<R>* future, const FUNC& func, const T1& arg1, const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R(T1)> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    NLIB_ASYNC_HOLDER(T1, BF) arg1_(arg1);
    e = th.Start(settings, task, arg1_.value, move_tag());
    return e;
}

template<class FUNC, class R, class BF, class T1, class T2>
errno_t Async_(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
               const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R(T1, T2)> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    NLIB_ASYNC_HOLDER(T1, BF) arg1_(arg1);
    NLIB_ASYNC_HOLDER(T2, BF) arg2_(arg2);
    e = th.Start(settings, task, arg1_.value, arg2_.value, move_tag());
    return e;
}

template<class FUNC, class R, class BF, class T1, class T2, class T3>
errno_t Async_(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3,
               const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R(T1, T2, T3)> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    NLIB_ASYNC_HOLDER(T1, BF) arg1_(arg1);
    NLIB_ASYNC_HOLDER(T2, BF) arg2_(arg2);
    NLIB_ASYNC_HOLDER(T3, BF) arg3_(arg3);
    e = th.Start(settings, task, arg1_.value, arg2_.value, arg3_.value, move_tag());
    return e;
}

template<class FUNC, class R, class BF, class T1, class T2, class T3, class T4>
errno_t Async_(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3,
               const T4& arg4, const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R(T1, T2, T3, T4)> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    NLIB_ASYNC_HOLDER(T1, BF) arg1_(arg1);
    NLIB_ASYNC_HOLDER(T2, BF) arg2_(arg2);
    NLIB_ASYNC_HOLDER(T3, BF) arg3_(arg3);
    NLIB_ASYNC_HOLDER(T4, BF) arg4_(arg4);
    e = th.Start(settings, task, arg1_.value, arg2_.value, arg3_.value, arg4_.value, move_tag());
    return e;
}

template<class FUNC, class R, class BF, class T1, class T2, class T3, class T4, class T5>
errno_t Async_(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3,
               const T4& arg4, const T5& arg5, const ThreadSettings& settings, BF) {
    if (!future || !settings.GetDetachState()) return EINVAL;
    PackagedTask<R(T1, T2, T3, T4, T5)> task;
    NLIB_ASYNC_HOLDER_F(FUNC, BF) func_(func);
    errno_t e = task.Init(func_.value);
    if (e != 0) return e;
    e = task.GetFuture(future);
    if (e != 0) return e;
    Thread th;
    NLIB_ASYNC_HOLDER(T1, BF) arg1_(arg1);
    NLIB_ASYNC_HOLDER(T2, BF) arg2_(arg2);
    NLIB_ASYNC_HOLDER(T3, BF) arg3_(arg3);
    NLIB_ASYNC_HOLDER(T4, BF) arg4_(arg4);
    NLIB_ASYNC_HOLDER(T5, BF) arg5_(arg5);
    e = th.Start(settings, task, arg1_.value, arg2_.value, arg3_.value, arg4_.value, arg5_.value,
                 move_tag());
    return e;
}

#undef NLIB_ASYNC_HOLDER
#undef NLIB_ASYNC_HOLDER_F

}  // namespace detail

template<class FUNC, class R>
inline errno_t Async(Future<R>* future, const FUNC& func, const ThreadSettings& settings) {
    return detail::Async_(future, func, settings, FalseType());
}

template<class FUNC, class R>
inline errno_t
Async(Future<R>* future, const FUNC& func, const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, settings, TrueType());
}

template<class FUNC, class R>
inline errno_t Async(Future<R>* future, const FUNC& func) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, settings);
}

template<class FUNC, class R>
inline errno_t Async(Future<R>* future, const FUNC& func, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, settings, move_tag());
}

template<class FUNC, class R, class T1>
inline errno_t
Async(Future<R>* future, const FUNC& func, const T1& arg1, const ThreadSettings& settings) {
    return detail::Async_(future, func, arg1, settings, FalseType());
}

template<class FUNC, class R, class T1>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1,
                     const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, arg1, settings, TrueType());
}

template<class FUNC, class R, class T1>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, settings);
}

template<class FUNC, class R, class T1>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, settings, move_tag());
}

template<class FUNC, class R, class T1, class T2>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const ThreadSettings& settings) {
    return detail::Async_(future, func, arg1, arg2, settings, FalseType());
}

template<class FUNC, class R, class T1, class T2>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, arg1, arg2, settings, TrueType());
}

template<class FUNC, class R, class T1, class T2>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, settings);
}

template<class FUNC, class R, class T1, class T2>
inline errno_t
Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, settings, move_tag());
}

template<class FUNC, class R, class T1, class T2, class T3>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const ThreadSettings& settings) {
    return detail::Async_(future, func, arg1, arg2, arg3, settings, FalseType());
}

template<class FUNC, class R, class T1, class T2, class T3>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, arg1, arg2, arg3, settings, TrueType());
}

template<class FUNC, class R, class T1, class T2, class T3>
inline errno_t
Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, settings);
}

template<class FUNC, class R, class T1, class T2, class T3>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, settings, move_tag());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4, const ThreadSettings& settings) {
    return detail::Async_(future, func, arg1, arg2, arg3, arg4, settings, FalseType());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4, const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, arg1, arg2, arg3, arg4, settings, TrueType());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, arg4, settings);
}

template<class FUNC, class R, class T1, class T2, class T3, class T4>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, arg4, settings, move_tag());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4, class T5>
inline errno_t
Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3,
      const T4& arg4, const T5& arg5, const ThreadSettings& settings) {
    return detail::Async_(future, func, arg1, arg2, arg3, arg4, arg5, settings, FalseType());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4, class T5>
inline errno_t
Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2, const T3& arg3,
      const T4& arg4, const T5& arg5, const ThreadSettings& settings, move_tag) {
    return detail::Async_(future, func, arg1, arg2, arg3, arg4, arg5, settings, TrueType());
}

template<class FUNC, class R, class T1, class T2, class T3, class T4, class T5>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4, const T5& arg5) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, arg4, arg5, settings);
}

template<class FUNC, class R, class T1, class T2, class T3, class T4, class T5>
inline errno_t Async(Future<R>* future, const FUNC& func, const T1& arg1, const T2& arg2,
                     const T3& arg3, const T4& arg4, const T5& arg5, move_tag) {
    ThreadSettings settings;
    settings.SetDetachState(true);
    return Async(future, func, arg1, arg2, arg3, arg4, arg5, settings, move_tag());
}

template<class R>
inline errno_t MakeReadyFuture(Future<R>* future, const R& value) {
    if (!future) return EINVAL;
    Promise<R> promise;
    errno_t e = promise.Init();
    if (e != 0) return e;
    e = promise.SetValue(value);
    if (e != 0) return e;
    e = promise.GetFuture(future);
    if (e != 0) return e;
    return 0;
}

namespace detail {
template<class FUTURE, class R, typename CONT>
class FutureContinuation : public FutureContinuationBase {
 public:
    FutureContinuation(FUTURE& f, CONT cont) NLIB_NOEXCEPT : status_(0), parent_(f), func_(cont) {}
    errno_t Init() NLIB_NOEXCEPT { return next_.Init(); }
    errno_t GetFuture(Future<R>* future) NLIB_NOEXCEPT { return next_.GetFuture(future); }
    virtual ~FutureContinuation() NLIB_NOEXCEPT NLIB_OVERRIDE {}
    virtual void DoContinuation() NLIB_OVERRIDE;

 private:
    int32_t status_;
    FUTURE& parent_;
    CONT func_;
    Promise<R> next_;
};

template<class FUTURE, class R, typename CONT>
void FutureContinuation<FUTURE, R, CONT>::DoContinuation() {
    int32_t expected = 0;
    if (nlib_atomic_compare_exchange32(&status_, &expected, 20120901L, 0, NLIB_ATOMIC_ACQUIRE,
                                       NLIB_ATOMIC_ACQUIRE)) {
        NLIB_TRY {
            R rval    = func_(parent_);
            errno_t e = GetFutureError(rval);
            if (e == 0) {
                (void)next_.SetValue(rval);
            } else {
                (void)next_.SetError(e);
            }
        }
#ifdef __cpp_exceptions
        NLIB_RETHROW_UNWIND {
            next_.SetError(ECANCELED);
            throw;
        }
        NLIB_CATCH(...) { (void)next_.SetError(EINVAL); }
#endif
    }
}

template<class Derived>
template<class R>
errno_t FutureBase<Derived>::Then(Future<R>* next, R (*cont)(Derived&)) NLIB_NOEXCEPT {
    Derived& ref = static_cast<Derived&>(*this);
    if (!ref.value_) return EINVAL;

    typedef R (*CONT)(Derived&);
    UniquePtr<FutureContinuation<Derived, R, CONT> > ptr(
        new (std::nothrow) FutureContinuation<Derived, R, CONT>(ref, cont));
    if (!ptr) return ENOMEM;
    errno_t e = ptr->Init();
    if (e != 0) return e;
    e = ptr->GetFuture(next);
    if (e != 0) return e;

    ref.value_->common.Lock();
    ref.value_->common.cont = ptr.release();
    bool isready            = ref.IsReady_();
    ref.value_->common.Unlock();
    if (isready) {
        ref.value_->common.cont->DoContinuation();
    }
    return 0;
}

}  // namespace detail

template<class R1 = None, class R2 = None, class R3 = None, class R4 = None, class R5 = None>
struct FutureTuple {
    Future<R1> f1;
    Future<R2> f2;
    Future<R3> f3;
    Future<R4> f4;
    Future<R5> f5;
};

template<>
struct FutureTuple<None, None, None, None, None>;
template<class R1>
struct FutureTuple<R1, None, None, None, None>;

template<class R1, class R2>
struct FutureTuple<R1, R2, None, None, None> {
    Future<R1> f1;
    Future<R2> f2;
};

template<class R1, class R2, class R3>
struct FutureTuple<R1, R2, R3, None, None> {
    Future<R1> f1;
    Future<R2> f2;
    Future<R3> f3;
};

template<class R1, class R2, class R3, class R4>
struct FutureTuple<R1, R2, R3, R4, None> {
    Future<R1> f1;
    Future<R2> f2;
    Future<R3> f3;
    Future<R4> f4;
};

template<class R1 = None, class R2 = None, class R3 = None, class R4 = None, class R5 = None>
class WhenFutureComplete {
 public:
    typedef UniquePtr<FutureTuple<R1, R2, R3, R4, R5> > ReturnType;
    typedef Future<ReturnType> FutureType;
    static errno_t All(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3,
                       Future<R4>* f4, Future<R5>* f5) {
        return DoWhen(f, f1, f2, f3, f4, f5, CallWhenAll);
    }
    static errno_t Any(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3,
                       Future<R4>* f4, Future<R5>* f5) {
        return DoWhen(f, f1, f2, f3, f4, f5, CallWhenAny);
    }

 private:
    typedef UniquePtr<WhenFutureComplete> ArgType;
    typedef void (*FuncType)(ArgType& ptr);
    static errno_t DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3,
                          Future<R4>* f4, Future<R5>* f5, FuncType func);
    static void Call(ArgType& ptr) {
        // called from Thread::Exec()
        ptr->func_(ptr);
    }
    static void CallWhenAny(ArgType& ptr) {
        for (;;) {
            if (ptr->tuple_->f1.IsReady() || ptr->tuple_->f2.IsReady() ||
                ptr->tuple_->f3.IsReady() || ptr->tuple_->f4.IsReady() ||
                ptr->tuple_->f5.IsReady()) {
                break;
            }
            this_thread::Sleep(TimeSpan(0, 0, 0, 200));
        }
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    static void CallWhenAll(ArgType& ptr) {
        ptr->tuple_->f1.Wait();
        ptr->tuple_->f2.Wait();
        ptr->tuple_->f3.Wait();
        ptr->tuple_->f4.Wait();
        ptr->tuple_->f5.Wait();
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    explicit WhenFutureComplete(FuncType f) : func_(f) {}
    errno_t Init(Future<ReturnType>* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3,
                 Future<R4>* f4, Future<R5>* f5) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        tuple_.reset(new (std::nothrow) FutureTuple<R1, R2, R3, R4, R5>());
        if (!tuple_) return ENOMEM;
        e = promise_.GetFuture(f);
        if (e != 0) return e;
        e = tuple_->f1.MakeSharedFrom(*f1);
        if (e != 0) return e;
        e = tuple_->f2.MakeSharedFrom(*f2);
        if (e != 0) return e;
        e = tuple_->f3.MakeSharedFrom(*f3);
        if (e != 0) return e;
        e = tuple_->f4.MakeSharedFrom(*f4);
        if (e != 0) return e;
        e = tuple_->f5.MakeSharedFrom(*f5);
        if (e != 0) return e;
        return 0;
    }

 private:
    FuncType func_;
    ReturnType tuple_;
    Promise<ReturnType> promise_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(WhenFutureComplete);
    friend class Thread;
};

template<class R1, class R2, class R3, class R4, class R5>
errno_t
WhenFutureComplete<R1, R2, R3, R4, R5>::DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2,
                                               Future<R3>* f3, Future<R4>* f4, Future<R5>* f5,
                                               FuncType func) {
    if (!f1 || !f2 || !f3 || !f4 || !f5) return EINVAL;
    ArgType ptr(new (std::nothrow) WhenFutureComplete(func));
    if (!ptr) return 0;
    errno_t e;
    e = ptr->Init(f, f1, f2, f3, f4, f5);
    if (e != 0) return e;
    Thread th;
    e = th.StartRaw(ptr);
    if (e != 0) return e;
    return 0;
}

template<>
class WhenFutureComplete<None, None, None, None, None>;
template<class R1>
class WhenFutureComplete<R1, None, None, None, None>;

template<class R1, class R2>
class WhenFutureComplete<R1, R2, None, None, None> {
 public:
    typedef UniquePtr<FutureTuple<R1, R2> > ReturnType;
    typedef Future<ReturnType> FutureType;
    static errno_t All(FutureType* f, Future<R1>* f1, Future<R2>* f2) {
        return DoWhen(f, f1, f2, CallWhenAll);
    }
    static errno_t Any(FutureType* f, Future<R1>* f1, Future<R2>* f2) {
        return DoWhen(f, f1, f2, CallWhenAny);
    }

 private:
    typedef UniquePtr<WhenFutureComplete> ArgType;
    typedef void (*FuncType)(ArgType& ptr);
    static errno_t DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2, FuncType func);
    static void Call(ArgType& ptr) {
        // called from Thread::Exec()
        ptr->func_(ptr);
    }
    static void CallWhenAny(ArgType& ptr) {
        for (;;) {
            if (ptr->tuple_->f1.IsReady() || ptr->tuple_->f2.IsReady()) {
                break;
            }
            this_thread::Sleep(TimeSpan(0, 0, 0, 200));
        }
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    static void CallWhenAll(ArgType& ptr) {
        ptr->tuple_->f1.Wait();
        ptr->tuple_->f2.Wait();
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    explicit WhenFutureComplete(FuncType f) : func_(f) {}
    errno_t Init(Future<ReturnType>* f, Future<R1>* f1, Future<R2>* f2) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        tuple_.reset(new (std::nothrow) FutureTuple<R1, R2>());
        if (!tuple_) return ENOMEM;
        e = promise_.GetFuture(f);
        if (e != 0) return e;
        e = tuple_->f1.MakeSharedFrom(*f1);
        if (e != 0) return e;
        e = tuple_->f2.MakeSharedFrom(*f2);
        if (e != 0) return e;
        return 0;
    }

 private:
    FuncType func_;
    ReturnType tuple_;
    Promise<ReturnType> promise_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(WhenFutureComplete);
    friend class Thread;
};

template<class R1, class R2>
errno_t inline WhenFutureComplete<R1, R2>::DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2,
                                                  FuncType func) {
    if (!f1 || !f2) return EINVAL;
    ArgType ptr(new (std::nothrow) WhenFutureComplete(func));
    if (!ptr) return 0;
    errno_t e;
    e = ptr->Init(f, f1, f2);
    if (e != 0) return e;
    Thread th;
    e = th.StartRaw(ptr);
    if (e != 0) return e;
    return 0;
}

template<class R1, class R2, class R3>
class WhenFutureComplete<R1, R2, R3, None, None> {
 public:
    typedef UniquePtr<FutureTuple<R1, R2, R3> > ReturnType;
    typedef Future<ReturnType> FutureType;
    static errno_t All(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3) {
        return DoWhen(f, f1, f2, f3, CallWhenAll);
    }
    static errno_t Any(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3) {
        return DoWhen(f, f1, f2, f3, CallWhenAny);
    }

 private:
    typedef UniquePtr<WhenFutureComplete> ArgType;
    typedef void (*FuncType)(ArgType& ptr);
    static errno_t
    DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3, FuncType func);
    static void Call(ArgType& ptr) {
        // called from Thread::Exec()
        ptr->func_(ptr);
    }
    static void CallWhenAny(ArgType& ptr) {
        for (;;) {
            if (ptr->tuple_->f1.IsReady() || ptr->tuple_->f2.IsReady() ||
                ptr->tuple_->f3.IsReady()) {
                break;
            }
            this_thread::Sleep(TimeSpan(0, 0, 0, 200));
        }
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    static void CallWhenAll(ArgType& ptr) {
        ptr->tuple_->f1.Wait();
        ptr->tuple_->f2.Wait();
        ptr->tuple_->f3.Wait();
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    explicit WhenFutureComplete(FuncType f) : func_(f) {}
    errno_t Init(Future<ReturnType>* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        tuple_.reset(new (std::nothrow) FutureTuple<R1, R2, R3>());
        if (!tuple_) return ENOMEM;
        e = promise_.GetFuture(f);
        if (e != 0) return e;
        e = tuple_->f1.MakeSharedFrom(*f1);
        if (e != 0) return e;
        e = tuple_->f2.MakeSharedFrom(*f2);
        if (e != 0) return e;
        e = tuple_->f3.MakeSharedFrom(*f3);
        if (e != 0) return e;
        return 0;
    }

 private:
    FuncType func_;
    ReturnType tuple_;
    Promise<ReturnType> promise_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(WhenFutureComplete);
    friend class Thread;
};

template<class R1, class R2, class R3>
inline errno_t WhenFutureComplete<R1, R2, R3>::DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2,
                                                      Future<R3>* f3, FuncType func) {
    if (!f1 || !f2 || !f3) return EINVAL;
    ArgType ptr(new (std::nothrow) WhenFutureComplete(func));
    if (!ptr) return 0;
    errno_t e;
    e = ptr->Init(f, f1, f2, f3);
    if (e != 0) return e;
    Thread th;
    e = th.StartRaw(ptr);
    if (e != 0) return e;
    return 0;
}

template<class R1, class R2, class R3, class R4>
class WhenFutureComplete<R1, R2, R3, R4, None> {
 public:
    typedef UniquePtr<FutureTuple<R1, R2, R3, R4> > ReturnType;
    typedef Future<ReturnType> FutureType;
    static errno_t
    All(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3, Future<R4>* f4) {
        return DoWhen(f, f1, f2, f3, f4, CallWhenAll);
    }
    static errno_t
    Any(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3, Future<R4>* f4) {
        return DoWhen(f, f1, f2, f3, f4, CallWhenAny);
    }

 private:
    typedef UniquePtr<WhenFutureComplete> ArgType;
    typedef void (*FuncType)(ArgType& ptr);
    static errno_t DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3,
                          Future<R4>* f4, FuncType func);
    static void Call(ArgType& ptr) {
        // called from Thread::Exec()
        ptr->func_(ptr);
    }
    static void CallWhenAny(ArgType& ptr) {
        for (;;) {
            if (ptr->tuple_->f1.IsReady() || ptr->tuple_->f2.IsReady() ||
                ptr->tuple_->f3.IsReady() || ptr->tuple_->f4.IsReady()) {
                break;
            }
            this_thread::Sleep(TimeSpan(0, 0, 0, 200));
        }
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    static void CallWhenAll(ArgType& ptr) {
        ptr->tuple_->f1.Wait();
        ptr->tuple_->f2.Wait();
        ptr->tuple_->f3.Wait();
        ptr->tuple_->f4.Wait();
        ptr->promise_.SetValue(ptr->tuple_.release());
    }
    explicit WhenFutureComplete(FuncType f) : func_(f) {}
    errno_t
    Init(Future<ReturnType>* f, Future<R1>* f1, Future<R2>* f2, Future<R3>* f3, Future<R4>* f4) {
        errno_t e = promise_.Init();
        if (e != 0) return e;
        tuple_.reset(new (std::nothrow) FutureTuple<R1, R2, R3, R4>());
        if (!tuple_) return ENOMEM;
        e = promise_.GetFuture(f);
        if (e != 0) return e;
        e = tuple_->f1.MakeSharedFrom(*f1);
        if (e != 0) return e;
        e = tuple_->f2.MakeSharedFrom(*f2);
        if (e != 0) return e;
        e = tuple_->f3.MakeSharedFrom(*f3);
        if (e != 0) return e;
        e = tuple_->f4.MakeSharedFrom(*f4);
        if (e != 0) return e;
        return 0;
    }

 private:
    FuncType func_;
    ReturnType tuple_;
    Promise<ReturnType> promise_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(WhenFutureComplete);
    friend class Thread;
};

template<class R1, class R2, class R3, class R4>
inline errno_t
WhenFutureComplete<R1, R2, R3, R4>::DoWhen(FutureType* f, Future<R1>* f1, Future<R2>* f2,
                                           Future<R3>* f3, Future<R4>* f4, FuncType func) {
    if (!f1 || !f2 || !f3 || !f4) return EINVAL;
    ArgType ptr(new (std::nothrow) WhenFutureComplete(func));
    if (!ptr) return 0;
    errno_t e;
    e = ptr->Init(f, f1, f2, f3, f4);
    if (e != 0) return e;
    Thread th;
    e = th.StartRaw(ptr);
    if (e != 0) return e;
    return 0;
}

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP_T_BEGIN3(nn, nlib, threading)
NLIB_DEFINE_STD_SWAP_T1(T, NLIB_NS::threading::Future)
NLIB_DEFINE_STD_SWAP_T1(R, NLIB_NS::threading::Promise)
NLIB_DEFINE_STD_SWAP_T1(T, NLIB_NS::threading::PackagedTask)
NLIB_DEFINE_STD_SWAP_T_END3(nn, nlib, threading)

#endif  // INCLUDE_NN_NLIB_THREADING_FUTURE_H_
