﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
#define INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_

#include <new>
#include <utility>

#include "nn/nlib/Swap.h"
#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

namespace internal {

class LspStatus NLIB_FINAL {
 public:
    NLIB_CEXPR LspStatus() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~LspStatus() NLIB_NOEXCEPT {}
    NLIB_DEFMOVE_PIMPL(LspStatus);
    errno_t ToShare(LspStatus* ptr) NLIB_NOEXCEPT {
        if (!prv_) {
            prv_ = new (std::nothrow) int32_t(2);
            if (!prv_) return ENOMEM;
        } else {
            nlib_atomic_add_fetch32(prv_, 1, NLIB_ATOMIC_ACQUIRE);
        }
        NLIB_ASSERT(!ptr->prv_);
        ptr->prv_ = prv_;
        return 0;
    }
    bool ToUnshare() NLIB_NOEXCEPT {
        if (!prv_) return false;
        if (nlib_atomic_sub_fetch32(prv_, 1, NLIB_ATOMIC_ACQUIRE) == 0) {
            delete prv_;
            prv_ = nullptr;
            return false;
        }
        prv_ = nullptr;
        return true;
    }
    bool IsUnique() NLIB_NOEXCEPT {
        if (!prv_) return true;
        return nlib_atomic_load32(prv_, NLIB_ATOMIC_ACQUIRE) == 1;
    }
    void Reset() NLIB_NOEXCEPT { ToUnshare(); }

 private:
    int32_t* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LspStatus);
};

}  // namespace internal

// Use of shared_ptr is discouraged because of unexpected overhead,
// and You have to share the pointer by calling MakeSharedFrom() method.
template<class T, class DEL = DefaultDelete<T> >
class LimitedSharedPtr {
 public:
    NLIB_CEXPR LimitedSharedPtr() NLIB_NOEXCEPT : ptr_(nullptr) {}
    ~LimitedSharedPtr() NLIB_NOEXCEPT { release(); }
    explicit LimitedSharedPtr(nullptr_t) NLIB_NOEXCEPT : ptr_(nullptr) {}
    explicit LimitedSharedPtr(T* p) NLIB_NOEXCEPT : ptr_(p) {
        // Do not count up LspStatus because there is the risk of alloc failure.
    }
#ifdef __cpp_rvalue_references
    LimitedSharedPtr(LimitedSharedPtr&& rhs) NLIB_NOEXCEPT : ptr_(rhs.ptr_),
                                                             status_(std::move(rhs.status_)) {
        rhs.ptr_ = nullptr;
    }
    LimitedSharedPtr& operator=(LimitedSharedPtr&& rhs) NLIB_NOEXCEPT {
        if (!status_.ToUnshare()) DEL()(ptr_);
        ptr_     = rhs.ptr_;
        rhs.ptr_ = nullptr;
        status_  = std::move(rhs.status_);
        return *this;
    }
#endif
    LimitedSharedPtr(LimitedSharedPtr& rhs, move_tag) NLIB_NOEXCEPT
        : ptr_(rhs.ptr_),
          status_(rhs.status_, move_tag()) {
        rhs.ptr_ = nullptr;
    }
    LimitedSharedPtr& assign(LimitedSharedPtr& rhs, move_tag) NLIB_NOEXCEPT {
        if (!status_.ToUnshare()) DEL()(ptr_);
        ptr_     = rhs.ptr_;
        rhs.ptr_ = nullptr;
        status_.assign(rhs.status_, move_tag());
        return *this;
    }
    void reset(T* p = 0) NLIB_NOEXCEPT {
        if (!status_.ToUnshare()) DEL()(ptr_);
        ptr_ = p;
    }
    T& operator*() const {
        NLIB_ASSERT(ptr_ != nullptr);
        return *ptr_;
    }
    T* operator->() const NLIB_NOEXCEPT {
        NLIB_ASSERT(ptr_ != nullptr);
        return ptr_;
    }
    T* get() const NLIB_NOEXCEPT { return ptr_; }
    errno_t MakeSharedFrom(const LimitedSharedPtr& p) NLIB_NOEXCEPT;
    bool IsUnique() const NLIB_NOEXCEPT { return status_.IsUnique(); }
    bool MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT;
    LimitedSharedPtr& operator=(nullptr_t) NLIB_NOEXCEPT {
        release();
        return *this;
    }
    NLIB_SAFE_BOOL(LimitedSharedPtr, ptr_ != nullptr);

 private:
    void release() NLIB_NOEXCEPT {
        if (!status_.ToUnshare()) DEL()(ptr_);
        ptr_ = nullptr;
    }

 private:
    T* ptr_;
    mutable internal::LspStatus status_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};

template<class T, class DEL>
errno_t LimitedSharedPtr<T, DEL>::MakeSharedFrom(const LimitedSharedPtr& p) NLIB_NOEXCEPT {
    LimitedSharedPtr tmp;
    if (p) {
        errno_t e = p.status_.ToShare(&tmp.status_);
        if (e != 0) return e;
        tmp.ptr_ = p.ptr_;
    }
    this->assign(tmp, move_tag());
    return 0;
}

template<class T, class DEL>
bool LimitedSharedPtr<T, DEL>::MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT {
    if (IsUnique()) {
        p.reset(ptr_);
        ptr_ = nullptr;
        release();
        return true;
    } else {
        return false;
    }
}

template<class T, class DEL>
class LimitedSharedPtr<T[], DEL> {
 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};  // not allowed for now

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP_T_BEGIN3(nn, nlib, threading)
NLIB_DEFINE_STD_SWAP_T2(T, DEL, NLIB_NS::threading::LimitedSharedPtr)
NLIB_DEFINE_STD_SWAP_T_END3(nn, nlib, threading)

#endif  // INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
