﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_
#define INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_

#include "nn/nlib/Nqueue.h"
#include "nn/nlib/threading/Future.h"
#include "nn/nlib/LockFree.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

namespace detail {
class NLIB_VIS_PUBLIC ThPoolTaskBase {
 public:
    ThPoolTaskBase() NLIB_NOEXCEPT {}
    virtual ~ThPoolTaskBase() NLIB_NOEXCEPT {}
    virtual void operator()() NLIB_NOEXCEPT = 0;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThPoolTaskBase);
};

template<class R>
class ThPoolTask : public ThPoolTaskBase {
 public:
    explicit ThPoolTask(PackagedTask<R()>& task, move_tag) NLIB_NOEXCEPT : task_(task, move_tag()) {
    }
    virtual ~ThPoolTask() NLIB_NOEXCEPT NLIB_OVERRIDE {}
    virtual void operator()() NLIB_NOEXCEPT NLIB_OVERRIDE { task_(); }

 private:
    PackagedTask<R()> task_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThPoolTask);
};
}  // namespace detail

class NLIB_VIS_PUBLIC ThreadPool NLIB_FINAL {
 public:
    ThreadPool() NLIB_NOEXCEPT : prv_(nullptr) {}
    errno_t Init(size_t thread_count, const ThreadSettings& settings) NLIB_NOEXCEPT;
    errno_t Init(size_t thread_count) NLIB_NOEXCEPT {
        ThreadSettings settings;
        return this->Init(thread_count, settings);
    }
    errno_t Init() NLIB_NOEXCEPT { return this->Init(GetHardwareConcurrency()); }
    ~ThreadPool() NLIB_NOEXCEPT;
    template<class R, class FUNC>
    errno_t Submit(Future<R>* future, FUNC& func) NLIB_NOEXCEPT;
    size_t Count() NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN void JoinThreads(Nlist<nlib_thread>& L) NLIB_NOEXCEPT;
    static NLIB_VIS_HIDDEN void WorkerThread(void* p);
    errno_t Submit_(detail::ThPoolTaskBase* p) NLIB_NOEXCEPT;
    struct ThreadPoolPrivate;
    ThreadPoolPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThreadPool);
};

template<class R, class FUNC>
inline errno_t ThreadPool::Submit(Future<R>* future, FUNC& func) NLIB_NOEXCEPT {
    if (!prv_) return EINVAL;
    PackagedTask<R()> task;
    errno_t e = task.Init(func);
    if (e != 0) return e;
    if (future) {
        e = task.GetFuture(future);
        if (e != 0) return e;
    }
    detail::ThPoolTaskBase* ptask = new (std::nothrow) detail::ThPoolTask<R>(task, move_tag());
    if (!ptask) return ENOMEM;
    return Submit_(ptask);
}

namespace detail {

class NLIB_VIS_PUBLIC LockFreeThreadPoolBase {
 protected:
    LockFreeThreadPoolBase(void (*primary_worker_thread)(void*), void (*worker_thread)(void*)) :
        primary_worker_thread_(primary_worker_thread),
        worker_thread_(worker_thread),
        status_(0),
        thread_count_(0),
        thread_list_(nullptr) {}
    ~LockFreeThreadPoolBase() NLIB_NOEXCEPT;
    typedef void* (*Func)(void**, size_t)NLIB_NOEXCEPT_FUNCPTR;
    typedef void (*Callback)(void*) NLIB_NOEXCEPT_FUNCPTR;
    errno_t InitBase(size_t thread_count, const ThreadSettings& settings) NLIB_NOEXCEPT;
    bool PrimaryWorkerThread_(Func func, Callback callback, void** args, size_t n) NLIB_NOEXCEPT;
    bool WorkerThread_(Func func, Callback callback, void** args, size_t n,
                       uint32_t* counter) NLIB_NOEXCEPT;
    void JoinThreads() NLIB_NOEXCEPT;

 public:
    size_t GetThreadCount() NLIB_NOEXCEPT { return thread_count_; }

 private:
    NLIB_VIS_HIDDEN void JoinThreads(nlib_thread* L, size_t cnt) NLIB_NOEXCEPT;
    void (*primary_worker_thread_)(void*);
    void (*worker_thread_)(void*);
    int32_t status_;
    uint32_t thread_count_;
    nlib_thread* thread_list_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(LockFreeThreadPoolBase);
};

}  // namespace detail

template<size_t N = 5>
class LockFreeThreadPool NLIB_FINAL : public NLIB_NS::threading::detail::LockFreeThreadPoolBase {
 public:
#ifdef NLIB_DOXYGEN
    typedef void* (*Func)(void** args, size_t n);
    typedef void (*Callback)(void* result);
#else
    typedef NLIB_NS::threading::detail::LockFreeThreadPoolBase::Func Func;
    typedef NLIB_NS::threading::detail::LockFreeThreadPoolBase::Callback Callback;
#endif

    LockFreeThreadPool() NLIB_NOEXCEPT : LockFreeThreadPoolBase(PrimaryWorkerThread, WorkerThread) {
    }
    ~LockFreeThreadPool() NLIB_NOEXCEPT { this->JoinThreads(); }
    errno_t
    Init(size_t work_queue_size, size_t thread_count, const ThreadSettings& settings) NLIB_NOEXCEPT;
    errno_t SubmitVarArgs(Func func, Callback callback, size_t n, ...) NLIB_NOEXCEPT;

#ifdef NLIB_DOXYGEN
    errno_t Submit(Func func, Callback callback);
    errno_t Submit(Func func, Callback callback, void* arg0);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4, void* arg5);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4, void* arg5, void* arg6);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4, void* arg5, void* arg6, void* arg7);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4, void* arg5, void* arg6, void* arg7, void* arg8);
    errno_t Submit(Func func, Callback callback, void* arg0, void* arg1, void* arg2, void* arg3,
                   void* arg4, void* arg5, void* arg6, void* arg7, void* arg8, void* arg9);
    size_t GetThreadCount();
#endif
    errno_t Submit(Func func, Callback callback) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        // memset(&info.arg[0], 0, sizeof(info.arg[0]) * N);
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t
    Submit(X func, Callback callback,
           typename EnableIf<N >= 1 && IsSame<X, Func>::value, void*>::type arg0) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        // memset(&info.arg[0] + 1, 0, sizeof(info.arg[0]) * (N - 1));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 2 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        // memset(&info.arg[0] + 2, 0, sizeof(info.arg[0]) * (N - 2));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 3 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        // memset(&info.arg[0] + 3, 0, sizeof(info.arg[0]) * (N - 3));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 4 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2, void* arg3) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        // memset(&info.arg[0] + 4, 0, sizeof(info.arg[0]) * (N - 4));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 5 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2, void* arg3, void* arg4) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        // memset(&info.arg[0] + 5, 0, sizeof(info.arg[0]) * (N - 5));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 6 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2, void* arg3, void* arg4, void* arg5) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        info.arg[5]   = arg5;
        // memset(&info.arg[0] + 6, 0, sizeof(info.arg[0]) * (N - 6));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t
    Submit(X func, Callback callback,
           typename EnableIf<N >= 7 && IsSame<X, Func>::value, void*>::type arg0, void* arg1,
           void* arg2, void* arg3, void* arg4, void* arg5, void* arg6) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        info.arg[5]   = arg5;
        info.arg[6]   = arg6;
        // memset(&info.arg[0] + 7, 0, sizeof(info.arg[0]) * (N - 7));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t
    Submit(X func, Callback callback,
           typename EnableIf<N >= 8 && IsSame<X, Func>::value, void*>::type arg0, void* arg1,
           void* arg2, void* arg3, void* arg4, void* arg5, void* arg6, void* arg7) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        info.arg[5]   = arg5;
        info.arg[6]   = arg6;
        info.arg[7]   = arg7;
        // memset(&info.arg[0] + 8, 0, sizeof(info.arg[0]) * (N - 8));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 9 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2, void* arg3, void* arg4, void* arg5, void* arg6,
                   void* arg7, void* arg8) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        info.arg[5]   = arg5;
        info.arg[6]   = arg6;
        info.arg[7]   = arg7;
        info.arg[8]   = arg8;
        // memset(&info.arg[0] + 9, 0, sizeof(info.arg[0]) * (N - 9));
        return work_queue_.Enqueue(info);
    }
    template<class X>
    errno_t Submit(X func, Callback callback,
                   typename EnableIf<N >= 10 && IsSame<X, Func>::value, void*>::type arg0,
                   void* arg1, void* arg2, void* arg3, void* arg4, void* arg5, void* arg6,
                   void* arg7, void* arg8, void* arg9) NLIB_NOEXCEPT {
        LockFreeThreadPoolExecInfo info;
        info.func     = func;
        info.callback = callback;
        info.arg[0]   = arg0;
        info.arg[1]   = arg1;
        info.arg[2]   = arg2;
        info.arg[3]   = arg3;
        info.arg[4]   = arg4;
        info.arg[5]   = arg5;
        info.arg[6]   = arg6;
        info.arg[7]   = arg7;
        info.arg[8]   = arg8;
        info.arg[9]   = arg9;
        // memset(&info.arg[0] + 10, 0, sizeof(info.arg[0]) * (N - 10));
        return work_queue_.Enqueue(info);
    }

 private:
    static void PrimaryWorkerThread(void* p) NLIB_NOEXCEPT;
    static void WorkerThread(void* p) NLIB_NOEXCEPT;

 private:
    struct LockFreeThreadPoolExecInfo {
        Func func;
        Callback callback;
        void* arg[N];
    };
    LockFreeQueue<LockFreeThreadPoolExecInfo> work_queue_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(LockFreeThreadPool);
};

template<size_t N>
errno_t LockFreeThreadPool<N>::Init(size_t work_queue_size, size_t thread_count,
                                    const ThreadSettings& settings) NLIB_NOEXCEPT {
    errno_t e = work_queue_.Init(work_queue_size);
    if (e != 0) return e;
    e = InitBase(thread_count, settings);
    if (e != 0) {
        LockFreeQueue<LockFreeThreadPoolExecInfo> q;
        q.SwapUnsafe(work_queue_);
        return e;
    }
    return 0;
}

template<size_t N>
errno_t
LockFreeThreadPool<N>::SubmitVarArgs(Func func, Callback callback, size_t n, ...) NLIB_NOEXCEPT {
    if (n > N) {
        NLIB_ASSERT(n <= N);
        return EINVAL;
    }

    va_list ap;
    LockFreeThreadPoolExecInfo info;
    info.func     = func;
    info.callback = callback;

    va_start(ap, n);
    for (size_t i = 0; i < n; ++i) {
        info.arg[i] = va_arg(ap, void*);
    }
    va_end(ap);
    return work_queue_.Enqueue(info);
}

template<size_t N>
void LockFreeThreadPool<N>::PrimaryWorkerThread(void* p) NLIB_NOEXCEPT {
    LockFreeThreadPool* This = reinterpret_cast<LockFreeThreadPool*>(p);
    LockFreeThreadPoolExecInfo info;

    for (;;) {
        errno_t e = This->work_queue_.Dequeue(&info);
        if (!This->PrimaryWorkerThread_(info.func, info.callback, e == 0 ? &info.arg[0] : nullptr,
                                        N))
            break;
    }
}

template<size_t N>
void LockFreeThreadPool<N>::WorkerThread(void* p) NLIB_NOEXCEPT {
    LockFreeThreadPool* This = reinterpret_cast<LockFreeThreadPool*>(p);
    LockFreeThreadPoolExecInfo info;
    uint32_t counter = 0;

    for (;;) {
        errno_t e = This->work_queue_.Dequeue(&info);
        if (!This->WorkerThread_(info.func, info.callback, e == 0 ? &info.arg[0] : nullptr, N,
                                 &counter))
            break;
    }
}

}  // namespace threading
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_
