﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_TLS_H_
#define INCLUDE_NN_NLIB_THREADING_TLS_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

template<class T>
class Tls NLIB_FINAL {
 public:
    NLIB_CEXPR Tls() NLIB_NOEXCEPT : idx_(NLIB_TLS_INVALID) {}
    errno_t Init() NLIB_NOEXCEPT {
        if (idx_ != NLIB_TLS_INVALID) return EALREADY;
        return nlib_tls_alloc(&idx_, destr_func);
    }
    ~Tls() NLIB_NOEXCEPT {
        if (idx_ != NLIB_TLS_INVALID) nlib_tls_free(idx_);
    }
#ifdef __cpp_rvalue_references
    Tls(Tls&& rhs) NLIB_NOEXCEPT : idx_(rhs.idx_) { rhs.idx_ = NLIB_TLS_INVALID; }
    Tls& operator=(Tls&& rhs) NLIB_NOEXCEPT {
        idx_     = rhs.idx_;
        rhs.idx_ = NLIB_TLS_INVALID;
        return *this;
    }
#endif
    Tls(Tls& rhs, move_tag) NLIB_NOEXCEPT {
        idx_     = rhs.idx_;
        rhs.idx_ = NLIB_TLS_INVALID;
    }
    Tls& assign(Tls& rhs, move_tag) NLIB_NOEXCEPT {
        idx_     = rhs.idx_;
        rhs.idx_ = NLIB_TLS_INVALID;
        return *this;
    }
    T* Get() const NLIB_NOEXCEPT {
        void* p;
        errno_t e = nlib_tls_getvalue(idx_, &p);
        if (e != 0) return nullptr;
        return reinterpret_cast<T*>(p);
    }
    T* operator->() const NLIB_NOEXCEPT { return this->Get(); }
    T& operator*() const NLIB_NOEXCEPT { return *this->Get(); }
    T* Release() NLIB_NOEXCEPT {
        T* obj = this->Get();
        nlib_tls_setvalue(idx_, nullptr);
        return obj;
    }
    errno_t Reset(T* p = 0) NLIB_NOEXCEPT {
        T* obj = this->Get();
        if (obj) {
            delete obj;
        }
        return nlib_tls_setvalue(idx_, p);
    }

 private:
    static void destr_func(void* p) NLIB_NOEXCEPT { delete reinterpret_cast<T*>(p); }
    nlib_tls idx_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Tls);
};

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP_T_BEGIN3(nn, nlib, threading)
NLIB_DEFINE_STD_SWAP_T1(T, NLIB_NS::threading::Tls)
NLIB_DEFINE_STD_SWAP_T_END3(nn, nlib, threading)

#endif  // INCLUDE_NN_NLIB_THREADING_TLS_H_
