﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#ifndef SAMPLES_SERIALIZER_DEF_SERIALIZER_H_  // NOLINT
#define SAMPLES_SERIALIZER_DEF_SERIALIZER_H_

#include <map>
#include <string>
#include <vector>

#include "nn/nlib/exi/exi.h"
#include "nn/nlib/UniquePtr.h"

#define N(x) NLIB_EXI_LITERAL(x)
#define M(x) NLIB_EXI_UTF8(x)
typedef std::basic_string< ::nlib_ns::exi::ExiChar> StdString;

namespace serializer {
// Non-Intrusive style is possible if you define serialize() in serializer namespace.
template <class Archive, class T>
void serialize(Archive& ar, T& x) {  // NOLINT
    ar.serialize(x);
}
}  // namespace serializer

class SimpleSerializer {
    nlib_ns::UniquePtr<nlib_ns::exi::XmlStreamWriter> writer_;

 public:
    explicit SimpleSerializer(nlib_ns::OutputStream* os);
    ~SimpleSerializer() {}

 public:
    template <class T>
    void operator&(const T& x) {
        // Calls serialize() member function via ::serializer::serialize().
        ::serializer::serialize(*this, x);
    }

    // Writes the object in binary XML.
    template <class T>
    void operator<<(const T& x);

    // Is serialization successful?
    bool IsSuccess() const { return !!*writer_; }

 private:
    template <class T>
    void serialize(const T& x);

    template <class T, std::size_t N>
    void serialize(const T (&seq)[N]);

    template <class T, class Al>
    void serialize(const std::vector<T, Al>& seq);

    template <class Key, class Value, class Pred, class Al>
    void serialize(const std::map<Key, Value, Pred, Al>& m);

 private:
    template <class T>
    void operator&(const T* /* x */);

    template <class Archive, class T>
    friend void serializer::serialize(Archive& ar, T& x);  // NOLINT
};

template <class T>
void SimpleSerializer::operator<<(const T& x) {
    writer_->WriteStartDocument();
    ::serializer::serialize(*this, x);
    writer_->WriteEndDocument();
    writer_->Flush();
}

template <class T>
void SimpleSerializer::serialize(const T& x) {
    writer_->WriteStartElement(N("o"));
    const_cast<T&>(x).serialize(*this);
    writer_->WriteEndElement();
}

// Writes <s>...</s>
template <>
void SimpleSerializer::serialize(const StdString& x);

// Writes <i>...</i>
template <>
void SimpleSerializer::serialize(const int& x);

// Writes <f>...</f>
template <>
void SimpleSerializer::serialize(const float& x);

// Writes <v>...</v>
template <class T, std::size_t N>
void SimpleSerializer::serialize(const T (&seq)[N]) {
    writer_->WriteStartElement(N("v"));
    std::size_t i;
    for (i = 0; i < N; ++i) {
        (*this) & seq[i];
    }
    writer_->WriteEndElement();
}

// Writes <v>...</v>
template <class T, class Al>
void SimpleSerializer::serialize(const std::vector<T, Al>& seq) {
    writer_->WriteStartElement(N("v"));
    std::size_t i;
    std::size_t size = seq.size();
    for (i = 0; i < size; ++i) {
        (*this) & seq[i];
    }
    writer_->WriteEndElement();
}

// Writes <m>...</m>
template <class Key, class Value, class Pred, class Al>
void SimpleSerializer::serialize(const std::map<Key, Value, Pred, Al>& m) {
    writer_->WriteStartElement(N("m"));
    typename std::map<Key, Value, Pred, Al>::const_iterator it;
    typename std::map<Key, Value, Pred, Al>::const_iterator itend = m.end();
    for (it = m.begin(); it != itend; ++it) {
        (*this) & it->first;
        (*this) & it->second;
    }
    writer_->WriteEndElement();
}

class SimpleDeserializer {
    nlib_ns::exi::XmlStreamReader::XmlStreamConstants cache_;
    nlib_ns::UniquePtr<nlib_ns::exi::XmlStreamReader> reader_;
    bool is_failed_;
    char _[3];

 public:
    explicit SimpleDeserializer(nlib_ns::InputStream* is);
    ~SimpleDeserializer() {}

 public:
    template <class T>
    void operator&(T& x) {  // NOLINT
        // Calls serialize() member function via ::serializer::serialize().
        ::serializer::serialize(*this, x);
    }

    template <class T>
    void operator>>(T& x);

    bool IsSuccess() const { return !!*reader_ && !is_failed_; }

 private:
    template <class T>
    void serialize(T& x);  // NOLINT

    template <class T, std::size_t N>
    void serialize(T (&seq)[N]);

    template <class T, class Al>
    void serialize(std::vector<T, Al>& seq);  // NOLINT

    template <class Key, class Value, class Pred, class Al>
    void serialize(std::map<Key, Value, Pred, Al>& m);  // NOLINT

    void Next(nlib_ns::exi::XmlStreamReader::XmlStreamConstants expect);
    bool Peek(nlib_ns::exi::XmlStreamReader::XmlStreamConstants expect);

    template <class T>
    void operator&(const T* /* x */);

    template <class Archive, class T>
    friend void serializer::serialize(Archive& ar, T& x);  // NOLINT
};

template <class T>
void SimpleDeserializer::operator>>(T& x) {
    using nlib_ns::exi::XmlStreamReader;
    this->Next(XmlStreamReader::START_DOCUMENT);
    ::serializer::serialize(*this, x);
    this->Next(XmlStreamReader::END_DOCUMENT);
}

template <class T>
void SimpleDeserializer::serialize(T& x) {  // NOLINT
    reader_->Next();                       // StartElement("o")
    x.serialize(*this);
    reader_->Next();  // EndElement("o");
}

template <>
void SimpleDeserializer::serialize(StdString& x);  // NOLINT

template <>
void SimpleDeserializer::serialize(int& x);  // NOLINT

template <>
void SimpleDeserializer::serialize(float& x);  // NOLINT

template <class T, std::size_t N>
void SimpleDeserializer::serialize(T (&seq)[N]) {
    using nlib_ns::exi::XmlStreamReader;
    this->Next(XmlStreamReader::START_ELEMENT);  // v
    std::size_t i;
    for (i = 0; i < N; ++i) {
        (*this) & seq[i];
    }
    this->Next(XmlStreamReader::END_ELEMENT);  // v
}

template <class T, class Al>
void SimpleDeserializer::serialize(std::vector<T, Al>& seq) {  // NOLINT
    using nlib_ns::exi::XmlStreamReader;
    this->Next(XmlStreamReader::START_ELEMENT);  // v
    seq.clear();
    while (this->Peek(XmlStreamReader::START_ELEMENT)) {
        T x;
        (*this) & x;
        seq.push_back(x);
    }
    this->Next(XmlStreamReader::END_ELEMENT);  // v
}

template <class Key, class Value, class Pred, class Al>
void SimpleDeserializer::serialize(std::map<Key, Value, Pred, Al>& m) {  // NOLINT
    using nlib_ns::exi::XmlStreamReader;
    this->Next(XmlStreamReader::START_ELEMENT);  // m
    m.clear();
    while (this->Peek(XmlStreamReader::START_ELEMENT)) {
        Key key;
        Value value;
        (*this) & key;
        (*this) & value;
        m[key] = value;
    }
    this->Next(XmlStreamReader::END_ELEMENT);  // m
}

#endif  // SAMPLES_SERIALIZER_DEF_SERIALIZER_H_ // NOLINT
