﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#include <string>
#include "nn/nlib/Platform.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/NativePathMapper.h"
#include "nn/nlib/DateTime.h"

#ifdef _MSC_VER
#pragma execution_character_set("utf-8")
#endif

#define SUCCEED_IF(cond) if (!(cond)) return false

NLIB_PATHMAPPER_FORSAMPLE

namespace {
nlib_thread_id GetMyThreadId() {
    nlib_thread_id id;
    nlib_thread_getid(&id);
    return id;
}

namespace barrier {
//! [barrier]
bool Barrier() {
    const int kNumThread = 4;
    nlib_thread th[kNumThread];
    nlib_barrier barrier;
    errno_t e = nlib_barrier_init(&barrier, kNumThread);
    SUCCEED_IF(e == 0);
    for (int i = 0; i < kNumThread; ++i) { th[i] = NLIB_THREAD_INVALID; }
    for (int i = 0; i < kNumThread; ++i) {
        e = nlib_thread_create(&th[i], nullptr, [](void* p) {
            nlib_barrier* pbarrier = reinterpret_cast<nlib_barrier*>(p);
            nlib_printf("Thread %d Step1\n", GetMyThreadId());
            nlib_barrier_wait(pbarrier);  // blocks until kNumThread threads comes here
            nlib_printf("Thread %d Step2\n", GetMyThreadId());
            nlib_barrier_wait(pbarrier);  // blocks until kNumThread threads comes here
        }, &barrier);
        SUCCEED_IF(e == 0);
    }
    for (int i = 0; i < kNumThread; ++i) { (void)nlib_thread_join(th[i]); }
    (void)nlib_barrier_destroy(&barrier);
    return true;
}
/*
    Output:
    Thread 9008 Step1
    Thread 19996 Step1
    Thread 9892 Step1
    Thread 14120 Step1
    Thread 14120 Step2
    Thread 9008 Step2
    Thread 19996 Step2
    Thread 9892 Step2
*/
//! [barrier]

}  // namespace barrier

namespace once {
//! [once]
void ThreadFunc(void*) {
    static nlib_onceflag onceflag = NLIB_ONCE_INIT;  // static initialization required
    nlib_printf("Thread #%d: start\n", GetMyThreadId());
    nlib_once(&onceflag, []() { // nlib_oncefunc
        nlib_printf("OnceFunc start\n");
        nlib_sleep(nlib_ns::TimeSpan(0, 2).ToTimeValue().tick);  // 2000 msec
        nlib_printf("OnceFunc end\n");
    });
    nlib_printf("Thread #%d: end\n", GetMyThreadId());
}
bool Once() {
    const int kNumThread = 4;
    nlib_thread th[kNumThread];
    for (int i = 0; i < kNumThread; ++i) { th[i] = NLIB_THREAD_INVALID; }
    for (int i = 0; i < kNumThread; ++i) {
        errno_t e = nlib_thread_create(&th[i], nullptr, ThreadFunc, nullptr);
        SUCCEED_IF(e == 0);
    }
    for (int i = 0; i < kNumThread; ++i) { (void)nlib_thread_join(th[i]); }
    ThreadFunc(nullptr);
    return true;
}
/*
    Output:
    Thread #19972: start
    Thread #892: start
    OnceFunc start
    Thread #13624: start
    Thread #10668: start
    OnceFunc end
    Thread #19972: end
    Thread #10668: end
    Thread #13624: end
    Thread #892: end
    Thread #19176: start
    Thread #19176: end
*/
//! [once]
}  // namespace once

namespace datetime {
bool TimeSpec() {
//! [timespec]
    struct timespec tm;
    nlib_time t;
    nlib_epochtime_timespec(&tm);

    // timespec -> nlib_time
    NLIB_FROM_TIMESPEC(&tm, t);

    // timespec <- nlib_time
    NLIB_TO_TIMESPEC(&tm, t);
//! [timespec]
    return true;
}
}  // namespace datetime

// nlib_virtual_alloc()
// message queue
// debug output
// strtol etc.
bool skipws() {
//! [skipws]
    const char* text = " \t\n \n string";
    size_t cnt_lf;
    const char* last_lf;
    const char* target = nlib_skipws(&cnt_lf, &last_lf, text, nlib_strlen(text));
    nlib_printf("Target String: %s\n", target);
    nlib_printf("LF found: %" PRIuS "\n", cnt_lf);
    nlib_printf("Last LF pos: %" PRIdS "\n", last_lf - text);
/*
    Output:
    Target String: string
    LF found: 2
    Last LF pos: 4
*/ 
//! [skipws]
    return true;
}

bool sleep() {
//! [sleep]
    nlib_sleep(10000);  // sleep for 1 msec
    nlib_sleep(nlib_ns::TimeSpan(0, 0, 1).ToTimeValue().tick);  // sleep for 1 msec
//! [sleep]
    return true;
}

bool clz32() {
//! [clz32]
    nlib_printf("nlib_clz32(0x80000000) -> %d\n", nlib_clz32(0x80000000));
    nlib_printf("nlib_clz32(0x00000001) -> %d\n", nlib_clz32(0x00000001));
/*
    Output:
    nlib_clz32(0x80000000) -> 0
    nlib_clz32(0x00000001) -> 31
*/ 
//! [clz32]
    return true;
}

bool clz64() {
//! [clz64]
    nlib_printf("nlib_clz64(0x8000000000000000ULL) -> %d\n", nlib_clz64(0x8000000000000000ULL));
    nlib_printf("nlib_clz64(0x0000000000000001ULL) -> %d\n", nlib_clz64(0x0000000000000001ULL));
/*
    Output:
    nlib_clz64(0x8000000000000000ULL) -> 0
    nlib_clz64(0x0000000000000001ULL) -> 63
*/
//! [clz64]
    return true;
}

bool ctz32() {
//! [ctz32]
    nlib_printf("nlib_ctz32(0x80000000) -> %d\n", nlib_ctz32(0x80000000));
    nlib_printf("nlib_ctz32(0x00000001) -> %d\n", nlib_ctz32(0x00000001));
/*
    Output:
    nlib_ctz32(0x80000000) -> 31
    nlib_ctz32(0x00000001) -> 0
*/
//! [ctz32]
    return true;
}

bool ctz64() {
    //! [ctz64]
    nlib_printf("nlib_ctz64(0x8000000000000000ULL) -> %d\n", nlib_ctz64(0x8000000000000000ULL));
    nlib_printf("nlib_ctz64(0x0000000000000001ULL) -> %d\n", nlib_ctz64(0x0000000000000001ULL));
/*
    Output:
    nlib_ctz64(0x8000000000000000ULL) -> 63
    nlib_ctz64(0x0000000000000001ULL) -> 0
*/
    //! [ctz64]
    return true;
}

bool memccpy() {
    //! [memccpy]
    const size_t kBufSize = 1024;
    char buf[kBufSize];
    const char* src = "line1\nline2\n";
    char* p = (char*)nlib_memccpy(buf, kBufSize, src, nlib_strlen(src), '\n');
    SUCCEED_IF(p != nullptr);  // nullptr if the line is too long
    *(p - 1) = '\0';  // buf is not null-terminated, replace '\n' with '\0'
    nlib_printf("Copied text is: %s\n", buf);
/*
    Output:
    Copied text is: line1
*/
    //! [memccpy]
    return true;
}

bool timer() {
    //! [timer]
    errno_t e;
    nlib_timer timer;
    int counter = 0;
    nlib_duration _100msec = nlib_ns::TimeSpan(0, 0, 100).ToTimeValue().tick;
    e = nlib_timer_create(&timer, [](nlib_timer t, void* p) {
        // NOTE:
        // callback function can be called in parallel
        int32_t* pcounter = reinterpret_cast<int*>(p);
        int32_t cnt = nlib_atomic_add_fetch32(pcounter, 1, NLIB_ATOMIC_SEQ_CST);
        nlib_printf("counter = %d\n", cnt);
        if (cnt == 5) {
            nlib_timer_delete(t, 1, NULL);
        }
    }, &counter, 0);
    SUCCEED_IF(e == 0);
    nlib_timerspec spec = {
        _100msec * 2,  // due_time
        _100msec       // interval
    };
    e = nlib_timer_settime(timer, &spec, NULL);
    if (e != 0) {
        nlib_timer_delete(timer, 1, NULL);
        SUCCEED_IF(0);
    }
    while (counter < 5) {
        nlib_sleep(_100msec);
    }
/*
    Output:
    counter = 1
    counter = 2
    counter = 3
    counter = 4
    counter = 5
*/
    //! [timer]
    return true;
}

bool from_chars() {
    //! [from_chars]
    int32_t result;
    const char* text;
    const char* endptr;
    errno_t e;

    // returns 0 and set 'result' 12345. endptr also moves forward.
    text = "12345;";
    e = nlib_int32_from_chars(&result, &endptr, text, text + nlib_strlen(text), 10);
    nlib_printf("result=%d, ERROR=%s, *endptr='%c'\n", result, nlib_error_string(e), *endptr);

    // returns ERANGE if overflow, and result_i8 unchanged, but endptr moves forward.
    int8_t result_i8;
    e = nlib_int8_from_chars(&result_i8, &endptr, text, text + nlib_strlen(text), 10);
    nlib_printf("ERROR=%s, *endptr='%c'\n", nlib_error_string(e), *endptr);

    // returns EINVAL and spaces are not skipped. endptr does not move forwad.
    text = ";12345";
    e = nlib_int32_from_chars(&result, &endptr, text, text + nlib_strlen(text), 10);
    nlib_printf("ERROR=%s, *endptr='%c'\n", nlib_error_string(e), *endptr);
/*
    Output:
    result=12345, ERROR=OK, *endptr=';'
    ERROR=ERANGE, *endptr=';'
    ERROR=EINVAL, *endptr=';'
*/
    //! [from_chars]
    return true;
}

bool logging() {
    //! [logging]
    nlib_log_attr_setint(
        kNlibLogWarn | kNlibLogLevelEqualOrAbove,  // priority: output if warn, error, fatal level
        kNlibLogAttrStderr,                        // channel: stdout,stderr,fd, ...
        1                                          // 0: OFF, otherwise: ON
    );
    const char* tag = "tag_to_grep";
    nlib_log_print(kNlibLogVerbose, tag, "Verbose level not displayed");
    nlib_log_print(kNlibLogDebug, tag, "Debug level not displayed");
    nlib_log_print(kNlibLogWarn, tag, "Warn level displayed");
    nlib_log_print(kNlibLogError, tag, "Error level displayed");
    nlib_log_print(kNlibLogError, tag, "Fatal level displayed");
/*
    Output:
    tag_to_grep: Warn level displayed
    tag_to_grep: Error level displayed
    tag_to_grep: Fatal level displayed
*/
    //! [logging]
    return true;
}

bool find_break() {
    //! [find_break]
    // 1 0xE2 0x83 0xA3 (1⃣, Emoji one, 1 U+20E3)
    // 0xF0 0x9F 0x8F 0x83 0xF0 0x9F 0x8F 0xBF (🏃🏿, running man with skin tone selector, U+1F3C3 U+1F3FF)
    // 0xF0 0x9F 0x87 0xAF 0xF0 0x9F 0x87 0xB5 (🇯🇵, JP, U+1F1EF U+1F1F5)
    const nlib_utf8_t* str = "1\xE2\x83\xA3\xF0\x9F\x8F\x83\xF0\x9F\x8F\xBF\xF0\x9F\x87\xAF\xF0\x9F\x87\xB5";
    const nlib_utf8_t* first = str;
    const nlib_utf8_t* last = str + nlib_strlen(str);
    while (first != last) {
        const nlib_utf8_t* p;
        nlib_find_break(&p, first, last, kUnicodeBreakPropertyGrapheme, kUnicodeBreakDefault);
        nlib_printf("'%s'\n", std::string(first, p).c_str());
        first = p;
    }
/*
    Output:
    '1⃣'
    '🏃🏿'
    '🇯🇵'
*/
    //! [find_break]
    return true;
}

}  // namespace

#define EXEC_SAMPLE(text, func) \
    nlib_printf("==== %s ====\n", text); \
    if (!func()) { nlib_printf("%s\n", text); return false; }

static bool capi() {
    EXEC_SAMPLE("Barrier", barrier::Barrier);
    EXEC_SAMPLE("Once", once::Once);
    EXEC_SAMPLE("TimeSpec", datetime::TimeSpec);
    EXEC_SAMPLE("nlib_skipws", skipws);
    EXEC_SAMPLE("nlib_sleep", skipws);
    EXEC_SAMPLE("nlib_clz32", clz32);
    EXEC_SAMPLE("nlib_clz64", clz64);
    EXEC_SAMPLE("nlib_ctz32", ctz32);
    EXEC_SAMPLE("nlib_ctz64", ctz64);
    EXEC_SAMPLE("nlib_memccpy", memccpy);
    EXEC_SAMPLE("nlib_timer", timer);
    EXEC_SAMPLE("nlib_xxxx_from_chars", from_chars);
    EXEC_SAMPLE("logging", logging);
    EXEC_SAMPLE("sleep", sleep);
    EXEC_SAMPLE("nlib_find_break", find_break);
    return true;
}

extern bool cppapi();
bool SampleMain(int, char**) {
    return capi() && cppapi();
}

NLIB_MAINFUNC
