﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#include <thread>
#include <algorithm>
#include <random>

#include "nn/nlib/Platform.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/NativePathMapper.h"
#include "nn/nlib/DateTime.h"
#include "nn/nlib/threading/Barrier.h"
#include "nn/nlib/LockFree.h"
#include "nn/nlib/Uri.h"
#include "nn/nlib/UriTemplate.h"
#include "nn/nlib/StringView.h"
#include "nn/nlib/Base64.h"
#include "nn/nlib/Nflags.h"
#include "nn/nlib/DateTime.h"
#include "nn/nlib/MemoryInputStream.h"
#include "nn/nlib/MemoryOutputStream.h"
#include "nn/nlib/BinaryReader.h"
#include "nn/nlib/BinaryWriter.h"
#include "nn/nlib/SmartBitmap.h"
#include "nn/nlib/ConsoleOutputStream.h"
#include "nn/nlib/TextWriter.h"
#include "nn/nlib/TextReader.h"
#include "nn/nlib/TaggedTextParser.h"
#include "nn/nlib/InputConverterStream.h"
#include "nn/nlib/Base64InputStream.h"
#include "nn/nlib/Base64OutputStream.h"
#include "nn/nlib/WcharInputStream.h"
#include "nn/nlib/CurlInputStream.h"

#define SUCCEED_IF(cond) if (!(cond)) return false

namespace {
nlib_thread_id GetMyThreadId() {
    nlib_thread_id id;
    nlib_thread_getid(&id);
    return id;
}

bool Barrier() {
    //! [barrier]
    const int kNumThread = 4;
    std::thread th_list[kNumThread];
    nlib_ns::threading::Barrier barrier;
    errno_t e = barrier.Init(kNumThread);
    SUCCEED_IF(e == 0);
    for (auto& th : th_list) {
        th = std::thread([](void* p) {
            nlib_barrier* pbarrier = reinterpret_cast<nlib_barrier*>(p);
            nlib_printf("Thread %d Step1\n", GetMyThreadId());
            nlib_barrier_wait(pbarrier);  // blocks until kNumThread threads comes here
            nlib_printf("Thread %d Step2\n", GetMyThreadId());
            nlib_barrier_wait(pbarrier);  // blocks until kNumThread threads comes here
        }, &barrier);
    }
    for (auto& th : th_list) {
        th.join();
    }
/*
    Output:
    Thread 2388 Step1
    Thread 15328 Step1
    Thread 8488 Step1
    Thread 16144 Step1
    Thread 16144 Step2
    Thread 15328 Step2
    Thread 8488 Step2
    Thread 2388 Step2
*/
    //! [barrier]
    return true;
}

bool LockFreePipe() {
    //! [LockFreePipe]
    nlib_ns::LockFreePipe<32> pipe;
    std::thread sender([&]() {
        // send strlen(text), text
        const int kTextListLength = 4;
        const char* text_list[kTextListLength] = {
            "012",
            "short text",
            "0123456789012345678901234567890",
            "#end"
        };
        for (int i = 0; i < kTextListLength; ++i) {
            const char* text = text_list[i];
            uint8_t text_len = static_cast<uint8_t>(nlib_strlen(text));
            while (pipe.Write(&text_len, sizeof(text_len)) != 0) {
                std::this_thread::yield();
            }
            while (pipe.Write(text, text_len) != 0) { std::this_thread::yield(); }
        }
    });
    std::thread receiver([&]() {
        for (;;) {
            uint8_t text_len;
            char buf[256];
            while (pipe.Read(&text_len, sizeof(text_len)) != 0) {
                std::this_thread::yield();
            }
            while (pipe.Read(&buf[0], text_len) != 0) { std::this_thread::yield(); }
            buf[text_len] = '\0';
            if (text_len == 4 && nlib_memcmp("#end", buf, text_len) == 0) { return; }
            nlib_printf("Received: %s\n", buf);
        }
    });
    sender.join();
    receiver.join();
/*
    Output:
    Received: 012
    Received: short text
    Received: 0123456789012345678901234567890
*/
    //! [LockFreePipe]
    return true;
}

bool LockFreeInit() {
    //! [LockFreeInit]
    struct MyObject {
        MyObject() {
            nlib_printf("MyObject(%p) is constructed by thread %d\n",
                this, GetMyThreadId());
            nlib_sleep(10000);
        }
        ~MyObject() {
            nlib_printf("MyObject(%p) is destroyed by thread %d\n",
                this, GetMyThreadId());
        }
        void UseObject() {
            nlib_printf("Object(%p) used from thread %d\n",
                this, GetMyThreadId());
        }
    };
    static MyObject* myobj = NULL;
    const int kNumThread = 8;
    std::thread th_list[kNumThread];
    for (auto& th : th_list) {
        th = std::thread([&]() {
            errno_t e = nlib_ns::LockFreeInit(&myobj);
            if (e == 0) { myobj->UseObject(); }
        });
    }
    for (auto& th : th_list) { th.join(); }
    MyObject* p = myobj;
    myobj = nullptr;  // use nlib_atomic_compare_exchangeptr if used in other places.
    delete p;
/*
    Output:
    MyObject(0000017EF4EF16F0) is constructed by thread 16972
    MyObject(0000017EF4EF1670) is constructed by thread 5744
    MyObject(0000017EF4EF1430) is constructed by thread 9036
    MyObject(0000017EF4EF0FF0) is constructed by thread 232
    MyObject(0000017EF4EF1470) is constructed by thread 12732
    MyObject(0000017EF4EF14B0) is constructed by thread 15228
    MyObject(0000017EF4F2DB80) is constructed by thread 5792
    MyObject(0000017EF4F2D640) is constructed by thread 7100
    Object(0000017EF4EF1670) used from thread 5744
    MyObject(0000017EF4EF1430) is destroyed by thread 9036
    MyObject(0000017EF4F2DB80) is destroyed by thread 5792
    MyObject(0000017EF4EF0FF0) is destroyed by thread 232
    MyObject(0000017EF4EF14B0) is destroyed by thread 15228
    Object(0000017EF4EF1670) used from thread 232
    MyObject(0000017EF4EF1470) is destroyed by thread 12732
    Object(0000017EF4EF1670) used from thread 9036
    Object(0000017EF4EF1670) used from thread 15228
    Object(0000017EF4EF1670) used from thread 12732
    Object(0000017EF4EF1670) used from thread 5792
    MyObject(0000017EF4F2D640) is destroyed by thread 7100
    MyObject(0000017EF4EF16F0) is destroyed by thread 16972
    Object(0000017EF4EF1670) used from thread 7100
    Object(0000017EF4EF1670) used from thread 16972
    MyObject(0000017EF4EF1670) is destroyed by thread 15912
*/
    //! [LockFreeInit]
    return true;
}

bool LockFreeStack() {
    //! [LockFreeStack]
    struct Pod { nlib_thread_id threadid; };
    nlib_ns::LockFreeStack<Pod> stack;
    errno_t e = stack.Init(128);
    SUCCEED_IF(e == 0);
    const int kNumThread = 8;
    std::thread th_list[kNumThread];
    for (auto& th : th_list) {
        th = std::thread([&]() {
            Pod pod;
            pod.threadid = GetMyThreadId();
            for (int i = 0; i < 2; ++i) {
                while (stack.Push(pod) != 0) {
                    nlib_yield();
                }
                nlib_sleep(10000);
            }
        });
    }
    for (auto& th : th_list) { th.join(); }
    Pod pod;
    for (int i = 0; i < kNumThread * 2; ++i) {
        e = stack.PopUnsafe(&pod);  // you can use PushUnsafe/PopUnsafe if single threaded
        SUCCEED_IF(e == 0);
        nlib_printf("pop: threadid = %d\n", pod.threadid);
    }
    SUCCEED_IF(stack.PopUnsafe(&pod) == EAGAIN);  // empty
/*
    Output:
    pop: threadid = 24248
    pop: threadid = 14012
    pop: threadid = 21972
    pop: threadid = 7384
    pop: threadid = 17956
    pop: threadid = 19628
    pop: threadid = 21608
    pop: threadid = 16848
    pop: threadid = 24248
    pop: threadid = 14012
    pop: threadid = 17956
    pop: threadid = 19628
    pop: threadid = 7384
    pop: threadid = 21972
    pop: threadid = 21608
    pop: threadid = 16848
*/
    //! [LockFreeStack]
    return true;
}

bool LockFreeQueue() {
    //! [LockFreeQueue]
    struct Pod { nlib_thread_id threadid; };
    nlib_ns::LockFreeQueue<Pod> queue;
    errno_t e = queue.Init(128);
    SUCCEED_IF(e == 0);
    const int kNumThread = 8;
    std::thread th_list[kNumThread];
    for (auto& th : th_list) {
        th = std::thread([&]() {
            Pod pod;
            pod.threadid = GetMyThreadId();
            for (int i = 0; i < 2; ++i) {
                while (queue.Enqueue(pod) != 0) {
                    nlib_yield();
                }
                nlib_sleep(10000);
            }
        });
    }
    for (auto& th : th_list) { th.join(); }
    Pod pod;
    for (int i = 0; i < kNumThread * 2; ++i) {
        e = queue.DequeueUnsafe(&pod);  // you can use DequeueUnsafe/DequeueUnsafe if single threaded
        SUCCEED_IF(e == 0);
        nlib_printf("dequeue: threadid = %d\n", pod.threadid);
    }
    SUCCEED_IF(queue.DequeueUnsafe(&pod) == EAGAIN);  // empty
/*
    Output:
    dequeue: threadid = 18992
    dequeue: threadid = 10128
    dequeue: threadid = 19896
    dequeue: threadid = 12040
    dequeue: threadid = 10120
    dequeue: threadid = 23808
    dequeue: threadid = 4872
    dequeue: threadid = 19440
    dequeue: threadid = 19896
    dequeue: threadid = 4872
    dequeue: threadid = 19440
    dequeue: threadid = 10128
    dequeue: threadid = 23808
    dequeue: threadid = 18992
    dequeue: threadid = 12040
    dequeue: threadid = 10120
*/
    //! [LockFreeQueue]
    return true;
}

bool LockFreePriorityQueue() {
    //! [LockFreePriorityQueue]
    struct Task {
        const char* text;
        int threadid;
    };
    nlib_ns::LockFreePriorityQueue<Task> queue;
    errno_t e = queue.Init(512);
    SUCCEED_IF(e == 0);
    const int kNumThread = 3;
    std::thread receiver_thread_list[kNumThread];
    for (auto& th : receiver_thread_list) {
        th = std::thread([&]() {
            for (int i = 0; i < 8; ++i) {
                int prio;
                nlib_ns::LockFreePriorityQueue<Task>::DequeueType item;
                while (queue.Dequeue(item, &prio) != 0) { nlib_yield(); }
                nlib_printf("Receiver(%d) receives '%s' from Sender(%d)\n",
                    GetMyThreadId(), item->text, item->threadid);
            }
        });
    }
    std::thread sender_thread_list[kNumThread];
    for (auto& th : sender_thread_list) {
        th = std::thread([&]() {
            const char* text_list[] = {
                "Low priority task",
                "Normal priority task",
                "High priority task"
            };
            std::random_device rd;
            std::mt19937 mt(rd());
            std::uniform_int_distribution<int> dice(0, 2);
            for (int i = 0; i < 8; ++i) {
                int prio = dice(mt);
                std::unique_ptr<Task> task(new (std::nothrow) Task());
                task->text = text_list[prio];
                task->threadid = GetMyThreadId();
                Task* p = task.release();
                while (queue.Enqueue(p, prio) != 0) { nlib_yield(); }
            }
        });
    }
    for (auto& th : sender_thread_list) { th.join(); }
    for (auto& th : receiver_thread_list) { th.join(); }
    (void)queue.Close();
/*
    Output:
    Receiver(23616) receives 'High priority task' from Sender(19036)
    Receiver(2472) receives 'High priority task' from Sender(19036)
    Receiver(16612) receives 'Normal priority task' from Sender(19036)
    Receiver(23616) receives 'High priority task' from Sender(19036)
    Receiver(2472) receives 'High priority task' from Sender(19036)
    Receiver(16612) receives 'High priority task' from Sender(1612)
    Receiver(23616) receives 'High priority task' from Sender(1612)
    Receiver(2472) receives 'High priority task' from Sender(1612)
    Receiver(16612) receives 'High priority task' from Sender(23708)
    Receiver(23616) receives 'High priority task' from Sender(23708)
    Receiver(2472) receives 'Normal priority task' from Sender(19036)
    Receiver(16612) receives 'Normal priority task' from Sender(19036)
    Receiver(23616) receives 'Normal priority task' from Sender(19036)
    Receiver(2472) receives 'Normal priority task' from Sender(1612)
    Receiver(16612) receives 'Normal priority task' from Sender(1612)
    Receiver(23616) receives 'Normal priority task' from Sender(1612)
    Receiver(2472) receives 'Normal priority task' from Sender(23708)
    Receiver(16612) receives 'Normal priority task' from Sender(23708)
    Receiver(23616) receives 'Normal priority task' from Sender(23708)
    Receiver(2472) receives 'Low priority task' from Sender(1612)
    Receiver(23616) receives 'Low priority task' from Sender(23708)
    Receiver(16612) receives 'Low priority task' from Sender(1612)
    Receiver(2472) receives 'Low priority task' from Sender(23708)
    Receiver(16612) receives 'Low priority task' from Sender(23708)
*/
    //! [LockFreePriorityQueue]
    return true;
}

bool LockFreeBroadcastQueue() {
    //! [LockFreeBroadcastQueue]
    struct MyMessage { const char* text; };
    nlib_ns::LockFreeBroadcastQueue<MyMessage> queue;
    const int kNumReceivers = 3;
    errno_t e = queue.Init(128, kNumReceivers);
    SUCCEED_IF(e == 0);
    std::unique_ptr<MyMessage> obj;
    const char* message_list[] = {
        "Hello, this is the queue for broadcasting messages",
        "The messages are shared among the listeners.",
        "They are preserved until all the listeners read them.",
        "They are to be deleted automatically.",
        nullptr
    };
    for (auto& message : message_list) {
        obj.reset(new (std::nothrow) MyMessage());
        SUCCEED_IF(!!obj);
        obj->text = message;
        queue.Enqueue(obj.release());
    }
    std::thread th_list[kNumReceivers];
    for (int i = 0; i < kNumReceivers; ++i) {
        th_list[i] = std::thread([&](int listender_id) {
            nlib_ns::LockFreeBroadcastQueue<MyMessage>::DequeueType ptr;
            for (;;) {
                while (queue.Dequeue(listender_id, ptr) != 0) { nlib_yield(); }
                if (!ptr->text) break;
                nlib_printf("thread %d receives '%s'\n", GetMyThreadId(), ptr->text);
            }
        }, i);
    }
    for (auto& th : th_list) { th.join(); }
/*
    Output:
    thread 20568 receives 'Hello, this is the queue for broadcasting messages'
    thread 20568 receives 'The messages are shared among the listeners.'
    thread 20516 receives 'Hello, this is the queue for broadcasting messages'
    thread 20568 receives 'They are preserved until all the listeners read them.'
    thread 20516 receives 'The messages are shared among the listeners.'
    thread 20568 receives 'They are to be deleted automatically.'
    thread 22760 receives 'Hello, this is the queue for broadcasting messages'
    thread 20516 receives 'They are preserved until all the listeners read them.'
    thread 22760 receives 'The messages are shared among the listeners.'
    thread 20516 receives 'They are to be deleted automatically.'
    thread 22760 receives 'They are preserved until all the listeners read them.'
    thread 22760 receives 'They are to be deleted automatically.'
*/
    //! [LockFreeBroadcastQueue]
    return true;
}

bool LockFreeUnitHeap() {
    //! [LockFreeUnitHeap]
    struct MemoryBlock { char data[32]; };
    nlib_ns::LockFreeUnitHeap heap;
    nlib_ns::LockFreeStack<uintptr_t> stack;
    errno_t e;
    const int kBlockCount = 1024;
    e = heap.Init(sizeof(MemoryBlock), 32, kBlockCount);
    SUCCEED_IF(e == 0);
    e = stack.Init(kBlockCount);
    SUCCEED_IF(e == 0);

    const int kNumThread = 16;
    std::thread th_list[kNumThread];
    for (auto& th : th_list) {
        th = std::thread([&]() {
            for (int i = 0; i < kBlockCount / kNumThread; ++i) {
                auto holder = heap.Alloc();
                stack.Push(reinterpret_cast<uintptr_t>(holder));
            }
        });
    }
    for (auto& th : th_list) { th.join(); }
    uintptr_t vec[kBlockCount];
    for (int i = 0; i < kBlockCount; ++i) {
        uintptr_t item = 0;
        e = stack.PopUnsafe(&item);
        SUCCEED_IF(e == 0);
        vec[i] = item;
    }
    std::sort(&vec[0], &vec[0] + kBlockCount);
    auto it = std::unique(&vec[0], &vec[0] + kBlockCount);
    SUCCEED_IF(it == &vec[0] + kBlockCount);  // confirm they are all different
    for (auto& ptr : vec) {
        heap.FreeUnsafe(reinterpret_cast<decltype(heap.Alloc())>(ptr));
    }
    //! [LockFreeUnitHeap]
    return true;
}

bool Uri() {
    //! [Uri]
    nlib_ns::Uri uri;
    bool result = uri.Parse("http://www.example.com/dir/index.html?query=test#part1");
    SUCCEED_IF(!!result);
    nlib_printf("scheme: %s\n", uri.GetScheme());
    nlib_printf("host: %s\n", uri.GetHost());
    nlib_printf("path: %s\n", uri.GetPath());
    nlib_printf("query: %s\n", uri.GetQuery());
    nlib_printf("fragment: %s\n", uri.GetFragment());
/*
    Output:
    scheme: http
    host: www.example.com
    path: /dir/index.html
    query: query=test
    fragment: part1
*/
    //! [Uri]
    return true;
}

bool UriHostName() {
    //! [Uri::IsHostName]
    nlib_printf("IsHostName(\"www.example.com\"): %d\n",
        nlib_ns::Uri::IsHostName("www.example.com"));
    nlib_printf("IsHostName(\"www.-example-.com\"): %d\n",
        nlib_ns::Uri::IsHostName("www.-example-.com"));
    nlib_printf("IsHostName(\"www.example-host.com\"): %d\n",
        nlib_ns::Uri::IsHostName("www.example-host.com"));
    nlib_printf("IsHostName(\"www.example_host.com\"): %d\n",
        nlib_ns::Uri::IsHostName("www.example_host.com"));
    nlib_printf("IsHostName(\"www.example..com\"): %d\n",
        nlib_ns::Uri::IsHostName("www.example..com"));
/*
    Output:
    IsHostName("www.example.com"): 1
    IsHostName("www.-example-.com"): 0
    IsHostName("www.example-host.com"): 1
    IsHostName("www.example_host.com"): 0
    IsHostName("www.example..com"): 0
*/
    //! [Uri::IsHostName]
    return true;
}

bool UriEmailAddress() {
    //! [Uri::IsEmailAddress]
    nlib_printf("IsEmailAddress(foo@example.com): %d\n",
        nlib_ns::Uri::IsEmailAddress("foo@example.com"));
    nlib_printf("IsEmailAddress(foo.bar@example.com): %d\n",
        nlib_ns::Uri::IsEmailAddress("foo.bar@example.com"));
    nlib_printf("IsEmailAddress(.foo.@example.com): %d\n",
        nlib_ns::Uri::IsEmailAddress(".foo.@example.com"));
    nlib_printf("IsEmailAddress(\".(foo:).\"@example.com): %d\n",
        nlib_ns::Uri::IsEmailAddress("\".(foo:).\"@example.com"));
/*
    Output:
    IsEmailAddress(foo@example.com): 1
    IsEmailAddress(foo.bar@example.com): 1
    IsEmailAddress(.foo.@example.com): 0
    IsEmailAddress(".(foo:)."@example.com): 1
*/
    //! [Uri::IsEmailAddress]
    return true;
}

bool DecodeUriComponent() {
    //! [Uri::DecodeUriComponent]
    char buf[256];
    size_t written;
    nlib_ns::Uri::DecodeUriComponent(&written, buf, "https://ja.wikipedia.org/wiki/%E3%83%A1%E3%82%A4%E3%83%B3%E3%83%9A%E3%83%BC%E3%82%B8");
    nlib_printf("%s\n", buf);
/*
    Output:
    https://ja.wikipedia.org/wiki/メインページ
*/
    //! [Uri::DecodeUriComponent]
    return true;
}

/*
bool EncodeUriComponent() {
    //! [Uri::EncodeUriComponent]
    //! [Uri::EncodeUriComponent]
    return true;
}
*/

bool UriQueryDecoder() {
    //! [UriQueryDecoder]
    nlib_ns::Uri uri;
    bool result = uri.Parse("http://www.example.com?date=today");
    SUCCEED_IF(!!result);
    nlib_ns::UriQueryDecoder<> decoder;
    errno_t e = decoder.Init(uri);
    SUCCEED_IF(e == 0);
    while (decoder.HasNext()) {
        auto k = decoder.GetKey();
        SUCCEED_IF(k.first == 0);
        auto v = decoder.GetValue();
        SUCCEED_IF(v.first == 0);
        nlib_printf("%s = %s\n", k.second, v.second);
        decoder.MoveNext();
    }
/*
    Output:
    date = today
*/
    //! [UriQueryDecoder]
    return true;
}

bool UriQueryEncoder() {
    //! [UriQueryEncoder]
    nlib_ns::UriQueryEncoder<> encoder;
    errno_t e = encoder.Append("date", "Mon Dec 25 10:00:00 DST 2017");
    SUCCEED_IF(e == 0);
    nlib_printf("%s\n", encoder.c_str());
/*
    Output:
    date=Mon%20Dec%2025%2010%3A00%3A00%20DST%202017
*/
    //! [UriQueryEncoder]
    return true;
}

bool UriTemplate() {
    //! [UriTemplate]
    nlib_ns::UriTemplate templ;
    errno_t e = templ.SetTemplate("http://www.example.com/foo{?query,number}");
    SUCCEED_IF(e == 0);
    e = templ.SetParameter("query", "mycelium");
    SUCCEED_IF(e == 0);
    e = templ.SetParameter("number", "100");
    SUCCEED_IF(e == 0);
    char str[256];
    auto rval = templ.Resolve(str);
    SUCCEED_IF(rval.first == 0);
    nlib_printf("Uri: %s\n", str);
/*
    Output:
    Uri: http://www.example.com/foo?query=mycelium&number=100
*/
    //! [UriTemplate]
    return true;
}

bool StringView() {
    //! [StringView]
    nlib_ns::StringView view("this is my input string");
    auto substr = view.substr(5, 2);
    char buf[16];
    (void)nlib_ns::ToCstring(buf, substr);
    nlib_printf("'this is my input string'.substr(5, 2): '%s'\n", buf);
    nlib_printf("StartsWith(view, \"this\"): %d\n", nlib_ns::StartsWith(view, "this"));
    nlib_printf("EndsWith(view, \"string\"): %d\n", nlib_ns::EndsWith(view, "string"));

    view = nlib_ns::StringView("  12.345  ");
    nlib_ns::Trim(view);  // remove space characters
    double val;
    (void)nlib_ns::ToDouble(&val, view);
    nlib_printf("ToDouble(\"12.345\"): %f\n", val);
/*
    Output:
    'this is my input string'.substr(5, 2): 'is'
    StartsWith(view, "this"): 1
    EndsWith(view, "string"): 1
    ToDouble("12.345"): 12.345000
*/
    //! [StringView]
    return true;
}

bool DynamicAlignedStorage() {
    //! [DynamicAlignedStorage]
    nlib_ns::DynamicAlignedStorage storage;
    errno_t e = storage.Init(64 * 1024, 64);  // 64 kbytes, 64 bytes aligned
    SUCCEED_IF(e == 0);
    void* ptr = storage.Get();  // ptr is what you want
    //! [DynamicAlignedStorage]
    (void)ptr;
    return true;
}

bool Base64EncoderEncode() {
    //! [Base64Encoder::Encode]
    const char* data = "Data to encode";
    auto result = nlib_ns::Base64Encoder::Encode(data, nlib_strlen(data));
    SUCCEED_IF(result.first == 0);
    nlib_printf("base64: %s\n", result.second.get());  // result.second is std::unique_ptr<char[]>
/*
    Output:
    base64: RGF0YSB0byBlbmNvZGU
*/
    //! [Base64Encoder::Encode]
    return true;
}

bool Base64EncoderStepEncode() {
    //! [Base64Encoder::StepEncode]
    nlib_ns::Base64Encoder encoder;
    char dst[1024];
    char* p = &dst[0];
    size_t dstsize = 1024;

    const char* step1 = "Data to encode step1\n";
    auto result = encoder.StepEncode(p, dstsize, step1, nlib_strlen(step1));
    SUCCEED_IF(result.first == 0);
    p += result.second;
    dstsize -= result.second;

    const char* step2 = "Data to encode step2\n";
    result = encoder.StepEncode(p, dstsize, step2, nlib_strlen(step2));
    SUCCEED_IF(result.first == 0);
    p += result.second;
    dstsize -= result.second;

    result = encoder.Close(p, dstsize, false);
    SUCCEED_IF(result.first == 0);
    p += result.second;
    dstsize -= result.second;
    SUCCEED_IF(dstsize > 0);
    *p = '\0';
    nlib_printf("base64: %s\n", dst);
/*
    Output:
    base64: RGF0YSB0byBlbmNvZGUgc3RlcDEKRGF0YSB0byBlbmNvZGUgc3RlcDIK
*/
    //! [Base64Encoder::StepEncode]
    return true;
}

bool Base64DecoderDecode() {
    //! [Base64Decoder::Decode]
    const char* base64 = "RGF0YSB0byBlbmNvZGU";
    auto result = nlib_ns::Base64Decoder::Decode(base64);
    SUCCEED_IF(std::get<0>(result) == 0);
    auto data = std::move(std::get<1>(result));
    size_t size = std::get<2>(result);
    const char* p = reinterpret_cast<const char*>(data.get());
    std::string str(p, p + size);
    nlib_printf("decoded: %s\n", str.c_str());
/*
    Output:
    decoded: Data to encode
*/
    //! [Base64Decoder::Decode]
    return true;
}

bool Base64DecoderStepDecode() {
    //! [Base64Decoder::StepDecode]
    char dst[256];
    char* p = &dst[0];
    size_t dstsize = 256;
    const char* base64[] = {
        "RGF0YS", "B0byBlbmNvZ", "GU", "gc3RlcDEKRGF0YSB", "0byBlbm", "NvZGUgc3RlcDIK"
    };
    nlib_ns::Base64Decoder decoder;
    for (auto& str : base64) {
        auto result = decoder.StepDecode(p, dstsize, str, nlib_strlen(str));
        SUCCEED_IF(result.first == 0);
        p += result.second;
        dstsize -= result.second;
    }
    auto result = decoder.Close(p, dstsize);
    SUCCEED_IF(result.first == 0);
    p += result.second;
    dstsize -= result.second;
    SUCCEED_IF(dstsize > 0);
    *p = '\0';
    nlib_printf("decoded:\n%s", dst);
/*
    Output:
    decoded:
    Data to encode step1
    Data to encode step2
*/
    //! [Base64Decoder::StepDecode]
    return true;
}

bool Base64DecoderDecodeInplace() {
    //! [Base64Decoder::DecodeInplace]
    char base64[] = "RGF0YSB0byBlbmNvZGUgc3RlcDEKRGF0YSB0byBlbmNvZGUgc3RlcDIK";
    size_t written;
    errno_t e = nlib_ns::Base64Decoder::DecodeInplace(&written, base64);
    SUCCEED_IF(e == 0);
    base64[written] = '\0';
    nlib_printf("decoded:\n%s", base64);
/*
    Output:
    decoded:
    Data to encode step1
    Data to encode step2
*/
    //! [Base64Decoder::DecodeInplace]
    return true;
}

bool NflagsGetStringCommaList() {
    //! [Nflags::GetStringCommaList]
    char comma_list[] = "text1,text2,text3";
    char* x[16];
    size_t len;
    errno_t e = nlib_ns::Nflags::GetStringCommaList(comma_list, x, &len);
    SUCCEED_IF(e == 0);
    for (size_t i = 0; i < len; ++i) {
        nlib_printf("%" PRIuS ": %s\n", i, x[i]);
    }
/*
    Output:
    0: text1
    1: text2
    2: text3
*/
    //! [Nflags::GetStringCommaList]
    return true;
}

bool DateTimeInit() {
    char w3cdtf[32];
    //! [DateTime::Init]
    nlib_ns::DateTime dt;
    errno_t e = dt.Init(1973, 11, 10);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("%s\n", w3cdtf);
    e = dt.Init(2018, 2, 29);
    SUCCEED_IF(e == EINVAL);
/*
    Output:
    1973-11-10T00:00:00.00Z
*/
    //! [DateTime::Init]
    return true;
}

bool DateTimeGetNow() {
    char w3cdtf[32];
    //! [DateTime::GetNow]
    nlib_ns::DateTime dt;
    nlib_ns::DateTime::GetNow(&dt);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("%s\n", w3cdtf);
/*
    Output:
    2017-12-24T21:00:00.00Z
*/
    //! [DateTime::GetNow]
    return true;
}

bool DateTimeParse() {
    //! [DateTime::Parse]
    nlib_ns::DateTime dt;
    nlib_ns::TimeSpan delta;
    errno_t e = nlib_ns::DateTime::Parse("1994-11-05T08:15:30-05:00", &dt, &delta);
    SUCCEED_IF(e == 0);
    nlib_ns::DateTimeParams params;
    dt.Get(&params);
    nlib_printf("Year: %d\nMonth: %d\nDay: %d\n", params.year, params.month, params.day);
    nlib_printf("Hour: %d\nMinute: %d\nSecond: %d\n", params.hour, params.min, params.sec);
/*
    Output:
    Year: 1994
    Month: 11
    Day: 5
    Hour: 8
    Minute: 15
    Second: 30
*/
    //! [DateTime::Parse]
    return true;
}

// GetDaysOfWeek
// GetDaysOfYear

bool DateTimeAddYears() {
    char w3cdtf[32];
    //! [DateTime::AddYears]
    nlib_ns::DateTime dt;
    nlib_ns::DateTime dt2;
    errno_t e = dt.Init(2016, 2, 29);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("Base: %s\n", w3cdtf);
    for (int i = 1; i <= 5; ++i) {
        dt2 = dt;
        dt2.AddYears(-i);
        dt2.ToW3cDtf(w3cdtf);
        nlib_printf("%d year(s) ago: %s\n", i, w3cdtf);
    }
    nlib_printf("\n");
    for (int i = 1; i <= 5; ++i) {
        dt2 = dt;
        dt2.AddYears(i);
        dt2.ToW3cDtf(w3cdtf);
        nlib_printf("%d year(s) from the base: %s\n", i, w3cdtf);
    }
/*
    Output:
    Base: 2016-02-29T00:00:00.00Z
    1 year(s) ago: 2015-02-28T00:00:00.00Z
    2 year(s) ago: 2014-02-28T00:00:00.00Z
    3 year(s) ago: 2013-02-28T00:00:00.00Z
    4 year(s) ago: 2012-02-29T00:00:00.00Z
    5 year(s) ago: 2011-02-28T00:00:00.00Z

    1 year(s) from the base: 2017-02-28T00:00:00.00Z
    2 year(s) from the base: 2018-02-28T00:00:00.00Z
    3 year(s) from the base: 2019-02-28T00:00:00.00Z
    4 year(s) from the base: 2020-02-29T00:00:00.00Z
    5 year(s) from the base: 2021-02-28T00:00:00.00Z
*/
    //! [DateTime::AddYears]
    return true;
}

bool DateTimeAddMonths() {
    char w3cdtf[32];
    //! [DateTime::AddMonths]
    nlib_ns::DateTime dt;
    nlib_ns::DateTime dt2;
    errno_t e = dt.Init(2015, 12, 31);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("Base: %s\n", w3cdtf);
    for (int i = 1; i <= 15; ++i) {
        dt2 = dt;
        dt2.AddMonths(i);
        dt2.ToW3cDtf(w3cdtf);
        nlib_printf("%d months(s) from the base: %s\n", i, w3cdtf);
    }
/*
    Output:
    Base: 2015-12-31T00:00:00.00Z
    1 months(s) from the base: 2016-01-31T00:00:00.00Z
    2 months(s) from the base: 2016-02-29T00:00:00.00Z
    3 months(s) from the base: 2016-03-31T00:00:00.00Z
    4 months(s) from the base: 2016-04-30T00:00:00.00Z
    5 months(s) from the base: 2016-05-31T00:00:00.00Z
    6 months(s) from the base: 2016-06-30T00:00:00.00Z
    7 months(s) from the base: 2016-07-31T00:00:00.00Z
    8 months(s) from the base: 2016-08-31T00:00:00.00Z
    9 months(s) from the base: 2016-09-30T00:00:00.00Z
    10 months(s) from the base: 2016-10-31T00:00:00.00Z
    11 months(s) from the base: 2016-11-30T00:00:00.00Z
    12 months(s) from the base: 2016-12-31T00:00:00.00Z
    13 months(s) from the base: 2017-01-31T00:00:00.00Z
    14 months(s) from the base: 2017-02-28T00:00:00.00Z
    15 months(s) from the base: 2017-03-31T00:00:00.00Z
*/
    //! [DateTime::AddMonths]
    return true;
}

bool DateTimeAddDays() {
    char w3cdtf[32];
    //! [DateTime::AddDays]
    nlib_ns::DateTime dt;
    errno_t e = nlib_ns::DateTime::GetNow(&dt);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("Now: %s\n", w3cdtf);
    dt.AddDays(1.5);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("1.5 days from now: %s\n", w3cdtf);
/*
    Output:
    Now: 2017-12-26T04:40:29.00Z
    1.5 days from now: 2017-12-27T16:40:29.00Z
*/
    //! [DateTime::AddDays]
    return true;
}

bool DateTimeAddHours() {
    char w3cdtf[32];
    //! [DateTime::AddHours]
    nlib_ns::DateTime dt;
    errno_t e = nlib_ns::DateTime::GetNow(&dt);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("Now: %s\n", w3cdtf);
    dt.AddHours(24 * 365);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("24x365 hours from now: %s\n", w3cdtf);
/*
    Output:
    Now: 2017 - 12 - 26T04 : 57 : 15.00Z
    24x365 hours from now : 2018 - 12 - 26T04 : 57 : 15.00Z
*/
    //! [DateTime::AddHours]
    return true;
}

bool DateTimeToRfc2822() {
    char rfc2822[32];
    //! [DateTime::ToRfc2822]
    nlib_ns::DateTime dt;
    errno_t e = dt.Init(1973, 11, 10);
    SUCCEED_IF(e == 0);
    dt.ToRfc2822(rfc2822);
    nlib_printf("%s\n", rfc2822);
/*
    Output:
    Sat, 10 Nov 1973 00:00:00 +0000
*/
    //! [DateTime::ToRfc2822]
    return true;
}

bool DateTimeToAscTime() {
    char asctime[26];
    //! [DateTime::ToAscTime]
    nlib_ns::DateTime dt;
    errno_t e = dt.Init(1973, 11, 10);
    SUCCEED_IF(e == 0);
    dt.ToAscTime(asctime);
    nlib_printf("%s\n", asctime);
/*
    Output:
    Sat Nov 10 00:00:00 1973
*/
    //! [DateTime::ToAscTime]
    return true;
}

bool TimeSpan() {
    char w3cdtf[32];
    //! [TimeSpan]
    nlib_ns::DateTime dt;
    errno_t e = nlib_ns::DateTime::GetNow(&dt);
    SUCCEED_IF(e == 0);
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("Now: %s\n", w3cdtf);
    nlib_ns::TimeSpan span(0, 60 * 60 * 10);
    dt += span;
    dt.ToW3cDtf(w3cdtf);
    nlib_printf("10 hours from now: %s\n", w3cdtf);
/*
    Output:
    Now: 2017-12-26T06:17:48.00Z
    10 hours from now: 2017-12-26T16:17:48.00Z
*/
    //! [TimeSpan]
    return true;
}

bool TimeSpanToTimeValue() {
    //! [TimeSpan::ToTimeValue]
    nlib_sleep(nlib_ns::TimeSpan(0, 0, 100).ToTimeValue().tick);
    //! [TimeSpan::ToTimeValue]
    return true;
}

bool BinaryReader() {
    //! [BinaryReader]
    unsigned char buf[] = {
        0x00,
        0x00, 0x01,
        0x00, 0x01, 0x02, 0x03,
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07
    };
    errno_t e;
    nlib_ns::MemoryInputStream s;
    e = s.Init(buf);
    SUCCEED_IF(e == 0);
    nlib_ns::BinaryReader r;
    e = r.Init(nlib_ns::BinaryReader::kEndianLittle);
    SUCCEED_IF(e == 0);
    e = r.Open(&s);
    SUCCEED_IF(e == 0);

    int8_t v8;
    int16_t v16;
    int32_t v32;
    int64_t v64;
    r.Read(&v8);
    r.Read(&v16);
    r.Read(&v32);
    r.Read(&v64);
    SUCCEED_IF(!!r);
    (void)r.Close();
    nlib_printf("v8: 0x%" PRIx8 "\n", v8);
    nlib_printf("v16: 0x%" PRIx16 "\n", v16);
    nlib_printf("v32: 0x%" PRIx32 "\n", v32);
    nlib_printf("v64: 0x%" PRIx64 "\n", v64);
/*
    Output:
    v8: 0x0
    v16: 0x100
    v32: 0x3020100
    v64: 0x706050403020100
*/
    //! [BinaryReader]
    return true;
}

bool BinaryWriter() {
    //! [BinaryWriter]
    unsigned char buf[64];
    errno_t e;
    nlib_ns::MemoryOutputStream s;
    e = s.Init(buf);
    SUCCEED_IF(e == 0);
    nlib_ns::BinaryWriter w;
    e = w.Init(nlib_ns::BinaryWriter::kEndianLittle);
    SUCCEED_IF(e == 0);
    e = w.Open(&s);
    SUCCEED_IF(e == 0);
    w.Write((uint8_t)0x00);
    w.Write((uint16_t)0x0100);
    w.Write((uint32_t)0x03020100);
    w.Write((uint64_t)0x0706050403020100ULL);
    SUCCEED_IF(!!w);
    bool result = w.Close();
    SUCCEED_IF(!!result);
    uint64_t pos = s.Pos64();
    result = s.Close();
    SUCCEED_IF(!!result);
    for (uint64_t i = 0; i < pos; ++i) {
        nlib_printf("%02x ", buf[i]);
    }
    nlib_printf("\n");
/*
    Output:
    00 00 01 00 01 02 03 00 01 02 03 04 05 06 07
*/
    //! [BinaryWriter]
    return true;
}

bool UniquePtr() {
    struct MyObject {};
    //! [UniquePtr]
    nlib_ns::UniquePtr<MyObject> myObjPtr(new MyObject());
    nlib_ns::UniquePtr<char[]> myArrayPtr(new char[1024]);
    nlib_ns::UniquePtr<void, void(*)(void*)> myMallocFreePtr(nlib_malloc(1024), nlib_free);
    //! [UniquePtr]
    return true;
}

bool SmartBitmap() {
    //! [SmartBitmap]
    nlib_ns::SmartBitmap<10> bmp;
    bool result;
    result = bmp.Set(0);
    SUCCEED_IF(!!result);
    result = bmp.Set(4);
    SUCCEED_IF(!!result);
    result = bmp.Set(7);
    SUCCEED_IF(!!result);

    unsigned int rank;
    rank = bmp.Rank1(4);
    nlib_printf("bmp.Rank1(4): %u\n", rank);
    rank = bmp.Rank1(10000);
    nlib_printf("bmp.Rank1(10000): %u\n", rank);

    int select;
    select = bmp.Select1(2);
    nlib_printf("bmp.Select1(2): %d\n", select);
    select = bmp.Select1(3);
    nlib_printf("bmp.Select1(3): %d\n", select);
/*
    Output:
    bmp.Rank1(4): 2
    bmp.Rank1(10000): 3
    bmp.Select1(2): 7
    bmp.Select1(3): -1
*/
    //! [SmartBitmap]
    return true;
}

bool MemoryInputStream() {
    const int kBufSize = 1024;
    unsigned char buf[kBufSize];
    //! [MemoryInputStream]
    nlib_ns::MemoryInputStream is1(&buf[0], kBufSize);
    nlib_ns::MemoryInputStream is2(buf);
    nlib_ns::MemoryInputStream is3;
    is3.Init(&buf[0], kBufSize);
    nlib_ns::MemoryInputStream is4;
    is4.Init(buf);
    //! [MemoryInputStream]
    return true;
}

bool MemoryOutputStream() {
    const int kBufSize = 1024;
    unsigned char buf[kBufSize];
    //! [MemoryOutputStream]
    nlib_ns::MemoryOutputStream os1(&buf[0], kBufSize);
    nlib_ns::MemoryOutputStream os2(buf);
    nlib_ns::MemoryOutputStream os3;
    os3.Init(&buf[0], kBufSize);
    nlib_ns::MemoryOutputStream os4;
    os4.Init(buf);
    //! [MemoryOutputStream]
    return true;
}

bool ConsoleOutputStream() {
    //! [ConsoleOutputStream]
    nlib_ns::ConsoleOutputStream ostr;
    nlib_ns::TextWriter writer;
    errno_t e = writer.Init();
    SUCCEED_IF(e == 0);
    e = writer.Open(&ostr);
    SUCCEED_IF(e == 0);
    (void)writer.Write("text to output\n");
    (void)writer.Close();
    (void)ostr.Close();
/*
    Output:
    text to output
*/
    //! [ConsoleOutputStream]
    return true;
}

namespace input_transform {
//! [InputTransform]
class NoInputTransform : public nlib_ns::InputTransform {
public:
    NoInputTransform() {}
    virtual errno_t Transform(nlib_ns::InputStream* is, void* p, size_t size, size_t* nbytes) override {
        *nbytes = is->Read(p, size);
        if (*nbytes == 0) return is->GetErrorValue();
        return 0;
    }
    virtual void* GetWorkBuffer(size_t* n) override {
        *n = 128;
        return buf_;
    }
    virtual errno_t OnSetStream(nlib_ns::InputStream*) override {
        return 0;
    }
    virtual errno_t OnClose() override {
        return 0;
    }
private:
    unsigned char buf_[128];
    NLIB_DISALLOW_COPY_AND_ASSIGN(NoInputTransform);
};

typedef nlib_ns::InputConverterStreamTempl<NoInputTransform> NoConvertInputStream;
//! [InputTransform]
}  // namespace input_transform

namespace output_transform {
//! [OutputTransform]
class NoOutputTransform : public nlib_ns::OutputTransform {
public:
    NoOutputTransform() {}
    virtual errno_t Transform(nlib_ns::OutputStream* os, const void* p, size_t nbytes, bool do_flush) override {
        if (!os->Write(p, nbytes)) { return os->GetErrorValue(); }
        if (do_flush) {
            if (!os->Flush()) { return os->GetErrorValue(); }
        }
        return 0;
    }
    virtual void* GetWorkBuffer(size_t* n) override {
        *n = 128;
        return buf_;
    }

    virtual errno_t OnSetStream(nlib_ns::OutputStream*) override {
        return 0;
    }
    virtual errno_t OnClose(nlib_ns::OutputStream*) override {
        return 0;
    }
private:
    unsigned char buf_[128];
    NLIB_DISALLOW_COPY_AND_ASSIGN(NoOutputTransform);
};

typedef nlib_ns::OutputConverterStreamTempl<NoOutputTransform> NoConvertOutputStream;
//! [OutputTransform]
}  // namespace output_transform

bool Utf32InputStream() {
    //! [Utf32InputStream]
    const nlib_utf32_t* str = U"\x1F929\x1F9DD";  // Star-Struck, Elf
    nlib_ns::MemoryInputStream is(str, sizeof(*str) * 2);
    nlib_ns::Utf32InputStream stream;
    errno_t e = stream.Init();
    SUCCEED_IF(e == 0);
    e = stream.SetStream(&is);
    SUCCEED_IF(e == 0);
    int c;
    while ((c = stream.Read()) > 0) {
        nlib_printf("%02x ", c);
    }
    nlib_printf("\n");
/*
    Output:
    f0 9f a4 a9 f0 9f a7 9d
*/
    //! [Utf32InputStream]
    return true;
}

bool Utf16InputStream() {
    //! [Utf16InputStream]
    const nlib_utf16_t* str = u"\xD83E\xDD29\xD83E\xDDDD";  // Star-Struck, Elf
    nlib_ns::MemoryInputStream is(str, sizeof(*str) * 4);
    nlib_ns::Utf16InputStream stream;
    errno_t e = stream.Init();
    SUCCEED_IF(e == 0);
    e = stream.SetStream(&is);
    SUCCEED_IF(e == 0);
    int c;
    while ((c = stream.Read()) > 0) {
        nlib_printf("%02x ", c);
    }
    nlib_printf("\n");
/*
    Output:
    f0 9f a4 a9 f0 9f a7 9d
*/
    //! [Utf16InputStream]
    return true;
}

bool Base64InputStream() {
    //! [Base64InputStream]
    const char* base64 = "RGF0YSB0byBlbmNvZGUgc3RlcDEKRGF0YSB0byBlbmNvZGUgc3RlcDIK";
    nlib_ns::MemoryInputStream is(base64, nlib_strlen(base64));
    nlib_ns::Base64InputStream base64_stream;
    errno_t e = base64_stream.Init();
    SUCCEED_IF(e == 0);
    e = base64_stream.SetStream(&is);
    SUCCEED_IF(e == 0);

    char buf[128];
    size_t nread = base64_stream.Read(buf, 128);
    SUCCEED_IF(nread > 0);
    buf[nread] = '\0';  // nread not checked for simplicity

    (void)base64_stream.Close();
    (void)is.Close();

    nlib_printf("converted:\n%s", buf);
/*
    Output:
    Data to encode step1
    Data to encode step2
*/
    //! [Base64InputStream]
    return true;
}

bool Base64OutputStream() {
    //! [Base64OutputStream]
    char base64[128];
    nlib_ns::MemoryOutputStream os(base64);
    nlib_ns::Base64OutputStream base64_stream;
    errno_t e = base64_stream.Init();
    SUCCEED_IF(e == 0);
    e = base64_stream.SetStream(&os);
    SUCCEED_IF(e == 0);
    const char* str = "Data to encode step1\nData to encode step2\n";
    bool result = base64_stream.Write(str, nlib_strlen(str));
    SUCCEED_IF(!!result);

    result = base64_stream.Close();
    SUCCEED_IF(!!result);
    base64[os.Pos64()] = '\0';
    result = os.Close();
    SUCCEED_IF(!!result);
    nlib_printf("converted:\n%s\n", base64);
/*
    Output:
    RGF0YSB0byBlbmNvZGUgc3RlcDEKRGF0YSB0byBlbmNvZGUgc3RlcDIK
*/
    //! [Base64OutputStream]
    return true;
}

bool CurlInputStream() {
#ifdef NLIB_HAS_LIBCURL
    //! [CurlInputStream]
    nlib_ns::CurlInputStream stream;
    errno_t e = stream.Init();
    SUCCEED_IF(e == 0);
    SUCCEED_IF(stream.SetUrl("https://www.nintendo.co.jp") == CURLE_OK);
    SUCCEED_IF(stream.StartDownload());
    char buf[256];
    size_t nread = stream.Read(buf, 255);
    if (nread == 0) {
        nlib_printf("CURLCode: %d\n", stream.GetCurlError());
        nlib_printf("CURLMcode :%d\n", stream.GetCurlMultiError());
        return false;
    }
    buf[nread] = '\0';
    (void)stream.Close();  // reusable after Close()
    nlib_printf("%s\n", buf);
/*
    Output:
    ......
*/
    //! [CurlInputStream]
#endif
    return true;
}

bool TextReader() {
    //! [TextReader]
    const char* str = "\xf0\x9f\xa4\xa9\xf0\x9f\xa7\x9d multibyte";
    nlib_ns::MemoryInputStream is(str, nlib_strlen(str));
    nlib_ns::TextReader reader;
    SUCCEED_IF(reader.Init() == 0);
    SUCCEED_IF(reader.Open(&is) == 0);
    int c;
    while ((c = reader.Read()) >= 0) {
        nlib_printf("U+%05x ", c);
    }
    nlib_printf("\n");
    (void)reader.Close();
    (void)is.Close();
/*
    Output:
    U+1f929 U+1f9dd U+00020 U+0006d U+00075 U+0006c U+00074 U+00069 U+00062 U+00079 U+00074 U+00065
*/
    //! [TextReader]
    return true;
}

bool TaggedTextParser() {
    //! [TaggedTextParser]
    // text is not const because TaggedTextParser rewrite it.
    char text[] = "<doc > my text </toc >";
    nlib_ns::TaggedTextParser parser(text);
    SUCCEED_IF(parser.Next() == nlib_ns::TaggedTextParser::kStartElement);
    nlib_printf("kStartElement='%s'\n", parser.GetText());
    SUCCEED_IF(parser.Next() == nlib_ns::TaggedTextParser::kCharacters);
    nlib_printf("kCharacters='%s'\n", parser.GetText());
    SUCCEED_IF(parser.Next() == nlib_ns::TaggedTextParser::kEndElement);
    nlib_printf("kEndElement='%s'\n", parser.GetText());
    SUCCEED_IF(parser.Next() == nlib_ns::TaggedTextParser::kEndDocument);
/*
    Output:
    kStartElement='doc'
    kCharacters='my text'
    kEndElement='toc'
*/
    //! [TaggedTextParser]
    return true;
}

}  // namespace

#define EXEC_SAMPLE(text, func) \
    nlib_printf("==== %s ====\n", text); \
    if (!func()) { nlib_printf("%s\n", text); return false; }

bool cppapi() {
    EXEC_SAMPLE("threading::Barrier", Barrier);
    EXEC_SAMPLE("LockFreePipe", LockFreePipe);
    EXEC_SAMPLE("LockFreeInit", LockFreeInit);
    EXEC_SAMPLE("LockFreeStack", LockFreeStack);
    EXEC_SAMPLE("LockFreeQueue", LockFreeQueue);
    EXEC_SAMPLE("LockFreePriorityQueue", LockFreePriorityQueue);
    EXEC_SAMPLE("LockFreeBroadcastQueue", LockFreeBroadcastQueue);
    EXEC_SAMPLE("LockFreeUnitHeap", LockFreeUnitHeap);
    EXEC_SAMPLE("Uri", Uri);
    EXEC_SAMPLE("Uri::IsHostName", UriHostName);
    EXEC_SAMPLE("Uri::IsEmailAddress", UriEmailAddress);
    EXEC_SAMPLE("Uri::DecodeUriComponent", DecodeUriComponent);
    EXEC_SAMPLE("UriQueryDecoder", UriQueryDecoder);
    EXEC_SAMPLE("UriQueryEncoder", UriQueryEncoder);
    EXEC_SAMPLE("UriTemplate", UriTemplate);
    EXEC_SAMPLE("StringView", StringView);
    EXEC_SAMPLE("DynamicAlignedStorage", DynamicAlignedStorage);
    EXEC_SAMPLE("Base64Encoder::Encode", Base64EncoderEncode);
    EXEC_SAMPLE("Base64Encoder::StepEncode", Base64EncoderStepEncode);
    EXEC_SAMPLE("Base64Decoder::Decode", Base64DecoderDecode);
    EXEC_SAMPLE("Base64Decoder::StepDecode", Base64DecoderStepDecode);
    EXEC_SAMPLE("Base64Decoder::DecodeInplace", Base64DecoderDecodeInplace);
    EXEC_SAMPLE("NflagsGetStringCommaList", NflagsGetStringCommaList);
    EXEC_SAMPLE("DateTime::Init", DateTimeInit);
    EXEC_SAMPLE("DateTime::GetNow", DateTimeGetNow);
    EXEC_SAMPLE("DateTime::Parse", DateTimeParse);
    EXEC_SAMPLE("DateTime::AddYears", DateTimeAddYears);
    EXEC_SAMPLE("DateTime::AddMonths", DateTimeAddMonths);
    EXEC_SAMPLE("DateTime::AddDays", DateTimeAddDays);
    EXEC_SAMPLE("DateTime::AddHours", DateTimeAddHours);
    EXEC_SAMPLE("DateTime::ToRfc2822", DateTimeToRfc2822);
    EXEC_SAMPLE("DateTime::ToAscTime", DateTimeToAscTime);
    EXEC_SAMPLE("TimeSpan", TimeSpan);
    EXEC_SAMPLE("TimeSpan::ToTimeValue", TimeSpanToTimeValue)
    EXEC_SAMPLE("BinaryReader", BinaryReader);
    EXEC_SAMPLE("BinaryWriter", BinaryWriter);
    EXEC_SAMPLE("UniquePtr", UniquePtr);
    EXEC_SAMPLE("SmartBitmap", SmartBitmap);
    EXEC_SAMPLE("MemoryInputStream", MemoryInputStream);
    EXEC_SAMPLE("MemoryInputStream", MemoryOutputStream);
    EXEC_SAMPLE("ConsoleOutputStream", ConsoleOutputStream);
    EXEC_SAMPLE("Base64InputStream", Base64InputStream);
    EXEC_SAMPLE("Base64OutputStream", Base64OutputStream);
    EXEC_SAMPLE("Utf32InputStream", Utf32InputStream);
    EXEC_SAMPLE("Utf16InputStream", Utf16InputStream);
    EXEC_SAMPLE("CurlInputStream", CurlInputStream);
    EXEC_SAMPLE("TextReader", TextReader);
    EXEC_SAMPLE("TaggedTextParser", TaggedTextParser);
    return true;
}
