﻿
/*--------------------------------------------------------------------------------*
Project: CrossRoad
Copyright (C)Nintendo All rights reserved.

These coded instructions, statements, and computer programs contain proprietary
information of Nintendo and/or its licensed developers and are protected by
national and international copyright laws. They may not be disclosed to third
parties or copied or duplicated in any form, in whole or in part, without the
prior written consent of Nintendo.

The content herein is highly confidential and should be handled accordingly.
*--------------------------------------------------------------------------------*/

#include <thread>
#include <algorithm>
#include <random>
#include <vector>

#include "nn/nlib/msgpack/MpObject.h"
#include "nn/nlib/msgpack/JsonStreamParser.h"
#include "nn/nlib/msgpack/JsonStreamGenerator.h"
#include "nn/nlib/msgpack/MpWalker.h"
#include "nn/nlib/msgpack/MpDirectWriter.h"
#include "nn/nlib/msgpack/CsvReader.h"
#include "nn/nlib/msgpack/JsonPatch.h"
#include "nn/nlib/msgpack/JsonSchema.h"
#include "nn/nlib/MemoryInputStream.h"
#include "nn/nlib/MemoryOutputStream.h"
#include "nn/nlib/msgpack/jsonrpc/JsonRpcRequest.h"

using nlib_ns::msgpack::MpObject;
using nlib_ns::msgpack::MpObjectKv;
using nlib_ns::msgpack::JsonStreamParser;
using nlib_ns::msgpack::JsonStreamParserSettings;
using nlib_ns::msgpack::JsonStreamGenerator;
using nlib_ns::msgpack::JsonStreamGeneratorSettings;
using nlib_ns::msgpack::MpWalker;
using nlib_ns::msgpack::MpDirectWriter;
using nlib_ns::msgpack::CsvReader;
using nlib_ns::msgpack::JsonPatch;
using nlib_ns::msgpack::JsonPatchBuilder;
using nlib_ns::msgpack::JsonSchemaConverter;
using nlib_ns::msgpack::JsonSchemaValidator;
using nlib_ns::msgpack::jsonrpc::JsonRpcRequestWriter;
using nlib_ns::msgpack::jsonrpc::JsonRpcRequest;

using nlib_ns::MemoryInputStream;
using nlib_ns::MemoryOutputStream;

#define SUCCEED_IF(cond) if (!(cond)) return false


//! [msgpack.CustomData1]
struct Vec3 { float x, y, z; };
NLIB_NAMESPACE_BEGIN
namespace msgpack {
template<>
errno_t Box(msgpack::MpObject* obj, const Vec3& v) {
    errno_t e;
    e = obj->InitArray(3);
    if (e != 0) return e;
    auto ar = obj->AsArray();
    ar[0].Box(v.x);
    ar[1].Box(v.y);
    ar[2].Box(v.z);
    return 0;
}
template<>
errno_t Unbox(const MpObject* obj, Vec3* v) {
    if (!obj->IsArray()) return EACCES;
    if (obj->GetSize() != 3) return EACCES;
    auto ar = obj->AsArray();
    if (ar[0].Unbox(&v->x) != 0) return EACCES;
    if (ar[1].Unbox(&v->y) != 0) return EACCES;
    if (ar[2].Unbox(&v->z) != 0) return EACCES;
    return 0;
}
}  // namespace msgpack
NLIB_NAMESPACE_END
//! [msgpack.CustomData1]

namespace {

//! [msgpack::HelperFunction]
template<size_t N>
void ToJson(nlib_utf8_t(&buf)[N], const MpObject& obj) {
    auto gen = JsonStreamGenerator::Generate(buf, obj);
    NLIB_ASSERT(gen.first == 0);
    (void)gen;
}
template<size_t N>
void ToJson(nlib_utf8_t(&buf)[N], const nlib_byte_t* first, const nlib_byte_t* last) {
    auto result = JsonStreamParser::Parse(first, last);
    NLIB_ASSERT(result.first == 0);
    auto gen = JsonStreamGenerator::Generate(buf, *result.second.get());
    NLIB_ASSERT(gen.first == 0);
    (void)gen;
}
std::unique_ptr<MpObject> ToMpObject(const nlib_utf8_t* json) {
    auto result = JsonStreamParser::Parse(json);
    NLIB_ASSERT(result.first == 0);
    return std::move(result.second);
}
template<size_t N>
size_t ToMsgpack(nlib_byte_t(&buf)[N], const nlib_utf8_t* json) {
    auto obj = ToMpObject(json);
    JsonStreamGenerator gen;
    JsonStreamGeneratorSettings settings;
    settings.msgpack = true;
    errno_t e = gen.Init(settings);
    NLIB_ASSERT(e == 0);
    size_t written;
    e = JsonStreamGenerator::Generate(&written, (nlib_utf8_t*)&buf[0], N, *obj, settings);
    NLIB_ASSERT(e == 0);
    return written;
}
//! [msgpack::HelperFunction]


namespace mpobject {
//! [msgpack::MpObject::BoxCustomData]
bool BoxCustomData() {
    Vec3 vec = { 1.f, 2.f, 3.f };
    MpObject obj;
    SUCCEED_IF(obj.Box(vec) == 0);
    float x, y, z;
    auto ar = obj.AsArray();
    ar[0].Unbox(&x);
    ar[1].Unbox(&y);
    ar[2].Unbox(&z);
    nlib_printf("obj[0]: %f, obj[1]: %f, obj[2]: %f\n", x, y, z);
    return true;
/*
    Output:
    obj[0]: 1.000000, obj[1]: 2.000000, obj[2]: 3.000000
*/
}
bool UnboxCustomData() {
    Vec3 vec;
    MpObject obj;
    obj.InitArray(3);
    auto ar = obj.AsArray();
    ar[0].Box(1.f);
    ar[1].Box(2.f);
    ar[2].Box(3.f);
    SUCCEED_IF(obj.Unbox(&vec) == 0);
    nlib_printf("Vec.x: %f, Vec.y: %f, Vec.z: %f\n", vec.x, vec.y, vec.z);
    return true;
/*
    Output:
    Vec.x: 1.000000, Vec.y: 2.000000, Vec.z: 3.000000
*/
}
//! [msgpack::MpObject::BoxCustomData]
bool BoxCustomData2() {
    Vec3 vec = { 1.f, 2.f, 3.f };
    MpObject obj;
    SUCCEED_IF(Box(&obj, vec) == 0);
    // Box(&obj, 1.f);  // sorry, error for now
    return true;
}
bool UnboxCustomData2() {
    Vec3 vec;
    MpObject obj;
    obj.InitArray(3);
    auto ar = obj.AsArray();
    ar[0].Box(1.f);
    ar[1].Box(2.f);
    ar[2].Box(3.f);
    SUCCEED_IF(Unbox(&obj, &vec) == 0);
    return true;
}

bool Box() {
//! [msgpack::MpObject::Box]
    int ar[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
    MpObject ar_obj;
    SUCCEED_IF(ar_obj.Box(ar) == 0);
    NLIB_ASSERT(ar_obj.IsArray());
    NLIB_ASSERT(ar_obj.GetSize() == 10);

    std::vector<int> vec(&ar[0], &ar[0] + 10);
    MpObject vec_obj;
    SUCCEED_IF(vec_obj.Box(vec) == 0);
    NLIB_ASSERT(vec_obj.IsArray());
    NLIB_ASSERT(vec_obj.GetSize() == 10);

    MpObject str_obj;
    const char str[] = "string data";
    SUCCEED_IF(str_obj.Box(str) == 0);
    SUCCEED_IF(str_obj.IsString());
    SUCCEED_IF(str_obj.GetSize() == 11);
//! [msgpack::MpObject::Box]
    return true;
}
bool Unbox() {
//! [msgpack::MpObject::Unbox]
    int intval;
    MpObject intobj = 1;
    intobj.Unbox(&intval);
    NLIB_ASSERT(intval == 1);

    std::vector<int> vec;
    auto x = JsonStreamParser::Parse("[1, 2, 3, 4, 5]");
    SUCCEED_IF(x.first == 0);
    x.second->Unbox(&vec);
    NLIB_ASSERT(vec.size() == 5);
//! [msgpack::MpObject::Unbox]
    return true;
}
bool AssignmentOperator() {
    MpObject obj;
//! [msgpack::MpObject::Assignment]
    obj = 1;         // same as obj.Box(1)
    obj = 1.2f;      // same as obj.Box(1.2f)
    // obj = "abc";  // error
    obj.Box("abc");  // ok
//! [msgpack::MpObject::Assignment]
    return true;
}
bool Resize() {
    MpObject obj;
    errno_t e;
//! [msgpack::MpObject::Resize]
    // array
    SUCCEED_IF(obj.InitArray(10) == 0);
    auto ar = obj.AsArray();
    nlib_printf("array size: %" PRIuS "\n", ar.GetSize());
    SUCCEED_IF(ar.Resize(100) == 0);
    nlib_printf("array size: %" PRIuS "\n", ar.GetSize());
    ar[99] = 1;

    // map
    SUCCEED_IF(obj.InitMap(10) == 0);
    nlib_printf("map size: %u\n", obj.GetSize());
    SUCCEED_IF(obj.Resize(100) == 0);
    nlib_printf("map size: %u\n", obj.GetSize());
    auto& kv = obj.AsMap()[99];
    SUCCEED_IF(kv.first.Box("key") == 0);
    SUCCEED_IF(kv.second.Box("value") == 0);

    // string
    SUCCEED_IF(obj.InitString("mystring") == 0);
    e = obj.Resize(100);
    nlib_printf("cannot resize string: err=%s\n", nlib_error_string(e));

    // binary
    SUCCEED_IF(obj.InitBinary("test", 4) == 0);
    e = obj.Resize(100);
    nlib_printf("cannot resize binary: err=%s\n", nlib_error_string(e));
/*
    Output:
    array size: 10
    array size: 100
    map size: 10
    map size: 100
    cannot resize string: err=EACCES
    cannot resize binary: err=EACCES
*/
//! [msgpack::MpObject::Resize]
    return true;
}

bool IsJsonPointer() {
//! [msgpack::MpObject::IsJsonPointer]
    const char* valid[] = {
        "",        // the whole document
        "/foo",
        "/foo/0",
        "/foo/",   // keys = ["foo", ""]
        "/",       // key=""
        "/a~1b",   // key="a/b"
        "/ ",      // key=" "
        "/m~0n"    // key="m~n"
    };
    for (auto& str : valid) {
        SUCCEED_IF(MpObject::IsJsonPointer(str));
    }
    const char* invalid[] = {
        "foo",    // no relative path
        "/m~2n"   // only ~0 and ~1
        "/m~xn"   // use "/m~0xn"
    };
    for (auto& str : invalid) {
        SUCCEED_IF(!MpObject::IsJsonPointer(str));
    }
//! [msgpack::MpObject::IsJsonPointer]
    return true;
}

bool ResolveJsonPointer() {
    MpObject* target;
    char buf[1024];
//! [msgpack::MpObject::ResolveJsonPointer]
    const char* json = R"({
        "foo": ["bar", "baz"],
        "": 0,
        "a/b": 1,
        "c%d": 2,
        "e^f": 3,
        "g|h": 4,
        "i\\j": 5,
        "k\"l": 6,
        " ": 7,
        "m~n": 8
    })";
    auto obj = ToMpObject(json);
    const char* json_pointers[] = {
        "", "/foo", "/foo/0", "/", "/a~1b", "/c%d", "/e^f", "/g|h", "/i\\j", "/k\"l", "/ ", "/m~0n"
    };
    for (auto& str : json_pointers) {
        SUCCEED_IF(MpObject::ResolveJsonPointer(&target, obj.get(), str) == 0);
        ToJson(buf, *target);
        nlib_printf("Resolve('%s') --> %s\n", str, buf);
    }
/*
    Output:
    Resolve('') --> {"foo":["bar","baz"],"":0,"a\/b":1,"c%d":2,"e^f":3,"g|h":4,"i\\j":5,"k\"l":6," ":7,"m~n":8}
    Resolve('/foo') --> ["bar","baz"]
    Resolve('/foo/0') --> "bar"
    Resolve('/') --> 0
    Resolve('/a~1b') --> 1
    Resolve('/c%d') --> 2
    Resolve('/e^f') --> 3
    Resolve('/g|h') --> 4
    Resolve('/i\j') --> 5
    Resolve('/k"l') --> 6
    Resolve('/ ') --> 7
    Resolve('/m~0n') --> 8
*/
//! [msgpack::MpObject::ResolveJsonPointer]
    return true;
}

bool DigByJsonPointer() {
    errno_t e;
    MpObject* node;
    char to[1024];
//! [msgpack::MpObject::DigByJsonPointer]
    const char* json1 = R"({ "a": { "foo": 1 } })";
    auto obj = ToMpObject(json1);
    e = MpObject::DigByJsonPointer(&node, obj.get(), "/a/b");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    nlib_printf("Dig(%s, '/a/b') --> %s\n", json1, to);

    obj = ToMpObject(json1);
    e = MpObject::DigByJsonPointer(&node, obj.get(), "/a/foo");
    ToJson(to, *obj);
    nlib_printf("Dig(%s, '/a/foo') --> %s(err = %s)\n", json1, to, nlib_error_string(e));

    const char* json2 = R"([0,1,2])";
    obj = ToMpObject(json2);
    e = MpObject::DigByJsonPointer(&node, obj.get(), "/-");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    nlib_printf("Dig(%s, '/-') --> %s\n", json2, to);

    obj = ToMpObject(json2);
    e = MpObject::DigByJsonPointer(&node, obj.get(), "/1");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    nlib_printf("Dig(%s, '/1') --> %s\n", json2, to);

    const char* json3 = R"({ "q": { "bar": 2 } })";
    obj = ToMpObject(json3);
    e = MpObject::DigByJsonPointer(&node, obj.get(), "/a/b");
    ToJson(to, *obj);
    nlib_printf("Dig(%s, '/a/b') --> %s(err = %s)\n", json3, to, nlib_error_string(e));
/*
    Output:
    Dig({ "a": { "foo": 1 } }, '/a/b') --> {"a":{"foo":1,"b":null}}
    Dig({ "a": { "foo": 1 } }, '/a/foo') --> {"a":{"foo":1}}(err = EEXIST)
    Dig([0,1,2], '/-') --> [0,1,2,null]
    Dig([0,1,2], '/1') --> [0,null,1,2]
    Dig({ "q": { "bar": 2 } }, '/a/b') --> {"q":{"bar":2}}(err = ENOENT)
*/
//! [msgpack::MpObject::DigByJsonPointer]
    return true;
}

bool RemoveByJsonPointer() {
    errno_t e;
    MpObjectKv kv;
    char to[1024];
    char key[1024];
    char value[1024];
//! [msgpack::MpObject::RemoveByJsonPointer]
    const char* json1 = R"({ "a": { "foo": 1 } })";
    auto obj = ToMpObject(json1);
    e = MpObject::RemoveByJsonPointer(&kv, obj.get(), "/a/foo");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    ToJson(key, kv.first);
    ToJson(value, kv.second);
    nlib_printf("Remove(%s, '/a/foo') --> %s, key=%s, value=%s\n", json1, to, key, value);

    obj = ToMpObject(json1);
    e = MpObject::RemoveByJsonPointer(&kv, obj.get(), "/a");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    ToJson(key, kv.first);
    ToJson(value, kv.second);
    nlib_printf("Remove(%s, '/a') --> %s, key=%s, value=%s\n", json1, to, key, value);

    const char* json2 = R"([0, 1, 2])";
    obj = ToMpObject(json2);
    e = MpObject::RemoveByJsonPointer(&kv, obj.get(), "/1");
    SUCCEED_IF(e == 0);
    ToJson(to, *obj);
    ToJson(key, kv.first);
    ToJson(value, kv.second);
    nlib_printf("Remove(%s, '/1') --> %s, key=%s, value=%s\n", json1, to, key, value);
/*
    Output:
    Remove({ "a": { "foo": 1 } }, '/a/foo') --> {"a":{}}, key="foo", value=1
    Remove({ "a": { "foo": 1 } }, '/a') --> {}, key="a", value={"foo":1}
    Remove({ "a": { "foo": 1 } }, '/1') --> [0,2], key=1, value=1
*/
//! [msgpack::MpObject::RemoveByJsonPointer]
    return true;
}
}  // namespace mpobject

namespace mpobjectasx {
std::unique_ptr<MpObject> ToMpObject(const nlib_utf8_t* json) {
    auto result = JsonStreamParser::Parse(json);
    NLIB_ASSERT(result.first == 0);
    return std::move(result.second);
}
bool AsArray() {
//! [msgpack::MpObjectAsX::Array]
    const char* json = R"([1,2,3])";
    auto obj = ToMpObject(json);
    for (auto& item : obj->AsArray()) {
        int val;
        item.Unbox(&val);
        nlib_printf("%d ", val);
    }
    nlib_printf("\n");
/*
    Output:
    1 2 3
*/
//! [msgpack::MpObjectAsX::Array]
    return true;
}
bool AsMap() {
//! [msgpack::MpObjectAsX::Map]
    const char* json = R"({"key1" : 1, "key2" : 2, "key3" : 3})";
    auto obj = ToMpObject(json);
    for (auto& item : obj->AsMap()) {
        int val;
        item.second.Unbox(&val);
        nlib_printf("%s : %d\n", item.first.GetString(), val);
    }
/*
    Output:
    key1 : 1
    key2 : 2
    key3 : 3
*/
//! [msgpack::MpObjectAsX::Map]
    return true;
}
}  // namespace mpobjectasx

namespace mpwalker {
bool GetSpecificKey() {
    nlib_byte_t buf[1024];
    nlib_utf8_t parsed_value[1024];
//! [msgpack::MpWalker::GetSpecificKey]
    const nlib_utf8_t* json = R"({ "a": { "key": "bad" }, "b" : [1, "key", 3], "key": ["That", "is", "correct"] })";
    size_t n = ToMsgpack(buf, json);
    MpWalker root(&buf[0], n);
    MpWalker x = root["key"];  // moving to right after the toplevel key 'key'
    SUCCEED_IF(x);
    auto r = JsonStreamParser::Parse(x);
    SUCCEED_IF(r.first == 0);
    ToJson(parsed_value, *r.second.get());
    nlib_printf("parsed value: %s\n", parsed_value);
/*
    Output:
    parsed value: ["That","is","correct"]
*/
//! [msgpack::MpWalker::GetSpecificKey]
    return true;
}

bool ArrayAccess() {
    nlib_byte_t buf[1024];
//! [msgpack::MpWalker::ArrayAccess]
    const nlib_utf8_t* json = R"(["a", "b", "c", "d", "e"])";
    size_t n = ToMsgpack(buf, json);
    MpWalker root(&buf[0], n);
    nlib_printf("root[3]: %s\n", root[3] ? "exists" : "not exists");
    nlib_printf("root[5]: %s\n", root[5] ? "exists" : "not exists");
//! [msgpack::MpWalker::ArrayAccess]
/*
    Output:
    root[3]: exists
    root[5]: not exists
*/
    return true;
}

bool MapAccess() {
    nlib_byte_t buf[1024];
//! [msgpack::MpWalker::MapAccess]
    const nlib_utf8_t* json = R"({"key1": 1, "key2" : 2, "key3" : 3})";
    size_t n = ToMsgpack(buf, json);
    MpWalker root(&buf[0], n);
    nlib_printf("root[key2]: %s\n", root["key2"] ? "exists" : "not exists");
    nlib_printf("root[key4]: %s\n", root["key4"] ? "exists" : "not exists");
//! [msgpack::MpWalker::MapAccess]
/*
    Output:
    root[key2]: exists
    root[key4]: not exists
*/
    return true;
}

bool ArrayCount() {
    nlib_byte_t buf[1024];
    nlib_utf8_t parsed_value[1024];
//! [msgpack::MpWalker::ArrayCount]
    const nlib_utf8_t* json = R"([[1,2,3], "a", { "b" : 1 }, false, null])";
    size_t n = ToMsgpack(buf, json);
    MpWalker root(&buf[0], n);
    auto ac_result = root.GetArrayCount();
    SUCCEED_IF(ac_result.first);
    nlib_printf("Array Count: %u\n", ac_result.second);
    for (uint32_t i = 0; i < ac_result.second; ++i) {
        auto r = JsonStreamParser::Parse(ac_result.first);
        SUCCEED_IF(r.first == 0);
        ToJson(parsed_value, *r.second.get());
        nlib_printf("root[%u]: %s\n", i, parsed_value);
        ac_result.first = ac_result.first.Skip();
        SUCCEED_IF(ac_result.first);
    }
/*
    Output:
    Array Count: 5
    root[0]: [1,2,3]
    root[1]: "a"
    root[2]: {"b":1}
    root[3]: false
    root[4]: null
*/
//! [msgpack::MpWalker::ArrayCount]
    return true;
}

bool MapCount() {
    nlib_byte_t buf[1024];
    nlib_utf8_t parsed_key[1024];
    nlib_utf8_t parsed_value[1024];
    //! [msgpack::MpWalker::MapCount]
    const nlib_utf8_t* json = R"({"key1" : "value", "key2" : [1,2,3], "key3" : [true, null]})";
    size_t n = ToMsgpack(buf, json);
    MpWalker root(&buf[0], n);
    auto ac_result = root.GetMapCount();
    SUCCEED_IF(ac_result.first);
    nlib_printf("Map Count: %u\n", ac_result.second);
    for (uint32_t i = 0; i < ac_result.second; ++i) {
        auto r = JsonStreamParser::Parse(ac_result.first);
        SUCCEED_IF(r.first == 0);
        ToJson(parsed_key, *r.second.get());
        ac_result.first = ac_result.first.Skip();
        SUCCEED_IF(ac_result.first);

        r = JsonStreamParser::Parse(ac_result.first);
        SUCCEED_IF(r.first == 0);
        ToJson(parsed_value, *r.second.get());
        nlib_printf("root[%s]: %s\n", parsed_key, parsed_value);
        ac_result.first = ac_result.first.Skip();
        SUCCEED_IF(ac_result.first);
    }
    /*
    Output:
    Map Count: 3
    root["key1"]: "value"
    root["key2"]: [1,2,3]
    root["key3"]: [true,null]
    */
    //! [msgpack::MpWalker::MapCount]
    return true;
}
}  // namespace mpwalker

namespace mpdirectwriter {
bool Write() {
    nlib_utf8_t str[1024];
//! [msgpack::MpDirectWriter]
    // make {"key1" : "value", "key2" : [1,2,3], "key3" : [true, null]}
    nlib_byte_t buf[1024];
    nlib_byte_t* p = &buf[0];
    nlib_byte_t* pend = &buf[0] + sizeof(buf);
    p = MpDirectWriter::WriteMapHeader(p, pend, 3);
    p = MpDirectWriter::WriteString(p, pend, "key1");
    p = MpDirectWriter::WriteString(p, pend, "value");
    p = MpDirectWriter::WriteString(p, pend, "key2");
    {
        p = MpDirectWriter::WriteArrayHeader(p, pend, 3);
        p = MpDirectWriter::WriteInt(p, pend, 1);           // write 0x01 (select best width)
        p = MpDirectWriter::WriteUnsignedInt(p, pend, 2U);  // write 0x02 (select best width, faster if unsigned)
        p = MpDirectWriter::WriteUint16(p, pend, 3);        // write 0xcd, 0x00, 0x03(fixed width)
    }
    p = MpDirectWriter::WriteString(p, pend, "key3");
    {
        p = MpDirectWriter::WriteArrayHeader(p, pend, 2);
        p = MpDirectWriter::WriteBoolean(p, pend, true);
        p = MpDirectWriter::WriteNil(p, pend);
    }
    ToJson(str, &buf[0], p);
    nlib_printf("%s\n", str);
/*
    Output:
    {"key1":"value","key2":[1,2,3],"key3":[true,null]}
*/
//! [msgpack::MpDirectWriter]
    return true;
}
}  // mpdirectwriter

namespace csvreader {
bool Read() {
    nlib_utf8_t buf[1024];
//! [msgpack::CsvReader]
    const nlib_utf8_t* csv = R"(10,-10,1.2,0x100
true,false,
"aaa","b
bb","ccc"
zzz,yyy,xxx
"aaa","b""bb","ccc")";
    auto r = CsvReader::Read(csv);
    SUCCEED_IF(r.first == 0);
    ToJson(buf, *r.second);
    nlib_printf("%s\n", buf);
/*
    Output:
    [[10,-10,1.2,"0x100"],[true,false,null,null],["aaa","b\nbb","ccc",null],["zzz","yyy","xxx",null],["aaa","b\"bb","ccc",null]]
*/
//! [msgpack::CsvReader]
    return true;
}
}  // csvreader

namespace jsonstreamparser {
bool Parse1() {
#define EV_FOUND(str) nlib_printf(str " found\n"); break
    //! [msgpack::JsonStreamParser::Parse1]
    const nlib_utf8_t* json = R"({"key1" : "value", "key2" : [-1,2,12345678901,1.2], "key3" : [true, false, null]})";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    while (parser.HasNext()) {
        JsonStreamParser::Event ev = parser.Next();
        SUCCEED_IF(!parser.GetToken().token_toobig);
        switch (ev) {
        case JsonStreamParser::kEventNull: EV_FOUND("null");
        case JsonStreamParser::kEventTrue: EV_FOUND("true");
        case JsonStreamParser::kEventFalse: EV_FOUND("false");
        case JsonStreamParser::kEventString:
            nlib_printf("String(%s) found\n", parser.GetToken().buf); break;
        case JsonStreamParser::kEventNumberInt32:
            nlib_printf("Int32(%" PRId32 ") found\n", parser.GetToken().number.i32); break;
        case JsonStreamParser::kEventNumberUint32:
            nlib_printf("Uint32(%" PRIu32 ") found\n", parser.GetToken().number.u32); break;
        case JsonStreamParser::kEventNumberInt64:
            nlib_printf("Int64(%" PRId64 ") found\n", parser.GetToken().number.i64); break;
        case JsonStreamParser::kEventNumberUint64:
            nlib_printf("Uint64(%" PRIu64 ") found\n", parser.GetToken().number.u64); break;
        case JsonStreamParser::kEventNumberFloat:
            nlib_printf("Float(%f) found\n", parser.GetToken().number.f32); break;
        case JsonStreamParser::kEventNumberDouble:
            nlib_printf("Double(%f) found\n", parser.GetToken().number.f64); break;
        case JsonStreamParser::kEventStartArray: EV_FOUND("array open");
        case JsonStreamParser::kEventEndArray: EV_FOUND("array close");
        case JsonStreamParser::kEventStartMap: EV_FOUND("map open");
        case JsonStreamParser::kEventEndMap: EV_FOUND("map close");
        case JsonStreamParser::kEventKeyName:
            nlib_printf("Key(%s) found\n", parser.GetToken().buf); break;
        case JsonStreamParser::kEventEndDocument:
            // unreachable because parser.HasNext() becomes false
            nlib_printf("End document found\n");
            break;
        default:
            SUCCEED_IF(false);  // error
        }
    }
    SUCCEED_IF(!!parser);
    parser.Close();
    stream.Close();
/*
    Output:
    map open found
    Key(key1) found
    String(value) found
    Key(key2) found
    array open found
    Int32(-1) found
    Uint32(2) found
    Uint64(12345678901) found
    Double(1.200000) found
    array close found
    Key(key3) found
    array open found
    true found
    false found
    null found
    array close found
    map close found
*/
    //! [msgpack::JsonStreamParser::Parse1]
#undef EV_FOUND
    return true;
}
bool Parse2() {
    //! [msgpack::JsonStreamParser::Parse2]
    const nlib_utf8_t* json = R"({"key1" : "value", "key2" : [-1,2,12345678901,1.2], "key3" : [true, false, null]})";
    auto result = JsonStreamParser::Parse(json);
    SUCCEED_IF(result.first == 0);
    auto& mpobj_ptr = result.second;  // std::unique_ptr<MpObject>
    //! [msgpack::JsonStreamParser::Parse2]
    (void)mpobj_ptr;
    return true;
}
bool ToInt32() {
    //! [msgpack::JsonStreamParser::ToInt32]
    const nlib_utf8_t* json = R"([100, "string", 2147483648, 1.23456])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToInt32(parser.GetToken());
    nlib_printf("ToInt32: %s, %" PRId32 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToInt32(parser.GetToken());
    nlib_printf("ToInt32: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToInt32(parser.GetToken());
    nlib_printf("ToInt32: %s, %" PRId32 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToInt32(parser.GetToken());
    nlib_printf("ToInt32: %s, %" PRId32 "\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToInt32: OK, 100
    ToInt32: EINVAL, invalid
    ToInt32: ERANGE, -2147483648
    ToInt32: EDOM, 1
*/
    //! [msgpack::JsonStreamParser::ToInt32]
    return true;
}
bool ToUint32() {
    //! [msgpack::JsonStreamParser::ToUint32]
    const nlib_utf8_t* json = R"([100, "string", -1, 1.23456])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToUint32(parser.GetToken());
    nlib_printf("ToUint32: %s, %" PRIu32 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToUint32(parser.GetToken());
    nlib_printf("ToUint32: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToUint32(parser.GetToken());
    nlib_printf("ToUint32: %s, %" PRIu32 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToUint32(parser.GetToken());
    nlib_printf("ToUint32: %s, %" PRIu32 "\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToUint32: OK, 100
    ToUint32: EINVAL, invalid
    ToUint32: ERANGE, -1
    ToUint32: EDOM, 1
*/
    //! [msgpack::JsonStreamParser::ToUint32]
    return true;
}
bool ToInt64() {
    //! [msgpack::JsonStreamParser::ToInt64]
    const nlib_utf8_t* json = R"([100, "string", 9223372036854775808, 1.23456])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToInt64(parser.GetToken());
    nlib_printf("ToInt64: %s, %" PRId64 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToInt64(parser.GetToken());
    nlib_printf("ToInt64: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToInt64(parser.GetToken());
    nlib_printf("ToInt64: %s, %" PRId64 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToInt64(parser.GetToken());
    nlib_printf("ToInt64: %s, %" PRId64 "\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToInt64: OK, 100
    ToInt64: EINVAL, invalid
    ToInt64: ERANGE, -9223372036854775808
    ToInt64: EDOM, 1
*/
    //! [msgpack::JsonStreamParser::ToInt64]
    return true;
}
bool ToUint64() {
    //! [msgpack::JsonStreamParser::ToUint64]
    const nlib_utf8_t* json = R"([100, "string", -1, 1.23456])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToUint64(parser.GetToken());
    nlib_printf("ToUint64: %s, %" PRIu64 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToUint64(parser.GetToken());
    nlib_printf("ToUint64: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToUint64(parser.GetToken());
    nlib_printf("ToUint64: %s, %" PRIu64 "\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToUint64(parser.GetToken());
    nlib_printf("ToUint64: %s, %" PRIu64 "\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToUint64: OK, 100
    ToUint64: EINVAL, invalid
    ToUint64: ERANGE, 18446744073709551615
    ToUint64: EDOM, 1
*/
    //! [msgpack::JsonStreamParser::ToUint64]
    return true;
}
bool ToFloat() {
    //! [msgpack::JsonStreamParser::ToFloat]
    const nlib_utf8_t* json = R"([123.4, "string", 1677216, 3.502823466e+38, 1.8976931348623158e+308])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToFloat(parser.GetToken());
    nlib_printf("ToFloat: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToFloat(parser.GetToken());
    nlib_printf("ToFloat: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToFloat(parser.GetToken());
    nlib_printf("ToFloat: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToFloat(parser.GetToken());
    nlib_printf("ToFloat: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToFloat(parser.GetToken());
    nlib_printf("ToFloat: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToFloat: EDOM, 123.400002
    ToFloat: EINVAL, invalid
    ToFloat: OK, 1677216.000000
    ToFloat: ERANGE, inf
    ToFloat: EINVAL, 0.000000
*/
    //! [msgpack::JsonStreamParser::ToFloat]
    return true;
}
bool ToDouble() {
    //! [msgpack::JsonStreamParser::ToDouble]
    const nlib_utf8_t* json = R"([123.4, "string", 9007199254740992, 1.8976931348623158e+308])";
    MemoryInputStream stream(json, nlib_strlen(json));
    JsonStreamParser parser;
    SUCCEED_IF(parser.Init() == 0);
    SUCCEED_IF(parser.Open(&stream) == 0);
    SUCCEED_IF(parser.Next() == JsonStreamParser::kEventStartArray);

    parser.Next();
    auto r = JsonStreamParser::ToDouble(parser.GetToken());
    nlib_printf("ToDouble: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToDouble(parser.GetToken());
    nlib_printf("ToDouble: %s, invalid\n", nlib_error_string(r.first));

    parser.Next();
    r = JsonStreamParser::ToDouble(parser.GetToken());
    nlib_printf("ToDouble: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Next();
    r = JsonStreamParser::ToDouble(parser.GetToken());
    nlib_printf("ToDouble: %s, %f\n", nlib_error_string(r.first), r.second);

    parser.Close();
    stream.Close();
/*
    Output:
    ToDouble: OK, 123.400000
    ToDouble: EINVAL, invalid
    ToDouble: EDOM, 9007199254740992.000000
    ToDouble: EINVAL, 0.000000
*/
    //! [msgpack::JsonStreamParser::ToDouble]
    return true;
}
}  // namespace jsonstreamparser

namespace jsonstreamgenerator {
std::unique_ptr<MpObject> ToMpObject(const nlib_utf8_t* json) {
    auto result = JsonStreamParser::Parse(json);
    NLIB_ASSERT(result.first == 0);
    return std::move(result.second);
}
bool Generate1() {
    nlib_utf8_t buf[8192];
//! [msgpack::JsonStreamGenerator::Generate1]
    // make {"key1" : "value", "key2" : [1,2,3], "key3" : [true, null]}
    JsonStreamGenerator gen;
    MemoryOutputStream stream(buf);
    SUCCEED_IF(gen.Init() == 0);
    SUCCEED_IF(gen.Open(&stream) == 0);

    gen.StartMap(3);
    gen.Key("key1").String("value");
    gen.Key("key2").StartArray(3).Int32(1).Int32(2).Int32(3).EndArray();
    gen.Key("key3").StartArray(2).Boolean(true).Null().EndArray();
    gen.EndMap();

    SUCCEED_IF(gen.Close() == 0);
    size_t n = stream.Pos();
    SUCCEED_IF(stream.Close());
    buf[n] = '\0';  // note that JsonStreamGenerator does not append '\0'
    nlib_printf("%s\n", buf);
/*
    Output:
    {"key1":"value","key2":[1,2,3],"key3":[true,null]}
*/
//! [msgpack::JsonStreamGenerator::Generate1]
    return true;
}
bool Generate2() {
    const nlib_utf8_t* json = R"({"key1" : "value", "key2" : [1,2,3], "key3" : [true, null]})";
    auto p = ToMpObject(json);
    MpObject& obj = *p;
    nlib_utf8_t buf[1024];
//! [msgpack::JsonStreamGenerator::Generate2]
    auto r = JsonStreamGenerator::Generate(buf, obj);
    SUCCEED_IF(r.first == 0);
    nlib_printf("output size = %" PRIdS "\n", r.second);
    nlib_printf("%s\n", buf);
/*
    Output:
    output size = 50
    {"key1":"value","key2":[1,2,3],"key3":[true,null]}
*/
//! [msgpack::JsonStreamGenerator::Generate2]
    return true;
}
}  // namespace jsonstreamgenerator

namespace jsonpatch {
template<size_t N>
void ToJson(nlib_utf8_t(&buf)[N], const MpObject& obj) {
    JsonStreamGeneratorSettings settings;
    settings.pretty_print = false;
    auto gen = JsonStreamGenerator::Generate(buf, obj, settings);
    NLIB_ASSERT(gen.first == 0);
    (void)gen;
}
std::unique_ptr<MpObject> ToMpObject(const nlib_utf8_t* json) {
    auto result = JsonStreamParser::Parse(json);
    NLIB_ASSERT(result.first == 0);
    return std::move(result.second);
}
bool Build() {
    nlib_utf8_t buf[1024];
//! [msgpack::JsonPatch::Build]
    JsonPatchBuilder builder;
    {
        MpObject v;
        SUCCEED_IF(v.InitString("foo") == 0);
        SUCCEED_IF(builder.Test("/a/b/c", std::move(v)) == 0);
    }
    builder.Remove("/a/b/c");
    {
        MpObject v;
        SUCCEED_IF(v.InitArray(2) == 0);
        SUCCEED_IF(v[0].InitString("foo") == 0);
        SUCCEED_IF(v[1].InitString("bar") == 0);
        SUCCEED_IF(builder.Add("/a/b/c", std::move(v)) == 0);
    }
    {
        MpObject v(42);
        SUCCEED_IF(builder.Replace("/a/b/c", std::move(v)) == 0);
    }
    SUCCEED_IF(builder.Move("/a/b/d", "/a/b/c") == 0);
    SUCCEED_IF(builder.Copy("/a/b/e", "/a/b/d") == 0);
    auto r = builder.Export();
    SUCCEED_IF(r.first == 0);
    ToJson(buf, *r.second);
    nlib_printf("%s\n", buf);

    auto obj = ToMpObject(R"({"a": {"b" : {"c" : "foo"}}})");
    SUCCEED_IF(JsonPatch::Apply(obj.get(), r.second.get(), nullptr) == 0);
    ToJson(buf, *obj);
    nlib_printf("%s\n", buf);
/*
    Output:
    [{"op":"test","path":"\/a\/b\/c","value":"foo"},
     {"op":"remove","path":"\/a\/b\/c"},
     {"op":"add","path":"\/a\/b\/c","value":["foo","bar"]},
     {"op":"replace","path":"\/a\/b\/c","value":42},
     {"op":"move","from":"\/a\/b\/c","path":"\/a\/b\/d"},
     {"op":"copy","from":"\/a\/b\/d","path":"\/a\/b\/e"}]
    {"a":{"b":{"d":42,"e":42}}}
*/
//! [msgpack::JsonPatch::Build]
    return true;
}
bool Retract() {
    nlib_utf8_t buf[1024];
//! [msgpack::JsonPatch::Retract]
    auto obj = ToMpObject(R"({"a": {"b" : {"c" : "foo"}}})");
    // To fail at "move" command (/a/b/d not present)
    auto patch = ToMpObject(R"(
    [{"op":"test","path":"/a/b/c","value":"foo"},
     {"op":"remove","path":"/a/b/c"},
     {"op":"add","path":"/a/b/c","value":["foo","bar"]},
     {"op":"replace","path":"/a/b/c","value":42},
     {"op":"move","from":"/a/b/d","path":"/a/b/c"},
     {"op":"copy","from":"/a/b/d","path":"/a/b/e"}])");
    size_t index;
    SUCCEED_IF(JsonPatch::Apply(obj.get(), patch.get(), &index) != 0);
    nlib_printf("fail at index = %" PRIuS "\n", index);
    ToJson(buf, *obj);
    nlib_printf("%s\n", buf);
/*
    Output:
    fail at index = 4
    {"a":{"b":{"c":"foo"}}}
*/
//! [msgpack::JsonPatch::Retract]
    return true;
}
}  // namespace jsonpatch

namespace jsonschema {
std::unique_ptr<MpObject> ToMpObject(const nlib_utf8_t* json) {
    auto result = JsonStreamParser::Parse(json);
    NLIB_ASSERT(result.first == 0);
    return std::move(result.second);
}
bool Validate() {
//! [msgpack::JsonSchema::Validate]
    auto schema = ToMpObject(R"({
    "$schema": "http://json-schema.org/draft-06/schema#",
    "title": "Product",
    "description": "A product from Acme's catalog",
    "type": "object",
    "properties": {
        "id": {
            "description": "The unique identifier for a product",
            "type": "integer"
        },
        "name": {
            "description": "Name of the product",
            "type": "string"
        },
        "price": {
            "type": "number",
            "exclusiveMinimum": 0
        },
        "tags": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1,
            "uniqueItems": true
        }
    },
    "required": ["id", "name", "price"]
    })");
    JsonSchemaConverter converter;
    SUCCEED_IF(converter.Init() == JsonSchemaConverter::kOk);
    SUCCEED_IF(converter.Convert(schema.get(), nullptr) == JsonSchemaConverter::kOk);
    auto bin_schema = converter.Export();
    SUCCEED_IF(std::get<0>(bin_schema) == JsonSchemaConverter::kOk);
    JsonSchemaValidator validator;
    SUCCEED_IF(validator.Init(std::get<1>(bin_schema).get(), std::get<2>(bin_schema)) == JsonSchemaValidator::kOk);
    auto conformant_json = ToMpObject(R"({
    "id": 1,
    "name": "A green door",
    "price": 12.50,
    "tags": ["home", "green"]
    })");
    SUCCEED_IF(validator.Validate(*conformant_json, nullptr) == JsonSchemaValidator::kOk);

    // 'price' is not conformat to the schema
    auto non_conformant_json = ToMpObject(R"({
    "id": 1,
    "name": "A green door",
    "price": -3.0,
    "tags": ["home", "green"]
    })");
    SUCCEED_IF(validator.Validate(*non_conformant_json, nullptr) == JsonSchemaValidator::kFail);
//! [msgpack::JsonSchema::Validate]
    return true;
}
}  // namespace jsonschema

namespace jsonrpc {
bool WriteRequest() {
    char tmp[1024];
    nlib_ns::msgpack::jsonrpc::reqid_t reqid = 1;
//! [msgpack::jsonrpc::WriteRequest]
    JsonRpcRequestWriter w;
    SUCCEED_IF(w.BeginWriteRequest(1) == 0);

    // http://www.jsonrpc.org/specification#examples
    JsonRpcRequest req;
    req.SetId(reqid);  // reqid from JsonRpcClient
    req.SetMethod("subtract");
    auto params = ToMpObject(R"([42,23])");
    req.MoveParamsFrom(std::move(*params));
    w.WriteRequest(req);

    auto result = w.EndWriteRequest();
    SUCCEED_IF(std::get<0>(result) == 0);

    // makes null terminated string and print it
    size_t bufsize = std::get<2>(result);
    memcpy(tmp, std::get<1>(result).get(), bufsize);
    tmp[bufsize] = '\0';
    nlib_printf("%s\n", tmp);
/*
    Output:
    {"id":1,"method":"subtract","params":[42,23],"jsonrpc":"2.0"}
*/
//! [msgpack::jsonrpc::WriteRequest]
    return true;
}

}  // namespace jsonrpc
}  // namespace

#define EXEC_SAMPLE(text, func) \
    nlib_printf("==== %s ====\n", text); \
    if (!func()) { nlib_printf("oops: %s\n", text); return false; }

bool cppapi() {
    EXEC_SAMPLE("msgpack::MpObject::BoxCustomData", mpobject::BoxCustomData);
    EXEC_SAMPLE("msgpack::MpObject::UnboxCustomData", mpobject::UnboxCustomData);
    EXEC_SAMPLE("msgpack::MpObject::BoxCustomData2", mpobject::BoxCustomData2);
    EXEC_SAMPLE("msgpack::MpObject::UnboxCustomData2", mpobject::UnboxCustomData2);
    EXEC_SAMPLE("msgpack::MpObject::Box", mpobject::Box);
    EXEC_SAMPLE("msgpack::MpObject::Unbox", mpobject::Unbox);
    EXEC_SAMPLE("msgpack::MpObject::Assignment", mpobject::AssignmentOperator);
    EXEC_SAMPLE("msgpack::MpObject::Resize", mpobject::Resize);
    EXEC_SAMPLE("msgpack::MpObject::IsJsonPointer", mpobject::IsJsonPointer);
    EXEC_SAMPLE("msgpack::MpObject::ResolveJsonPointer", mpobject::ResolveJsonPointer);
    EXEC_SAMPLE("msgpack::MpObject::DigByJsonPointer", mpobject::DigByJsonPointer);
    EXEC_SAMPLE("msgpack::MpObject::RemoveByJsonPointer", mpobject::RemoveByJsonPointer);

    EXEC_SAMPLE("msgpack::MpObjectAsX::Array", mpobjectasx::AsArray);
    EXEC_SAMPLE("msgpack::MpObjectAsX::Map", mpobjectasx::AsMap);

    EXEC_SAMPLE("msgpack::MpWalker::GetSpecificKey", mpwalker::GetSpecificKey);
    EXEC_SAMPLE("msgpack::MpWalker::ArrayAccess", mpwalker::ArrayAccess);
    EXEC_SAMPLE("msgpack::MpWalker::MapAccess", mpwalker::MapAccess);
    EXEC_SAMPLE("msgpack::MpWalker::ArrayCount", mpwalker::ArrayCount);
    EXEC_SAMPLE("msgpack::MpWalker::MapCount", mpwalker::MapCount);
    EXEC_SAMPLE("msgpack::MpDirectWriter", mpdirectwriter::Write);
    EXEC_SAMPLE("msgpack::CsvReader", csvreader::Read);
    EXEC_SAMPLE("msgpack::JsonStreamParser::Parse1", jsonstreamparser::Parse1);
    EXEC_SAMPLE("msgpack::JsonStreamParser::Parse2", jsonstreamparser::Parse2);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToInt32", jsonstreamparser::ToInt32);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToUint32", jsonstreamparser::ToUint32);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToInt64", jsonstreamparser::ToInt64);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToUint64", jsonstreamparser::ToUint64);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToFloat", jsonstreamparser::ToFloat);
    EXEC_SAMPLE("msgpack::JsonStreamParser::ToDouble", jsonstreamparser::ToDouble);

    EXEC_SAMPLE("msgpack::JsonStreamGenerator::Generate1", jsonstreamgenerator::Generate1);
    EXEC_SAMPLE("msgpack::JsonStreamGenerator::Generate2", jsonstreamgenerator::Generate2);

    EXEC_SAMPLE("msgpack::JsonPatch::build", jsonpatch::Build);
    EXEC_SAMPLE("msgpack::JsonPatch::retract", jsonpatch::Retract);

    EXEC_SAMPLE("msgpack::JsonSchema::Validate", jsonschema::Validate);

    EXEC_SAMPLE("msgpack::jsonrpc::WriteRequest", jsonrpc::WriteRequest);
    return true;
}

bool SampleMain(int, char**) {
    return cppapi();
}

NLIB_MAINFUNC
