﻿
/*--------------------------------------------------------------------------------*
Project: CrossRoad
Copyright (C)Nintendo All rights reserved.

These coded instructions, statements, and computer programs contain proprietary
information of Nintendo and/or its licensed developers and are protected by
national and international copyright laws. They may not be disclosed to third
parties or copied or duplicated in any form, in whole or in part, without the
prior written consent of Nintendo.

The content herein is highly confidential and should be handled accordingly.
*--------------------------------------------------------------------------------*/

#include <thread>
#include <algorithm>
#include <random>
#include <vector>
#include <map>
#include <string>
#include <set>

#include "nn/nlib/succinct/Trie.h"
#include "nn/nlib/succinct/Sbv.h"
#include "nn/nlib/succinct/Bp.h"
#include "nn/nlib/succinct/Louds.h"

using nlib_ns::succinct::Trie;
using nlib_ns::succinct::TrieBuilder;
using nlib_ns::succinct::SparseSet;
using nlib_ns::succinct::Louds;
using nlib_ns::succinct::Bp;

#define SUCCEED_IF(cond) if (!(cond)) return false

namespace {
namespace trie {
bool UseTrie() {
    //! [succinct::UseTrie]
    TrieBuilder builder;
    builder.Init();
    auto loop = [&](std::function<bool(const char*)> func) {
        char buf[5];
        buf[4] = '\0';
        for (char ch0 = 'a'; ch0 <= 'z'; ++ch0) {
            buf[0] = ch0;
            for (char ch1 = 'a'; ch1 <= 'z'; ++ch1) {
                buf[1] = ch1;
                for (char ch2 = 'a'; ch2 <= 'z'; ++ch2) {
                    buf[2] = ch2;
                    for (char ch3 = 'a'; ch3 <= 'z'; ++ch3) {
                        buf[3] = ch3;
                        if (!func(buf)) return false;
                    }
                }
            }
        }
        return true;
    };
    bool result = loop([&](const char* str) { return builder.AddWord(str); });
    SUCCEED_IF(result);
    auto trie = builder.Build2();
    nlib_printf("The size of the words([a-z][a-z][a-z][a-z]): %d bytes\n", 4 * 26 * 26 * 26 * 26);
    nlib_printf("The size of the trie for them: %" PRIuS " bytes\n", trie->MemSize());
    result = loop([&](const char* str) {
        int val = 0;
        trie->Match(str, [](const char* first, const char* last, uint32_t nodeid, void* user_obj) {
            NLIB_UNUSED(nodeid);
            NLIB_UNUSED(first);
            NLIB_UNUSED(last);
            *reinterpret_cast<int*>(user_obj) = 1;
            // nlib_printf("%s\n", std::string(first, last).c_str());
            return true;
        },
        &val);
        return val == 1;
    });
    SUCCEED_IF(result);
/*
    Output:
    The size of the words([a-z][a-z][a-z][a-z]): 1827904 bytes
    The size of the trie for them: 785007 bytes
    aaaa
    ....
    zzzz
*/
    //! [succinct::UseTrie]
    return true;
}
bool SuffixTrie() {
    //! [succinct::SuffixTrie]
    // text from https://en.wikipedia.org/wiki/Trie
    const char* str = R"(In computer science, a trie, also called digital tree and sometimes radix tree or prefix tree (as they can be searched by prefixes), is a kind of search tree - an ordered tree data structure that is used to store a dynamic set or associative array where the keys are usually strings. Unlike a binary search tree, no node in the tree stores the key associated with that node; instead, its position in the tree defines the key with which it is associated. All the descendants of a node have a common prefix of the string associated with that node, and the root is associated with the empty string. Values are not necessarily associated with every node. Rather, values tend only to be associated with leaves, and with some inner nodes that correspond to keys of interest. For the space-optimized presentation of prefix tree, see compact prefix tree.)";
    const size_t len = nlib_strlen(str);
    TrieBuilder builder;
    SUCCEED_IF(builder.Init());

    // construct the suffix trie for 'str'
    for (size_t i = 0; i < len - 1; ++i) {
        SUCCEED_IF(builder.AddWord(str + i));
    }
    auto trie = builder.Build2();

    // map nodeid -> pos in 'str'
    std::map<uint32_t, size_t> nodeid_to_pos;
    struct UserObj {
        std::map<uint32_t, size_t>* nodeid_to_pos;
        size_t i;
    } uobj;
    uobj.nodeid_to_pos = &nodeid_to_pos;
    for (size_t i = 0; i < len - 1; ++i) {
        uobj.i = i;
        trie->Match(str + i, [](const char*, const char*, uint32_t nodeid, void* user_obj) {
            UserObj* u = reinterpret_cast<UserObj*>(user_obj);
            (*u->nodeid_to_pos)[nodeid] = u->i;
            return true;
        }, &uobj);
    }

    // list the occurences of the specified pattern
    auto list_up = [&](const char* pat) {
        nlib_printf("---- %s ----\n", pat);
        size_t l = nlib_strlen(pat);
        auto result = trie->GetCommonPrefixWords(pat, l);
        if (result.first != 0) return;
        for (auto& nodeid : result.second) {
            size_t pos = nodeid_to_pos[nodeid];
            size_t first = pos >= 10 ? pos - 10 : 0;
            size_t last = pos + 10 + l < len ? pos + 10 + l : len;
            nlib_printf("%s\n", std::string(str + first, str + last).c_str());
        }
    };
    list_up("tree");
    list_up("of");
    /*
    Output:
    ---- tree ----
    or prefix tree (as they
    of search tree - an orde
    d digital tree and somet
    n ordered tree data stru
    on in the tree defines t
    mes radix tree or prefix
    de in the tree stores th
    ry search tree, no node
    of prefix tree, see comp
    ct prefix tree.
    ---- of ----
    scendants of a node ha
    d to keys of interest.
    sentation of prefix tr
    is a kind of search tr
    on prefix of the strin
    */
    //! [succinct::SuffixTrie]
    return true;
}
}  // namespace trie

namespace sbv {
bool UseSparseSet() {
    //! [succinct::SparseSet]
    SparseSet s;
    SUCCEED_IF(s.Init(100000000));
    std::random_device seed_gen;
    std::mt19937 engine(seed_gen());
    std::uniform_int_distribution<int> dice(0, 100000000 - 1);
    for (int i = 0; i < 1000000; ++i) {
        uint32_t result = dice(engine);
        s.TurnOn(result);  // result may duplicate, but ignore it for simplicity
    }
    SUCCEED_IF(s.Build());
    nlib_printf("Set of about 1000000 numbers requires %" PRIuS " bytes\n", s.MemSize());
    /*
    Output(it varies):
    Set of 1000000 random numbers requires 1652103 bytes
    */
    //! [succinct::SparseSet]
    return true;
}
}

namespace louds {
bool UseLouds() {
    //! [succinct::Louds]
    class MyTree {
    public:
        int GetNumChildren() {
            return static_cast<int>(data_[cur_]);
        }
        bool BfsMoveNext() {
            if (cur_ + 1 >= data_.size()) return false;
            ++cur_;
            return true;
        }
    private:
        std::vector<int> data_{ 3, 1, 0, 3, 0, 0, 1, 2, 0, 0, 0 };
        size_t cur_ = 0;
    } mytree;
    Louds louds;
    SUCCEED_IF(louds.Init(mytree, 11));
    std::function<void(Louds::NodeId)> traverse = [&](Louds::NodeId nodeid) {
        auto parent = louds.GetNodeIdOfParent(nodeid);
        if (parent.first) {
            nlib_printf("%u -> %u\n", parent.second, nodeid);
        }
        auto cld = louds.GetNodeIdOfFirstChild(nodeid);
        if (cld.first) {
            do {
                traverse(cld.second);
                cld = louds.GetNodeIdOfNextSibling(cld.second);
            } while (cld.first);
        }
    };
    traverse(0);
    /*
    Output:
    0 -> 1
    1 -> 4
    0 -> 2
    0 -> 3
    3 -> 5
    3 -> 6
    6 -> 8
    3 -> 7
    7 -> 9
    7 -> 10
    */
    //! [succinct::Louds]
    return true;
}
}  // namespace louds

namespace bp {
bool UseBp() {
    //! [succinct::Bp]
    class MyTree {
    public:
        const void* NodePtr() {
            return &data_[cur_];
        }
        const void* ParentNodePtr() {
            int parent = data_[cur_];
            return parent >= 0 ? &data_[parent] : nullptr;
        }
        bool MoveNext() {
            if (cur_ + 1 >= data_.size()) return false;
            ++cur_;
            return true;
        }

    private:
        std::vector<int> data_{ -1, 0, 1, 0, 0, 4, 4, 6, 4, 8, 8 };
        size_t cur_ = 0;
    } mytree;
    Bp bp;
    SUCCEED_IF(bp.Init(mytree, 11));
    std::function<void(Bp::NodeId)> traverse = [&](Bp::NodeId nodeid) {
        auto parent = bp.GetNodeIdOfParent(nodeid);
        if (parent.first) {
            nlib_printf("%u -> %u\n", parent.second, nodeid);
        }
        auto cld = bp.GetNodeIdOfFirstChild(nodeid);
        if (cld.first) {
            do {
                traverse(cld.second);
                cld = bp.GetNodeIdOfNextSibling(cld.second);
            } while (cld.first);
        }
    };
    traverse(0);
    /*
    Output:
    0 -> 1
    1 -> 2
    0 -> 3
    0 -> 4
    4 -> 5
    4 -> 6
    6 -> 7
    4 -> 8
    8 -> 9
    8 -> 10
    */
    //! [succinct::Bp]
    return true;
}
}  // namespace bp
}  // namespace

#define EXEC_SAMPLE(text, func) \
    nlib_printf("==== %s ====\n", text); \
    if (!func()) { nlib_printf("oops: %s\n", text); return false; }

bool cppapi() {
    EXEC_SAMPLE("succinct::UseTrie", trie::UseTrie);
    EXEC_SAMPLE("succinct::SuffixTrie", trie::SuffixTrie);
    EXEC_SAMPLE("succinct::UseSparseSet", sbv::UseSparseSet);
    EXEC_SAMPLE("succinct::Louds", louds::UseLouds);
    EXEC_SAMPLE("succinct::Bp", bp::UseBp);
    return true;
}

bool SampleMain(int, char**) {
    return cppapi();
}

NLIB_MAINFUNC
