﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#include <thread>
#include <algorithm>
#include <random>
#include <type_traits>

#include "nn/nlib/testing/testing.h"
#include "nn/nlib/NativePathMapper.h"
#include "nn/nlib/Cstring.h"

#define SUCCEED_IF(cond) if (!(cond)) return false

namespace {

//! [testing::Test]
namespace test {
int Factorial(int n) {
    if (n == 0) return 1;
    return n * Factorial(n - 1);
}
TEST(FactorialTest, Zero) {
    EXPECT_EQ(1, Factorial(0));
}
TEST(FactorialTest, Positive) {
    EXPECT_EQ(1, Factorial(1));
    EXPECT_EQ(2, Factorial(2));
    EXPECT_EQ(6, Factorial(3));
    EXPECT_EQ(40320, Factorial(8));
}
/*!
    Output:
    [ RUN      ] FactorialTest.Zero
    [       OK ] FactorialTest.Zero (0 ms)
    [ RUN      ] FactorialTest.Positive
    [       OK ] FactorialTest.Positive (0 ms)
    [----------] 2 tests from FactorialTest (0 ms total)
*/
//! [testing::Test]
}  // namespace test

namespace fixture {
//! [testing::TestFixture]
class MyTest : public ::testing::Test {
 protected:
    static void SetUpTestCase() { nlib_printf("SetUpTestCase called\n"); }
    static void TearDownTestCase() { nlib_printf("TearDownTestCase called\n"); }
    void SetUp() override { nlib_printf("SetUp called\n"); }
    void TearDown() override { nlib_printf("TearDown called\n"); }
    // You can add the member functions or the data members for MyTest
};
TEST_F(MyTest, test1) {
    nlib_printf("MyTest::test1\n");
    // Write your test
}
TEST_F(MyTest, test2) {
    nlib_printf("MyTest::test2\n");
    // Write your test
}
/*
    Output:
    SetUpTestCase called
    [ RUN      ] MyTest.test1
    SetUp called
    MyTest::test1
    TearDown called
    [       OK ] MyTest.test1 (0 ms)
    [ RUN      ] MyTest.test2
    SetUp called
    MyTest::test2
    TearDown called
    [       OK ] MyTest.test2 (0 ms)
    TearDownTestCase called
    [----------] 2 tests from MyTest (1 ms total)
*/
//! [testing::TestFixture]
}  // namespace fixture

namespace param {
//! [testing::Param]
bool IsEven(int n) { return n % 2 == 0; }
class ParamTestInt : public ::testing::TestWithParam<int> {
 protected:
    static void SetUpTestCase() { nlib_printf("SetUpTestCase called\n"); }
    static void TearDownTestCase() { nlib_printf("TearDownTestCase called\n"); }
    void SetUp() override {}
    void TearDown() override {}
    // You can add the member functions or the data members for ParamTestInt
};
TEST_P(ParamTestInt, IsEven) {
    int n = GetParam();  // returns int because ParamTestInt derives from TestWithParam<int>
    nlib_printf("IsEven(%d)\n", n);
    ASSERT_TRUE(IsEven(n));
}
INSTANTIATE_TEST_CASE_P(
    UseRange,                         // Name
    ParamTestInt,                     // Test Case
    ::testing::Range(0, 100, 10)      // The values retrieved by GetParam(first, last, step)
);
INSTANTIATE_TEST_CASE_P(
    UseValues,                        // Name
    ParamTestInt,                     // Test Case
    ::testing::Values(0, 2, 4, 6, 8)  // Specify the values directly
);
int values[] = { 12, 14, 16, 18 };
INSTANTIATE_TEST_CASE_P(
    UseValuesIn,                      // Name
    ParamTestInt,                     // Test Case
    ::testing::ValuesIn(values)       // Specify the container
);
/*
    Output:
    SetUpTestCase called
    [ RUN      ] ParamTestInt.IsEven
    IsEven(0)
    IsEven(10)
    IsEven(20)
    IsEven(30)
    IsEven(40)
    IsEven(50)
    IsEven(60)
    IsEven(70)
    IsEven(80)
    IsEven(90)
    IsEven(0)
    IsEven(2)
    IsEven(4)
    IsEven(6)
    IsEven(8)
    IsEven(12)
    IsEven(14)
    IsEven(16)
    IsEven(18)
    [       OK ] ParamTestInt.IsEven (2 ms)
    TearDownTestCase called
*/
//! [testing::Param]
}  // namespace param

namespace typed_test {
//! [testing::TypedTest]
template<class T>
class IsIntegralTest : public testing::Test {
 public:  // they must be public
    static void SetUpTestCase() { nlib_printf("SetUpTestCase called\n"); }
    static void TearDownTestCase() { nlib_printf("TearDownTestCase called\n"); }
 protected:
    void SetUp() override {}
    void TearDown() override {}
    // You can add the member functions or the data members for IsIntegralTest
};
typedef ::testing::Types<int, long, long long, char, int16_t> TestTypes;
TYPED_TEST_CASE(IsIntegralTest, TestTypes);
TYPED_TEST(IsIntegralTest, test) {
    ASSERT_TRUE(std::is_integral<TypeParam>::value);
}
/*
    Output:
    TearDownTestCase called
    [----------] 1 tests from ParamTestInt (5 ms total)

    SetUpTestCase called
    [ RUN      ] IsIntegralTest.test.1
    [       OK ] IsIntegralTest.test.1 (0 ms)
    TearDownTestCase called
    [----------] 1 tests from IsIntegralTest (1 ms total)

    SetUpTestCase called
    [ RUN      ] IsIntegralTest.test.2
    [       OK ] IsIntegralTest.test.2 (0 ms)
    TearDownTestCase called
    [----------] 1 tests from IsIntegralTest (1 ms total)

    SetUpTestCase called
    [ RUN      ] IsIntegralTest.test.3
    [       OK ] IsIntegralTest.test.3 (0 ms)
    TearDownTestCase called
    [----------] 1 tests from IsIntegralTest (1 ms total)

    SetUpTestCase called
    [ RUN      ] IsIntegralTest.test.4
    [       OK ] IsIntegralTest.test.4 (0 ms)
    TearDownTestCase called
    [----------] 1 tests from IsIntegralTest (1 ms total)

    SetUpTestCase called
    [ RUN      ] IsIntegralTest.test.5
    [       OK ] IsIntegralTest.test.5 (0 ms)
    TearDownTestCase called
    [----------] 1 tests from IsIntegralTest (1 ms total)
*/
//! [testing::TypedTest]
}  // namespace typed_test

namespace pred {
//! [testing::Pred]
bool IsEven(int n) { return n % 2 == 0; }
TEST(Pred, test) {
    ASSERT_PRED1(IsEven, 1 + 2);
}
/*
    Output:
    [ RUN      ] pred.test
    IsEven(1 + 2) evaluates to false, where
    1 + 2 evaluates to 3
    [  FAILED  ] pred.test (0 ms)
*/
//! [testing::Pred]
}  // namespace pred

namespace pred_assertionresult {
//! [testing::PredAssertionResult]
::testing::AssertionResult IsEven(int n) {
    if (n % 2 != 0)
        return ::testing::AssertionFailure() << n << " is not even";
    return ::testing::AssertionSuccess();
}
TEST(PredAssertionResult, test) {
    ASSERT_TRUE(IsEven(1 + 2));
}
/*
    Output:
    [ RUN      ] pred.test
    Value of: IsEven(1 + 2)
    Actual: false (3 is not even)
    Expected: true

    [  FAILED  ] pred.test (0 ms)
*/
//! [testing::PredAssertionResult]
}  // namespace pred_assertionresult

namespace pred_format {
//! [testing::PredFormat]
::testing::AssertionResult IsEven(const char* n_expr, int n) {
    if (n % 2 != 0)
        return ::testing::AssertionFailure() << n_expr << " is not even";
    return ::testing::AssertionSuccess();
}
TEST(PredFormat, test) {
    ASSERT_PRED_FORMAT1(IsEven, (1 + 2));
}
/*
    Output:
    [ RUN      ] PredFormat.test
    (1 + 2) is not even
    [  FAILED  ] PredFormat.test (0 ms)
    [----------] 1 tests from PredFormat (0 ms total)
*/
//! [testing::PredFormat]
}  // namespace pred_format
}  // namespace

NLIB_PATHMAPPER_FORSAMPLE

bool SampleMain(int argc, char** argv) {
    InitPathMapperForSample();
    char path[512];
    char buf[512];
    size_t count;
    g_pathmapper.ResolvePath(&count, path, "nlibpath:///readwrite/snippet.xml");
    nlib_snprintf(&count, buf, "xml:%s", path);

    ::testing::GTEST_FLAG(output) = buf;
    ::testing::InitNintendoTest(&argc, argv);

    // RUN_ALL_TESTS returns 0 if all the tests are successful.
    // You can use the return value of main() function.
    // return RUN_ALL_TESTS() == 0;
    RUN_ALL_TESTS();
    return true;
}

NLIB_MAINFUNC
