<#
    Copyright (C)Nintendo All rights reserved.

    These coded instructions, statements, and computer programs contain proprietary
    information of Nintendo and/or its licensed developers and are protected by
    national and international copyright laws. They may not be disclosed to third
    parties or copied or duplicated in any form, in whole or in part, without the
    prior written consent of Nintendo.

    The content herein is highly confidential and should be handled accordingly.
#>

<#
.SYNOPSIS
    Inspects the software installed and etc.
.DESCRIPTION
    It collects and outputs the information of the software installed on the computer.
.PARAMETER ParentOfNxSdkRoot
    The script checks the subdirectory of ${ParentOfNxSdkRoot} to find NintendoSDK NXAddon.
.PARAMETER All
    Equals to -Msvc -Cygwin -Git -Doxygen -CMake -NintendoSdk -Vagrant -Wsl -Ninja
.PARAMETER Msvc
    Checks Visual Studio 2013,2015,2017 and the libraries.
.PARAMETER Cygwin
    Checks Cygwin
.PARAMETER Git
    Checks git for windows
.PARAMETER Doxygen
    Checks Doxygen and GraphViz
.PARAMETER CMake
    Checks CMake
.PARAMETER NintendoSDK
    Checks NintendoSDK NXAddon 3.x+
.PARAMETER Wsl
    Checks Windows Subsystem for Linux
.PARAMETER Ninja
    Checks ninja.exe
.EXAMPLE
    .\Inspection.ps1 -All | ConvertTo-Json
#>

#Requires -Version 5
Param(
    [string]$ParentOfNxSdkRoot="C:\ndi",
    [switch]$All,
    [switch]$Msvc,
    [switch]$Cygwin,
    [switch]$Git,
    [switch]$Doxygen,
    [switch]$CMake,
    [switch]$NintendoSdk,
    [switch]$Vagrant,
    [switch]$Wsl,
    [switch]$Ninja
)

$table = @{}
Set-Variable drive_list -Option ReadOnly -Value `
    (Get-WmiObject Win32_LogicalDisk | Where-Object { $_.DriveType -eq 3 })
Set-Variable script_dir (Resolve-Path (Split-Path -Parent $MyInvocation.MyCommand.Path)).Path -Option ReadOnly
function DoInspection {
    Param(
        [string]$label,
        $script
    )
    try {
        $value = & $script
        $table.Add($label, $value)
    } catch {
        Write-Error "Error in inspecting $label"
        $table.Add($label, $null)
    }
}

function ScanNxSdkDir {
    Param([string]$root, [int]$major_ver)
    $target_dir = ""
    $target_ver = [System.Version]("{0}.99999.99999" -f @($major_ver - 1))
    $dirs = Get-ChildItem -Path $root -Attributes Directory | ForEach-Object { $_.FullName }
    $dirs += $Env:NINTENDO_SDK_ROOT
    $spec_nx = "Include\nn\detail\nn_Version-spec.NX.h"
    foreach ($dir in $dirs) {
        # dev repository
        $sdkdir = $dir
        $verfile = "$dir\Programs\Alice\$spec_nx"
        if (!(Test-Path $verfile)) {
            # $Env:NINTENDO_SDK_ROOT
            $verfile = "$dir\$spec_nx"
            if (!(Test-Path $verfile)) {
                # NDK at $ParentOfNxSdkRoot
                $sdkdir = "$dir\NintendoSdk"
                $verfile = "$dir\NintendoSdk\$spec_nx"
                if (!(Test-Path $verfile)) {
                    continue
                }
            }
        }
        $sdkver_major = Select-String "NN_NX_ADDON_VERSION_MAJOR\s+([0-9]+)" $verfile | ForEach-Object {
            $_.Matches.Groups[1].Value -as [int]
        }
        if ($sdkver_major -eq $major_ver) {
            $sdkver_minor = Select-String "NN_NX_ADDON_VERSION_MINOR\s+([0-9]+)" $verfile | ForEach-Object {
                $_.Matches.Groups[1].Value -as [int]
            }
            $sdkver_micro = Select-String "NN_NX_ADDON_VERSION_MICRO\s+([0-9]+)" $verfile | ForEach-Object {
                $_.Matches.Groups[1].Value -as [int]
            }
            $ver = ("{0}.{1}.{2}" -f @($sdkver_major, $sdkver_minor, $sdkver_micro))
            if ([System.Version]$ver -ge $target_ver) {
                if ([System.Version]$ver -eq $target_ver) {
                    $sdkdir_time = (Get-ItemProperty $sdkdir).LastWriteTime
                    $target_dir_time = (Get-ItemProperty $target_dir).LastWriteTime
                    if ($sdkdir_time -le $target_dir_time) {
                        continue
                    }
                }
                $target_dir = $sdkdir
                $target_ver = $ver
            }
        }
    }
    if ($target_dir) {
        $rynda_ver = "error"
        if (Test-Path "$target_dir\ToolChains") {
            $rynda_ver = Get-ChildItem "$target_dir\ToolChains" -Directory | Where-Object {
                $_.Name -match "clang-for-nx"
            } | ForEach-Object -Begin { $ver = "" } -End { Write-Output $ver } {
                if ($_.Name -match "clang-for-nx.([0-9\.]+)") {
                    $ver = $Matches[1]
                }
            } 
        } else {
            $text = & "$target_dir\Compilers\NX\bin\nx-clang.exe" --version
            if ($text[0] -match "Rynda clang version ([0-9\.]+)") {
                $rynda_ver = $Matches[1]
            }
        }
        Write-Output @{version = $target_ver; path = $target_dir; rynda_version = $rynda_ver}
    } else {
        Write-Output $null
    }
}

if ($All) {
    DoInspection "Windows" {
        $version = ([System.Environment]::OSVersion).Version -as [string]
        Write-Output @{version = $version}
    }
    DoInspection "PowerShell" {
        $version = $PSVersionTable.PSVersion -as [string]
        Write-Output @{version = $version}
    }
}
if ($Doxygen -or $All) {
    DoInspection "Doxygen" {
        $path = (Get-Command doxygen -ErrorAction SilentlyContinue).Source
        if ($path) {
            $version = & doxygen --version
            Write-Output @{version = $version; path = $path}
        } else {
            Write-Output $null
        }
    }
    DoInspection "GraphViz" {
        $path = (Get-Command dot -ErrorAction SilentlyContinue).Source
        if ($path) {
            $text = & dot -V 2>&1
            if ($text -match "version ([0-9][0-9\.]*[0-9])") {
                $version = $Matches[1]
            } else {
                $version = $text
            }
            Write-Output @{version = $version; path = $path}
        } else {
            Write-Output $null
        }
    }
}
if ($Msvc -or $All) {
    DoInspection "Msvc" {
        $vs2017 = $null
        try {
            if (Test-Path "${Env:ProgramFiles(x86)}\Microsoft Visual Studio\Installer\vswhere.exe") {
                $data = & "${Env:ProgramFiles(x86)}\Microsoft Visual Studio\Installer\vswhere.exe" -nologo -version "[15.1,16.0)" -format json | ConvertFrom-Json
                if ($data) {
                    $vs2017 = @{version = $data.installationVersion; path = $data.installationPath}
                }
            }
        } catch {
        }
        $vs2015 = $null
        try {
            if (Test-Path Env:VS140COMNTOOLS) {
                $version = & "$Env:VS140COMNTOOLS\..\..\VC\bin\cl.exe" 2>&1
                $path = Resolve-Path "$Env:VS140COMNTOOLS\..\.."
                $vs2015 = @{version = $Version.TargetObject[0]; path = $path.Path}
            }
        } catch {
        }
        $vs2013 = $null
        try {
            if (Test-Path Env:VS120COMNTOOLS) {
                $version = & "$Env:VS120COMNTOOLS\..\..\VC\bin\cl.exe" 2>&1
                $path = Resolve-Path "$Env:VS120COMNTOOLS\..\.."
                $vs2013 = @{version = $Version.TargetObject[0]; path = $path.Path}
            }
        } catch {
        }

        # zlib
        $zlib_version = $null
        $zlib_path = $null
        if (${Env:VCPKG_ROOT} -and (Test-Path "${Env:VCPKG_ROOT}\installed\x64-windows\bin\zlib1.dll")) {
            $zlib_path = "${Env:VCPKG_ROOT}\installed\x64-windows\bin\zlib1.dll"
        } elseif (Test-Path "${script_dir}\.cache\vcpkg\installed\x64-windows\bin\zlib1.dll") {
            $zlib_path = "${script_dir}\.cache\vcpkg\installed\x64-windows\bin\zlib1.dll"
        } elseif (Test-Path "${Env:ProgramFiles}\zlib\bin\zlib.dll") {
            $zlib_path = "${Env:ProgramFiles}\zlib\bin\zlib.dll"
        }
        if ($zlib_path) {
            $zlib_version = (Get-ItemProperty $zlib_path).VersionInfo.FileVersion
        }

        # curl
        $curl_version = $null
        $curl_path = $null
        if (${Env:VCPKG_ROOT} -and (Test-Path "${Env:VCPKG_ROOT}\installed\x64-windows\bin\libcurl.dll")) {
            $curl_path = "${Env:VCPKG_ROOT}\installed\x64-windows\bin\libcurl.dll"
        } elseif (Test-Path "${script_dir}\.cache\vcpkg\installed\x64-windows\bin\libcurl.dll") {
            $curl_path = "${script_dir}\.cache\vcpkg\installed\x64-windows\bin\libcurl.dll"
        } elseif (Test-Path "${Env:ProgramFiles}\CURL\bin\libcurl.dll") {
            $curl_path = "${Env:ProgramFiles}\CURL\bin\libcurl.dll"
        }
        if ($curl_path) {
            $curl_version = (Get-ItemProperty $curl_path).VersionInfo.FileVersion
        }

        $cmake_default_generator = $null
        if ($vs2017) {
            $cmake_default_generator = "Visual Studio 15 2017"
        } elseif ($vs2015) {
            $cmake_default_generator = "Visual Studio 14 2015"
        } elseif ($vs2013) {
            $cmake_default_generator = "Visual Studio 12 2013"
        }
        if ($vs2017 -or $vs2015 -or $vs2013) {
            Write-Output @{
                Vs2017 = $vs2017;
                Vs2015 = $vs2015;
                Vs2013 = $vs2013;
                zlib_version = $zlib_version;
                zlib_path = $zlib_path;
                curl_version = $curl_version;
                curl_path = $curl_path;
                cmake_default_generator = $cmake_default_generator
            }
        }
    }
}
if ($Git -or $All) {
    DoInspection "Git" {
        $git_prog_list = @(
            "git.exe",
            "${Env:ProgramFiles}\Git\bin\git.exe",
            "${Env:ProgramFiles(x86)}\Git\bin\git.exe")
        # CMake's FindGit does not use this
        #if ($table.Msvc.Vs2017) {
        #    $git_prog_list += ("{0}\Common7\IDE\CommonExtensions\Microsoft\TeamFoundation\Team Explorer\Git\cmd\git.exe" -f $table.Msvc.Vs2017.path)
        #}
        foreach ($git_prog in $git_prog_list) {
            $comm = Get-Command $git_prog -ErrorAction SilentlyContinue
            if ($comm) {
                $path = $comm.Source
                $version = $comm.Version -as [string]
                Write-Output @{version = $version; path = $path}
                return
            }
        }
        Write-Output $null
    }
}
if ($CMake -or $All) {
    DoInspection "CMake" {
        $cmake_prog_list = @(
            "cmake.exe",
            "${Env:ProgramFiles}\CMake\bin\cmake.exe",
            "${Env:ProgramFiles(x86)}\CMake\bin\cmake.exe")
        if ($table.Msvc.Vs2017) {
            $cmake_prog_list += ("{0}\Common7\IDE\CommonExtensions\Microsoft\CMake\CMake\bin\cmake.exe" -f $table.Msvc.Vs2017.path)
        }
        foreach ($cmake_prog in $cmake_prog_list) {
            $path = (Get-Command $cmake_prog -ErrorAction SilentlyContinue).Source
            if ($path) {
                $data = & $cmake_prog --version
                Write-Output @{version = ($data[0] -replace "cmake version ", ""); path = $path}
                return
            }
        }
        Write-Output $null
    }
}
if ($Ninja -or $All) {
    DoInspection "Ninja" {
        $path = (Get-Command "ninja.exe" -ErrorAction SilentlyContinue).Source
        if ($path) {
            $version = & $path --version
            Write-Output @{version = $version; path = $path}
            return
        }
        $path = ("{0}\Common7\IDE\CommonExtensions\Microsoft\CMake\Ninja\ninja.exe" -f $table.Msvc.Vs2017.path)
        if (Test-Path $path) {
            $version = & $path --version
            $Env:Path += (";{0}" -f (Get-Item $path).DirectoryName)
            Write-Output @{version = $version; path = $path}
            return
        }
        Write-Output $null
    }
}
if ($NintendoSdk -or $All) {
    DoInspection "NxAddon" {
        $nxaddon6 = ScanNxSdkDir $ParentOfNxSdkRoot 6
        $nxaddon5 = ScanNxSdkDir $ParentOfNxSdkRoot 5
        $nxaddon4 = ScanNxSdkDir $ParentOfNxSdkRoot 4
        $nxaddon3 = ScanNxSdkDir $ParentOfNxSdkRoot 3
        if ($nxaddon6 -or $nxaddon5 -or $nxaddon4 -or $nxaddon3) {
            Write-Output @{
                v6 = $nxaddon6;
                v5 = $nxaddon5;
                v4 = $nxaddon4;
                v3 = $nxaddon3
            }
        }
    }
}
if ($Cygwin -or $All) {
    DoInspection "Cygwin" {
        # must be \cygwin\bin\bash.exe or \cygwin64\bin\bash.exe
        foreach ($path in @("\cygwin64", "\cygwin")) {
            foreach ($drive in $drive_list) {
                $x = ("{0}{1}" -f @($drive.DeviceID, $path))
                if ((Test-Path "$x\bin\bash.exe") -and (Test-Path "$x\bin\cygcheck.exe") -and (Test-Path "$x\bin\uname.exe")) {
                    $arch = & "$x\bin\uname.exe" -m
                    if ($arch -ne "x86_64") { continue } # 64bit only
                    $lines = & "$x\bin\cygcheck.exe" -c cygwin
                    $lines[2] -match "([1-9][^ ]+)" | Out-Null
                    $version = $Matches[1]
                    $Bash = "$x\bin\bash.exe"

                    $gcc_version = $null
                    $text = @(& $Bash -c "/usr/bin/gcc --version" 2>$null)
                    if ($LastExitCode -eq 0) {
                        if ($text[0] -match "([0-9\.]+)") { $gcc_version = $Matches[1] }
                    }

                    $cmake_version = $null
                    $text = @(& $Bash -c "/usr/bin/cmake --version" 2>$null)
                    if ($LastExitCode -eq 0) {
                        if ($text[0] -match "cmake version ([0-9\.]+)") { $cmake_version = $Matches[1] }
                    }

                    $ninja_version = $null
                    $text = @(& $Bash -c "/usr/bin/ninja --version" 2>$null)
                    if ($LastExitCode -eq 0) {
                        if ($text[0] -match "([0-9\.]+)") { $ninja_version = $Matches[1] }
                    }

                    $ccache_version = $null
                    $text = @(& $Bash -c "/usr/bin/ccache --version" 2>$null)
                    if ($LastExitCode -eq 0) {
                        if ($text[0] -match "ccache version ([0-9\.]+)") { $ccache_version = $Matches[1] }
                    }

                    Write-Output @{
                        version = $version;
                        path = $Bash;
                        gcc_version = $gcc_version;
                        cmake_version = $cmake_version;
                        ccache_version = $ccache_version;
                        ninja_version = $ninja_version
                    }
                    return
                }
            }
        }
    }
}
if ($Wsl -or $All) {
    DoInspection "Wsl" {
        $Bash = "${Env:windir}\System32\bash.exe"
        $version = & $Bash -c "uname -r"
        if ($LastExitCode -ne 0) { $version = $null }

        $gcc_version = $null
        $text = @(& $Bash -c "gcc --version" 2>$null)
        if ($LastExitCode -eq 0) {
            if ($text[0] -match "Ubuntu ([0-9\.]+)") { $gcc_version = $Matches[1] }
        }

        $clang_version = $null
        $text = @(& $Bash -c "clang --version" 2>$null)
        if ($LastExitCode -eq 0) {
            if ($text[0] -match "clang version ([0-9\.]+)") { $clang_version = $Matches[1] }
        }

        $cmake_version = $null
        $text = @(& $Bash -c "cmake --version" 2>$null)
        if ($LastExitCode -eq 0) {
            if ($text[0] -match "cmake version ([0-9\.]+)") { $cmake_version = $Matches[1] }
        }

        $ccache_version = $null
        $text = @(& $Bash -c "ccache --version" 2>$null)
        if ($LastExitCode -eq 0) {
            if ($text[0] -match "ccache version ([0-9\.]+)") { $ccache_version = $Matches[1] }
        }

        $ninja_version = $null
        $text = @(& $Bash -c "ninja --version" 2>$null)
        if ($LastExitCode -eq 0) {
            if ($text[0] -match "([0-9\.]+)") { $ninja_version = $Matches[1] }
        }

        Write-Output @{
            version = $version;
            path = $Bash;
            gcc_version = $gcc_version;
            clang_version = $clang_version;
            cmake_version = $cmake_version;
            ccache_version = $ccache_version;
            ninja_version = $ninja_version
        }
    }
}
if ($Vagrant -or $All) {
    DoInspection "VirtualBox" {
        $path = (Get-Command "${Env:ProgramFiles}\Oracle\VirtualBox\VBoxManage.exe" -ErrorAction SilentlyContinue).Source
        if ($path) {
            $version = & "${Env:ProgramFiles}\Oracle\VirtualBox\VBoxManage.exe" --version
            Write-Output @{version = $version; path = $path}
        }
    }
    DoInspection "Vagrant" {
        $path = (Get-Command vagrant -ErrorAction SilentlyContinue).Source
        if ($path) {
            $version = (& vagrant --version) -replace "Vagrant ", ""
            $box_list = & vagrant box list
            Write-Output @{version = $version; box_list = $box_list; path = $path}
        }
    }
}

Write-Output $table
