<#
    Copyright (C)Nintendo All rights reserved.

    These coded instructions, statements, and computer programs contain proprietary
    information of Nintendo and/or its licensed developers and are protected by
    national and international copyright laws. They may not be disclosed to third
    parties or copied or duplicated in any form, in whole or in part, without the
    prior written consent of Nintendo.

    The content herein is highly confidential and should be handled accordingly.
#>

<#
.SYNOPSIS
    Updates CMake
.DESCRIPTION
    Downloads and installs the latest CMake from https://cmake.org/download/ if the older one is installed.
.PARAMETER Inspection
    The output of Inspection.ps1
#>

#Requires -Version 5
Param($Inspection, [switch]$ForceCheck)

Set-Variable script_dir (Resolve-Path (Split-Path -Parent $MyInvocation.MyCommand.Path)).Path -Option ReadOnly
if (!$Inspection) {
    $Inspection = & $script_dir\Inspection.ps1 -CMake
}
$prop = Get-ItemProperty $script_dir\Update-CMake.ps1
$wtime = $prop.LastWriteTime
$atime = $prop.LastAccessTime
if (!$ForceCheckk -and $Inspection.CMake -and ($atime -gt $wtime) -and ($atime.AddMinutes(360) -gt (Get-Date))) {
    Write-Output $Inspection.CMake.version
    return
}

[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12
$links = @((Invoke-WebRequest "https://cmake.org/download/").Links | Where-Object {
    $_.href -match "win64-x64.msi"
} | ForEach-Object {
    if ($_.href -match "cmake-([0-9][0-9\.]*[0-9])") {
        $v = $Matches[1]
        if ($_.href -match "-rc([0-9]+)") {
            $rc = $Matches[1] -as [int]
        } else {
            $rc = 1000000
        }
        $ver = [System.Version]("{0}.{1}" -f @($v, $rc))
        Write-Output @{ href = $_.href; ver = $ver }
    }
} | Sort-Object -Descending @{expression = {$_.ver}})
$href = $links[0].href
$link_ver = $links[0].ver
if (!$href) { throw "href broken" }
Write-Verbose "Newest version is: $link_ver, $href"

if ($Inspection.CMake.version) {
    $Inspection.CMake.version -match "([0-9][0-9\.]*[0-9])" | Out-Null
    $ver = $Matches[1]
    $rc = 1000000
    if ($Inspection.CMake.version -match "-rc([0-9]+)") {
        $rc = $Matches[1] -as [int]
    }
    $installed_ver = [System.Version]("{0}.{1}" -f @($ver, $rc))
} else {
    $installed_ver = [System.Version]"0.0.0.0"
}
Write-Verbose "Installed version is: $installed_ver"

if ($installed_ver -lt $link_ver) {
    Write-Host "Updating cmake ..."
    if (!(Test-Path $script_dir\.cache -PathType Container)) {
        New-Item $script_dir\.cache -ItemType Directory
    }
    if (Test-Path $script_dir\.cache\cmake-package.msi) {
        Remove-Item $script_dir\.cache\cmake-package.msi -Force
    }
    Invoke-WebRequest -Uri "https://cmake.org$href" -OutFile $script_dir\.cache\cmake-package.msi
    $proc = Start-Process -Wait -PassThru -FilePath msiexec.exe `
        -ArgumentList @("/i", "cmake-package.msi", "/passive") `
        -WorkingDirectory $script_dir\.cache
    if ($proc.ExitCode -eq 0) {
        $Inspection = & $script_dir\Inspection.ps1 -CMake
    } else {
        throw ("oops!: msiexec returns {0}" -f $proc.ExitCode)
    }
}

Set-ItemProperty -Path $script_dir\Update-CMake.ps1 -Name LastAccessTime -Value $(Get-Date)
Write-Output $Inspection.CMake.version
