
<#
    Copyright (C)Nintendo All rights reserved.

    These coded instructions, statements, and computer programs contain proprietary
    information of Nintendo and/or its licensed developers and are protected by
    national and international copyright laws. They may not be disclosed to third
    parties or copied or duplicated in any form, in whole or in part, without the
    prior written consent of Nintendo.

    The content herein is highly confidential and should be handled accordingly.
#>

<#
.SYNOPSIS
    Installs/Updates zlib and curl for x86/x64 with vcpkg by Microsoft.
.DESCRIPTION
    This script uses vcpkg(https://github.com/Microsoft/vcpkg) as the package manager.
    Git for Windows and Visual Studio 2015+ have to be installed beforehand.
    Read https://blogs.msdn.microsoft.com/vcblog/2016/09/19/vcpkg-a-tool-to-acquire-and-build-c-open-source-libraries-on-windows/ for more information.
#>

#Requires -Version 5
Set-Variable script_dir (Resolve-Path (Split-Path -Parent $MyInvocation.MyCommand.Path)).Path -Option ReadOnly
if (${Env:VCPKG_ROOT}) {
    Set-Variable vcpkg_dir "${Env:VCPKG_ROOT}" -Option ReadOnly
    Set-Variable vcpkg_exe "${Env:VCPKG_ROOT}\vcpkg.exe" -Option ReadOnly
} else {
    Set-Variable vcpkg_dir "${script_dir}\.cache\vcpkg" -Option ReadOnly
    Set-Variable vcpkg_exe "${script_dir}\.cache\vcpkg\vcpkg.exe" -Option ReadOnly
}

Get-Command git -ErrorAction SilentlyContinue | Out-Null
if (!$?) {
    Write-Error "Please install git"
    exit 1
}

if (!(Test-Path "${vcpkg_dir}\.git")) {
    if (Test-Path "${vcpkg_dir}") {
        Write-Error "${vcpkg_dir} exists, remove it."
        exit 1
    }
    Write-Host "Get vcpkg"
    git clone https://github.com/Microsoft/vcpkg ${vcpkg_dir}
    if (!$?) {
        Write-Error "git clone https://github.com/Microsoft/vcpkg failed"
        exit 1
    }
}

Push-Location . -StackName "Update-MsvcLibraries"
try {
    Set-Location ${vcpkg_dir}
    $src_update = git pull | Select-String toolsrc

    if ($src_update -or !(Test-Path "${vcpkg_exe}")) {
        Write-Host "Build(Rebuild) vcpkg.exe"
        powershell -exec bypass ${vcpkg_dir}\scripts\bootstrap.ps1
        if (!$?) {
            Write-Error "Build vcpkg.exe failed"
            exit 1
        }
    }

    & ${vcpkg_exe} install zlib:x64-windows
    & ${vcpkg_exe} install zlib:x86-windows
    & ${vcpkg_exe} install curl:x64-windows
    & ${vcpkg_exe} install curl:x86-windows
    & ${vcpkg_exe} update
    & ${vcpkg_exe} upgrade --no-dry-run

    if (Test-Path "${Env:ProgramFiles(x86)}/zlib") {
        Write-Warning "zlib at ${Env:ProgramFiles(x86)}/zlib might make trouble with cmake's library detection. Remove it if you can."
    }
    if (Test-Path "${Env:ProgramFiles}/zlib") {
        Write-Warning "zlib at ${Env:ProgramFiles}/zlib might make trouble with cmake's library detection. Remove it if you can."
    }
    if (Test-Path "${Env:ProgramFiles(x86)}/CURL") {
        Write-Warning "libcurl at ${Env:ProgramFiles(x86)}/CURL might make trouble with cmake's library detection. Remove it if you can."
    }
    if (Test-Path "${Env:ProgramFiles}/CURL") {
        Write-Warning "libcurl at ${Env:ProgramFiles}/CURL might make trouble with cmake's library detection. Remove it if you can."
    }
} finally {
    Pop-Location -StackName "Update-MsvcLibraries"
}
