<#
    Copyright (C)Nintendo All rights reserved.

    These coded instructions, statements, and computer programs contain proprietary
    information of Nintendo and/or its licensed developers and are protected by
    national and international copyright laws. They may not be disclosed to third
    parties or copied or duplicated in any form, in whole or in part, without the
    prior written consent of Nintendo.

    The content herein is highly confidential and should be handled accordingly.
#>

<#
.SYNOPSIS
    Updates ninja.exe
.DESCRIPTION
    Downloads and installs the latest ninja.exe from https://github.com/ninja-build/ if the older one is installed.
.PARAMETER Inspection
    The output of Inspection.ps1
#>

#Requires -Version 5
Param($Inspection, [switch]$ForceCheck)

Set-Variable script_dir (Resolve-Path (Split-Path -Parent $MyInvocation.MyCommand.Path)).Path -Option ReadOnly
if (!$Inspection) {
    $Inspection = & $script_dir\Inspection.ps1 -Ninja
}
$prop = Get-ItemProperty $script_dir\Update-Ninja.ps1
$wtime = $prop.LastWriteTime
$atime = $prop.LastAccessTime
if (!$ForceCheck -and $Inspection.Ninja -and ($atime -gt $wtime) -and ($atime.AddMinutes(360) -gt (Get-Date))) {
    Write-Output $Inspection.Ninja.version
    return
}

[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12
$links = @((Invoke-WebRequest "https://github.com/ninja-build/ninja/releases").Links | Where-Object {
    $_.href -match "-win.zip$"
} | ForEach-Object {
    if ($_.href -match "/download/v([0-9][0-9\.]*[0-9])") {
        $ver = [System.Version]$Matches[1]
        Write-Output @{ href = $_.href; ver = $ver }
    }
} | Sort-Object -Descending @{expression = {$_.ver}})
$link_ver = $links[0].ver
$href = $links[0].href
if (!$href) { throw "href broken" }
Write-Verbose "Newest version is: $link_ver, $href"

if ($Inspection.Ninja.version) {
    $Inspection.Ninja.version -match "([0-9][0-9\.]*[0-9])" | Out-Null
    $installed_ver = [System.Version]$Matches[1]
} else {
    $installed_ver = [System.Version]"0.0.0"
}
Write-Verbose "Installed version is: $installed_ver"

if ($installed_ver -lt $link_ver) {
    if (!(Test-Path $script_dir\.cache -PathType Container)) {
        New-Item $script_dir\.cache -ItemType Directory
    }
    $zipfile = "$script_dir\.cache\ninja-win-$link_ver.zip"
    if (Test-Path $zipfile) {
        Remove-Item $zipfile -Force
    }
    if (Test-Path $script_dir\.cache\ninja.exe) {
        Remove-Item $script_dir\.cache\ninja.exe -Force
    }

    Invoke-WebRequest -Uri "https://github.com$href" -OutFile $zipfile
    Expand-Archive $zipfile -Force -DestinationPath $script_dir\.cache
    if ($Inspection.Ninja.path) {
        $dest = $Inspection.Ninja.path
    } else {
        $dest = "${Env:windir}\System32"
    }
    $ninjabin = (Resolve-Path "$script_dir\.cache\ninja.exe").Path
    Start-Process PowerShell.exe -ArgumentList "Copy-Item $ninjabin $dest" -Wait -Verb RunAs
    $Inspection = & $script_dir\Inspection.ps1 -Ninja
}

Set-ItemProperty -Path $script_dir\Update-Ninja.ps1 -Name LastAccessTime -Value $(Get-Date)
Write-Output $Inspection.Ninja.version
